/*
 *  File:        bmtable.h
 *  Purpose:     Declaration of Bitmap Tables
 *  Author:      Alejandro Aguilar Sierra <asierra@servidor.unam.mx> 
 *  Created:     November 1995
 *  Description: A BitmapTable simulates a 2D array of bitmap buttons.
 *               Can be used to build 2D bitmap menues.
 * 
 *  Dependencies: Xlib, XForms
 *               
 *  Copyright: (c) 1995, Alejandro Aguilar Sierra 
 *
 *  Please report bugs and suggestions.
 * 
 *  You are free to use and modify this code under the terms of
 *  the GNU General Public Licence version 2 or later. 
 */ 

#ifndef __BITMAPTABLE_H__
#define __BITMAPTABLE_H__ 

#include "forms.h"

/*  If you have created your own xforms objects,  
 *  maybe you'll need to change this number.
 */
#define FL_BITMAPTABLE 1500

/*  A flat bitmap table */
#define FL_BMTABLE_FLAT 0

/* A boxed bitmaptable */
#define FL_BMTABLE_BOX 1 

/* These two are similar to other xforms functions */
FL_OBJECT *fl_create_bitmaptable(int, FL_Coord, FL_Coord, 
				 FL_Coord, FL_Coord, const char *);

FL_OBJECT *fl_add_bitmaptable(int, FL_Coord, FL_Coord,FL_Coord, FL_Coord, 
			      const char *);

/* Number of columns (nx) and rows (ny), and the background bitmap */
void fl_set_bitmaptable_data(FL_OBJECT *, int nx, int ny, int, int , char *);

void fl_set_bitmaptable_file(FL_OBJECT *, int nx, int ny, const char *);

/* Adjust bitmap origin (ox, oy) and cell dimensions (dx, dy)
 * These values increment current values.
 */
void fl_set_bitmaptable_adjust(FL_OBJECT *, int ox, int oy, int dx, int dy);

/* You can set the number of items */
void fl_set_bitmaptable_maxitems(FL_OBJECT *, int);

int fl_get_bitmaptable_maxitems(FL_OBJECT *); 

/* Returns the index of the selected item or -1 if none is selected */
int fl_get_bitmaptable(FL_OBJECT *ob);

/* Change a item's sub_bitmap at id position */
void fl_replace_bitmaptable_item(FL_OBJECT *ob, int id, int  cw, int ch, char *data); 

/* Get the sub_bitmap from id position. You must alloc enough memory for data. 
 */ 
void fl_get_bitmaptable_item(FL_OBJECT *ob, int id, int *cw, int *ch, char *data);


#endif  /* __BITMAPTABLE_H__ */

