/* This file is part of
 * ======================================================
 * 
 *           LyX, the High Level Word Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *
 *======================================================*/

#include <stdio.h>

#define LYX_UNCHANGED 0
#define LYX_NEED_LITTLE_REFRESH 1
#define LYX_NEED_MORE_REFRESH 2
#define LYX_NEED_VERY_LITTLE_REFRESH 3

/* the footnoteflag */ 
#define LYX_NO_FOOTNOTE 0
#define LYX_OPEN_FOOTNOTE 1
#define LYX_CLOSED_FOOTNOTE 2

/* the footnotekinds */ 
#define LYX_FOOTNOTE 0
#define LYX_MARGIN 1
#define LYX_FIG 2
#define LYX_TAB 3

#define LYX_ROMAN_FAMILY 0
#define LYX_SANS_FAMILY 1
#define LYX_TYPEWRITER_FAMILY 2
#define LYX_MODERN_FAMILY 3	       /* obsolete */
#define LYX_MEDIUM_SERIES 0
#define LYX_BOLD_SERIES 1
#define LYX_UP_SHAPE 0
#define LYX_ITALIC_SHAPE 1
#define LYX_SMALLCAPS_SHAPE 2
#define LYX_LAYOUT_DEFAULT 99
#define LYX_NO_FONT_CHANGE 98

#define LYX_SIZE_TINY 0
#define LYX_SIZE_SMALL 1
#define LYX_SIZE_NORMAL 2
#define LYX_SIZE_LARGE 3
#define LYX_SIZE_LARGER 4
#define LYX_SIZE_LARGEST 5
#define LYX_SIZE_HUGE 6
#define LYX_SIZE_GIANT 7

#define LYX_NO_BAR 0
#define LYX_UNDER_BAR 1

#define LYX_NO_LATEX 0
#define LYX_LATEX 1
#define LYX_FORMULA_LATEX 2
#define LYX_PROTECTED_LATEX 3
#define LYX_PROTECTED_FORMULA_LATEX 4

#define LYX_SMALL_ZOOM 0
#define LYX_NORMAL_ZOOM 1
#define LYX_LARGE_ZOOM 2

#define LYX_META_FOOTNOTE 1
#define LYX_META_MARGIN 2
#define LYX_META_FIG 3
#define LYX_META_TAB 4
#define LYX_META_HFILL 5
#define LYX_META_NEWLINE 6
#define LYX_META_PROTECTED_SEPARATOR 7
#define LYX_META_INSET 8

#define LYX_ALIGN_BLOCK 1
#define LYX_ALIGN_LEFT 2
#define LYX_ALIGN_RIGHT 4
#define LYX_ALIGN_CENTER 8
#define LYX_ALIGN_FORMULA 16
#define LYX_ALIGN_LAYOUT 32

#define LYX_DUMMY_LAYOUT 99


#define LYX_PARSEP_INDENT 0
#define LYX_PARSEP_SKIP 1


#define PAPER_A4PAPER 0
#define PAPER_A4 1
#define PAPER_A4WIDE 2
#define PAPER_WIDEMARGINSA4 3
#define PAPER_USLETTER 4
#define PAPER_A5PAPER 5
#define PAPER_B5PAPER 6
#define PAPER_EXECUTIVEPAPER 7
#define PAPER_LEGALPAPER 8

#define MODE_FINAL 0
#define MODE_DRAFT 1

#define ORIENTATION_PORTRAIT 0
#define ORIENTATION_LANDSCAPE 1


#define PAGESTYLE_PLAIN 0
#define PAGESTYLE_HEADINGS 1
#define PAGESTYLE_EMPTY 2

#define GERMAN 0
#define ENGLISH 1
#define FRENCH 2

/* LyX code for Insets */
#define LYX_NO_CODE 0
#define LYX_TOC_CODE 1
#define LYX_QUOTE_CODE 2
#define LYX_MARK_CODE 3
#define LYX_REF_CODE 4
#define LYX_URL_CODE 5
#define LYX_HTMLURL_CODE 6
#define LYX_SEPARATOR_CODE 7
#define LYX_ENDING_CODE 8
#define LYX_LABEL_CODE 9
#define LYX_IGNORE_CODE 10

struct LyXFont  {
  char family;
  char series;
  char shape;
  char size;
   
  char latex; /* a boolean flag */ 
  char bar;
};

LyXFont GetLyXFont(char family, char series, char shape, char size, 
		   char latex, char bar);


/* insets */ 
class Inset {
public:
  virtual ~Inset(){};
   
  virtual int Ascent(LyXFont font)=0;
  virtual int Descent(LyXFont font)=0;
  virtual int Width(LyXFont font)=0;
  virtual LyXFont ConvertFont(LyXFont font);
  virtual void Draw(LyXFont font, unsigned long  pm, int baseline, float &x)=0;
  virtual void Edit();
  virtual char Editable();
  virtual char AutoDelete();
  virtual void Write(FILE *file)=0;
  virtual void Read(FILE *file)=0;
  virtual void Latex(FILE *file)=0;
  virtual char Deletable();
  virtual Inset* Clone()=0;

  /* returns LyX code associated with the inset. Used for TOC, ...) */
  virtual char LyxCode();
};




class LyXParagraph  {
private:
  
   struct FontTable  {
     int pos;
     int pos_end;
     LyXFont font;
     FontTable *next;
   };
   
  struct InsetTable {
    int pos;
    Inset *inset;
    InsetTable *next;
   };
   
   
  FontTable *fonttable;
  InsetTable *insettable;
   

public:
  int size;
  /* allocates more memory for the specified paragraph
   * pos is needed to specify the paragraph correctly. Remember the
   * closed footnotes */ 
  void Enlarge(int pos, int number);
 
  /* make the allocated memory fit to the needed size
   * used to make a paragraph smaller */
  void FitSize();
      
   
   
  LyXParagraph* Clone();
  char HasSameLayout(LyXParagraph *par);
  void MakeSameLayout(LyXParagraph *par);

  char *text;
  char layout;
  char footnoteflag;		       /* no footnote, closed footnote, 
					* open footnote, where footnote
					* means footnote-environement */
   
  char footnotekind;		       /* footnote, margin, fig, tab  */
   
   
  /* the lyrix- DTP-switches */ 
  char line_top;
  char line_bottom;
   
  char fill_top;
  char fill_bottom;
   
  float added_space_top;	       /* in cm   */
  float added_space_bottom;	       /* in cm   */
   
   
  char pagebreak_top;
  char pagebreak_bottom;
   
  char align;
   
  char depth;
  char noindent;
   
  char counter[10];
  char enumdepth;
  char itemdepth;
   
  char* labelstring;
  char* labelwidthstring;
   
  int last;
  LyXParagraph *next;
  LyXParagraph *previous;
 
   
  /* these function are able to hide closed footnotes */
  LyXParagraph *Next();
  LyXParagraph *Previous();
   
  /* these function are able to hide open and closed footnotes */ 
  LyXParagraph *NextAfterFootnote();  
  LyXParagraph *LastPhysicalPar();    
  LyXParagraph *FirstPhysicalPar();
   
  /* returns the physical paragraph */ 
  LyXParagraph *ParFromPos(int pos);
  /* returns the position in the physical par */
  int PositionInParFromPos(int pos);

  /* for the environements */ 
  LyXParagraph* DepthHook(int depth);
   
  int BeginningOfMainBody();
  char* GetLabelString();

  /* the next two functions are for the manual labels */ 
  char* GetLabelWidthString();
  void SetLabelWidthString(char* s);
   
  int GetLayout();
  char GetAlign();
  char GetDepth();
  void SetLayout(char new_layout);
  void SetOnlyLayout(char new_layout);
   
  char GetCounter(int i);
   
   
  int Last();
   
   
  LyXParagraph();
  /* this konstruktor inserts the new paragraph in a list */ 
  LyXParagraph(LyXParagraph *par);
  /* the destruktors removes the new paragraph from the list */ 
  ~LyXParagraph();
   
  /* this one resets all layout and dtp switches but not the font
   * of the single characters */ 
  void Clear();
   

   
   
   
  void Erase(int pos);

  /* the flag determines wether the layout should be copied */ 
  void BreakParagraph(int pos, int flag);
  void BreakParagraphConservative(int pos);
   
  /* paste this paragraph with the next one
   * be carefull, this doesent make any check at all */ 
  void PasteParagraph();
   
   
  /* returns the difference between the characters font
   * and the layoutfont . This is what is stored in the fonttable */ 
  LyXFont GetFontSettings(int pos);
   
  char GetChar(int pos);
  void SetFont(int pos, LyXFont font);
  void InsertChar(int pos, char c);
  void InsertInset(int pos, Inset *inset);
  Inset* GetInset(int pos);
  /* used to remove the error messages */
  int  AutoDeleteInsets();

  Inset* ReturnNextInsetPointer(int &pos);
   
  /* ok and now some footnote functions */ 
  void OpenFootnotes();
  void OpenFootnote(int pos);
  void CloseFootnotes();
  void CloseFootnote(int pos);
   

  /* important for cut and paste */
  void CopyIntoMinibuffer(int pos);
  void CutIntoMinibuffer(int pos);
  void InsertFromMinibuffer(int pos);
   
   
};




struct Row {
  LyXParagraph *par;
  int pos;
  unsigned short  baseline;
  int fill; /* what is missing to a full row 
	     * can be negative.
	     * needed for hfills, flushright, block etc. */
  unsigned short  height;
   
   
  Row* next;
  Row* previous;
};

struct LyXCursor {
  LyXParagraph *par;
  int pos;
  int x;
  int x_fix;
  long y;
  Row *row;
};


struct LyXTextParameters {
  LyXParagraph *paragraph;
  char paragraph_separation;	       /* wether paragraphs are separated 
					* by using a indent like in articles
					* or by using a little skip like
					* in letters */
   
  char quotes_language;
  char quotes_times;
  char papersize;
  char fontsize;
   
  char textclass;
   
  /* some LaTeX options */ 
  char* epsfig;
  char* fonts;
   
  float baseline;
  signed char secnumdepth;
  signed char tocdepth;
  char* language;
  char* fontenc;
  char* inputenc;
  char* preamble;
  char* options;
  
  char columns;
  char sides;
  char orientation;
  char pagestyle;
   
  LyXCursor cursor;
};



class LyXText {
private:
  Row* firstrow;
  Row* lastrow;
   
  /* the currentrow is needed to access rows faster*/ 
  Row* currentrow;		       /* pointer to the current row  */
  long  currentrow_y;		       /* position in the text   */
   
  unsigned short  paperwidth; /* width of the paper */
   
   
  /* inserts a new row behind the specified row, increments
   * the touched counters */
  void InsertRow(Row *row, LyXParagraph *par, int pos );

  /* removes the row and reset the touched counters */
  void RemoveRow(Row *row);
   
  /* remove all following rows of the paragraph of the specified row. */
  void RemoveParagraph(Row *row);
   
  /* insert the specified paragraph behind the specified row */
  void InsertParagraph(LyXParagraph *par, Row *row);
   
  /* appends  the implizit specified paragraph behind the specified row,
   * start at the implizit given position */
  void AppendParagraph(Row *row);
   
   
  void BreakAgain(Row *row);
   
  void BreakAgainOneRow(Row *row);
   
  void SetHeightOfRow(Row *row_ptr); /* calculate and set the height 
				      * of the row */
   
  /* this calculates the specified parameters. needed when setting
   * the cursor and when creating a visible row */ 
  void PrepareToPrint(Row *row, float &x, float &fill_separator, 
		      float &fill_hfill, float &fill_label_hfill);

  void DeleteEmptyParagraphMechanism(LyXCursor old_cursor);

  /* Updates all counters starting BEHIND the row. Changed paragraphs
   * with a dynamic left border will be rebroken. */ 
  void LyXText::UpdateCounters(Row *row);

  void LyXText::SetCounter(LyXParagraph *par);
   
   
  /* some low level functions */ 
  int SingleWidth(LyXParagraph *par, int pos);
  void Draw(Row *row, int &pos, unsigned long  pm, float &x);
  /* get the next breakpoint in a given paragraph */
  int NextBreakPoint(Row* row, int width);
  /* returns the minimum space a row needs on the screen in pixel */
  int Fill(Row *row, int paperwidth);
  /* returns the minimum space a manual label needs on the screen in pixel */ 
  int LabelFill(Row *row);
  int BeginningOfMainBody(LyXParagraph *par);
  /* Returns the left beginning of the text. 
   * This information cannot be taken from the layouts-objekt, because in 
   * LaTeX the beginning of the text fits in some cases (for example sections)
   * exactly the label-width. */
  int LeftBorder(Row *row);
  int RightBorder(Row *row);
  int LabelEnd (Row *row);
   
  /* returns the number of separators in the specified row. The separator 
   * on the very last column doesnt count */ 
  int LyXText::NumberOfSeparators(Row *row);
   
  /* returns the number of hfills in the specified row. The LyX-Hfill is
   * a LaTeX \hfill so that the hfills at the beginning and at the end were 
   * ignored. This is *MUCH* more usefull than not to ignore!  */
  int LyXText::NumberOfHfills(Row *row);
   
  /* like NumberOfHfills, but only those in the manual label! */ 
  int LyXText::NumberOfLabelHfills(Row *row);
   
  /* boolean. returns true, if a expansion is needed. Rules are given by 
   * LaTeX */ 
  int LyXText::HfillExpansion(Row *row_ptr, int pos);
   
   
  /* returns the paragraph position of the last character in the 
   * specified row */
  int RowLast(Row *row);

   
public:
  LyXTextParameters *parameters;
   
  int number_of_rows;
   
   
   
   
  long height;
   
  LyXFont current_font;	       /* the current font settings */
  LyXFont real_current_font;	       /* the current font  */

  LyXFont GetFont(LyXParagraph* par, int pos);
  void SetCharFont(LyXParagraph *par, int pos, LyXFont font);
   
  /* returns a pointer to the very first LyXParagraph */ 
  LyXParagraph *FirstParagraph();
  
  /* Konstruktor */
  LyXText(int pw, LyXTextParameters *p);
   
  /* Destruktor */ 
  ~LyXText();
   
  /* what you expect when pressing <enter> at cursor position */ 
  void BreakParagraph();
   

  /* set layout over selection and make a total rebreak of those  paragraphs */
  void SetLayout(char layout);
  /* used in setlayout */
  void LyXText::MakeFontEntriesLayoutSpecific(LyXParagraph *par);
   
  /* increment depth over selection and make a total rebreak of those 
   * paragraphs */
  void IncDepth();
  /* decrement depth over selection and make a total rebreak of those  
   * paragraphs */
   void DecDepth();
  /* set font over selection and make a total rebreak of those paragraphs */
  void SetFont(LyXFont font);
   
  /* deletes and inserts again all paragaphs between the cursor
   * and the specified par .The Cursor is needed to set the refreshing
   * parameters. 
   * This function is needed after SetLayout and SetFont etc. */
  void RedoParagraphs(LyXCursor cursor, LyXParagraph *end_par);
   
  void ToggleEmphasize();
  void ToggleFree(LyXFont font);
  void ToggleNoun();
  void ToggleLaTeX();
  void ToggleFormLaTeX();
   
   
  /* recalculates the heights of all previous rows of the specified paragraph.
   * needed, if the last characters font has changed. */ 
  void RedoHeightOfParagraph(LyXCursor cursor);
   
  /* forces the redrawing of a paragraph. Needed when manipulating a 
   * right address box */ 
  void RedoDrawingOfParagraph(LyXCursor cursor);
   

      
  /* insert a character, moves all the following breaks in the 
   * same Paragraph one to the right and make a little rebreak */
  void InsertChar(char c);
  void InsertInset(Inset *inset);
   
  /* completes the insertion with a full rebreak */ 
  int FullRebreak();
   

  /* may be important for the menu */
  char* GetLayout(int row);
  char GetFootnoteFlag(int row);

  Row* need_break_row;
   
  long  refresh_y;
  int refresh_height;
  int refresh_width;
  int refresh_x;
  Row *refresh_row;
   
   
  int refresh_pos;
  char status; /* wether the screen needs a refresh,
		* starting with refresh_y */
  
   
   
  /* returns a pointer to the row near the specified y-coordinate
   * (relative to the whole text). y is set to the real beginning
   * of this row */ 
  Row* GetRowNearY(long &y);

   
  /* returns the column near the specified x-coordinate of the row 
   * x is set to the real beginning of this column  */ 
  int GetColumnNearX(Row *row, int& x);
   
  /* returns a pointer to a specified row. y is set to the beginning
   * of the row */
  Row* GetRow(LyXParagraph *par, int pos, long &y);

   
   
   
  /* returns the height of a default row, needed  for scrollbar */
  int DefaultHeight();


   
  /* the cursor */
  LyXCursor cursor;
   
  /* the selection cursor */
  char selection;	       /* just a boolean flag*/
  char mark_set; 	       /* just a boolean flag*/
   
  LyXCursor sel_cursor;
  LyXCursor sel_start_cursor;
  LyXCursor sel_end_cursor;
  LyXCursor last_sel_cursor; /* needed for the toggling */ 
  LyXCursor toggle_cursor;
  LyXCursor toggle_end_cursor;
   
   
  /* need the selection cursor: */ 
  void SetSelection();
  void ClearSelection();

  /* just selects the word the cursor is in */ 
  void SelectWord();
   
  /* 'selects" the next word, where the cursor is not in 
   * and returns this word as string. THe cursor will be moved 
   * to the beginning of this word. 
   * With SelectSelectedWord can this be highlighted really */ 
  char* SelectNextWord(float &value);
  void SelectSelectedWord();
   
   
  void SetCursor(LyXParagraph *par, int pos);
  void SetCursorFromCoordinates(int x, long y);
   
  void CursorUp();
  void CursorDown();
  void CursorLeft();
  void CursorRight();
  void CursorLeftOneWord();
  void CursorRightOneWord();
  void CursorHome();
  void CursorEnd();
  void CursorTop();
  void CursorBottom();
  void Delete();
  void Backspace();
   
  /* returns a printed row in a pixmap. The y value is needed to decide, 
   * wether it is selected text or not. This is a strange solution but
   * faster. */ 
  unsigned long LyXText::GetVisibleRow(Row *row_ptr, long y);
					   
  unsigned long GetVisiblePos(Row *row_ptr, int pos);
   
  /* footnotes: */
   
   
  void ToggleFootnote();
  
  void OpenFootnotes();
  void OpenFootnote();
  void CloseFootnotes();
  void CloseFootnote();
   
  /* turn the selection into a new environement. If there is no selection,
   * create an empty environement */ 
  void InsertFootnoteEnvironement(char kind);
  void MeltFootnoteEnvironement();
   
   
  void CutSelection();
  void CopySelection();
  void PasteSelection();
   
   
  void InsertFootnote();
  void InsertMarginpar();
  void InsertFigure();
  void InsertTabular();
   
   
  /* the DTP switches for paragraphs. Lyrix will store the top settings
   * always in the first physical paragraph, the bottom settings in the
   * last. When a paragraph is broken, the top settings rest, the bottom 
   * settings are given to the new one. So I can make shure, they do not
   * duplicate themself (and you cannnot make dirty things with them! ) */ 
  void SetParagraph(char line_top, char line_bottom,
		    char pagebreak_top, char pagebreak_bottom,
		    float space_top, float space_bottom,
		    char fill_top, char fill_bottom,
		    char align, 
		    const char* labelwidthstring,
		    char noindent);


  /* these things are for search and replace */
   
  /* returns true if the specified string is at the specified position */
  char IsStringInText(LyXParagraph *par, int pos, char* string);

  /* sets the selection over the number of characters of string, no check!! */
  void SetSelectionOverString(char* string);

  /* simple replacing. The font of the first selected character is used */
  void ReplaceSelectionWithString(char* string);

  /* if the string can be found: return true and set the cursor to the new 
   * position */
  char SearchForward(char* string);
  char SearchBackward(char* string);
 
  /* needed to insert the selection */
  void InsertStringA(char* string);
  void InsertStringB(char* string);
   
  /* usefull when texing from within lyrix */ 
  char GotoNextError();

  /* just another features :) */
  char GotoNextNote();

  /* needed to switch between different classes 
   * this works for a list of paragraphs beginning with the specified par 
   * return value is the number of wrong conversions */ 
  int SwitchLayoutsBetweenClasses(char class1, char class2, LyXParagraph *par);
};






void SetZoom(char new_zoom);
char GetZoom();


/* load a style-file, still VERY dump */ 
void LyXSetStyle(char *path);

int StringEqual(const char* a, const char* b);
char* StringCopy(const char* a);
char* date();

