/* This file is part of
* ======================================================
* 
*           LyX, the High Level Word Processor
* 	 
*	    Copyright (C) 1995 Matthias Ettrich
*
*======================================================*/

#include "config.h"
#include "file.h"
#include <ctype.h>
#include <time.h>
#include "kbmap.h"
#include "keybind.h"
#include "lyxserver.h"
#include "tex-defs.h"

#define LYX_FORMAT 2.1

extern  WriteAlert(const char* s1, const char* s2, const char* s3);

extern char* roman_font_name;
extern char* sans_font_name;
extern char* typewriter_font_name;
extern char font_sizes[];
extern char* font_norm;

static char* string_size[9] = {
   "tiny", "small", "normal", "large", 
   "larger", "largest", "huge", "giant", ""
};

static char* string_footnotekinds[5] = {
   "footnote", "margin", "fig", "tab", ""
};


/* the modern is obsolete and there for compartibility */ 
static char* string_family[5] = {
   "roman", "sans", "typewriter", "modern", ""
};

static char* string_series[3] = {
  "medium", "bold", ""
};

static char* string_shape[4] = {
   "up", "italic", "smallcaps", ""
};

static char* string_latex[4] = {
   "no_latex", "latex", "formula_latex", ""
};

static char* string_bar[3] = {
   "no", "under", ""
};

static char* string_align[5] = {
   "block", "left", "right", "center", ""
};

static char* string_paragraph_separation[3] = {
   "indent", "skip", ""
};

static char* string_papersize[10] = {
   "a4paper", "a4", "a4wide", "widemarginsa4", "usletter", 
   "a5paper","b5paper","executivepaper","legalpaper",""
};

static char* string_orientation[3] = {
   "portrait", "landscape", ""
};
static char* string_pagestyle[4] = {
   "plain", "headings", "empty", ""
};
static char* string_quotes_language[4] = {
   "german", "english", "french", ""
};


 







static char SimpleTeXOnePar(FILE *file, LyXFont &font1, LyXParagraph *par);
static void TeXOnePar(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar);
static void TeXDeeper(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar);
static void TeXEnvironement(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar);
static void TeXFootnote(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar);

static FILE *myfile;

static char token_end_character;

struct TexRow {
  LyXParagraph *par;
  int pos;
  TexRow *next;
};

static int tex_code_break_column;

static char textclass = 0;

static TexRow *texrow = NULL;

char* date() 
{
   time_t tid;
   if ((tid=time(NULL)) == -1)
      return (char*)NULL;
   else
      return (ctime(&tid));
}

static void ResetTexRow() {
  TexRow *tmp=texrow;
  while (texrow) {
    tmp = texrow;
    texrow = texrow->next;
    delete(tmp);
  }
}

static void IncTexRow(LyXParagraph *par, int pos) {
  TexRow *tmp;
  tmp = new TexRow;
  tmp->pos = pos;
  tmp->par = par;
  tmp->next = texrow;
  texrow = tmp;
}

static char* stringtoken = new char[100];


int StringEqual(const char* a, const char* b) {
   int i;
   if (a == b)
     return True;

   if (!a || !b)
     return False;
   
   for (i=0;a[i] && b[i] && a[i]==b[i];i++);
   
   return (a[i]==b[i]);
}

char* StringCopy(const char* a) {
   int i,n;
   if (!a) return NULL;
   
   for (n=0; a[n]; n++);
   char* new_string = new char[n+2];
   for (i=0; i<=n; i++)
      new_string[i] = a[i];
   return new_string;
}


static LyXFont GetFont(LyXParagraph* par, int pos){
   LyXFont layoutfont, tmpfont;
   int main_body;
   
   LyXLayout *layout = style(textclass, par->GetLayout());
   
   if (layout->labeltype != LABEL_MANUAL)
     main_body = 0;
   else
     main_body = par->BeginningOfMainBody();
   
   if (pos < main_body)
     layoutfont = layout->labelfont;
   else
     layoutfont = layout->font;
   
   tmpfont = par->GetFontSettings(pos);
   
   if (tmpfont.family == LYX_LAYOUT_DEFAULT)
     tmpfont.family = layoutfont.family;
   if (tmpfont.series == LYX_LAYOUT_DEFAULT)
     tmpfont.series = layoutfont.series;
   if (tmpfont.shape == LYX_LAYOUT_DEFAULT)
     tmpfont.shape = layoutfont.shape;
   if (tmpfont.size == LYX_LAYOUT_DEFAULT)
     tmpfont.size = layoutfont.size;
   if (tmpfont.latex == LYX_LAYOUT_DEFAULT)
     tmpfont.latex = layoutfont.latex;
   if (tmpfont.bar == LYX_LAYOUT_DEFAULT)
     tmpfont.bar = layoutfont.bar;
   return tmpfont;
};


char IsLyXFilename(const char* filename) {
  int i;
  for (i=0; filename[i]; i++);
  return (
	  (i>4
	   && filename[i-4]=='.' 
	   && filename[i-3]=='l' 	
	   && filename[i-2]=='y' 	
	   && filename[i-1]=='x'
	   )
	  ||
	  (i>5
	   && filename[i-5]=='.' 
	   && filename[i-4]=='l' 	
	   && filename[i-3]=='y' 	
	   && filename[i-2]=='x'
	   && filename[i-1]=='#'
	   )
	  );
}

char IsSGMLFilename(const char* filename) {
  int i;
  for (i=0; filename[i]; i++);
  return (
	  (i>5
	   && filename[i-5]=='.' 
	   && filename[i-4]=='s' 	
	   && filename[i-3]=='g' 	
	   && filename[i-2]=='m'
           && filename[i-1]=='l'
	   )
	  ||
	  (i>6
	   && filename[i-6]=='.' 
	   && filename[i-5]=='s' 	
	   && filename[i-4]=='g' 	
	   && filename[i-3]=='m'
	   && filename[i-2]=='l'
	   && filename[i-1]=='#'
	   )
	  );
}


int OpenFile(const char *filename) {
  myfile = NULL;
  if (IsLyXFilename(filename))
    myfile = fopen(filename, "r");
  
  return (myfile != NULL);
}

int OpenAllFile(const char *filename) {
  myfile = fopen(filename, "r");
  return (myfile != NULL);
}




void CloseFile() {
   fclose(myfile);
}




char* NextToken() {
   char* token = NULL;
   char c = 0;
   int i = 0;
   
   if (!feof(myfile)) {
      token = new char[100];
      do {
	 c = fgetc(myfile);
	 token[i]=c;
	 i++;
      } while (!feof(myfile) && !isspace(c));
      
      token[i-1]='\0';	       /* just the end of a command  */
   }
   token_end_character = c;
   return token;
}

char* NextLongToken() {
     char* token = NULL;
   char c;
   int i = 0;
   
   if (!feof(myfile)) {
      token = new char[100];
      do {
	 c = fgetc(myfile);
	 token[i]=c;
	 i++;
      } while (!feof(myfile) && c!='\n');
      token[i-1]='\0';
   }
   return token;
}

int NextQStringToken(char *input, char *buf) {
   char qc;
   int i = 0;

   *buf = '\0';
   while(input[i] && input[i]<=' ') i++;
   qc = input[i++];
   if(!qc) return 0;
   while(input[i])
   {
      if(input[i]==qc)
      {
         *buf='\0';
         return i+1;
      }
      *buf++ = input[i++];
   }
   //!!!should complain about missing quote
   *buf = '\0';
   return i;
}

int NextStringToken(char* command, char* string[]) {
   char c;
   int i = 0;
   
   if (!feof(myfile)) {
      i = 0;
      do {
	 c = fgetc(myfile);
	 stringtoken[i]=c;
	 i++;
      } while (!feof(myfile) && c!=' ' && c!='\n');
      stringtoken[i-1]='\0';
      
      if (StringEqual(stringtoken, "default"))
      	 i = LYX_LAYOUT_DEFAULT;
      else {
	 for (i=0; string[i][0] && !StringEqual(string[i], stringtoken);i++);
	 if (!string[i][0]) {
	    fprintf(stderr, "LYX_ERROR: unknown argument %s for %s\n", stringtoken, command);
	    i = LYX_LAYOUT_DEFAULT;
	 }
      }
   }
   else
      fprintf(stderr, "LYX_ERROR: file ended while scanning string token \n");
   
   return i;
}


int NextIntToken() {
   char c;
   int number = 0;
   
   if (!feof(myfile)) {
      fscanf(myfile, "%d", &number);
      if (!feof(myfile))
      	c=fgetc(myfile);
   }
   return number;
}

float NextFloatToken() {
   char c;
   float number = 0;
   
   if (!feof(myfile)) {
      fscanf(myfile, "%f", &number);
      if (!feof(myfile))
      	c=fgetc(myfile);
   }
   
   return number;
}

   



/* the Paragraph must exist!*/
void ReadParagraph(LyXParagraph *par) {
   char c = 'a';
   char ac = 'b';
   while(!feof(myfile)) {

      /* read the contents: */
      int fertig = 0;
      
      while (!feof(myfile) && !fertig) {
	 ac=c;
	 c = fgetc(myfile);
	 if (c == ac && c == 10) fertig = 1;
/*	 if (c == 10) fertig = 1;*/ 
	 else {
	    if (c != 10) {
	       par->InsertChar(par->last,c);
	    }
	    else par->InsertChar(par->last,' ');
	 }
      }
      
      if (!feof(myfile)) {
	 par = new LyXParagraph(par);
      }
   }
}

void ReadLanguage(LyXTextParameters &p) {
   char *tmptok;
   char *test;
   int n=0;
   
   tmptok = NextLongToken();
   // check if tmptok is part of tex_babel in tex-defs.h
   while(1) {
      test = tex_babel[n++];
      
      if (strcmp(test,tmptok)==0) {
	 p.language = tmptok;
	 break;
      }
      else if (strcmp(test,"last_item")==0) {
	 fprintf(stderr,"Warning: language `%s' not recognized!",
		 tmptok);
	 fprintf(stderr,"         Setting language to `default'.\n");
	 p.language = StringCopy("default");
 	 break;	 
      }      
   }
}

void ReadEpsfig(LyXTextParameters &p) {
   char *tmptok;
   char *test;
   int n=0;
   
   tmptok = NextLongToken();
   // check if tmptok is part of tex_epsfig in tex_defs.h
   while(1) {
      test = tex_epsfig[n++];
      
      if (strcmp(test,tmptok)==0) {	 
	 p.epsfig = tmptok;
	 break;
      }      
      else if (strcmp(test, "last_item")==0) {
	 fprintf(stderr,"Warning: epsfig driver `%s' not recognized!\n",
		 tmptok);
	 fprintf(stderr,"         Setting epsfig driver to `none'.\n");
	 p.epsfig = StringCopy("none");
	 break;
      }      
   }
}


void ReadPreamble(LyXTextParameters &p) {
   char* token;
   char c;
   LyXParagraph *par = new LyXParagraph;
   int pos = 0;
   int i;
   if (p.preamble)
      delete(p.preamble);
   c = fgetc(myfile);
   while (!feof(myfile)) {
      
      /* is it a lyx command? */ 
      if (c=='\\') {
	 token = NextToken();
	 /*printf("Token: %s\n", token);*/
	 if (StringEqual(token, "end_preamble")) {
	    par->InsertChar(pos - 1, '\0');   /* the - 1 deletes the newline
					       * before end_preamble*/
	    p.preamble = par->text;
	    par->text = NULL; /* otherwise the destructor will delete it */
	    delete(par);
	    return;
	 }
	 else {
	    par->InsertChar(pos, '\\');
	    pos++;
	    for (i = 0; token[i]; i++) {
	       par->InsertChar(pos, token[i]);
	       pos++;
	    }
	    par->InsertChar(pos, token_end_character);
	    pos++;
	 }
	 
	 delete(token);
      }
      else {
	 /* simple text */
	 par->InsertChar(pos, c);
	 pos++;
      }
      
      c = fgetc(myfile);
      
   }
}
   


int ReadRCFile(char* filename, LyXRC &rc) {
   FILE *file = fopen(filename, "r");
   char* token;
   char c;
   char *tmptok;
   int tmpret;
   textclass = rc.parameters.textclass;
   
   if (!file) {
      fprintf(stderr,"No resource file `%s', using defaults. \n", filename);
      return -1;
   }

   myfile = file;
   
   c = fgetc(myfile);
   while (!feof(myfile)) {
      /* is it a lyx command? */ 
      if (c=='\\') {
	 token = NextToken();
	 /*printf("Token: %s\n", token);*/
	 if (StringEqual(token, "begin_preamble")) {
	    ReadPreamble(rc.parameters);
	 }
	 else if (StringEqual(token, "options")) {
	    tmptok = NextLongToken();
	    rc.parameters.options = tmptok;
	 }
	 else if (StringEqual(token, "textclass")) {
	    tmptok = NextLongToken();
	    textclass = NumberOfClass(tmptok);
	    rc.parameters.textclass = textclass;
	    delete(tmptok);
	 }
	 else if (StringEqual(token, "language")) {
	    // tmptok = NextLongToken();
	    // rc.parameters.language = tmptok;
	    ReadLanguage(rc.parameters);
	 }
	 else if (StringEqual(token, "fontencoding")) {
	    tmptok = NextLongToken();
	    rc.parameters.fontenc = tmptok;
	 }
	 else if (StringEqual(token, "inputencoding")) {
	    tmptok = NextLongToken();
	    rc.parameters.inputenc = tmptok;
	 }
	 else if (StringEqual(token, "fontscheme")) {
	    tmptok = NextLongToken();
	    rc.parameters.fonts = tmptok;
	 }
	 else if (StringEqual(token, "printer")) {
	    tmptok = NextLongToken();
	    rc.printer = tmptok;
	 }
	 else if (StringEqual(token, "print_command")) {
	    tmptok = NextLongToken();
	    rc.print_command = tmptok;
	 }
	 else if (StringEqual(token, "print_evenpage_flag")) {
	    tmptok = NextLongToken();
	    rc.print_evenpage_flag = tmptok;
	 }
	 else if (StringEqual(token, "print_oddpage_flag")) {
	    tmptok = NextLongToken();
	    rc.print_oddpage_flag = tmptok;
	 }
	 else if (StringEqual(token, "print_reverse_flag")) {
	    tmptok = NextLongToken();
	    rc.print_reverse_flag = tmptok;
	 }
	 else if (StringEqual(token, "print_landscape_flag")) {
	    tmptok = NextLongToken();
	    rc.print_landscape_flag = tmptok;
	 }
	 else if (StringEqual(token, "print_to_printer")) {
	    tmptok = NextLongToken();
	    rc.print_to_printer = tmptok;
	 }
	 else if (StringEqual(token, "print_to_file")) {
	    tmptok = NextLongToken();
	    rc.print_to_file = tmptok;
	 }
	 else if (StringEqual(token, "print_file_extension")) {
	    tmptok = NextLongToken();
	    rc.print_file_extension = tmptok;
	 }
	 else if (StringEqual(token, "print_extra_options")) {
	    tmptok = NextLongToken();
	    rc.print_extra_options = tmptok;
	 }
	 else if (StringEqual(token, "latex_command")) {
	    tmptok = NextLongToken();
	    rc.latex_command = tmptok;
	 }
	 else if (StringEqual(token, "normal_fontsize")) {
	    rc.default_zoom = NextIntToken();
	 }
	 else if (StringEqual(token, "fontsizes")) {
	    font_sizes[LYX_SIZE_TINY] = NextIntToken();
	    font_sizes[LYX_SIZE_SMALL] = NextIntToken();
	    font_sizes[LYX_SIZE_NORMAL] = NextIntToken();
	    font_sizes[LYX_SIZE_LARGE] = NextIntToken();
	    font_sizes[LYX_SIZE_LARGER] = NextIntToken();
	    font_sizes[LYX_SIZE_LARGEST] = NextIntToken();
	    font_sizes[LYX_SIZE_HUGE] = NextIntToken();
	    font_sizes[LYX_SIZE_GIANT] = NextIntToken();
	 }
	 else if (StringEqual(token, "autosave")) {
	    rc.autosave = NextIntToken();
	 }
	 else if (StringEqual(token, "xdvi_extra_options")) {
	    tmptok = NextLongToken();
	    rc.xdvi_extra_options = tmptok;
	 }
         else if (StringEqual(token, "sgml_extra_options")) {
            tmptok = NextLongToken();
            rc.sgml_extra_options = tmptok;
         }
         else if (StringEqual(token, "lyx_layout_lib")) {
            tmptok = NextLongToken();
            rc.lyx_layout_lib = tmptok;
         }
	 else if (StringEqual(token, "template_path")) {
	    tmptok = NextLongToken();
	    rc.template_path = tmptok;
	 }
	 else if (StringEqual(token, "lastfiles")) {
	    tmptok = NextLongToken();
	    rc.lastfiles = tmptok;
	 }
	 else if (StringEqual(token, "roman_font")) {
	    roman_font_name = NextLongToken();
	 }
	 else if (StringEqual(token, "sans_font")) {
	    sans_font_name = NextLongToken();
	 }
	 else if (StringEqual(token, "typewriter_font")) {
	    typewriter_font_name = NextLongToken();
	 }
	 else if (StringEqual(token, "font_norm")) {
	    font_norm = NextLongToken();
	 }
	 else if (StringEqual(token, "paragraph_separation")) {
	    tmpret = NextStringToken(token, string_paragraph_separation);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       rc.parameters.paragraph_separation = tmpret;
	 }
	 else if (StringEqual(token, "epsfig")) {
	   ReadEpsfig(rc.parameters);
	 }	 
	 else if (StringEqual(token, "quotes_language")) {
	    tmpret = NextStringToken(token, string_quotes_language);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       rc.parameters.quotes_language = tmpret;
	 }
	 else if (StringEqual(token, "quotes_times")) {
	    rc.parameters.quotes_times = NextIntToken();
	 }
	 else if (StringEqual(token, "papersize")) {
	    tmpret = NextStringToken(token, string_papersize);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       rc.parameters.papersize = tmpret;
	 }
	 else if (StringEqual(token, "paperfontsize")) {
	    rc.parameters.fontsize = NextIntToken();
	 }
	 else if (StringEqual(token, "paperorientation")) {
	    tmpret = NextStringToken(token, string_orientation);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       rc.parameters.orientation = tmpret;
	 }
	 else if (StringEqual(token, "papercolumns")) {
	    rc.parameters.columns = NextIntToken();
	 }
	 else if (StringEqual(token, "papersides")) {
	    rc.parameters.sides = NextIntToken();
	 }
	 else if (StringEqual(token, "paperpagestyle")) {
	    tmpret = NextStringToken(token, string_pagestyle);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       rc.parameters.pagestyle = tmpret;
	 }
	 else if (StringEqual(token, "baselinestretch")) {
	    rc.parameters.baseline = NextFloatToken();
	 }
	 else if (StringEqual(token, "secnumdepth")) {
	    rc.parameters.secnumdepth = NextIntToken();
	 }
	 else if (StringEqual(token, "tocdepth")) {
	    rc.parameters.tocdepth = NextIntToken();
	 }
         else if (StringEqual(token, "bind")) {
            // !!!chb, dynamic key binding...
            int i;
            kb_sequence tmpseq;
            kb_func_table *tab = lyx_func_table;
            char seq[100], cmd[100];
            tmptok = NextLongToken();
            i = NextQStringToken(tmptok, seq);
            i = NextQStringToken(tmptok+i, cmd);
            /* fprintf(stderr,"bind: /%s/ /%s/\n", seq, cmd); */
            delete tmptok;
            for (i = 0; tab[i].name; i++)   // lookup the command
            {
               if (StringEqual(cmd, tab[i].name)) {
                  if (0==tmpseq.parse(seq)) {
                     toplevel_keymap->defkey(&tmpseq, tab[i].action);
                  }
                  else {
                     fprintf(stderr,
                             "LYX_ERROR: invalid key sequence '%s'\n", seq);
                  }
                  tmpseq.delseq();
                  break;
               }
            }
            if (!tab[i].name)
               fprintf(stderr, "LYX_ERROR: unknown lyx func /%s/\n", cmd);
         }
         else if (StringEqual(token, "serverpipe")) {
            tmptok = NextLongToken();
            char name[256];
            NextQStringToken(tmptok, name);
            lyxpipes.pipename = new char[strlen(name)+1];
            strcpy(lyxpipes.pipename, name);
            delete tmptok;
         }
	 else {
	    fprintf(stderr, "LYX_ERROR: unknown rc command %s\n", token);
	 }
	 
	 delete(token);
      }
      else if (c=='#') {
	 tmptok = NextLongToken();
	 delete(tmptok);
      }
      
      c = fgetc(myfile);
      
   }
   fclose(file);
   return 0;
}

      
	  
void ReadFileOld(LyXTextParameters &p) {
   char* token;
   char c;
   char *tmptok;
   int obsolete;
   LyXParagraph *par = new LyXParagraph;
   Inset *inset;
   int pos = 0;
   textclass = p.textclass;
   LyXParagraph *return_par = NULL;
   LyXFont font = GetLyXFont(LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT);

   c = fgetc(myfile);
   while (!feof(myfile)) {
      /* is it a lyx command? */ 
      if (c=='\\') {
	 token = NextToken();
	 /*printf("Token: %s\n", token);*/
	 if (StringEqual(token, "layout")) {
	    if (!return_par) 
	      return_par = par;
	    else 
	      par = new LyXParagraph(par);
	    pos = 0;
	    tmptok = NextLongToken();
	    par->layout = NumberOfLayout(textclass, tmptok);
	    delete(tmptok);
	    font = GetLyXFont(LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT);
	    
	 }
	 else if (StringEqual(token, "begin_preamble")) {
	    ReadPreamble(p);
	 }
	 else if (StringEqual(token, "textclass")) {
	    tmptok = NextLongToken();
	    textclass = NumberOfClass(tmptok);
	    p.textclass = textclass;
	    delete(tmptok);
	 }
	 else if (StringEqual(token, "otions")) {
	    tmptok = NextLongToken();
	    p.options = tmptok;
	 }
	 else if (StringEqual(token, "language")) {
	    // tmptok = NextLongToken();
	    // p.language = tmptok;
	    ReadLanguage(p);	    
	 }
	 else if (StringEqual(token, "noindent")) {
	    par->noindent = 1;
	 }
	 else if (StringEqual(token, "fill_top")) {
	    par->fill_top = 1;
	 }
	 else if (StringEqual(token, "fill_bottom")) {
	    par->fill_bottom = 1;
	 }
	 else if (StringEqual(token, "line_top")) {
	    par->line_top = 1;
	 }
	 else if (StringEqual(token, "line_bottom")) {
	    par->line_bottom = 1;
	 }
	 else if (StringEqual(token, "pagebreak_top")) {
	    par->pagebreak_top = 1;
	 }
	 else if (StringEqual(token, "pagebreak_bottom")) {
	    par->pagebreak_bottom = 1;
	 }
	 else if (StringEqual(token, "footnote")) {
	    par->footnoteflag = LYX_CLOSED_FOOTNOTE;
	    par->footnotekind = NextIntToken();
	 }
	 else if (StringEqual(token, "depth")) {
	    par->depth = NextIntToken();
	 }
	 else if (StringEqual(token, "paragraph_separation")) {
	    p.paragraph_separation = NextIntToken();
	 }
	 else if (StringEqual(token, "quotes_language")) {
	    p.quotes_language = NextIntToken();
	 }
	 else if (StringEqual(token, "quotes_times")) {
	    p.quotes_times = NextIntToken();
	 }
	 else if (StringEqual(token, "papersize")) {
	    p.papersize = NextIntToken();
	 }
	 else if (StringEqual(token, "paperfontsize")) {
	    p.fontsize = NextIntToken();
	 }
	 else if (StringEqual(token, "paperorientation")) {
	    p.orientation = NextIntToken();
	 }
	 else if (StringEqual(token, "papercolumns")) {
	    p.columns = NextIntToken();
	 }
	 else if (StringEqual(token, "papersides")) {
	    p.sides = NextIntToken();
	 }
	 else if (StringEqual(token, "paperpagestyle")) {
	    p.pagestyle = NextIntToken();
	 }
	 else if (StringEqual(token, "secnumdepth")) {
	    p.secnumdepth = NextIntToken();
	 }
	 else if (StringEqual(token, "no_isolatin1")) {
	    obsolete= NextIntToken();
	 }
	 else if (StringEqual(token, "no_babel")) {
	    obsolete = NextIntToken();
	 }
	 else if (StringEqual(token, "no_epsfig")) {
	    obsolete = NextIntToken();
	 }
	 else if (StringEqual(token, "baselinestretch")) {
	    p.baseline = NextFloatToken();
	 }
	 else if (StringEqual(token, "cursor")) {
	    p.cursor.par = par;
	    p.cursor.pos = NextIntToken();
	 }
	 else if (StringEqual(token, "fontfamily")) {
	    font.family = NextIntToken();
	 }
	 else if (StringEqual(token, "fontseries")) {
	    font.series = NextIntToken();
	 }
	 else if (StringEqual(token, "fontshape")) {
	    font.shape = NextIntToken();
	 }
	 else if (StringEqual(token, "fontsize")) {
	    font.size = NextIntToken();
	 }
	 else if (StringEqual(token, "latex")) {
	    font.latex = NextIntToken();
	 }
	 else if (StringEqual(token, "align")) {
	    par->align = NextIntToken();
	 }
	 else if (StringEqual(token, "added_space_top")) {
	    par->added_space_top = NextFloatToken();
	 }
	 else if (StringEqual(token, "added_space_bottom")) {
	    par->added_space_bottom = NextFloatToken();
	 }
	 else if (StringEqual(token, "labelwidthstring")) {
	    par->labelwidthstring = NextLongToken();
	    /* do not delete this token, it is still needed! */ 
	 }
	 else if (StringEqual(token, "InsetQuotes")) {
	    inset = new InsetQuotes;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 
	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "InsetLatex")) {
	    inset = new InsetLatex;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 
	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "InsetLatexDel")) {
	    inset = new InsetLatexDel;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 
	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "InsetFormulaLatex")) {
	    inset = new InsetFormulaLatex;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 
	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "newline")) {
	    par->InsertChar(pos, LYX_META_NEWLINE);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "hfill")) {
	    par->InsertChar(pos, LYX_META_HFILL);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "protected_separator")) {
	    par->InsertChar(pos, LYX_META_PROTECTED_SEPARATOR);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "backslash")) {
	    par->InsertChar(pos, '\\');
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else {
	    fprintf(stderr, "LYX_ERROR: unknown command %s\n", token);
	 }
	 
	 delete(token);
      }
      else {
	 /* simple text */
	 if (c!='\n') {
	    par->InsertChar(pos, c);
	    par->SetFont(pos, font);
	    pos++;
	 }
      }
      
      c = fgetc(myfile);
      
   }
   
   if (!return_par)
     return_par = par;

   p.paragraph = return_par;
}
	    
	    
void ReadLyxformat2(LyXTextParameters &p) {
   char* token;
   char c;
   char *tmptok;
   LyXParagraph *par = new LyXParagraph;
   Inset *inset;
   int pos = 0;
   int obsolete;
   char *obsolete_char;
   int tmpret, tmpret2;
   char depth = 0;
   char footnoteflag = LYX_NO_FOOTNOTE;
   char footnotekind = LYX_FOOTNOTE;
   textclass = p.textclass;
   LyXParagraph *return_par = NULL;
   LyXFont font = GetLyXFont(LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT,
			     LYX_LAYOUT_DEFAULT);

   c = fgetc(myfile);
   while (!feof(myfile)) {
      /* is it a lyx command? */ 
      if (c=='\\') {
	 token = NextToken();
	 /*printf("Token: %s\n", token);*/
	 if (StringEqual(token, "layout")) {
	    if (!return_par) 
	      return_par = par;
	    else 
	      par = new LyXParagraph(par);
	    pos = 0;
	    tmptok = NextLongToken();
	    par->layout = NumberOfLayout(textclass, tmptok);
	    par->footnoteflag = footnoteflag;
	    par->footnotekind = footnotekind;
	    par->depth = depth;
	    delete(tmptok);
	    font = GetLyXFont(LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT);
	    
	 }
	 else if (StringEqual(token, "end_float")) {
	    if (!return_par) 
	      return_par = par;
	    else 
	      par = new LyXParagraph(par);
	    footnotekind = LYX_FOOTNOTE;
	    footnoteflag = LYX_NO_FOOTNOTE;
	    pos = 0;
	    tmptok = NextLongToken();
	    par->layout = LYX_DUMMY_LAYOUT;
	    delete(tmptok);
	    font = GetLyXFont(LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT,
			      LYX_LAYOUT_DEFAULT);
	 }
	 else if (StringEqual(token, "begin_float")) {
	    tmpret = NextStringToken(token, string_footnotekinds);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       footnotekind = tmpret;
	    footnoteflag = LYX_CLOSED_FOOTNOTE;
	 }
	 else if (StringEqual(token, "begin_deeper")) {
	    depth++;
	 }
	 else if (StringEqual(token, "end_deeper")) {
	    if (!depth) {
	       fprintf(stderr, "LYX_ERROR: end_deeper: depth is already null \n");
	    }
	    else
	       depth--;
	 }
	 else if (StringEqual(token, "begin_preamble")) {
	    ReadPreamble(p);
	 }
	 else if (StringEqual(token, "textclass")) {
	    tmptok = NextLongToken();
	    textclass = NumberOfClass(tmptok);
	    p.textclass = textclass;
	    delete(tmptok);
	 }
	 else if (StringEqual(token, "otions")) {
	    tmptok = NextLongToken();
	    p.options = tmptok;
	 }
	 else if (StringEqual(token, "language")) {
	    // tmptok = NextLongToken();
	    // p.language = tmptok;
	    ReadLanguage(p);	    
	 }
	 else if (StringEqual(token, "fontencoding")) {
	    obsolete_char = NextLongToken();
	    /* p.fontenc = tmptok; */
	 }
	 else if (StringEqual(token, "inputencoding")) {
	    tmptok = NextLongToken();
	    p.inputenc = tmptok;
	 }
	 else if (StringEqual(token, "epsfig")) {
	   ReadEpsfig(p);
	 }	 
	 else if (StringEqual(token, "fontscheme")) {
	    tmptok = NextLongToken();
	    p.fonts = tmptok;
	 }
	 else if (StringEqual(token, "noindent")) {
	    par->noindent = 1;
	 }
	 else if (StringEqual(token, "fill_top")) {
	    par->fill_top = 1;
	 }
	 else if (StringEqual(token, "fill_bottom")) {
	    par->fill_bottom = 1;
	 }
	 else if (StringEqual(token, "line_top")) {
	    par->line_top = 1;
	 }
	 else if (StringEqual(token, "line_bottom")) {
	    par->line_bottom = 1;
	 }
	 else if (StringEqual(token, "pagebreak_top")) {
	    par->pagebreak_top = 1;
	 }
	 else if (StringEqual(token, "pagebreak_bottom")) {
	    par->pagebreak_bottom = 1;
	 }
	 else if (StringEqual(token, "paragraph_separation")) {
	    tmpret = NextStringToken(token, string_paragraph_separation);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       p.paragraph_separation = tmpret;
	 }
	 else if (StringEqual(token, "no_isolatin1")) {
	    obsolete= NextIntToken();
	 }
	 else if (StringEqual(token, "no_babel")) {
	    obsolete = NextIntToken();
	 }
	 else if (StringEqual(token, "no_epsfig")) {
	    obsolete = NextIntToken();
	 }
	 else if (StringEqual(token, "quotes_language")) {
	    tmpret = NextStringToken(token, string_quotes_language);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       p.quotes_language = tmpret;
	 }
	 else if (StringEqual(token, "quotes_times")) {
	    p.quotes_times = NextIntToken();
	 }
	 else if (StringEqual(token, "papersize")) {
	    tmpret = NextStringToken(token, string_papersize);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       p.papersize = tmpret;
	 }
	 else if (StringEqual(token, "paperfontsize")) {
	    p.fontsize = NextIntToken();
	 }
	 else if (StringEqual(token, "paperorientation")) {
	    tmpret = NextStringToken(token, string_orientation);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       p.orientation = tmpret;
	 }
	 else if (StringEqual(token, "papercolumns")) {
	    p.columns = NextIntToken();
	 }
	 else if (StringEqual(token, "papersides")) {
	    p.sides = NextIntToken();
	 }
	 else if (StringEqual(token, "paperpagestyle")) {
	    tmpret = NextStringToken(token, string_pagestyle);
	    if (tmpret != LYX_LAYOUT_DEFAULT) 
	       p.pagestyle = tmpret;
	 }
	 else if (StringEqual(token, "secnumdepth")) {
	    p.secnumdepth = NextIntToken();
	 }
	 else if (StringEqual(token, "tocdepth")) {
	    p.tocdepth = NextIntToken();
	 }
	 else if (StringEqual(token, "baselinestretch")) {
	    p.baseline = NextFloatToken();
	 }
	 else if (StringEqual(token, "cursor")) {
	    p.cursor.par = par;
	    p.cursor.pos = NextIntToken();
	 }
	 else if (StringEqual(token, "family")) { 
	    font.family = NextStringToken(token, string_family);
	    if (font.family == LYX_MODERN_FAMILY)
	       font.family = LYX_SANS_FAMILY;
	 }
	 else if (StringEqual(token, "series")) {
	    font.series = NextStringToken(token, string_series);
	 }
	 else if (StringEqual(token, "shape")) {
	    font.shape = NextStringToken(token, string_shape);
	 }
	 else if (StringEqual(token, "size")) {
	    font.size = NextStringToken(token, string_size);
	 }
	 else if (StringEqual(token, "latex")) {
	    font.latex = NextStringToken(token, string_latex);
	 }
	 else if (StringEqual(token, "bar")) {
	    font.bar = NextStringToken(token, string_bar);
	 }
	 else if (StringEqual(token, "align")) {
	    tmpret = NextStringToken(token, string_align);
	    if (tmpret != LYX_LAYOUT_DEFAULT) {
	       tmpret2 = 1;
	       for (; tmpret>0; tmpret--)
	       	  tmpret2 = tmpret2 * 2;
	       par->align = tmpret2;
	    }
	 }
	 else if (StringEqual(token, "added_space_top")) {
	    par->added_space_top = NextFloatToken();
	 }
	 else if (StringEqual(token, "added_space_bottom")) {
	    par->added_space_bottom = NextFloatToken();
	 }
	 else if (StringEqual(token, "labelwidthstring")) {
	    par->labelwidthstring = NextLongToken();
	    /* do not delete this token, it is still needed! */ 
	 }
	 else if (StringEqual(token, "end_inset")) {
	    /* simple ignore this. The insets do not have to read this */ 
	 }
	 else if (StringEqual(token, "begin_inset")) {
	    tmptok = NextToken();
	    /* test the different insets */ 
	    if (StringEqual(tmptok, "Quotes")) {
	       inset = new InsetQuotes;
	       inset->Read(myfile);
	       par->InsertChar(pos, LYX_META_INSET); 
	       par->InsertInset(pos, inset);
	       par->SetFont(pos, font);
	       pos++;
	    }
	    else if (StringEqual(tmptok, "Latex")) {
	       inset = new InsetLatex;
	       inset->Read(myfile);
	       par->InsertChar(pos, LYX_META_INSET); 
	       par->InsertInset(pos, inset);
	       par->SetFont(pos, font);
	       pos++;
	    }
	    else if (StringEqual(tmptok, "LatexDel")) {
	       inset = new InsetLatexDel;
	       inset->Read(myfile);
	       par->InsertChar(pos, LYX_META_INSET); 
	       
	       par->InsertInset(pos, inset);
	       par->SetFont(pos, font);
	       pos++;
	    }
	    else if (StringEqual(tmptok, "FormulaLatex")) {
	       inset = new InsetFormulaLatex;
	       inset->Read(myfile);
	       par->InsertChar(pos, LYX_META_INSET); 
	       par->InsertInset(pos, inset);
	       par->SetFont(pos, font);
	       pos++;
	    }
	    else if (StringEqual(tmptok, "Label")) {
	       inset = new InsetLabel;
	       inset->Read(myfile);
	       par->InsertChar(pos, LYX_META_INSET); 
	       par->InsertInset(pos, inset);
	       par->SetFont(pos, font);
	       pos++;
	    }
	    else if (StringEqual(tmptok, "Info")) {
	       inset = new InsetInfo;
	       inset->Read(myfile);
	       par->InsertChar(pos, LYX_META_INSET); 
	       par->InsertInset(pos, inset);
	       par->SetFont(pos, font);
	       pos++;
	    }
	    delete(tmptok);
	 }
	 else if (StringEqual(token, "InsetQuotes")) {
	    inset = new InsetQuotes;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 
	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "InsetLatex")) {
	    inset = new InsetLatex;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 
	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "InsetLatexDel")) {
	    inset = new InsetLatexDel;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 

	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "InsetFormulaLatex")) {
	    inset = new InsetFormulaLatex;
	    inset->Read(myfile);
	    par->InsertChar(pos, LYX_META_INSET); 
	    par->InsertInset(pos, inset);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "newline")) {
	    par->InsertChar(pos, LYX_META_NEWLINE);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "hfill")) {
	    par->InsertChar(pos, LYX_META_HFILL);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "protected_separator")) {
	    par->InsertChar(pos, LYX_META_PROTECTED_SEPARATOR);
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else if (StringEqual(token, "backslash")) {
	    par->InsertChar(pos, '\\');
	    par->SetFont(pos, font);
	    pos++;
	 }
	 else {
	    fprintf(stderr, "LYX_ERROR: unknown command %s\n", token);
	 }
	 
	 delete(token);
      }
      else {
	 /* simple text */
	 if (c!='\n') {
	    par->InsertChar(pos, c);
	    par->SetFont(pos, font);
	    pos++;
	 }
      }
      
      c = fgetc(myfile);
      
   }
   
   if (!return_par)
     return_par = par;

   p.paragraph = return_par;
}
	    
void ReadFile(LyXTextParameters &p) {
   char* token;
   char c;
   
   c = fgetc(myfile);
   if (c != '#') {
      ungetc(c, myfile);
      ReadFileOld(p);
      return;
   }
   while (!feof(myfile)) {
      /* is it a lyx command? */ 
      if (c=='\\') {
	 token = NextToken();
	 if (StringEqual(token, "lyxformat")) {
	    delete(token);
	    float format = NextFloatToken();
	    if (format > 1) {
	       if (LYX_FORMAT - format > 0.05) {
		  printf("Warning: read lyxformat %.2f but found %.2f\n",
			 LYX_FORMAT, format);
	       }
	       if (format - LYX_FORMAT > 0.05) {
		  printf("ERROR: read lyxformat %.2f but found %.2f\n",
			 LYX_FORMAT, format);
	       }
	       ReadLyxformat2(p);
	       return;
	    }
	    else {
	       ReadFileOld(p);
	       return;
	    }
	 }
	 delete (token);
      }
      c = fgetc(myfile);
   }
   WriteAlert("LYX_ERROR:", "couldn't find out lyxformat\n", "");
}
	    	    
	      
void WriteFile(const char* filename, LyXTextParameters p){
   FILE *file = fopen(filename, "w");
   LyXFont font1, font2;
   Inset *inset;
   LyXParagraph *par = p.paragraph;
   char c;
   int i;
   
   char footnoteflag = 0;
   char depth = 0;

   int h;
   
   textclass = p.textclass;
   
   if (!file) {
      WriteAlert("LYX_ERROR:", "Cannot write file:", filename);
      return;
   }
   
   /* some very important information */ 
   // yes...but let's change it a bit...
   fprintf(file, "#This file was created by <%s> %s",getenv("USER"),date());
   fprintf(file, "#LyX 0.8 (C) 1995 Matthias Ettrich\n");
   

   /* at the very beginning the used lyx format */
   fprintf(file, "\\lyxformat %.2f\n", LYX_FORMAT);
   
   /* the textclass */
   fprintf(file, "\\textclass %s\n", NameOfClass(p.textclass));
   
   /* then the the preamble */
   if (p.preamble && p.preamble[0]) {
      fprintf(file, "\\begin_preamble\n");
	{
	   for (i = 0; p.preamble[i];i++) {
	      fprintf(file, "%c", p.preamble[i]);
	   }
	   fprintf(file, "\n\\end_preamble\n");
	}
   }
      
   /* the options */ 
   if (p.options[0]) {
      fprintf(file, "\\options %s\n", p.options);
   }
   
   
   
   
   /* then the text parameters */ 
   fprintf(file, "\\language %s\n", p.language);
   fprintf(file, "\\inputencoding %s\n", p.inputenc);
   fprintf(file, "\\fontscheme %s\n", p.fonts);
   fprintf(file, "\\epsfig %s\n", p.epsfig);
   fprintf(file, "\\papersize %s ", string_papersize[p.papersize]);
   fprintf(file, "\\paperfontsize %d ", p.fontsize);
   fprintf(file, "\\baselinestretch %.2f ", p.baseline);
   fprintf(file, "\\secnumdepth %d ", p.secnumdepth);
   fprintf(file, "\\tocdepth %d ", p.tocdepth);
   fprintf(file, "\\paragraph_separation %s ", string_paragraph_separation[p.paragraph_separation]);
   fprintf(file, "\\quotes_language %s ", string_quotes_language[p.quotes_language]);
   fprintf(file, "\\quotes_times %d ", p.quotes_times);
   fprintf(file, "\\paperorientation %s ", string_orientation[p.orientation]);
   fprintf(file, "\\papercolumns %d ", p.columns);
   fprintf(file, "\\papersides %d ", p.sides);
   fprintf(file, "\\paperpagestyle %s ", string_pagestyle[p.pagestyle]);
   
   
   
   while (par) {
      
      if (par->layout != LYX_DUMMY_LAYOUT) {
	 
	 /* begins a footnote environement ? */ 
	 if (footnoteflag != par->footnoteflag) {
	    footnoteflag = par->footnoteflag;
	    if (footnoteflag) {
	       fprintf(file, "\n\\begin_float %s ", 
		       string_footnotekinds[par->footnotekind]);
	    }
	    else {
	       fprintf(file, "\n\\end_float ");
	    }
	 }
	 
	 /* begins or ends a deeper area ?*/ 
	 if (depth != par->depth) {
	    if (par->depth > depth) {
	       while (par->depth > depth) {
		  fprintf(file, "\n\\begin_deeper ");
		  depth++;
	       }
	    }
	    else {
	       while (par->depth < depth) {
		  fprintf(file, "\n\\end_deeper ");
		  depth--;
	       }
	    }
	 }
	 
	 
	 /* First write the layout */ 
	 fprintf(file, "\n\\layout %s\n", NameOfLayout(textclass,par->layout));
	 
	 /* maybe some vertical spaces */ 
	 if (par->added_space_top != 0)
	    fprintf(file, "\\added_space_top %.2f ", par->added_space_top);
	 if (par->added_space_bottom != 0)
	    fprintf(file, "\\added_space_bottom %.2f ",
		    par->added_space_bottom);
	 
	 /* the labelwidthstring used in lists */ 
	 if (par->labelwidthstring)
	    fprintf(file, "\\labelwidthstring %s\n", par->labelwidthstring);
	 
	 /* some lines? */ 
	 if (par->line_top)
	    fprintf(file, "\\line_top ");
	 if (par->line_bottom)
	    fprintf(file, "\\line_bottom ");
	 
	 /* some pagebreaks? */ 
	 if (par->pagebreak_top)
	    fprintf(file, "\\pagebreak_top ");
	 if (par->pagebreak_bottom)
	    fprintf(file, "\\pagebreak_bottom ");
	 
	 /* noindent ? */ 
	 if (par->noindent)
	    fprintf(file, "\\noindent ");
	 
	 /* some fills ? */ 
	 if (par->fill_top)
	    fprintf(file, "\\fill_top ");
	 if (par->fill_bottom)
	    fprintf(file, "\\fill_bottom ");
	 
	 /* what about the alignment */ 
	 if (par->align != LYX_ALIGN_LAYOUT) {
	    switch (par->align) {
	     case LYX_ALIGN_LEFT: h = 1; break;
	     case LYX_ALIGN_RIGHT: h = 2; break;
	     case LYX_ALIGN_CENTER: h = 3; break;
	     default: h = 0; break;
	    }
	    fprintf(file, "\\align %s ", string_align[h]);
	 }
      }
      else {
	 /* dummy layout, that means a footnote ended */ 
	 fprintf(file, "\n\\end_float ");
	 footnoteflag = LYX_NO_FOOTNOTE;
      }
      
      /* ok, and now the contents */
      font1 = GetLyXFont(LYX_LAYOUT_DEFAULT,
			 LYX_LAYOUT_DEFAULT,
			 LYX_LAYOUT_DEFAULT,
			 LYX_LAYOUT_DEFAULT,
			 LYX_LAYOUT_DEFAULT,
			 LYX_LAYOUT_DEFAULT);
      
      if (par == p.cursor.par)
      	 fprintf(file, "\\cursor %d ", p.cursor.pos);
      
      for (i = 0; i < par->last; i++) {
	 if (!i)
	    fprintf(file, "\n");
	 font2 = par->GetFontSettings(i);
	 if (font1.family != font2.family) {
	    if (font2.family == LYX_LAYOUT_DEFAULT)
	       fprintf(file, "\\family default ");
	    else
	       fprintf(file, "\\family %s ", string_family[font2.family]);
	 }
	 if (font1.series != font2.series) {
	    if (font2.series == LYX_LAYOUT_DEFAULT)
	       fprintf(file, "\\series default ");
	    else
	       fprintf(file, "\\series %s ", string_series[font2.series]);
	 }
	 if (font1.shape != font2.shape) {
	    if (font2.shape == LYX_LAYOUT_DEFAULT)
	       fprintf(file, "\\shape default ");
	    else
	       fprintf(file, "\\shape %s ", string_shape[font2.shape]);
	 }
	 if (font1.size != font2.size) {
	    if (font2.size == LYX_LAYOUT_DEFAULT)
	       fprintf(file, "\\size default ");
	    else
	       fprintf(file, "\\size %s ", string_size[font2.size]);
	 }
	 if (font1.latex != font2.latex) {
	    if (font2.latex == LYX_LAYOUT_DEFAULT)
	       fprintf(file, "\\latex default ");
	    else
	       fprintf(file, "\\latex %s ", string_latex[font2.latex]);
	 }
      
	 if (font1.bar != font2.bar) {
	    if (font2.bar == LYX_LAYOUT_DEFAULT)
	       fprintf(file, "\\bar default ");
	    else
	       fprintf(file, "\\bar %s ", string_bar[font2.bar]);
	 }
      
	 
	 
	 c = par->GetChar(i);
	 
	 switch (c) {
	  case LYX_META_INSET:
	    inset = par->GetInset(i);
	    if (inset) {
	       fprintf(file, "\\begin_inset ");
	       inset->Write(file);
	       fprintf(file, "\\end_inset ");
	    }
	    break;
	  case LYX_META_NEWLINE: 
	    fprintf(file, "\\newline ");
	    break;
	  case LYX_META_HFILL: 
	    fprintf(file, "\\hfill ");
	    break;
	  case LYX_META_PROTECTED_SEPARATOR: 
	    fprintf(file, "\\protected_separator ");
	    break;
	  case '\\': 
	    fprintf(file, "\\backslash ");
	    break;
	  default:
	    fprintf(file, "%c", c);
	    break;
	 }
	 
	 font1 = font2;
      }
      
      par = par->next;
   }
   
   fprintf(file, "\n");
   fclose(file);
}

	       

static char SimpleTeXOnePar(FILE *file, LyXFont &font1, LyXParagraph *par) {
   LyXFont font2;
   char c;
   Inset *inset;
   int main_body;
   int column;
   
   char return_value = 0;
   
   char underbar = 0;

   IncTexRow(par,0);

   /* maybe we have to create a optional argument */ 
   if (style(textclass, par->GetLayout())->labeltype != LABEL_MANUAL)
     main_body = 0;
   else
     main_body = par->BeginningOfMainBody();
   
   if (main_body > 0) {
      fprintf(file, "[");
      font1 = style(textclass, par->GetLayout())->labelfont;
   }
   else
      font1 = style(textclass, par->GetLayout())->font;

   
   column = 0;
   
   if (main_body >= 0 && !par->last && par->layout != LYX_DUMMY_LAYOUT){
     
     if (style(textclass, par->GetLayout())->latextype == LATEX_COMMAND)
       fprintf(file,"{");
     else if (par->GetFontSettings(par->Last()-1).size != LYX_LAYOUT_DEFAULT
	      ||
	      par->align != LYX_ALIGN_LAYOUT)
       {
	 fprintf(file, "{");
	 return_value = 1;
       }
     
   }
   
   for (int i = 0; i < par->last; i++) {
     column++;
      if (i == main_body && par->layout != LYX_DUMMY_LAYOUT) {
	 if (main_body > 0) {
	    if (font1.latex == LYX_FORMULA_LATEX)
	       fprintf(file, "\\)");
	    fprintf(file, "]");
	    font1 = style(textclass, par->GetLayout())->font;
	 }
	 if (style(textclass, par->GetLayout())->latextype == LATEX_COMMAND)
	   fprintf(file,"{");
	 else if (par->GetFontSettings(par->Last()-1).size != LYX_LAYOUT_DEFAULT
		  ||
		  par->align != LYX_ALIGN_LAYOUT)
	   {
	     fprintf(file, "{");
	     return_value = 1;
	   }

	 if (par->noindent)
	   fprintf(file, "\\noindent ");
	 switch (par->align) {
	  case LYX_ALIGN_LEFT:
	    fprintf(file, "\\raggedright ");
	    column+=13;
	    break;
	  case LYX_ALIGN_RIGHT:
	    fprintf(file, "\\raggedleft ");
	    column+=13;
	    break;
	  case LYX_ALIGN_CENTER:
	    fprintf(file, "\\centering ");
	    column+=13;
	    break;
	 }	 
	 
      }


      c = par->GetChar(i);
      
      /* blanks are printed before any fontswitch. With this
	 the brackets after a fontswitch can be removed */

      if (c==' '){
	/* do not print the separation of the optional argument */ 
	if (i != main_body - 1) {
	  if (column > tex_code_break_column
	      && i && par->GetChar(i-1) != ' ') {
	    column = 0;
	    fprintf(file, "\n");
	    IncTexRow(par, i+1);
	    column = 0;
	  }
	  else { 
	    fprintf(file, " ");
	  }
	}
      }


      font2 = GetFont(par, i);      
      
      if (font1.family != font2.family) {
	 switch (font2.family) {
	  case LYX_ROMAN_FAMILY:
	    fprintf(file, "\\rmfamily ");
	    column+=13;
	    break;
	  case LYX_SANS_FAMILY:
	    fprintf(file, "\\sffamily ");
	    column+=13;
	    break;
	  case LYX_TYPEWRITER_FAMILY:
	    fprintf(file, "\\ttfamily ");
	    column+=13;
	    break;
	 }
      }
      if (font1.series != font2.series) {
	 switch (font2.series) {
	  case LYX_MEDIUM_SERIES:
	    fprintf(file, "\\mdseries ");
	    column+=13;
	    break;
	  case LYX_BOLD_SERIES:
	    fprintf(file, "\\bfseries ");
	    column+=13;
	    break;
	 }
      }
      if (font1.shape != font2.shape) {
	 switch (font2.shape) {
	   // case LYX_UP_SHAPE:
	   //  if (font1.shape == LYX_ITALIC_SHAPE)
	   //    fprintf(file, "\\/");     /* italic correction  */
	   // fprintf(file, "\\upshape{}");
	   // column+=13;
	   // break;
	   // case LYX_ITALIC_SHAPE:
	   // fprintf(file, "\\itshape{}");
	   // column+=13;
	   //  break;
	 case LYX_ITALIC_SHAPE:
	   fprintf(file, "\\em ");
	   column+=4;
	   break;
	 case LYX_UP_SHAPE:
	   if (font1.shape == LYX_ITALIC_SHAPE){
	     fprintf(file, "\\em ");
	     column+=4;
	   }
	   else {
	     fprintf(file, "\\upshape ");
	     column+=10;
	   }
	   break;
	  case LYX_SMALLCAPS_SHAPE:
	    if (font1.shape == LYX_ITALIC_SHAPE){
	      fprintf(file, "\\em ");
	      column+=4;
	    // fprintf(file, "\\/");     /* italic correction  */
	    }
	    fprintf(file, "\\scshape ");
	    column+=13;
	    break;
	 }
      }
      if (font1.size != font2.size) {
	 switch (font2.size) {
	  case LYX_SIZE_SMALL:
	    fprintf(file, "\\small ");
	    column+=8;
	    break;
	  case LYX_SIZE_NORMAL:
	    fprintf(file, "\\normalsize ");
	    column+=13;
	    break;
	  case LYX_SIZE_LARGE:
	    fprintf(file, "\\large ");
	    column+=8;
	    break;
	  case LYX_SIZE_LARGER:
	    fprintf(file, "\\Large ");
	    column+=8;
	    break;
	  case LYX_SIZE_LARGEST:
	    fprintf(file, "\\LARGE ");
	    column+=8;
	    break;
	  case LYX_SIZE_HUGE:
	    fprintf(file, "\\huge ");
	    column+=8;
	    break;
	  case LYX_SIZE_GIANT:
	    fprintf(file, "\\Huge ");
	    column+=8;
	    break;
	 }
      }

      if (font1.latex != font2.latex) {
	if (font2.latex != LYX_NO_LATEX &&
	    style(textclass, par->GetLayout())->latextype == LATEX_COMMAND) {
	  fprintf(file, "\\protect ");
	  column+=8;
	}
	switch (font2.latex) {
	case LYX_FORMULA_LATEX:
	  if (style(textclass, par->GetLayout())->font.latex != LYX_FORMULA_LATEX) {
	       fprintf(file, "\\(");
	    }
	    break;
	  default: 
	    if (font1.latex == LYX_FORMULA_LATEX) {
	       if (style(textclass, par->GetLayout())->font.latex != LYX_FORMULA_LATEX)
	       	  fprintf(file, "\\)");
	    }
	    break;
	 }
      }
      
      


      
      if (c == LYX_META_INSET) {
	inset = par->GetInset(i);
	if (inset)
	  inset->Latex(file);	     
      }
      else {
	 
	 if (font2.latex != LYX_NO_LATEX) {
	   switch (c) {
	   case LYX_META_NEWLINE: 
	     if (font2.latex == LYX_LATEX
		   || !style(textclass, par->GetLayout())->newline_allowed)
	       fprintf(file, "\n");
	     else
	       fprintf(file, "\\\\\n");
	     IncTexRow(par, i+1);
	     column = 0;
	     break;  
	   case LYX_META_HFILL: 
	     fprintf(file, "\\hfill{}");
	     column+=8;
	     break;
	     case LYX_META_PROTECTED_SEPARATOR: 
	       fprintf(file, "~");
	       break;
	   default:
	       fprintf(file, "%c", c);
	       break;
	    }
	 }
	 else {
	   if (font2.bar == LYX_UNDER_BAR
	       && c != ' ' 
	       && c != LYX_META_NEWLINE
	       && c != LYX_META_HFILL){
	     fprintf(file, "\\underbar{");
	     column += 11;
	     underbar = 1;
	   }
	   else 
	     underbar = 0;

	   switch (c) {
	   case LYX_META_NEWLINE: 
	     if (!style(textclass, par->GetLayout())->newline_allowed)
	       fprintf(file, "\n");
	     else
	       fprintf(file, "\\\\\n");
	     IncTexRow(par, i+1);
	     column = 0;
	     break;
	   case LYX_META_HFILL: 
	     fprintf(file, "\\hfill{}");
	     column+=8;
	     break;
	   case LYX_META_PROTECTED_SEPARATOR: 
	       fprintf(file, "~");
	       break;
	     case '\\': 
	       if (style(textclass, par->GetLayout())->latextype == LATEX_COMMAND) {
		  fprintf(file, "\\protect\\(\\backslash\\protect\\)");
		  column+=32;
	       }
	       else {
		  fprintf(file, "\\(\\backslash\\)");
		  column+=13;
	       }
	       break;
	     case '<': 
	       if (style(textclass, par->GetLayout())->latextype == LATEX_COMMAND) {
		  fprintf(file, "\\protect\\(<\\protect\\)");
		  column+=26;
	       }
	       else {
		  fprintf(file, "\\(<\\)");
		  column+=6;
	       }
	       break;
	     case '>': 
	       if (style(textclass, par->GetLayout())->latextype == LATEX_COMMAND) {
		  fprintf(file, "\\protect\\(>\\protect\\)");
		  column+=26;
	       }
	       else {
		  fprintf(file, "\\(>\\)");
		  column+=6;
	       }
	       break;
	     case '\"': 
	       fprintf(file, "\\lyxquote{}");
	       break;
	     case '$': 
	       fprintf(file, "\\$");
	       break;
	     case '&': 
	       fprintf(file, "\\&");
	       break;
	     case '%': 
	       fprintf(file, "\\%%");
	       break;
	     case '#': 
	       fprintf(file, "\\#");
	       break;
	     case '{': 
	       fprintf(file, "\\{");
	       break;
	     case '}':
	       fprintf(file, "\\}");
	       break;
	     case '_': 
	       fprintf(file, "\\_");
	       break;
	     case '^': 
	       fprintf(file, "\\^{}");
	       break;
	     case '~': 
	       fprintf(file, "\\~{}");
	       break;
	     case ' ':
	       /* blanks are printed bevor font switching */
	       break;
	     default:
	       fprintf(file, "%c", c);
	       break;
	   }
	   if (underbar){
	     fprintf(file, "}");
	     column += 1;
	     underbar = 0;
	   }
	 }
      } 
      
      font1 = font2;
   }
   
   /* needed if there is an optional argument but no contents */ 
   if (main_body > 0 && main_body == par->last) {
      if (font1.latex == LYX_FORMULA_LATEX)
      	 fprintf(file, "\\)");
      fprintf(file, "]~");
      font1 = style(textclass, par->GetLayout())->font;
      return_value = 0;
   }
   
   if (font1.latex == LYX_FORMULA_LATEX
       && style(textclass, par->GetLayout())->font.latex != LYX_FORMULA_LATEX) {
      fprintf(file, "\\)");
      font1.latex = LYX_NO_LATEX;
   }


   /* restore the font if necessary */
   /* it is not necessary if the return_value is >0, since then
      the paragraph will be put into brackets.
      This is for example the case, if there are different font sizes
      at the end or another alignment */
   if (!return_value) {
     font2 = style(textclass, par->GetLayout())->font;
      
     if (font1.family != font2.family) {
       switch (font2.family) {
       case LYX_ROMAN_FAMILY:
	 fprintf(file, "\\rmfamily ");
	 column+=13;
	 break;
       case LYX_SANS_FAMILY:
	 fprintf(file, "\\sffamily ");
	 column+=13;
	 break;
       case LYX_TYPEWRITER_FAMILY:
	 fprintf(file, "\\ttfamily ");
	 column+=13;
	 break;
       }
     }
     if (font1.series != font2.series) {
       switch (font2.series) {
       case LYX_MEDIUM_SERIES:
	 fprintf(file, "\\mdseries ");
	 column+=13;
	 break;
       case LYX_BOLD_SERIES:
	 fprintf(file, "\\bfseries ");
	 column+=13;
	 break;
       }
     }
     if (font1.shape != font2.shape) {
       switch (font2.shape) {
	 // case LYX_UP_SHAPE:
	 //  if (font1.shape == LYX_ITALIC_SHAPE)
	 //    fprintf(file, "\\/");     /* italic correction  */
	 // fprintf(file, "\\upshape{}");
	 // column+=13;
	 // break;
	 // case LYX_ITALIC_SHAPE:
	 // fprintf(file, "\\itshape{}");
	 // column+=13;
	 //  break;
       case LYX_ITALIC_SHAPE:
	 fprintf(file, "\\em ");
	 column+=4;
	 break;
       case LYX_UP_SHAPE:
	 if (font1.shape == LYX_ITALIC_SHAPE){
	   fprintf(file, "\\em ");
	   column+=4;
	 }
	 else {
	   fprintf(file, "\\upshape ");
	   column+=10;
	 }
	 break;
       case LYX_SMALLCAPS_SHAPE:
	 if (font1.shape == LYX_ITALIC_SHAPE){
	   fprintf(file, "\\em ");
	   column+=4;
	   // fprintf(file, "\\/");     /* italic correction  */
	 }
	 fprintf(file, "\\scshape ");
	 column+=13;
	 break;
       }
      }
   }
   return return_value;
}


static void TeXOnePar(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar) {
   LyXFont font;
   
   LyXParagraph *thispar = par;
   LyXParagraph *tmppar;
   char further_blank_line = 0;
   if (par->layout == LYX_DUMMY_LAYOUT)
     fprintf(stderr, "Error in TeXOnePar  \n");

   if (tex_code_break_column &&
       style(textclass, par->layout)->latextype == LATEX_COMMAND){
     fprintf(file, "\n");
     IncTexRow(par, 0);
   }

   if (thispar->pagebreak_top) {
     fprintf(file, "\\newpage");
     further_blank_line = 1;
   }

   if (thispar->fill_top) {
     fprintf(file, "\\vspace*{\\fill}");
     further_blank_line = 1;
   }
      
   if (thispar->added_space_top) {
     fprintf(file, "\\vspace{%.2fcm}", thispar->added_space_top);
     further_blank_line = 1;
   }
      
   if (thispar->line_top) {
     switch (GetFont(thispar, 0).size) {
     case LYX_SIZE_SMALL:
       fprintf(file, "\\lyxline{\\small}");
       break;
     case LYX_SIZE_NORMAL:
       fprintf(file, "\\lyxline{\\normalsize}");
       break;
     case LYX_SIZE_LARGE:
       fprintf(file, "\\lyxline{\\large}");
       break;
     case LYX_SIZE_LARGER:
       fprintf(file, "\\lyxline{\\Large}");
       break;
     case LYX_SIZE_LARGEST:
       fprintf(file, "\\lyxline{\\LARGE}");
       break;
     case LYX_SIZE_HUGE:
       fprintf(file, "\\lyxline{\\huge}");
       break;
     case LYX_SIZE_GIANT:
       fprintf(file, "\\lyxline{\\Huge}");
       break;
      }
     fprintf(file, "\\vspace{-1\\parskip}");
     further_blank_line = 1;
   }
   if (further_blank_line){
     fprintf(file, "\n");
     IncTexRow(thispar, 0);
   }

   switch (style(textclass, par->layout)->latextype) {
   case LATEX_COMMAND:
     fprintf(file, "\\%s", style(textclass, par->layout)->latexname);
     break;
   case LATEX_ITEM_ENVIRONEMENT:
     fprintf(file, "\\item ");
     break;
   case LATEX_LIST_ENVIRONEMENT:
     fprintf(file, "\\item ");
     break;
   }
   
   font = style(textclass, par->layout)->font;
   
   char need_par = SimpleTeXOnePar(file, font, par);
   tmppar = par;
   prevpar = par;
   par = par->next;
   
   while (par && par->footnoteflag != LYX_NO_FOOTNOTE
	  && par->footnoteflag != thispar->footnoteflag) {
	     TeXFootnote(file, par, prevpar);
	     SimpleTeXOnePar(file, font, par);
	     prevpar = par;
	     par = par->next;
	  }
   
   
   //if (style(textclass, thispar->GetLayout())->latextype != LATEX_COMMAND
   //    && !(thispar->footnoteflag != LYX_NO_FOOTNOTE
   //&& par && par->footnoteflag == LYX_NO_FOOTNOTE)
   //) {
   
     //if (style(textclass, thispar->GetLayout())->latextype != LATEX_PARAGRAPH
     //||
     //thispar->GetFontSettings(thispar->Last()-1).size != LYX_LAYOUT_DEFAULT
     //||
     //thispar->align != LYX_ALIGN_LAYOUT
     //||
     //par && style(textclass, par->GetLayout())->latextype == LATEX_PARAGRAPH)
     //{
     //fprintf(file, "\\par");
     //}

   if (need_par)
     fprintf(file, "\\par}");
   else if (style(textclass, thispar->GetLayout())->latextype == LATEX_COMMAND)
     fprintf(file, "}");

   
   fprintf(file, "\n");
   further_blank_line = 0;
   if (thispar->line_bottom) {
      switch (GetFont(tmppar, tmppar->Last() - 1).size) {
       case LYX_SIZE_SMALL:
	 fprintf(file, "\\lyxline{\\small}");
	 break;
       case LYX_SIZE_NORMAL:
	 fprintf(file, "\\lyxline{\\normalsize}");
	 break;
       case LYX_SIZE_LARGE:
	 fprintf(file, "\\lyxline{\\large}");
	 break;
       case LYX_SIZE_LARGER:
	 fprintf(file, "\\lyxline{\\Large}");
	 break;
       case LYX_SIZE_LARGEST:
	 fprintf(file, "\\lyxline{\\LARGE}");
	 break;
       case LYX_SIZE_HUGE:
	 fprintf(file, "\\lyxline{\\huge}");
	 break;
       case LYX_SIZE_GIANT:
	 fprintf(file, "\\lyxline{\\Huge}");
	 break;
      }
      further_blank_line = 1;
   }

   if (thispar->added_space_bottom) {
     fprintf(file, "\\vspace{%.2fcm}", thispar->added_space_bottom);
      further_blank_line = 1;
   }
      
   if (thispar->fill_bottom) {
     fprintf(file, "\\vspace*{\\fill}");
      further_blank_line = 1;
   }

   if (thispar->pagebreak_bottom) {
     fprintf(file, "\\newpage");
      further_blank_line = 1;
   }

   if (further_blank_line){
     fprintf(file, "\n");
     IncTexRow(thispar, thispar->Last());
   }
   
   if (thispar->footnoteflag != LYX_NO_FOOTNOTE
       && par && par->footnoteflag == LYX_NO_FOOTNOTE) {
   }
   else {
     fprintf(file, "\n");
     IncTexRow(thispar, thispar->Last());
   }
}

static void TeXDeeper(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar) {
   LyXParagraph *thispar = par;
   prevpar = par;
   while (par && par->depth == thispar->depth) {
      if (par->layout == LYX_DUMMY_LAYOUT)
      	 fprintf(stderr, "Error in TeXDeeper \n");
      if (style(textclass, par->layout)->latextype == LATEX_ENVIRONEMENT
	  || style(textclass, par->layout)->latextype == LATEX_ITEM_ENVIRONEMENT
	  || style(textclass, par->layout)->latextype == LATEX_LIST_ENVIRONEMENT
	  )
      	TeXEnvironement(file, par, prevpar);
      else
      	TeXOnePar(file, par, prevpar);
      
   }
}

static void TeXEnvironement(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar) {
   LyXParagraph *thispar = par;
   prevpar = par;
   if (par->layout == LYX_DUMMY_LAYOUT)
     fprintf(stderr, "Error in TeXEnvironement \n");
   
   if (style(textclass, par->layout)->latextype == LATEX_LIST_ENVIRONEMENT)
     fprintf(file, "\\begin{%s}{%s}\n", 
	     style(textclass, par->layout)->latexname,
	     par->labelwidthstring
	     );
   else
     fprintf(file, "\\begin{%s}\n", style(textclass, par->layout)->latexname);
   IncTexRow(par, 0);
   if (tex_code_break_column){
     fprintf(file, "\n");
     IncTexRow(par, 0);
   }

   do {
      TeXOnePar(file, par, prevpar);
      if (par && par->depth > thispar->depth)
      	TeXDeeper(file, par, prevpar);    
   } while (par && par->layout == thispar->layout
	    && par->depth == thispar->depth); 
   
   fprintf(file, "\\end{%s}\n", style(textclass, thispar->layout)->latexname);
   IncTexRow(prevpar, prevpar->Last());
   if (tex_code_break_column){
     fprintf(file, "\n");
     IncTexRow(prevpar, prevpar->Last());
   }
}

static void TeXFootnote(FILE* file, LyXParagraph* &par, LyXParagraph* &prevpar) {
   LyXParagraph *thispar = par;
   prevpar = par;
   if (par->footnoteflag == LYX_NO_FOOTNOTE)
     fprintf(stderr, "Error in TeXFootnote \n");
   
   if (style(textclass, thispar->previous->GetLayout())->latextype == LATEX_COMMAND) {
      	  fprintf(stderr, "LyX Warning: Footnoteenvironement in command will be ignored \n");
	  while (par->footnoteflag != LYX_NO_FOOTNOTE) {
	     par = par->next;
	  }
	  return;
       }
   
   switch (thispar->footnotekind) {
    case LYX_FOOTNOTE:
      fprintf(file, "\\footnote{\n");
      break;
    case LYX_MARGIN:
      fprintf(file, "\\marginpar{\n");
      break;
    case LYX_FIG:
      fprintf(file, "\\begin{figure}\n");
      break;
    case LYX_TAB:
      fprintf(file, "\\begin{table}\n");
      break;
   }
   
   do {
   if (par->layout == LYX_DUMMY_LAYOUT)
      	 fprintf(stderr, "Error in TeXFootnote \n");
      if (style(textclass, par->layout)->latextype == LATEX_ENVIRONEMENT
	  || style(textclass, par->layout)->latextype == LATEX_ITEM_ENVIRONEMENT
	  || style(textclass, par->layout)->latextype == LATEX_LIST_ENVIRONEMENT
	  )
      	 TeXEnvironement(file, par, prevpar);
      else
      	 TeXOnePar(file, par, prevpar);
      if (par && par->depth > thispar->depth)
      	 TeXDeeper(file, par, prevpar);
   } while (par && par->footnoteflag != LYX_NO_FOOTNOTE);
   
   switch (thispar->footnotekind) {
    case LYX_FOOTNOTE:
      fprintf(file, "}");
      break;
    case LYX_MARGIN:
      fprintf(file, "}");
      break;
    case LYX_FIG:
      fprintf(file, "\\end{figure}");
      break;
    case LYX_TAB:
      fprintf(file, "\\end{table}");
      break;
   }
}

void MakeLaTeXFile(const char* filename, LyXTextParameters p, int column){
   int i;
   LyXParagraph *par = p.paragraph;
   LyXParagraph *prevpar = par;
   textclass = p.textclass;
   tex_code_break_column = column; 


   FILE *file = fopen(filename, "w");

   if (!file) {
      WriteAlert("LYX_ERROR:","Cannot write file",filename);
      return;
   }
   
   ResetTexRow();

   fprintf(file, "%% This LaTeX-file was created by <%s> %s",
	   getenv("USER"),date());
   fprintf(file, "%% LyX 0.8 (C) 1995 by Matthias Ettrich\n");

   IncTexRow(par, 0);
   IncTexRow(par, 0);
   fprintf(file, "\\nonstopmode\n");
   IncTexRow(par, 0);

   fprintf(file, "\\documentclass[");
   
   fprintf(file, "%dpt", p.fontsize);
   
   switch (p.papersize) {
   case PAPER_A4PAPER:
      fprintf(file, ",a4paper");
      break;
   case PAPER_USLETTER:
      fprintf(file, ",letterpaper");
      break;
   case PAPER_A5PAPER:
      fprintf(file, ",a5paper");
      break;
   case PAPER_B5PAPER:
      fprintf(file, ",b5paper");
      break;
   case PAPER_EXECUTIVEPAPER:
      fprintf(file, ",executivepaper");
      break;
   case PAPER_LEGALPAPER:
      fprintf(file, ",legalpaper");
      break;
   }
   
   if (p.sides == 2)
      fprintf(file, ",twoside");
   else
      fprintf(file, ",oneside");
   
   if (p.columns == 2)
      fprintf(file, ",twocolumn");
   else
      fprintf(file, ",onecolumn");
   
   if (p.orientation == ORIENTATION_LANDSCAPE)
      fprintf(file, ",landscape");
	   
	   
   /* the user-defined options  */
   char* tmpchar = p.options;
   while (tmpchar[0] == ' ')
      tmpchar++;
   
   if (tmpchar[0])
      fprintf(file, ",%s", p.options);
   
   fprintf(file, "]{%s}\n", NameOfClass(textclass));
   IncTexRow(par, 0);
   
   fprintf(file, "\\usepackage{lyx}\n");
   IncTexRow(par, 0);
   if (!StringEqual(p.fontenc, "default")) {
      fprintf(file, "\\usepackage[%s]{fontenc}\n", p.fontenc);;
      IncTexRow(par, 0);
   }
   if (!StringEqual(p.inputenc, "default")) {
      fprintf(file, "\\usepackage[%s]{inputenc}\n", p.inputenc);;
      IncTexRow(par, 0);
   }
   if (!StringEqual(p.language, "default")) {
      fprintf(file, "\\usepackage[%s]{babel}\n", p.language);
      IncTexRow(par, 0);
   }
   if (!StringEqual(p.fonts, "default")) {
      fprintf(file, "\\usepackage{%s}\n", p.fonts);;
      IncTexRow(par, 0);
   }
   if (!StringEqual(p.epsfig, "none")) {
      fprintf(file, "\\usepackage[%s]{epsfig}\n", p.epsfig);;
      IncTexRow(par, 0);
   }
   
   /* at the very beginning the text parameters */
   switch (p.papersize) {
   case PAPER_A4:
      fprintf(file, "\\usepackage{a4}\n");
      IncTexRow(par, 0);
      break;
   case PAPER_A4WIDE:
      fprintf(file, "\\usepackage{a4wide}\n");
      IncTexRow(par, 0);
      break;
   case PAPER_WIDEMARGINSA4:
      fprintf(file, "\\usepackage[widemargins]{a4}\n");
      IncTexRow(par, 0);
      break;
   }
   
   switch (p.pagestyle) {
   case PAGESTYLE_EMPTY:
      fprintf(file, "\\pagestyle{empty}\n");
      IncTexRow(par, 0);
      break;
   case PAGESTYLE_PLAIN:
      fprintf(file, "\\pagestyle{plain}\n");
      IncTexRow(par, 0);
      break;
   case PAGESTYLE_HEADINGS:
      fprintf(file, "\\pagestyle{headings}\n");
      IncTexRow(par, 0);
      break;
   }
   
   fprintf(file, "\\setcounter{secnumdepth}{%d}\n", p.secnumdepth);
   IncTexRow(par, 0);
   fprintf(file, "\\setcounter{tocdepth}{%d}\n", p.tocdepth);
   IncTexRow(par, 0);
   
   
   if (p.paragraph_separation) {
      fprintf(file, "\\lyxletterstyle\n");
      IncTexRow(par, 0);
   }
   
   if (p.baseline >= 1.01) {
      fprintf(file, "\\renewcommand{\\baselinestretch}{%.2f}\n", 
              p.baseline);
      IncTexRow(par, 0);
   }
   
   /* the user-defined praeamble */ 
   for (i = 0; p.preamble[i]; i++) {
      fprintf(file, "%c", p.preamble[i]);
      if (p.preamble[i]=='\n')
	IncTexRow(par, 0);
   }

   if (p.preamble[0]) {
      fprintf(file, "\n");
      IncTexRow(par, 0);
   }
      
   fprintf(file, "\\begin{document}\n\n");
   IncTexRow(par, 0);
   IncTexRow(par, 0);
   
   while (par) {
      if (par->layout == LYX_DUMMY_LAYOUT)
         fprintf(stderr, "Error in MakeLateXFile \n");
      if (style(textclass, par->layout)->latextype == LATEX_ENVIRONEMENT
	  || style(textclass, par->layout)->latextype 
             == LATEX_ITEM_ENVIRONEMENT
          || style(textclass, par->layout)->latextype 
             == LATEX_LIST_ENVIRONEMENT)
         TeXEnvironement(file, par, prevpar);
      else
      	 TeXOnePar(file, par, prevpar);
      
   }
   
   fprintf(file, "\\end{document}\n");
   IncTexRow(prevpar, prevpar->Last());
   fclose(file);
}

void ReadSimpleWholeFile(LyXParagraph *par) {
  char c;
   long file_size = 0;
   while (!feof(myfile)) {
      c = fgetc(myfile);
      file_size++;
   }
   /* it is horrible, I know, but faster.
    * I should not use InsertString for that :-( */
   
   /* I will write a better insertion in the future */ 
   par->Enlarge(0, file_size + 10);
   
   rewind(myfile);
   
   if (!feof(myfile)) {
      do {
	 c = fgetc(myfile);
	 par->InsertChar(par->last,c);
      } while (!feof(myfile));
      
   }

}


static void readline(FILE *file, char* s) {
   int i=0;
   do  {
      s[i] = fgetc(file);
      i++;
   } while (!feof(file) && s[i-1]!='\n');
   s[i] = '\0';
}

/* scans the latex-logfile and inserts error insets */
// Lgb: ScanLogFile now returns all status information
//      in the TeXErrors struct errors (pointer.)
// int ScanLogFile(const char* filename) {
void ScanLogFile(const char* filename, TeXErrors* errors) {
   
   FILE *file = fopen(filename, "r");

   if (!file) {
      fprintf(stderr, "LYX_ERROR: Cannot read file `%s' \n", filename);
      errors->status = NO_LOGFILE;
      return;
   }
   
   TexRow *tmp, *tmp2;
   
   InsetError *new_inset; 
   int i=0;
   int i2;
   int number_of_tex_rows;
   int row;
   errors->number_of_errors = 0;
   // int number_of_errors = 0;
   

   tmp = texrow;
   while (tmp) {
      i++;
      tmp=tmp->next;
   }

   
   number_of_tex_rows = i;
   
   char *s1 = new char[400];
   char *s2 = new char[200];

   
   while (!feof(file)) {
      do {
	 readline(file, s1);
      }while (!feof(file) && (s1[0]!='!' || s1[1]!=' '));
      if (s1[0]=='!' && s1[1]==' ') {
	 do {
	    readline(file, s2);
	 }while (!feof(file) && (s2[0]!='l' || s2[1]!='.'));
	 
	 if (s2[0]=='l' && s2[1]=='.') {
	    sscanf(s2, "l.%d", &row);
	    for (i=0; s1[i]; i++);
	    for (i2=0; s2[i2]; i2++) {
	       s1[i+i2]=s2[i2];
	    }
	    s1[i+i2]='\0';
	    
	    /* insert errorstring for row number row */ 
	    tmp = texrow;
	    i = number_of_tex_rows;
	    while (tmp && tmp->next && i > row) {
	      tmp = tmp->next;
	      i--;
	    }
	    
	      
	    new_inset = new InsetError(s1);
	    tmp->par->InsertChar(tmp->pos, LYX_META_INSET);
	    tmp->par->InsertInset(tmp->pos, new_inset);
	    errors->number_of_errors++;
	    
	    /* may be necessary */
	    tmp2 = texrow;
	    while (tmp2) {
	       if (tmp2 != tmp && tmp2->par == tmp->par && tmp2->pos >= tmp->pos)
	       	 tmp2->pos++;
	       tmp2 = tmp2->next;
	    }
	 }
      }
   }

   fclose(file);
   delete(s1);
   delete(s2);
   // what about returning a struct in stead?
   errors->status = RUN_OK;
   return;
   
   //return number_of_errors;
   
}
   



/*
 * LinuxDoc SGML support functions.
 *
 * version 0.5
 * written by Pascal Andre (andre@via.ecp.fr)
 *
 */

/* constants used for LinuxDoc SGML handling */
enum _linuxdoc_tags {
  LINUXDOC_TITLE, LINUXDOC_AUTHOR, LINUXDOC_DATE, LINUXDOC_ABSTRACT,
  LINUXDOC_SECT, LINUXDOC_SECT1, LINUXDOC_SECT2, LINUXDOC_SECT3, 
  LINUXDOC_SECT4, LINUXDOC_PAR, LINUXDOC_CODE, LINUXDOC_VERB, LINUXDOC_ITEM, 
  LINUXDOC_ENUM, LINUXDOC_DESCR, LINUXDOC_NONE 
};

const char *linuxdoc_styles[LINUXDOC_NONE] = {
  "Title", "Author", "Date", "Abstract", "Section", "Subsection", 
  "Subsubsection", "Paragraph", "Subparagraph", "Standard", "Code", 
  "Verbatim", "Itemize", "Enumerate", "Description"
};

const char *linuxdoc_tags[LINUXDOC_NONE] = {
  "title", "author", "date", "abstract", "sect", "sect1", "sect2", "sect3",
  "sect4", "p", "code", "verb", "itemize", "enumerate", "descrip"
};

enum _special_tags {
  LDS_NONE, LDS_QUOTE, LDS_TSCREEN, LDS_LAST
};

const char *special_tags[LDS_LAST] = {
  "", "\n<quote>", "\n<tscreen>"
};

const char *special_close[LDS_LAST] = {
  "" ,"\n</quote>", "\n<tscreen>"
};

/* print an error message */
static void LinuxDocError(LyXParagraph *par, int pos, char *message) 
{
  InsetError *new_inset;

  /* insert an error marker in text */
  new_inset = new InsetError(message);
  par->InsertChar(pos, LYX_META_INSET);
  par->InsertInset(pos, new_inset);
}

/* gets tag number from layout name */
static int LinuxDocTag(char *layout) {
  int i;

  for (i=0; i<LINUXDOC_NONE; i++)
    if (StringEqual(linuxdoc_styles[i], layout)) return i;

  return LINUXDOC_NONE;
}

/* push a tag in a style stack */
static void push_tag(FILE *file, char *tag, int& pos, char stack[5][3]) {
  int j;

  /* pop all previous tags */
  for (j=pos; j>=0; j--)
    fprintf(file, "</%s>", stack[j]); 

  /* add new tag */
  sprintf(stack[++pos], "%s", tag);

  /* push all tags */
  for (j=0; j<=pos; j++)
    fprintf(file, "<%s>", stack[j]);
};

/* pop a tag from a style stack */
static void pop_tag(FILE *file, char * tag, int& pos, char stack[5][3]) {
  int j;

  /* pop all tags till specified one */
  for (j=pos; (j>=0) && (strcmp(stack[j], tag)); j--)
    fprintf(file, "</%s>", stack[j]);

  /* closes the tag */
  fprintf(file, "</%s>", tag);

  /* push all tags, but the specified one */
  for (j=j+1; j<=pos; j++) {
    fprintf(file, "<%s>", stack[j]);
    strcpy(stack[j-1], stack[j]);
  };
  pos --;
};

/* gets next latex parameter */
static int GetNextParameter(LyXParagraph *par, int pos, char *buffer, 
			    int size)
{
  int i, flag;
  char c;
  Inset *inset;

  flag = 1;
  bzero(buffer, size);
  for (i=pos+1; i<=par->last && flag; i++) {
    c = par->GetChar(i);
    if (c == LYX_META_INSET) {
      inset = par->GetInset(i);
      switch (inset->LyxCode()) {
      case LYX_QUOTE_CODE:
        c='\"';
        break;
      case LYX_SEPARATOR_CODE:
      case LYX_ENDING_CODE:
	flag = 0;
	break;
      case LYX_TOC_CODE:
	LinuxDocError(par, i, "Error : table of contents not allowed "
		      "in body.\n");
	break;
       case LYX_IGNORE_CODE:
	 /* ignore */ 
	 break;
      default:
	LinuxDocError(par, i, "Error : unhandled inset.\n");
	break;
      }
    }
    if (flag) *buffer=c;
    buffer++;
  }

  if (flag) {
    LinuxDocError(par, pos, "Error : incomplete macro.\n");
    return 0;
  }
  return i;
}

/* handle internal paragraph parsing -- layout already processed */
static void SimpleLinuxDocOnePar(FILE *file, LyXFont &font1, 
                                 LyXParagraph *par, int &verb_on, 
                                 int &desc_on) 
{
  LyXFont font2;
  char c;
  Inset *inset;
  int main_body, j;

  int is_tt = 0;               /* typewriter font flag */
  int is_bf = 0;               /* bold font flag */
  int is_em = 0;               /* emphasis (italic) font flag */
  int is_quote = 0;            /* quote (paragraph indenting) flag */

  int stack_num = -1;          /* style stack position */
  char stack[5][3];    	       /* style stack */

  if (style(textclass, par->GetLayout())->labeltype != LABEL_MANUAL)
    main_body = 0;
  else
    main_body = par->BeginningOfMainBody();
   
  /* gets paragraph main font */
  if (main_body > 0)
    font1 = style(textclass, par->GetLayout())->labelfont;
  else
    font1 = style(textclass, par->GetLayout())->font;

  
  /* parsing main loop */
  for (int i = 0; i < par->last; i++) {

    /* handle quote tag */
    if (i == main_body && par->layout != LYX_DUMMY_LAYOUT) {
      if (main_body > 0)
	font1 = style(textclass, par->GetLayout())->font;
      if ((!verb_on) && (par->align == LYX_ALIGN_RIGHT)) {
	fprintf(file, "<quote>");
	is_quote = 1;
      }
    }

    font2 = GetFont(par, i);      

    /* handle <tt> tag */
    if (font1.family != font2.family) {
      if (verb_on)
	LinuxDocError(par, 0, "Error : TypeWriter font not allow in "
		      "Verbatim.\n");
      else {
	if (font2.family == LYX_TYPEWRITER_FAMILY) {
	  push_tag(file, "tt", stack_num, stack);
	  is_tt = 1;
	}
	else if (is_tt == 1) {
	  pop_tag(file, "tt", stack_num, stack);
	  is_tt = 0;
	}
      }
    }

    /* handle <bf> tag */
    if (font1.series != font2.series) {
      if (verb_on)
	LinuxDocError(par, 0, "Error : Bold font not allow in Verbatim.\n");
      else {
	if (font2.series == LYX_BOLD_SERIES) {
	  push_tag(file, "bf", stack_num, stack);
	  is_bf = 1;
	}
	else if (is_bf == 1) {
	  pop_tag(file, "bf", stack_num, stack);
	  is_bf = 0;
	}
      }
    }

    /* handle <em> tag */
    if ((!verb_on) && (font1.shape != font2.shape)) {
      if (verb_on)
	LinuxDocError(par, 0, "Error : Emphasis (italic font)  not allow in "
		      "Verbatim.\n");
      else {
	if (font2.shape == LYX_ITALIC_SHAPE) {
	  push_tag(file, "em", stack_num, stack);
	  is_em = 1;
	}
	else if (is_em == 1) {
	  pop_tag(file, "em", stack_num, stack);
	  is_em = 0;
	}
      }
    }
      
    c = par->GetChar(i);
      
    if (c == LYX_META_INSET) {
      inset = par->GetInset(i);
      switch (inset->LyxCode()) {
      case LYX_TOC_CODE:
	LinuxDocError(par, i, "Error : table of contents  not allowed "
		      "in body.\n");
	break;
      case LYX_QUOTE_CODE:
        fprintf(file, "\"");
        break;
      case LYX_MARK_CODE:
	{
	  int j;
	  char label[64];

	  if ((j = GetNextParameter(par, i, label, 64))) {
	    i = j;
	    fprintf(file, "<label id=\"%s\">", label);
	  }
	  i++;
	}
	break;
      case LYX_LABEL_CODE:
	   {
	      fprintf(file, "<label id=\"%s\">", ((InsetLabel*) inset)->GetLabel());
	   }
	 break;
      case LYX_REF_CODE:
	{
	  int j;
	  char label[64], name[64];

	  if ((j = GetNextParameter(par, i, label, 64)) && 
	      (j = GetNextParameter(par, j-1, name, 64))) {
	    i = j;
	    fprintf(file, "<ref id=\"%s\" name=\"%s\">", label, name);
	  }
	}
	break;
      case LYX_URL_CODE:
	{
	  int j;
	  char url[64], name[64];

	  if ((j = GetNextParameter(par, i, url, 64)) && 
	      (j = GetNextParameter(par, j-1, name, 64))) {
	    i = j;
	    fprintf(file, "<url url=\"%s\" name=\"%s\">", url, name);
	  }
	}
	break;
      case LYX_HTMLURL_CODE:
	{
	  int j;
	  char url[64], name[64];

	  if ((j = GetNextParameter(par, i, url, 64)) && 
	      (j = GetNextParameter(par, j-1, name, 64))) {
	    i = j;
	    fprintf(file, "<htmlurl url=\"%s\" name=\"%s\">", url, name);
	  }
	}
	break;
       case LYX_IGNORE_CODE:
	 /* ignore */ 
	 break;
       default:
	 LinuxDocError(par, i, "Error : unhandled inset.\n");
	 i++;
	 break;
      }
    }
    else {
      switch (c) {
      case LYX_META_HFILL:
	break;
      case LYX_META_PROTECTED_SEPARATOR: 
	fprintf(file, " "); break;
      case LYX_META_NEWLINE:
	fprintf(file, "\n"); break;
      case '&': 
	fprintf(file, "&amp;"); break;
      case '<': 
	fprintf(file, "&lt;"); break;
      case '>':
	fprintf(file, "&gt;"); break;
      case '$': 
	fprintf(file, "&dollar;"); break;
      case '#': 
	fprintf(file, "&num;"); break;
      case '%': 
	fprintf(file, "&percnt;"); break;
      case '~': 
	fprintf(file, "&tilde;"); break;
      case '\"': 
	/* should work says the doc, but ... fprintf(file, "&dquote;"); */
	fprintf(file, "\"");
	break;
#if 0
      case '.':
	switch(par->GetChar(i+1)) {
	case LYX_META_PROTECTED_SEPARATOR:
	case ' ':
	  if (verb_on) fprintf(file, ".");
	  else fprintf(file, ".\n"); 
	  break;
	default:
	  fprintf(file, ".");
	  break;
	}
	break;
#endif
      case ' ':
	if (desc_on == 1) {
	  fprintf(file, "</tag>");
	  desc_on = 2;
	}
	else fprintf(file, " ");
	break;
      default:
	fprintf(file, "%c", c); break;
      }
	 
    }
      
    font1 = font2;
  }
   
  /* needed if there is an optional argument but no contents */ 
  if (main_body > 0 && main_body == par->last) {
    font1 = style(textclass, par->GetLayout())->font;
  }

  /* pop all defined styles (non verbatim) */
  if (!verb_on) for (j=stack_num; j>=0; j--)
    fprintf(file, "</%s>", stack[j]);

  /* closes quote tag if necessary */
  if (is_quote) fprintf(file, "</quote>");

  /* resets description flag correctly */
  if (desc_on == 2) desc_on = 1;
}

/* handle the layout constraints for a given paragraph */
static void LinuxDocOnePar(FILE* file, LyXParagraph* &par, 
                           LyXParagraph* &prevpar, int is_init) 
{
  LyXFont font;
  LyXParagraph *thispar = par;
  LyXParagraph *tmppar;
  int i = 0;
  // InsetError *new_inset;

  int error;                     /* error flag */
  char error_message[64];        /* error message */
  int tag;                       /* number of the layout tag */
  static int depth;              /* paragraph depth */
  static int enum_flags[10];     /* enum/itemize status */
  static int desc_on;            /* description mode on/off */
  static int verb_on;            /* verbatim mode on/off */
  static int is_quote;           /* quote mode on/off */
  int special;                   /* special tag */

  /* first time initialisation */
  if (is_init) {
    depth = -1;
    enum_flags[0] = LINUXDOC_NONE;
    desc_on = 0;
    is_quote = 0;
  }   

  /* standard initialisation */
  is_quote = 0;
  verb_on = 0;
  special = LDS_NONE;
  if (par->layout == LYX_DUMMY_LAYOUT) {
    error = 1;
    fprintf(stderr, "Error in LinuxDocOnePar  \n");
  }
  par->AutoDeleteInsets();

  error = 0;
  tag = LinuxDocTag(NameOfLayout(textclass, par->GetLayout()));

  if (depth==-1) {
    /* special handling for <toc> tag (\tableofcontents TeX macro) */
    if (par->GetChar(0) == LYX_META_INSET) {
      Inset *inset = par->GetInset(0);
      if (inset->LyxCode() == LYX_TOC_CODE) {
	fprintf(file, "\n<toc>");
	error = 1;                 /* fake : no futher processing required */
      }
    }
    else {
      /* before article body : limited tag set */
      switch(tag) {
      case LINUXDOC_TITLE:
      case LINUXDOC_AUTHOR:
      case LINUXDOC_DATE:
      case LINUXDOC_ABSTRACT:
        fprintf(file, "\n<%s>", linuxdoc_tags[tag]);
        break;
      case LINUXDOC_SECT:
        fprintf(file, "\n<%s>", linuxdoc_tags[tag]);
        depth = 0;
        break;
      default:
        sprintf(error_message, "Style %s not allowed before first Section.\n",
	        linuxdoc_styles[tag]);
        LinuxDocError(par, 0, error_message);
        error = 1;
        break;
      }
    }
  }
  else {
    /* handle enumerate, itemize and quote SGML tags closing */
    switch(enum_flags[depth]) {
    case LINUXDOC_ENUM:
    case LINUXDOC_ITEM:
      if (((tag != enum_flags[depth]) && (tag!=LINUXDOC_VERB) 
	   && (tag!=LINUXDOC_CODE)) || (depth > par->depth)) {
	fprintf(file, "\n</%s>", linuxdoc_tags[enum_flags[depth]]);
	if (depth == 0) enum_flags[depth] = LINUXDOC_NONE;
      }
      break;
    default:
      if ((tag!=LINUXDOC_ITEM) && (tag!=LINUXDOC_ENUM) 
	  && (depth < par->depth)) special = LDS_QUOTE;
      break;
    }

    /* handle descriptions closing */
    if ((desc_on) && (tag != LINUXDOC_DESCR)) {
      fprintf(file, "\n</descrip>\n");
      desc_on = 0;
    }

    /* write opening SGML tags */
    switch(tag) {
    case LINUXDOC_TITLE:
    case LINUXDOC_AUTHOR:
    case LINUXDOC_DATE:
    case LINUXDOC_ABSTRACT:
      sprintf(error_message, "Style %s not allowed after first Section.\n",
	      linuxdoc_styles[tag]);
      LinuxDocError(par ,i, error_message);
      error = 1;
      break;
    case LINUXDOC_PAR:
      if (special == LDS_NONE) fprintf(file, "\n\n");
      else fprintf(file, "%s", special_tags[special]);
      break;
    case LINUXDOC_VERB:
      font = GetFont(par, 0);
      verb_on = 1;
      /* special type handling for tscreen */
      if ((font.family == LYX_TYPEWRITER_FAMILY)
	  && (depth < par->depth)) special = LDS_TSCREEN;
      fprintf(file, "%s\n<verb>", special_tags[special]);
      break;
    case LINUXDOC_CODE:
      fprintf(file, "%s\n<code>", special_tags[special]);
      break;
    case LINUXDOC_SECT:
    case LINUXDOC_SECT1:
    case LINUXDOC_SECT2:
    case LINUXDOC_SECT3:
    case LINUXDOC_SECT4:
      if (depth!=0) {
	for (i=depth; i>=0; i--) {
	  if (enum_flags[depth]!=LINUXDOC_NONE)
	    fprintf(file, "\n</%s>", linuxdoc_tags[enum_flags[depth]]);
	  enum_flags[i] = LINUXDOC_NONE;
	}
      }
      fprintf(file, "%s\n<%s>", special_tags[special], linuxdoc_tags[tag]);
      break;
    case LINUXDOC_ENUM:
    case LINUXDOC_ITEM:
      if (depth < par->depth) {
	depth = par->depth;
	enum_flags[depth] = LINUXDOC_NONE;
      }
      if (enum_flags[depth] != tag) 
	fprintf(file, "\n<%s>", linuxdoc_tags[tag]);
      fprintf(file, "\n<item>");
      depth = par->depth;
      enum_flags[depth] = tag;
      break;
    case LINUXDOC_DESCR:
      if (!desc_on) fprintf(file, "\n<descrip>");
      fprintf(file, "\n<tag>");
      desc_on = 1;
      break;
    default:
      fprintf(file, "\n<%s>", linuxdoc_tags[tag]);
      break;
    }
  }

  if (!error) {
    SimpleLinuxDocOnePar(file, font, par, verb_on, desc_on);
    
    /* write closing SGML tags */
    switch(tag) {
    case LINUXDOC_ENUM:
    case LINUXDOC_ITEM:
    case LINUXDOC_DESCR:
    case LINUXDOC_PAR:
    case LINUXDOC_DATE:
    case LINUXDOC_AUTHOR:
    case LINUXDOC_TITLE:
      break;
    case LINUXDOC_SECT:
    case LINUXDOC_SECT1:
    case LINUXDOC_SECT2:
    case LINUXDOC_SECT3:
    case LINUXDOC_SECT4:
      fprintf(file, "\n<p>");
      break;
    default:
      fprintf(file, "\n</%s>", linuxdoc_tags[tag]);
      break;
    }
    if (special != LDS_NONE) fprintf(file, "%s", special_close[special]);
  }

  if (depth != -1) depth = par->depth;
  tmppar = par;
  prevpar = par;
  par = par->next;


  while (par && par->footnoteflag != LYX_NO_FOOTNOTE
	 && par->footnoteflag != thispar->footnoteflag) {
    /* no footnote support yet */
  }
}

void MakeLinuxDocFile(const char* filename, LyXTextParameters p, int column){
   int is_init = 1;
   LyXParagraph *par = p.paragraph;
   LyXParagraph *prevpar = par;

   FILE *file = fopen(filename, "w");
   textclass = p.textclass;
   tex_code_break_column = column; 

   if (!file) {
      WriteAlert("LYX_ERROR:", "Cannot write file",filename);
      return;
   }
   
   ResetTexRow();

   fprintf(file, "<!-- LinuxDoc file was created by LyX 0.8 (C) 1995 by "
	"Matthias Ettrich -->\n");
   fprintf(file, "<!-- Export filter v0.5 by Pascal Andre               "
        "                 -->\n");
   
   fprintf(file, "\n<!doctype linuxdoc system>\n\n");
   fprintf(file, "<article>\n");   

   while (par) {
      if (par->layout == LYX_DUMMY_LAYOUT)
      	 fprintf(stderr, "Error in MakeLinuxDocFile \n");
      else
      	LinuxDocOnePar(file, par, prevpar, is_init);
      is_init = 0;
   }
   
   fprintf(file, "\n\n</article>\n");
   fclose(file);
}
