/* This file is part of
 * ======================================================
 * 
 *           LyX, the High Level Word Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *
 *======================================================*/

#include "inset.h"
#include <ctype.h>

extern Widget toplevel;
extern char* date();

static char* InsetNextToken(FILE* myfile, char &token_end_character) {
   char* token = NULL;
   char c = 0;
   int i = 0;
   
   if (!feof(myfile)) {
      token = new char[100];
      do {
	 c = fgetc(myfile);
	 token[i]=c;
	 i++;
      } while (!feof(myfile) && !isspace(c));
      
      token[i-1]='\0';	       /* just the end of a command  */
   }
   token_end_character = c;
   return token;
}



/* Insets default methods */

char Inset::Deletable() {
  return 1;
}


char Inset::Editable() {
  return 0;
}

char Inset::AutoDelete() {
  return 0;
}

void Inset::Edit() {
}

LyXFont Inset::ConvertFont(LyXFont font) {
  return font;
}

char Inset::LyxCode() {
  return LYX_NO_CODE;
}

/* Quotes. Used for the various quotes. German, English, French, all either
 * double or single */ 
   
InsetQuotes::InsetQuotes(char* string) {
  kind = new char[4];
  contents = new char[2];
  ParseString(string);
}

InsetQuotes::~InsetQuotes() {
  delete(kind);
  delete(contents);
}

void InsetQuotes::ParseString(char* string) {
  if (string[0])
    kind[0]=string[0];
  else
    kind[0]=' ';
  if (string[1])
    kind[1]=string[1];
  else
    kind[1]=' ';
  if (string[2])
    kind[2]=string[2];
  else
    kind[2]=' ';
  kind[3] = '\0';
   
  switch (kind[0]) {
  case 'g':			       /* german quotes  */
    if (kind[1]=='l')
      contents[0]=',';
    else
      contents[0]='`';
    break;
  case 'f':			       /* french quotes  */
    if (kind[1]=='l')
      contents[0]='<';
    else
      contents[0]='>';
    break;
  default:			       /* english quotes  */
    if (kind[1]=='l')
      contents[0]='`';
    else
      contents[0]='\'';
    break;
  }
   
  /* single or double quotes */ 
  if (kind[2]=='s')
    contents[1]='\0';
  else
    contents[1]=contents[0];
}

int InsetQuotes::Ascent(LyXFont font) {
  if (contents[1])
    return Maximum(LyXAscent(font, contents[0]),
		   LyXAscent(font, contents[1]));
  else
    return LyXAscent(font, contents[0]);
}

int InsetQuotes::Descent(LyXFont font) {
  if (contents[1])
    return Maximum(LyXDescent(font, contents[0]),
		   LyXDescent(font, contents[1]));
  else
    return LyXDescent(font, contents[0]);
}

int InsetQuotes::Width(LyXFont font) {
  int w = LyXWidth(font, contents[0]);
  
  if (contents[1])
    w += LyXWidth(font, ',');       /* double quotes are wider*/
   
  return w;
}

InsetQuotes::InsetQuotes() {
  kind = new char[4];
  contents = new char[2];
  ParseString("eld");
}

LyXFont InsetQuotes::ConvertFont(LyXFont font) {
  /* quotes-insets cannot be latex of any kind */
  font.latex = LYX_NO_LATEX;
  return font;
}

void InsetQuotes::Draw(LyXFont font, unsigned long  pm, int baseline, 
		       float &x) {
  /* draw it! */ 
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x, baseline,
	      contents, 1);
   
  if (contents[1]) {	       /* double quotes*/
    x += LyXWidth(font, ',');
    /* draw the second! */ 
    XDrawString(fl_display,
		pm,
		LyXGetGC(font),
		(int) x, baseline,
		contents+1, 1);      /* a little pointer hack!  */
  }
  
  /* calculate the width and set the new x position */ 
  x +=  LyXWidth(font, contents[0]);
}


void InsetQuotes::Write(FILE *file) {
  fprintf(file, "Quotes %s", kind);
}

void InsetQuotes::Read(FILE *file) {
  char string[4];
  if (!feof(file))
    string[0] = fgetc(file);
  else
    string[0] = ' ';
  if (!feof(file))
    string[1] = fgetc(file);
  else
    string[1] = ' ';
  if (!feof(file))
    string[2] = fgetc(file);
  else
    string[2] = ' ';
  string[3]='\0';
   
  ParseString(string);
}

void InsetQuotes::Latex(FILE *file){
  switch (kind[0]) {
  case 'g':			       /* german quotes  */
    fprintf(file, "\\%c%cq", kind[0], kind[1]);
    if (kind[2]!='s')
      fprintf(file, "q");
    fprintf(file,"{}");
    break;
  case 'f':			       /* french quotes  */
    fprintf(file, "\\%c%cq", kind[0], kind[1]);
    if (kind[2]!='s')
      fprintf(file, "q");
    fprintf(file,"{}");
    break;
  default:			       /* english quotes  */
    fprintf(file, "%c", contents[0]);
    if (kind[2]!='s')
      fprintf(file, "%c", contents[0]);
    break;
  }
}


Inset* InsetQuotes::Clone() {
  InsetQuotes *result;
  result = new InsetQuotes(kind);
  return result;
}

char InsetQuotes::LyxCode() {
  return LYX_QUOTE_CODE;
}



/* Latex. Used to insert Latex-Code automatically */

InsetLatex::InsetLatex(char* string) {
  contents = StringCopy(string);
}

InsetLatex::~InsetLatex() {
   if (contents)
      delete(contents);
}

int InsetLatex::Ascent(LyXFont font) {
  int max=0;
  int i;
  for (i=0; contents[i];i++) {
    if (LyXAscent(font, contents[i])> max)
      max = LyXAscent(font, contents[i]);
  }
  return max;
}

int InsetLatex::Descent(LyXFont font) {
  int max=0;
  int i;
  for (i=0; contents[i];i++) {
    if (LyXDescent(font, contents[i])> max)
      max = LyXDescent(font, contents[i]);
  }
  return max;
}

int InsetLatex::Width(LyXFont font) {
  int n;
  for (n=0; contents[n];n++);	/* get the size */
  
  int w = LyXTextWidth(font, contents, n );
   
  w +=2;
   
  return w;
}

InsetLatex::InsetLatex() {
  contents = new char[10];
  contents[0] = '\0';
}



void InsetLatex::Draw(LyXFont font, unsigned long  pm, int baseline, 
		      float &x) {
  int n;

  for (n=0; contents[n];n++);	/* get the size */
   
  /* Latex-insets are always LaTeX, so just correct the font */ 
  if (font.latex == LYX_NO_LATEX)
    font.latex = LYX_LATEX;
   
  /* draw it! */ 
   
  x += 1;
  XFillRectangle(fl_display,
		 pm,
		 LyXGetLightedGC(),
		 (int) x, baseline- Ascent(font),
		 Width(font)-2, (Ascent(font)+Descent(font)));
		 
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x, baseline,
	      contents, n);
  
  x +=  Width(font) - 1;
}


void InsetLatex::Write(FILE *file) {
  fprintf(file, "Latex %s\n", contents);
}

void InsetLatex::Read(FILE *file) {
  int i = 0;
  char c='\0';
  char tmp[100];
  while (!feof(file) && (c = fgetc(file)) != '\n') {
    tmp[i]=c;
    i++;
  }
  tmp[i]='\0';
  contents = StringCopy(tmp);
}

void InsetLatex::Latex(FILE *file){
  fprintf(file, "%s", contents);
};

char InsetLatex::Deletable() {
  return 0;
}


Inset* InsetLatex::Clone(){
  InsetLatex *result = new InsetLatex(contents);
  return result;
}

char InsetLatex::LyxCode() {
  if (!strcmp(contents, "\\tableofcontents")) return LYX_TOC_CODE;
  return LYX_NO_CODE;
};


/* FormulaLatex. Used to insert FormulaLatex-Code automatically */

InsetFormulaLatex::InsetFormulaLatex(char* string) {
  contents = StringCopy(string);
}

InsetFormulaLatex::~InsetFormulaLatex() {
   if (contents)
      delete(contents);
}

int InsetFormulaLatex::Ascent(LyXFont font) {
  int max=0;
  int i;
  for (i=0; contents[i];i++) {
    if (LyXAscent(font, contents[i])> max)
      max = LyXAscent(font, contents[i]);
  }
  return max;
}

int InsetFormulaLatex::Descent(LyXFont font) {
  int max=0;
  int i;
  for (i=0; contents[i];i++) {
    if (LyXDescent(font, contents[i])> max)
      max = LyXDescent(font, contents[i]);
  }
  return max;
}


int InsetFormulaLatex::Width(LyXFont font) {
  int n;
  for (n=0; contents[n];n++);	/* get the size */

  int w = LyXTextWidth(font, contents, n );
   
  w +=2;
   
  return w;
}


InsetFormulaLatex::InsetFormulaLatex() {
  contents = new char[10];
  contents[0] = '\0';
}

LyXFont InsetFormulaLatex::ConvertFont(LyXFont font) {
  /* FormulaLatex-insets are always formulas*/ 
  font.latex = LYX_FORMULA_LATEX;
  return font;
}

void InsetFormulaLatex::Draw(LyXFont font, unsigned long  pm, int baseline, 
			     float &x) {
  int n;
  for (n=0; contents[n];n++);	/* get the size */

  /* draw it! */ 
  
  x += 1;
  XFillRectangle(fl_display,
		 pm,
		 LyXGetLightedGC(),
		 (int) x, baseline- Ascent(font),
		 Width(font)-2, (Ascent(font)+Descent(font)));
		 
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x, baseline,
	      contents, n);
  
  x +=  Width(font) - 1;
}


void InsetFormulaLatex::Write(FILE *file) {
  fprintf(file, "FormulaLatex %s\n", contents);
}

void InsetFormulaLatex::Read(FILE *file) {
  int i = 0;
  char c='\0';
  char tmp[100];
  while (!feof(file) && (c = fgetc(file)) != '\n') {
    tmp[i]=c;
    i++;
  }
  tmp[i]='\0';
  contents = StringCopy(tmp);
}

void InsetFormulaLatex::Latex(FILE *file){
  fprintf(file, "%s", contents);
};

char InsetFormulaLatex::Deletable() {
  return 0;
}

Inset* InsetFormulaLatex::Clone(){
  InsetFormulaLatex *result = new InsetFormulaLatex(contents);
  return result;
}





InsetFormulaLatexDel::InsetFormulaLatexDel(char* string) {
  contents = StringCopy(string);
}

InsetFormulaLatexDel::~InsetFormulaLatexDel() {
   if (contents)
      delete(contents);
}

int InsetFormulaLatexDel::Ascent(LyXFont font) {
  int max=0;
  int i;
  for (i=0; contents[i];i++) {
    if (LyXAscent(font, contents[i])> max)
      max = LyXAscent(font, contents[i]);
  }
  return max;
}

int InsetFormulaLatexDel::Descent(LyXFont font) {
  int max=0;
  int i;
  for (i=0; contents[i];i++) {
    if (LyXDescent(font, contents[i])> max)
      max = LyXDescent(font, contents[i]);
  }
  return max;
}


int InsetFormulaLatexDel::Width(LyXFont font) {
  int n;
  for (n=0; contents[n];n++);	/* get the size */

  int w = LyXTextWidth(font, contents, n );
   
  w +=2;
   
  return w;
}

InsetFormulaLatexDel::InsetFormulaLatexDel() {
  contents = new char[10];
  contents[0] = '\0';
}

LyXFont InsetFormulaLatexDel::ConvertFont(LyXFont font) {
  /* FormulaLatex-insets are always formulas*/ 
  font.latex = LYX_FORMULA_LATEX;
  return font;
}

void InsetFormulaLatexDel::Draw(LyXFont font, unsigned long  pm, int baseline, 
				float &x) {
  int n;
  for (n=0; contents[n];n++);	/* get the size */

  /* draw it! */ 
  
  x += 1;
  XFillRectangle(fl_display,
		 pm,
		 LyXGetLightedGC(),
		 (int) x, baseline- Ascent(font),
		 Width(font)-2, (Ascent(font)+Descent(font)));
		 
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x, baseline,
	      contents, n);
  
  x +=  Width(font) - 1;
}


void InsetFormulaLatexDel::Write(FILE *file) {
  fprintf(file, "FormulaLatex %s\n", contents);
}

void InsetFormulaLatexDel::Read(FILE *file) {
  int i = 0;
  char c='\0';
  char tmp[100];
  while (!feof(file) && (c = fgetc(file)) != '\n') {
    tmp[i]=c;
    i++;
  }
  tmp[i]='\0';
  contents = StringCopy(tmp);
}

void InsetFormulaLatexDel::Latex(FILE *file){
  fprintf(file, "%s", contents);
};

char InsetFormulaLatexDel::Deletable() {
  return 1;
}

Inset* InsetFormulaLatexDel::Clone(){
  InsetFormulaLatexDel *result = new InsetFormulaLatexDel(contents);
  return result;
}



/* LatexDel. Used to insert Latex-Code automatically */

InsetLatexDel::InsetLatexDel(char* string) {
  contents = StringCopy(string);
}


InsetLatexDel::~InsetLatexDel() {
   if (contents)
      delete(contents);
}


int InsetLatexDel::Ascent(LyXFont font) {
  int max=0;
  int i;

  for (i=0; contents[i];i++) {
    if (LyXAscent(font, contents[i])> max)
      max = LyXAscent(font, contents[i]);
  }

  return max;
}

int InsetLatexDel::Descent(LyXFont font) {
  int max=0;
  int i;

  for (i=0; contents[i];i++) {
    if (LyXDescent(font, contents[i])> max)
      max = LyXDescent(font, contents[i]);
  }
  return max;
}

int InsetLatexDel::Width(LyXFont font) {
  int n;
  for (n=0; contents[n];n++);	/* get the size */

  int w = LyXTextWidth(font, contents, n );
   
  w +=2;
   
  return w;
}

InsetLatexDel::InsetLatexDel() {
  contents = new char[10];
  contents[0] = '\0';
}

void InsetLatexDel::Draw(LyXFont font, unsigned long  pm, int baseline, 
			 float &x) {
  int n;

  for (n=0; contents[n];n++);	/* get the size */
  
  /* Latex-insets are always LaTeX, so just correct the font */ 
  if (font.latex == LYX_NO_LATEX)
    font.latex = LYX_LATEX;

  /* draw it! */ 
  
  x += 1;
  XFillRectangle(fl_display,
		 pm,
		 LyXGetLightedGC(),
		 (int) x, baseline- Ascent(font),
		 Width(font)-2, (Ascent(font)+Descent(font)));
		 
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x, baseline,
	      contents, n);
  
  x +=  Width(font) - 1;
}


void InsetLatexDel::Write(FILE *file) {
  fprintf(file, "LatexDel %s\n", contents);
}

void InsetLatexDel::Read(FILE *file) {
  int i = 0;
  char c='\0';
  char tmp[100];
  while (!feof(file) && (c = fgetc(file)) != '\n') {
    tmp[i]=c;
    i++;
  }
  tmp[i]='\0';
  contents = StringCopy(tmp);
}

void InsetLatexDel::Latex(FILE *file){
  fprintf(file, "%s", contents);
};

char InsetLatexDel::Deletable() {
  return 1;
}

Inset* InsetLatexDel::Clone() {
  InsetLatexDel *result = new InsetLatexDel(contents);
  return result;
}

char InsetLatexDel::LyxCode() {
  if (!strcmp(contents, "\\tableofcontents")) return LYX_TOC_CODE;
  if (!strcmp(contents, "\\label{")) return LYX_MARK_CODE;
  if (!strcmp(contents, "\\ref{")) return LYX_REF_CODE;
  if (!strcmp(contents, "\\url{")) return LYX_URL_CODE;
  if (!strcmp(contents, "\\htmlurl{")) return LYX_HTMLURL_CODE;
  if (!strcmp(contents, "}{")) return LYX_SEPARATOR_CODE;
  if (!strcmp(contents, "}")) return LYX_ENDING_CODE;
  return LYX_NO_CODE;
}

/* Error, used for the LaTeX-Error Messages */

InsetError::InsetError(char* string) {
  contents = StringCopy(string);
}


InsetError::~InsetError() {
   if (contents)
      delete(contents);
}


int InsetError::Ascent(LyXFont font) {
  return LyXAscent(font, 'E')+3;
}

int InsetError::Descent(LyXFont font) {
  return LyXDescent(font, 'E')+3;
}

int InsetError::Width(LyXFont font) {
  int w = LyXTextWidth(font, "Error", 5 );
   
  w +=6;
   
  return w;
}

InsetError::InsetError() {
  contents = new char[10];
  contents[0] = '\0';
}

void InsetError::Draw(LyXFont font, unsigned long  pm, int baseline, float &x) {
   
  /* Latex-insets are always LaTeX, so just correct the font */ 
   
  font.latex = LYX_LATEX;
   
   
  /* draw it! */ 
   
  x += 1;
  XFillRectangle(fl_display,
		 pm,
		 LyXGetLightedGC(),
		 (int) x, baseline- Ascent(font),
		 Width(font)-2, (Ascent(font)+Descent(font)));
   
  XDrawRectangle(fl_display,
		 pm,
		 LyXGetCopyGC(),
		 (int) x-1, baseline- Ascent(font)-1,
		 Width(font)-1, (Ascent(font)+Descent(font))+1);
   
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x+3, baseline,
	      "Error", 5);
   
  x +=  Width(font) - 1;
}


void InsetError::Write(FILE *file) {
}

void InsetError::Read(FILE *file) {
}

void InsetError::Latex(FILE *file){
};

char InsetError::AutoDelete() {
  return 1;
}

char InsetError::Editable() {
  return 1;
}


static void CloseErrorCB(FL_OBJECT *ob, long data) {
   FL_FORM *form = (FL_FORM*) data;
   if (form)
     fl_hide_form(form);
}


FL_FORM* InsetError::form = NULL;
FL_OBJECT* InsetError::strobj = NULL;

void InsetError::Edit() {
   
  if (!form) {
    FL_OBJECT *obj;
    form = fl_bgn_form(FL_UP_BOX,400,140);
    strobj = fl_add_box(FL_FRAME_BOX,10,50,380,80,"");
    fl_set_object_color(strobj,FL_MCOL,FL_MCOL);
    obj = fl_add_button(FL_RETURN_BUTTON,160,10,120,30,"OK");
    fl_set_object_callback(obj, CloseErrorCB, (long)form);
    fl_end_form();
  }
  fl_set_object_label(strobj, contents);
  fl_show_form(form,FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER, 
	       "LaTeX-Error");
}
   
Inset* InsetError::Clone(){
  InsetError *result = new InsetError(contents);
  return result;
}



/* Info, used for the Info boxes */

InsetInfo::InsetInfo() {
   form = NULL;
   strobj = NULL;
   contents = new char[10];
   contents[0] = '\0';
}


InsetInfo::~InsetInfo() {
   if (contents)
      delete(contents);
   if (form){
      fl_hide_form(form);
      fl_free_form(form);
      form = NULL;
   }
}


int InsetInfo::Ascent(LyXFont font) {
  return LyXAscent(font, 'E')+3;
}

int InsetInfo::Descent(LyXFont font) {
  return LyXDescent(font, 'E')+3;
}

int InsetInfo::Width(LyXFont font) {
  int w = LyXTextWidth(font, "Note", 5 );
   
  w +=6;
   
  return w;
}

void InsetInfo::Draw(LyXFont font, unsigned long  pm, int baseline, float &x) {
   
  /* Latex-insets are never LaTeX, so just correct the font */ 
   
  font.latex = LYX_NO_LATEX;
   
   
  /* draw it! */ 
   
  x += 1;
  XFillRectangle(fl_display,
		 pm,
		 LyXGetYellowGC(),
		 (int) x, baseline- Ascent(font),
		 Width(font)-2, (Ascent(font)+Descent(font)));
   
  XDrawRectangle(fl_display,
		 pm,
		 LyXGetCopyGC(),
		 (int) x-1, baseline- Ascent(font)-1,
		 Width(font)-1, (Ascent(font)+Descent(font))+1);
   
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x+3, baseline,
	      "Note", 5);
   
  x +=  Width(font) - 1;
}


void InsetInfo::Write(FILE *file) {
   fprintf(file, "Info %s\n", contents);
}

void InsetInfo::Read(FILE *file) {
   char* token;
   char c;
   char token_end_character;
   LyXParagraph *par = new LyXParagraph;
   int pos = 0;
   int i;
   c = fgetc(file);
   while (!feof(file)) {
      /* is it a lyx command? */ 
      if (c=='\\') {
	 token = InsetNextToken(file, token_end_character);
	 if (StringEqual(token, "end_inset")) {
	    par->InsertChar(pos - 1, '\0');   /* the - 1 deletes the newline
					       * before end_preamble*/
	    Set(par->text);
	    delete(par);
	    return;
	 }
	 else {
	    par->InsertChar(pos, '\\');
	    pos++;
	    for (i = 0; token[i]; i++) {
	       par->InsertChar(pos, token[i]);
	       pos++;
	    }
	    par->InsertChar(pos, token_end_character);
	    pos++;
	 }
	 delete(token);
      }
      else {
	 /* simple text */
	 par->InsertChar(pos, c);
	 pos++;
      }
      c = fgetc(file);
   }
}
      

void InsetInfo::Latex(FILE *file){
};

char InsetInfo::Editable() {
  return 1;
}


static void CloseInfoCB(FL_OBJECT *ob, long data) {
   InsetInfo *inset = (InsetInfo*) data;
   char *tmp=StringCopy(fl_get_input(inset->strobj));
   inset->Set(tmp);
   delete (tmp);
   fl_hide_form(inset->form);
   fl_free_form(inset->form);
   inset->form = NULL;
}


void InsetInfo::Edit() {
   
  if (!form) {
    FL_OBJECT *obj;
    form = fl_bgn_form(FL_UP_BOX,400,180);
    strobj = obj = fl_add_input(FL_MULTILINE_INPUT,10,50,380,120,"");
    fl_set_object_color(obj,FL_MCOL,FL_MCOL);
    fl_set_object_resize(obj, FL_RESIZE_ALL, NorthWestGravity, SouthEastGravity);
    obj = fl_add_button(FL_NORMAL_BUTTON,160,10,120,30,"OK");
    fl_set_object_resize(obj, FL_RESIZE_ALL, SouthWestGravity, SouthEastGravity);
    fl_set_object_callback(obj, CloseInfoCB, (long)this);
    fl_end_form();
  }
  fl_set_input(strobj, contents);
  fl_show_form(form,FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER, 
	       "Note");
}

void InsetInfo::Set(char* string) {
   if (contents)
      delete(contents);
   contents = StringCopy(string);
}
   
Inset* InsetInfo::Clone(){
   InsetInfo *result = new InsetInfo();
   result->Set(contents);
   return result;
}

char InsetInfo::LyxCode() {
   return LYX_IGNORE_CODE;
}






/* Label. Used to insert a label automatically */

InsetLabel::InsetLabel(char* string) {
  label = StringCopy(string);
  int i;
  for (i=0; label[i]; i++);
  contents = new char [i+20];
  sprintf(contents, "\\label{%s}", label);
}


InsetLabel::~InsetLabel() {
   if (label)
      delete(label);
   if (contents)
      delete(contents);
}


int InsetLabel::Ascent(LyXFont font) {
  int max=0;
  int i;

  for (i=0; contents[i];i++) {
    if (LyXAscent(font, contents[i])> max)
      max = LyXAscent(font, contents[i]);
  }

  return max;
}

int InsetLabel::Descent(LyXFont font) {
  int max=0;
  int i;

  for (i=0; contents[i];i++) {
    if (LyXDescent(font, contents[i])> max)
      max = LyXDescent(font, contents[i]);
  }
  return max;
}

int InsetLabel::Width(LyXFont font) {
  int n;
  for (n=0; contents[n];n++);	/* get the size */

  int w = LyXTextWidth(font, contents, n );
   
  w +=2;
   
  return w;
}

InsetLabel::InsetLabel() {
  contents = new char[10];
  contents[0] = '\0';
  label = new char[10];
  label[0] = '\0';

}

void InsetLabel::Draw(LyXFont font, unsigned long  pm, int baseline, 
			 float &x) {
  int n;

  for (n=0; contents[n];n++);	/* get the size */
  
  /* Latex-insets are always LaTeX, so just correct the font */ 
  if (font.latex == LYX_NO_LATEX)
    font.latex = LYX_LATEX;

  /* draw it! */ 
  
  x += 1;
  XFillRectangle(fl_display,
		 pm,
		 LyXGetLightedGC(),
		 (int) x, baseline- Ascent(font),
		 Width(font)-2, (Ascent(font)+Descent(font)));
		 
  XDrawString(fl_display,
	      pm,
	      LyXGetGC(font),
	      (int) x, baseline,
	      contents, n);
  
  x +=  Width(font) - 1;
}


void InsetLabel::Write(FILE *file) {
  fprintf(file, "Label %s\n", label);
}

void InsetLabel::Read(FILE *file) {
  int i = 0;
  char c='\0';
  char tmp[100];
  while (!feof(file) && (c = fgetc(file)) != '\n') {
    tmp[i]=c;
    i++;
  }
  tmp[i]='\0';
  label = StringCopy(tmp);
  for (i=0; label[i]; i++);
  contents = new char [i+20];
  sprintf(contents, "\\label{%s}", label);
}

void InsetLabel::Latex(FILE *file){
  fprintf(file, "%s", contents);
};

char InsetLabel::Deletable() {
  return 1;
}

Inset* InsetLabel::Clone() {
  InsetLabel *result = new InsetLabel(label);
  return result;
}

char InsetLabel::LyxCode() {
  return LYX_LABEL_CODE;
}

char* InsetLabel::GetLabel(){
  return label;
}
