/* This file is part of
* ======================================================
* 
*           LyX, the High Level Word Processor
* 	 
*	    Copyright (C) 1995 Matthias Ettrich
*
*======================================================*/

#include"xdefinitions.h"

/* Quotes. Used for the various quotes. German, English, French, all either
 * double or single */ 

class InsetQuotes: public Inset {
private:
  char* contents;
  char* kind;
   
  void ParseString(char* string);
   
public:
  ~InsetQuotes();
   
  /* the construktor works like this:
   * 
   * fls <- french single quote left 
   * grd <- german double quote right 
   * 
   * etc. */ 
   
  InsetQuotes(char* string);
  InsetQuotes();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  LyXFont ConvertFont(LyXFont font);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  Inset* Clone();
  char LyxCode();
};




/* Latex. Used to insert LaTeX-Code automatically */

class InsetLatex: public Inset {
private:
  char* contents;
public:
  ~InsetLatex();
   
  InsetLatex(char* string);
  InsetLatex();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  char Deletable();
  Inset* Clone();
  char LyxCode();
};



/* FormulaLatex. Used to insert FormulaLatex-Code automatically */

class InsetFormulaLatex: public Inset {
private:
  char* contents;
   
public:
  ~InsetFormulaLatex();
  
  InsetFormulaLatex(char* string);
  InsetFormulaLatex();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  LyXFont ConvertFont(LyXFont font);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  char  Deletable();
  Inset* Clone(); 
};

/* FormulaLatexDel. Used to insert deletable FormulaLatex-Code automatically */

class InsetFormulaLatexDel: public Inset {
private:
  char* contents;
   
public:
  ~InsetFormulaLatexDel();
   
  InsetFormulaLatexDel(char* string);
  InsetFormulaLatexDel();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  LyXFont ConvertFont(LyXFont font);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  char  Deletable();
  Inset* Clone();
};

class InsetLatexDel: public Inset {
private:
  char* contents;
public:
  ~InsetLatexDel();
   
  InsetLatexDel(char* string);
  InsetLatexDel();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  char Deletable();
  Inset* Clone();
  char LyxCode();
};



class InsetError: public Inset {
private:
  char* contents;
  static   FL_FORM *form;
  static    FL_OBJECT *strobj;

public:
  ~InsetError();
   
  InsetError(char* string);
  InsetError();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  char AutoDelete();
  void Edit();
  char Editable();
  Inset* Clone();
};


class InsetInfo: public Inset {
private:
   
public:
  char* contents;
  FL_FORM *form;
  FL_OBJECT *strobj;

  ~InsetInfo();
  void Set(char* string);
  InsetInfo();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  void Edit();
  char Editable();
  char LyxCode();
  Inset* Clone();
};





class InsetLabel: public Inset {
private:
  char* contents;
  char* label;
public:
  ~InsetLabel();
   
  InsetLabel(char* string);
  InsetLabel();
  int Ascent(LyXFont font);
  int Descent(LyXFont font);
  int Width(LyXFont font);
  void Draw(LyXFont font, unsigned long  pm, int baseline, float &x);
  void Write(FILE *file);
  void Read(FILE *file);
  void Latex(FILE *file);
  char Deletable();
  Inset* Clone();
  char LyxCode();
  char* GetLabel();
};
