/*
	International support for LyX

*/



#include <unistd.h>
#include <forms.h>
#include "form1.h"
#include "tex-defs.h"
#include "inset.h"
#include "intl.h"

extern char system_lyxdir [255];

BOOL keymapon = FALSE;	/* disabled by default */
int chsetcode = 0;	/* default to 7-bit ascii */
int curkeymap = 0;	/* current keymap */
int otherkeymap;	/* use other keymap */

FD_KeyMap *fd_form_keymap;	/* keymap form */
FD_KeyMapErr *fd_form_keymaperr;/* keymap load error form */


// dead keys definition:

#define dead_acute	1	// '
#define dead_grave	2	// `
#define dead_macron	3	// -
#define dead_tilde	4	// ~
#define dead_underbar	5	// _
#define dead_cedilla	6	// ,
#define dead_underdot	7	// .
#define dead_circumflex	8	// ^
#define dead_circle	9	// o
#define dead_tie	10	//  ^
#define dead_breve	11	// v
#define dead_caron	12	// v
#define dead_hunguml	13	// "
#define dead_umlaut	14	// :
#define dead_dot	15	// .
#define max_dead	15

char* deadkeysyms[max_dead+1] = {"", "\\'", "\\`", "\\=", "\\~", "\\b ",
		"\\c ", "\\d ", "\\^", "\\r ", "\\t ", "\\u ", "\\v ",
		"\\H ", "\\\"", "\\."};
char* deadkeyname[max_dead+1] = {"", "acu", "gra", "mac", "til",
		"underb", "ced", "underd", "circu", "circl", "tie",
		"bre", "car", "hug", "uml", "dot"};

typedef struct KEYEXC {
	char c;		/* character to make exception */
	char *data;	/* exception data */
	struct KEYEXC *next;
} *keyexc;

char* keymap[256];	/* keymap for all 256 characters */
/* points to character definition or NULL if no translation defined */
/* character definition is what to return, if begins with 0 - then it's */
/* the dead key - found in next byte */

keyexc exclist[max_dead+1];	/* exception list for each deadkey */

int deadkey = 0;	/* no dead key pending */

extern void InsertCorrectQuote();

void InitKeymap()
	/* low-level init */
{
	int i;
	for (i = 0; i < 256; ++i) keymap[i] = NULL;
	for (i = 0; i <= max_dead; ++i) exclist[i] = NULL;
}

void freeexc(keyexc exclist)
{
	keyexc p;

	p = exclist;
	while (p) {
		p = exclist->next;
		free(exclist->data);
		free(exclist);
		exclist = p;
	}
}

void FreeKeymap()
	/* free key map description */
{
	int i;

	for (i = 0; i < 256; ++i)
		if (keymap[i]) free(keymap[i]);
	for (i = 0; i <= max_dead; ++i)
		freeexc(exclist[i]);

	InitKeymap();
}

FILE *kfopen(char *name)
	/* open key definition file */
{
	char buf[256];
	FILE *f;

	sprintf(buf, "%s/kbd/%s.kmap", system_lyxdir, name);
#ifdef debug
	printf("Looking for %s...\n", buf);
#endif
	if (f = fopen(buf, "r")) return f;
	sprintf(buf, "~/.%s.kmap", name);
#ifdef debug
	printf("Looking for %s...\n", buf);
#endif
	if (f = fopen(buf, "r")) return f;
	sprintf(buf, "%s.kmap", name);
#ifdef debug
	printf("Looking for %s...\n", buf);
#endif
	if (f = fopen(buf, "r")) return f;
	sprintf(buf, "%s/kbd/%s", system_lyxdir, name);
#ifdef debug
	printf("Looking for %s...\n", buf);
#endif
	if (f = fopen(buf, "r")) return f;
	sprintf(buf, "~/.%s", name);
#ifdef debug
	printf("Looking for %s...\n", buf);
#endif
	if (f = fopen(buf, "r")) return f;
#ifdef debug
	printf("Looking for %s...\n", name);
#endif
	if (f = fopen(name, "r")) return f;

#ifdef debug
	printf("not found\n");
#endif
	return NULL;
}

char getkeychar(char **p)
	/* get next char from *p, update *p */
{
	char c;

	if (!**p) return 0;
	c = *((*p)++);

	if (c == '\\') {
		if (!**p) return 0;
		c = *((*p)++);
	} else if (c == '\"') c = 0x80;

	return c;
}

int getkeymod(char **p)
	/* return modifier - decoded from p and update p */
{
	int i;

	for (i = 1; i <= max_dead; ++i)
	    if (strncmp(*p, deadkeyname[i], strlen(deadkeyname[i])) == 0) {
		/* found it! Update *p to next space or so */
		while (**p && **p != ' ' && **p != '\t' && **p != '\n') (*p)++;
		return i;
	    }

	return 0;
}

char *getkeydata(char *p)
	/* translate until end of string p */
{
	char buf[128], *p2;

	p2 = buf;

	while (*(p2++) = getkeychar(&p));

	p2 = (char *) malloc(p2 - buf);
	strcpy(p2, buf);
	return p2;
}

void insexc(keyexc *exclist, char c, char *data)
{
	keyexc p;

	p = (keyexc) malloc(sizeof(struct KEYEXC));
	p -> next = *exclist;
	p -> c = c;
	p -> data = getkeydata(data);

	*exclist = p;
}

void ParseKeymapFile(FILE *f, int depth)
	/* primitive for parsing keymap files */
{
	char bf[128], *p, c, m;
	int action;

	if (depth > 8) return;	/* recursive ??? */

	/* read string */
	while (fgets(bf, 128, f)) {
		if (bf[0] != '\\') continue;	/* not a control keyword */

		/* first remove comment, beginning by unescaped # */
		p = bf + 1;
		while (*p && (*p != '#' || *(p-1) == '\\')) ++p;

		if (!*p) {
			/* haven't found '#' - remove whitespaces at the end */
			p = bf + strlen(bf);
		}
		do {
			*(p--) = 0;
		} while (*p == ' ' || *p == '\t' || *p == '\n');
		/* now process the entry */
		p = bf;
		while (*p && *p != ' ' && *p != '\t' && *p != '\n') ++p;
		action = 0;
		switch (p - bf) {
			case 5:
				if (strncmp(bf, "\\kmod", 5) == 0) action = 1;
				else if (strncmp(bf, "\\kmap", 5) == 0) action = 2;
				break;
			case 6:
				if (strncmp(bf, "\\kxmod", 6) == 0) action = 3;
				break;
		}
		/* do the action - define a key */
#ifdef debug
		if (action) printf("Key definition: [%d;%s]\n", action, p);
#endif
		switch (action) {
		    case 1:	/* define a modifier */
			if (!*(p++)) goto error;
			if (!(c = getkeychar(&p))) goto error;
			if (!*(p++)) goto error;
			if (!(m = getkeymod(&p))) goto error;
			/* define a modifier */
			if (keymap[c]) free(keymap[c]);	/* free old def */
			keymap[c] = (char *) malloc(2);
			keymap[c][0] = 0;
			keymap[c][1] = m;
			break;
		    case 2:	/* define a key mapping */
			if (!*(p++)) goto error;
			if (!(c = getkeychar(&p))) goto error;
			if (!*(p++)) goto error;
			/* define mapping */
			if (keymap[c]) free(keymap[c]);	/* free old def */
			keymap[c] = getkeydata(p);
			break;
		    case 3:	/* define an exception */
			if (!*(p++)) goto error;
			if (!(m = getkeymod(&p))) goto error;
			if (!*(p++)) goto error;
			if (!(c = getkeychar(&p))) goto error;
			if (!*(p++)) goto error;
			/* define an exception for modifier m/char c */
			insexc(&exclist[m], c, p);
			break;
		}
		continue;
error:
		fprintf(stderr, "Error in keymap definition: [%s]", bf);
	}

}

int SelectKeyMap(char *keyfile)
	/* load keyboard map from keyfile, return 0 on success, -1 else */
{
	FILE *f;

	if (strcmp(keyfile, tex_babel[0]) == 0) {
		FreeKeymap();
		return 0;
	}

	f = kfopen(keyfile);
	if (!f) return -1;

	/* free old data and initialize new keyboard mapping */
	FreeKeymap();

	/* now parse the file & it's includes */
	ParseKeymapFile(f, 0);

	fclose(f);

	return 0;
}

void SelectCharset(int code)
	/* select charset to convert symbols to: 0-7-bit ascii (default),
	   1-10-ISOLatin-xx */
	/* NOT SUPPORTED YET */
{
	chsetcode = code;
}

void KeyMapOn(BOOL on)
	/* turn on/off key mappings, status in keymapon */
{
	keymapon = on;
	fl_set_button(fd_form_keymap->KeyOnBtn, on);
}

void ToggleKeyMap()
{
	KeyMapOn(!keymapon);
}

void ChangeKeymap(char *keyfile, BOOL on, BOOL warn, int lngidx)
{
	if (SelectKeyMap(keyfile) != 0) {
		if (warn) fl_show_form(fd_form_keymaperr->KeyMapErr, FL_PLACE_CENTER,
			FL_FULLBORDER, "Error");
		return;
	}

	fl_set_choice(fd_form_keymap->Language, lngidx);
	KeyMapOn(on);
}

void Translate(char *instr, char *outstr)
	/* translate instr and write to outstr, if translations are on */
	/* provide big enough buffer !!! Ie, 128 bytes */
{
	if (!keymapon) {
		strcpy(outstr, instr);
		return;
	}

cont:
	while (*instr) {
		if (keymap[*instr]) {
			/* found keymap entry */
			if (!keymap[*instr][0]) {
				/* dead-key */
				deadkey = keymap[*instr][1];
			} else {
				/* mapped character */
				strcpy(outstr, keymap[*instr]);
				outstr += strlen(outstr);
				deadkey = 0;
			}
			++instr;
		} else if (deadkey) {
			/* map character from instr */
			keyexc p = exclist[deadkey];
			char buf[32], *p2;
			while (p) {
				if (p->c == *instr) {
					deadkey = 0;
					strcpy(outstr, p->data);
					outstr += strlen(outstr);
					++instr;
					goto cont; // I hate this
				}
				p = p->next;
			}
			/* now compose the latex char */
			buf[0] = 0x80;
			p2 = buf+1;
			strcpy(p2, deadkeysyms[deadkey]);
			p2 += strlen(p2);
			if (*instr == ' ') {
				strcpy(p2, "{ }\200");
				p2 += strlen(p2);
			} else {
				*(p2++) = *instr;
				*(p2++) = 0x80;
			}
			*p2 = 0;
			deadkey = 0;
			instr++;
			strcpy(outstr, buf);
			outstr += strlen(outstr);
		} else *(outstr++) = *(instr++);
	}
	*outstr = 0;
}

void InitKeyMapper()
	/* initialize key mapper */
{
#ifdef debug
	printf("Initializing key mappings...\n");
#endif
	InitKeymap();

	fd_form_keymap = create_form_KeyMap();
	fd_form_keymaperr = create_form_KeyMapErr();

	fl_set_button(fd_form_keymap->KeyOnBtn, 0);
	int n=0;
	while (1) if (strcmp(tex_babel[n++], "last_item")==0) break;
	    else fl_addto_choice(fd_form_keymap->Language, tex_babel[n-1]);
	fl_addto_choice(fd_form_keymap->Language, "other...");
	otherkeymap = n - 1;
}

void Keymap(FL_OBJECT *obj, long code)
{
	int i;

#ifdef debug
	printf("KeyMap callback: %d\n", code);
#endif

	switch (code) {
	    case 0:
		/* accept */
		const char *p;
		fl_hide_form(fd_form_keymap->KeyMap);
		i = fl_get_choice(fd_form_keymap->Language) - 1;
		if (i == curkeymap && i != otherkeymap) break;
		if (i == 0) {
			FreeKeymap();
			break;
		}
		/* read text from choice */
		if (i == otherkeymap) p = fl_get_input(fd_form_keymap->OtherKeymap);
		    else p = fl_get_choice_text(fd_form_keymap->Language);
		if (!SelectKeyMap((char *) p))
			curkeymap = i;
		else {
#ifdef debug
			printf("Keymap file: %s\n", p);
#endif
			fl_set_choice(fd_form_keymap->Language, curkeymap + 1);
			/* error !!!! */
			fl_show_form(fd_form_keymaperr->KeyMapErr, FL_PLACE_CENTER,
				FL_FULLBORDER, "Error");
		}
		break;
	    case 1:
		/* cancel */
		fl_hide_form(fd_form_keymap->KeyMap);
		fl_set_choice(fd_form_keymap->Language, curkeymap + 1);
		break;
	    case 3:
		/* on/off */
		keymapon = fl_get_button(fd_form_keymap->KeyOnBtn);
#ifdef debug
		printf("KeyMapOn: %d\n", keymapon);
#endif
		break;
	    case 4:
		/* change keymap */
		i = fl_get_choice(fd_form_keymap->Language) - 1;
		printf("Table:%s\n", tex_babel[i]);
		break;
	    case 10:
		/* close error box */
		fl_hide_form(fd_form_keymaperr->KeyMapErr);
		break;
	}
}

void MenuKeymap()
{
	if (fd_form_keymap->KeyMap->visible) {
		XRaiseWindow(fl_display, fd_form_keymap->KeyMap->window);
	} else fl_show_form(fd_form_keymap->KeyMap, FL_PLACE_CENTER,
		FL_FULLBORDER, "Key Mappings");
}

void TranslateAndInsert(char c, LyXText *text)
	/* insert correct stuff into paragraph */
{
	char buf[2], obuf[128], tbuf[128];
	unsigned char *p, *p2;

	/* fill in instr */
	buf[0] = c;
	buf[1] = 0;

	/* translate the sequence */
	Translate(buf, obuf);

#ifdef debug
	printf("Translated string: %s to %s\n", buf, obuf);
#endif

	/* insert into buffer for text */
	p = obuf;

	while (*p) {
		if (*p == 0x80) {
			/* start Latex style */
			++p;
			p2 = tbuf;
			while (*p && *p != 0x80) *(p2++) = *(p++);
			*p2 = 0;
			Inset *new_inset = new InsetLatexDel(tbuf);
			text->InsertInset(new_inset);
			if (!*p) break;
		} else if (*p == '\"')
			InsertCorrectQuote();
		else text->InsertChar(*p);
		++p;
	}
}
