/* ======================================================================= *\
    File   : keybind.C, $Id: keybind.C,v 1.2 1995/11/08 08:01:26 chb Exp $
    Author : chb, 28.Oct.1995
    Docu   : see keybind.h
    Purpose: rebindable keys for LyX
    Note   : this file is included in lyx_cb.C (not anymore! Lgb.)
\* ======================================================================= */

#define  _KEYBIND_C_
#include <stdio.h>
#include "file.h"
#include "kbmap.h"
#include "keybind.h"
#include "lyxserver.h"
#include "lyx.h"
#include "intl.h"

// extern void show_symbols_form(void);
extern FD_form_main *fd_form_main;
extern FD_form_toc *fd_form_toc;

extern Buffer *buffer[];
extern int current_buffer;
extern int old_current_buffer;
extern char selection_possible;

extern void Update(signed char);
extern void SetMinibuffer(const char*, const char*, const char*);
extern void BeforeChange();
extern void LyXBell();
extern void UpdateScrollbar();
extern void MenuNew();
extern void MenuOpen();
extern void MenuWrite(int);
extern void MenuWriteAs();
extern void MenuRunLaTeX();
extern void CloseBuffer();
extern void MenuQuit();
extern void AutoSave();
extern void MenuSearch();
extern void CopyCB();
extern void GotoError();
extern void OpenStuff();
extern void HyphenationPoint();
extern void HFill();
extern void MenuLoadAutoSave();
extern void ResizeCurrentBufferPseudoExpose();
extern void InitMinibuffer();
extern void CursorPrevious();
extern void CursorNext();
extern void SmallUpdate(signed char);
extern void SetUpdateTimer();
extern void FreeUpdateTimer();
extern void InsertCorrectQuote();
extern void MenuPreview();
extern void MenuPreviewPS();

extern int work_area_height;

int WorkAreaKeyPress(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d);

/* === globals =========================================================== */

kb_keymap *toplevel_keymap;
kb_sequence keyseq;

// if we use WorkAreaKeyPress as function dispatcher, set this:

static kb_action dispatch_action = LFUN_NOACTION;
static char *dispatch_argument   = NULL;
static char *dispatch_result     = NULL;
static char  dispatch_buffer[100];

/* === code ============================================================== */

/* ---F+------------------------------------------------------------------ *\
    Function  : DispatchFunction
    Called by : [user]
    Purpose   : Call a LyX editing function manually
    Parameters: action   - the kb_action to call
                argument - a string argument
    Returns   : dispatch_result (set by WorkAreaKeyPress)
    Note      : dispatch_result will always point to a static string;
                so the caller should save it if it's to be used later.
\* ---F------------------------------------------------------------------- */

char *DispatchFunction(kb_action action, char *argument)
{
   // this is a dirty trick to use the callback function directly...
   dispatch_action   = action;
   dispatch_argument = argument;
   dispatch_result   = NULL;

   WorkAreaKeyPress(0, 0, 0, 0, 0, 0);

   return dispatch_result;
}


/* ---F+------------------------------------------------------------------ *\
    Function  : MoveCursorUpdate
    Called by : WorkAreaKeyPress
    Purpose   : update display after cursor movement
    Parameters: shift - 1 if *SEL command
    Returns   : nothing
\* ---F------------------------------------------------------------------- */

void MoveCursorUpdate(int shift)
{
   if (shift || buffer[current_buffer]->text->mark_set)
   {
      buffer[current_buffer]->text->SetSelection();
      buffer[current_buffer]->screen->ToggleToggle();
      Update(0);
   }
   else
   {
      Update(0);
      buffer[current_buffer]->text->sel_cursor =
         buffer[current_buffer]->text->cursor;
   }
}


/* ---F+------------------------------------------------------------------ *\
    Function  : WorkAreaKeyPress
    Called by : XForms Call back
    Purpose   : process key bindings for LyX
    Parameters: ob   - XForms object; passed but not used in called funcs
                win  - unused
                w, h - unused
                ev   - X Event
                d    - unused
    Returns   : 0
\* ---F------------------------------------------------------------------- */

int WorkAreaKeyPress(FL_OBJECT *ob,
                     Window win,         // unused
                     int w,              // unused
                     int h,              // unused
                     XEvent *ev,
                     void *d)            // unused
{
   static char s_r[10];
   static int num_bytes;
   kb_action action;
   char *argument;
   XKeyEvent *keyevent = &ev->xkey;
   KeySym keysym_return;
   int tmp;

   selection_possible = False;
   if (buffer[current_buffer]->screen)
      buffer[current_buffer]->screen->HideCursor();

   if(dispatch_action != LFUN_NOACTION)
   {
      action = dispatch_action;
      dispatch_action = LFUN_NOACTION;
      argument = dispatch_argument;
   }
   else
   {
     num_bytes = XLookupString(keyevent, s_r, 10, &keysym_return, NULL);
     // fprintf(stderr,"%lx %x\n", keysym_return, keyevent->state);
     // fprintf(stderr,"s_r: %s len: %d\n",s_r,num_bytes);
      if((keysym_return>=XK_Shift_L && keysym_return <= XK_Hyper_R)
	 || keysym_return == 0x0)
         return 0;
      // remove Caps Lock as a modifier
      action = (kb_action)keyseq.addkey(keysym_return, keyevent->state&~LockMask);
      if(action==0) action = LFUN_PREFIX;
      argument = NULL;
      // {
      //    char buf[100];
      //    keyseq.print(buf,100);
      //    fprintf(stderr,"Key: %s  Action: %d\n",buf,action);
      // }
   }

   if(!buffer[current_buffer]->screen)
   {
      switch(action)
      {
       case LFUN_PREFIX:
       case LFUN_MENUNEW:
       case LFUN_MENUOPEN:
       case LFUN_MENUQUIT:
         break;
       default:
         return 0;  // action not allowed
      }
   }

   if(keyseq.length>1 || keyseq.length<-1)
   {
      char buf[100];
      keyseq.print(buf,100);
      SetMinibuffer(buf, "", "");
   }

   switch(action)
   {
      // --- Misc ---------------------------------------------------------
    case LFUN_PREFIX:
      if(buffer[current_buffer]->screen)
         Update(-2);
      {
         char buf[100];
         keyseq.print(buf,100);
         SetMinibuffer(buf, "", "");
      }
      break;
    case LFUN_PREFIXARG:
      // !!! add prefixarg setting here
      break;
    case LFUN_CENTER:
      BeforeChange();
      Update(-2);
      if (buffer[current_buffer]->text->cursor.y > work_area_height / 2)
      {
         buffer[current_buffer]->screen->
            Draw(buffer[current_buffer]->text->cursor.y
                        - work_area_height / 2);
         UpdateScrollbar();
         SetMinibuffer("Screen centered!","","");
      }
      else
      {
         LyXBell();
         SetMinibuffer("Cannot center top of document.","","");
      }
      Update(0);
      break;
      // --- Menus --------------------------------------------------------
    case LFUN_MENUNEW:      MenuNew();       break;
    case LFUN_MENUOPEN:     MenuOpen();      break;
    case LFUN_MENUWRITE:    MenuWrite(current_buffer);      break;
    case LFUN_MENUWRITEAS:  MenuWriteAs();   break;
    case LFUN_RUNLATEX:     MenuRunLaTeX();  break;
    case LFUN_PREVIEW:      MenuPreview();   break;
    case LFUN_PREVIEWPS:    MenuPreviewPS(); break;
    case LFUN_CLOSEBUFFER:  CloseBuffer();   break;
    case LFUN_MENUQUIT:     MenuQuit();      break;
    case LFUN_TOCUPDATE:
      TocUpdateCB(ob, 0);
      fl_show_form(fd_form_toc->form_toc,
                   FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
                   "Table Of Contents");
      break;

    case LFUN_AUTOSAVE:     AutoSave();      break;
    case LFUN_MENULOADAUTO: MenuLoadAutoSave(); break;
    case LFUN_MENUSEARCH:   MenuSearch();    break;
    case LFUN_PASTE:        PasteCB(ob, 0);  break;
    case LFUN_CUT:          CutCB(ob, 0);    break;
    case LFUN_COPY:         CopyCB();        break;
    case LFUN_GOTOERROR:    GotoError();     break;
    case LFUN_OPENSTUFF:    OpenStuff();     break;
    case LFUN_HYPHENATION:  HyphenationPoint(); break;
    case LFUN_HFILL:        HFill();         break;
    case LFUN_DEPTH:        DepthCB(ob, 0);  break;
    case LFUN_FREE:         FreeCB(ob, 0);   break;
    case LFUN_TEX:          TexCB(ob, 0);    break;
    case LFUN_FORMULA:      FormulaCB(ob, 0); break;
    case LFUN_FOOTMELT:
      if (buffer[current_buffer]->text
          && !buffer[current_buffer]->text->selection
          && buffer[current_buffer]->text->cursor.par->footnoteflag
             != LYX_NO_FOOTNOTE)
         MeltCB(ob,0);
      else
         FootCB(ob, 0); 
      break;
    case LFUN_MARGINMELT:
      if (buffer[current_buffer]->text
          && !buffer[current_buffer]->text->selection
          && buffer[current_buffer]->text->cursor.par->footnoteflag
             != LYX_NO_FOOTNOTE)
         MeltCB(ob,0);
      else
         MarginCB(ob, 0); 
      break;
      

      // --- buffers ------------------------------------------------------
    case LFUN_PREVBUFFER:
      if (buffer[current_buffer]->text){	  
         BeforeChange();
         Update(-2);
      }
      tmp = current_buffer;
      current_buffer = old_current_buffer;
      old_current_buffer = tmp;
      ResizeCurrentBufferPseudoExpose();
      InitMinibuffer();
      break;

      // --- text style ---------------------------------------------------
      // this is really stupid, anyone has a better idea?...:
    case LFUN_LAYOUT1:   case LFUN_LAYOUT2:
    case LFUN_LAYOUT3:   case LFUN_LAYOUT4:
    case LFUN_LAYOUT5:   case LFUN_LAYOUT6:
    case LFUN_LAYOUT7:   case LFUN_LAYOUT8:
    case LFUN_LAYOUT9:   case LFUN_LAYOUT10:
    case LFUN_LAYOUT11:  case LFUN_LAYOUT12:
    case LFUN_LAYOUT13:  case LFUN_LAYOUT14:
    case LFUN_LAYOUT15:  case LFUN_LAYOUT16:
    case LFUN_LAYOUT17:  case LFUN_LAYOUT18:
    case LFUN_LAYOUT19:  case LFUN_LAYOUT20:
    case LFUN_LAYOUT21:  case LFUN_LAYOUT22:
    case LFUN_LAYOUT23:  case LFUN_LAYOUT24:
    case LFUN_LAYOUT25:  case LFUN_LAYOUT26:
    case LFUN_LAYOUT27:  case LFUN_LAYOUT28:
    case LFUN_LAYOUT29:  case LFUN_LAYOUT30:
    case LFUN_LAYOUT31:  case LFUN_LAYOUT32:
    case LFUN_LAYOUT33:  case LFUN_LAYOUT34:
    case LFUN_LAYOUT35:
      {
         int layoutno = (action - LFUN_LAYOUT1 + 1);
         if(layoutno > 0 && layoutno <= 35)
         {
            if (fl_get_choice(fd_form_main->choice_layouts) != layoutno)
            {
               Update(-2);
               fl_set_choice(fd_form_main->choice_layouts, layoutno);
               buffer[current_buffer]->text->SetLayout(layoutno-1);
            }
         }
      }
      Update(1);
      break;
    case LFUN_EMPH:   EmphCB(ob, 0);  break;
    case LFUN_NOUN:   NounCB(ob, 0);  break;

      // --- Cursor Movements ---------------------------------------------
    case LFUN_RIGHT:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorRight();
      MoveCursorUpdate(0);
      break;
    case LFUN_LEFT:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorLeft();
      MoveCursorUpdate(0);
      break;
    case LFUN_UP:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorUp();
      MoveCursorUpdate(0);
      break;
    case LFUN_DOWN:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorDown();
      MoveCursorUpdate(0);
      break;
    case LFUN_PRIOR:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      CursorPrevious();
      MoveCursorUpdate(0);
      break;
    case LFUN_NEXT:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      CursorNext();
      MoveCursorUpdate(0);
      break;
    case LFUN_HOME:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorHome();
      MoveCursorUpdate(0);
      break;
    case LFUN_END:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorEnd();
      MoveCursorUpdate(0);
      break;
    case LFUN_WORDRIGHT:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorRightOneWord();
      MoveCursorUpdate(0);
      break;
    case LFUN_WORDLEFT:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorLeftOneWord();
      MoveCursorUpdate(0);
      break;
    case LFUN_BEGINNINGBUF:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorTop();
      MoveCursorUpdate(0);
      break;
    case LFUN_ENDBUF:
      if(!buffer[current_buffer]->text->mark_set) BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->CursorBottom();
      MoveCursorUpdate(0);
      break;
    case LFUN_RIGHTSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorRight();
      MoveCursorUpdate(1);
      break;
    case LFUN_LEFTSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorLeft();
      MoveCursorUpdate(1);
      break;
    case LFUN_UPSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorUp();
      MoveCursorUpdate(1);
      break;
    case LFUN_DOWNSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorDown();
      MoveCursorUpdate(1);
      break;
    case LFUN_PRIORSEL:
      Update(-2);
      CursorPrevious();
      MoveCursorUpdate(1);
      break;
    case LFUN_NEXTSEL:
      Update(-2);
      CursorNext();
      MoveCursorUpdate(1);
      break;
    case LFUN_HOMESEL:
      Update(-2);
      buffer[current_buffer]->text->CursorHome();
      MoveCursorUpdate(1);
      break;
    case LFUN_ENDSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorEnd();
      MoveCursorUpdate(1);
      break;
    case LFUN_WORDRIGHTSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorRightOneWord();
      MoveCursorUpdate(1);
      break;
    case LFUN_WORDLEFTSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorLeftOneWord();
      MoveCursorUpdate(1);
      break;
    case LFUN_BEGINNINGBUFSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorTop();
      MoveCursorUpdate(1);
      break;
    case LFUN_ENDBUFSEL:
      Update(-2);
      buffer[current_buffer]->text->CursorBottom();
      MoveCursorUpdate(1);
      break;

      // --- text changing commands ---------------------------------------
    case LFUN_PROTECTEDNEWLINE:
      BeforeChange();
      buffer[current_buffer]->text->InsertChar(LYX_META_NEWLINE);
      SmallUpdate(1);
      SetUpdateTimer();
      break;
    case LFUN_PROTECTEDSPACE:
      BeforeChange();
      buffer[current_buffer]->text->InsertChar(LYX_META_PROTECTED_SEPARATOR);
      SmallUpdate(1);
      SetUpdateTimer();
      break;
    case LFUN_SETMARK:
      if(buffer[current_buffer]->text->mark_set)
      {
         BeforeChange();
         Update(0);
         SetMinibuffer("mark removed","","");
      }
      else
      {
         BeforeChange();
         buffer[current_buffer]->text->mark_set = 1;
         Update(0);
         SetMinibuffer("mark set","","");
      }
      buffer[current_buffer]->text->sel_cursor =
         buffer[current_buffer]->text->cursor;
      break;
    case LFUN_DELETE:
      FreeUpdateTimer();
      if (!buffer[current_buffer]->text->selection) {
         buffer[current_buffer]->text->Delete();
         buffer[current_buffer]->text->sel_cursor =
            buffer[current_buffer]->text->cursor;
         SmallUpdate(1);
      }
      else
      {
         buffer[current_buffer]->text->CutSelection();
         Update(1);
         SetMinibuffer("Cut!","","");
      }
      SetUpdateTimer();
      break;
    case LFUN_BACKSPACE:
      BeforeChange();
      buffer[current_buffer]->text->Backspace();
      SmallUpdate(1);
      SetUpdateTimer();
      buffer[current_buffer]->text->sel_cursor =
         buffer[current_buffer]->text->cursor;
      break;
    case LFUN_NEWLINE:
      BeforeChange();
      buffer[current_buffer]->text->BreakParagraph();
      SmallUpdate(1);
      SetUpdateTimer();
      buffer[current_buffer]->text->sel_cursor =
         buffer[current_buffer]->text->cursor;
      break;
    case LFUN_QUOTE:
      BeforeChange();
      buffer[current_buffer]->text->InsertChar('\"');
      SmallUpdate(1);
      SetUpdateTimer();
      break;

      // --- lyxserver commands -------------------------------------------

    case LFUN_CHARATCURSOR:
      {
         int pos;
         char chr;
         pos = buffer[current_buffer]->text->cursor.pos;
         if(pos < buffer[current_buffer]->text->cursor.par->last)
            chr = buffer[current_buffer]->text->cursor.par->text[pos];
         else
            chr = '\0';
         if(chr)
         {
            dispatch_buffer[0] = chr;
            dispatch_buffer[1] = '\0';
         }
         else
         {
            dispatch_buffer[0] = 'E';
            dispatch_buffer[1] = 'O';
            dispatch_buffer[2] = 'F';
            dispatch_buffer[3] = '\0';
         }
         dispatch_result = dispatch_buffer;
      }
      break;
    case LFUN_GETXY:
      sprintf(dispatch_buffer, "%d %ld",
              buffer[current_buffer]->text->cursor.x,
              buffer[current_buffer]->text->cursor.y);
      dispatch_result = dispatch_buffer;
      break;
    case LFUN_SETXY:
      {
         int  x;
         long y;
         sscanf(argument, " %d %ld", &x, &y);
         buffer[current_buffer]->text->SetCursorFromCoordinates(x, y);
      }
      break;
    case LFUN_GETLAYOUT:
      sprintf(dispatch_buffer, "%d",
              (int)buffer[current_buffer]->text->cursor.par->layout);
      dispatch_result = dispatch_buffer;
      break;
    case LFUN_GETFONT:
      {
         LyXFont *font = &(buffer[current_buffer]->text->current_font);
         if(font->shape == LYX_ITALIC_SHAPE)
            dispatch_buffer[0] = 'E';
         else if(font->shape == LYX_SMALLCAPS_SHAPE)
            dispatch_buffer[0] = 'N';
         else
            dispatch_buffer[0] = '0';
         dispatch_buffer[1] = '\0';
         dispatch_result = dispatch_buffer;
      }
      break;
    case LFUN_GETLATEX:
      {
         LyXFont *font = &(buffer[current_buffer]->text->current_font);
         if(font->latex == LYX_LATEX)
            dispatch_buffer[0] = 'L';
         else if(font->latex == LYX_FORMULA_LATEX)
            dispatch_buffer[0] = 'F';
         else
            dispatch_buffer[0] = '0';
         dispatch_buffer[1] = '\0';
         dispatch_result = dispatch_buffer;
      }
      break;
    case LFUN_GETNAME:
      dispatch_result = buffer[current_buffer]->filename;
      break;
    case LFUN_NOTIFY:
      keyseq.print(dispatch_buffer,100);
      LyxNotifyClient(dispatch_buffer);
      break;

      // --- insert characters --------------------------------------------

    case LFUN_DIAERESIS:
      {
         int l = keyseq.length;
         if(l<0) l = -l;
         KeySym tmp = keyseq.getsym();
         switch(tmp)
         {
          case XK_space:     tmp = XK_diaeresis; break;
          case XK_diaeresis: tmp = XK_diaeresis; break;
          case XK_A: tmp = XK_Adiaeresis; break;
          case XK_E: tmp = XK_Ediaeresis; break;
          case XK_I: tmp = XK_Idiaeresis; break;
          case XK_O: tmp = XK_Odiaeresis; break;
          case XK_U: tmp = XK_Udiaeresis; break;
          case XK_a: tmp = XK_adiaeresis; break;
          case XK_e: tmp = XK_ediaeresis; break;
          case XK_i: tmp = XK_idiaeresis; break;
          case XK_o: tmp = XK_odiaeresis; break;
          case XK_u: tmp = XK_udiaeresis; break;
          case XK_y: tmp = XK_ydiaeresis; break;
          default: return 0;
         }
         buffer[current_buffer]->text->InsertChar((int)tmp);
         SmallUpdate(1);
         SetUpdateTimer();
         buffer[current_buffer]->text->sel_cursor = 
            buffer[current_buffer]->text->cursor;
      }
      break;
    case LFUN_CIRCUMFLEX:
      {
         int l = keyseq.length;
         if(l<0) l = -l;
         KeySym tmp = keyseq.getsym();
         switch(tmp)
         {
          case XK_space:       tmp = XK_asciicircum; break;
          case XK_asciicircum: tmp = XK_asciicircum; break;
          case XK_A: tmp = XK_Acircumflex; break;
          case XK_E: tmp = XK_Ecircumflex; break;
          case XK_I: tmp = XK_Icircumflex; break;
          case XK_O: tmp = XK_Ocircumflex; break;
          case XK_U: tmp = XK_Ucircumflex; break;
          case XK_a: tmp = XK_acircumflex; break;
          case XK_e: tmp = XK_ecircumflex; break;
          case XK_i: tmp = XK_icircumflex; break;
          case XK_o: tmp = XK_ocircumflex; break;
          case XK_u: tmp = XK_ucircumflex; break;
          default: return 0;
         }
         buffer[current_buffer]->text->InsertChar((int)tmp);
         SmallUpdate(1);
         SetUpdateTimer();
         buffer[current_buffer]->text->sel_cursor = 
            buffer[current_buffer]->text->cursor;
      }
      break;
    case LFUN_GRAVE:
      {
         int l = keyseq.length;
         if(l<0) l = -l;
         KeySym tmp = keyseq.getsym();
         switch(tmp)
         {
          case XK_space: tmp = XK_grave; break;
          case XK_grave: tmp = XK_grave; break;
          case XK_A: tmp = XK_Agrave; break;
          case XK_E: tmp = XK_Egrave; break;
          case XK_I: tmp = XK_Igrave; break;
          case XK_O: tmp = XK_Ograve; break;
          case XK_U: tmp = XK_Ugrave; break;
          case XK_a: tmp = XK_agrave; break;
          case XK_e: tmp = XK_egrave; break;
          case XK_i: tmp = XK_igrave; break;
          case XK_o: tmp = XK_ograve; break;
          case XK_u: tmp = XK_ugrave; break;
          default: return 0;
         }
         buffer[current_buffer]->text->InsertChar((int)tmp);
         SmallUpdate(1);
         SetUpdateTimer();
         buffer[current_buffer]->text->sel_cursor = 
            buffer[current_buffer]->text->cursor;
      }
      break;
    case LFUN_ACUTE:
      {
         int l = keyseq.length;
         if(l<0) l = -l;
         KeySym tmp = keyseq.getsym();
         switch(tmp)
         {
          case XK_space: tmp = XK_acute; break;
          case XK_acute: tmp = XK_acute; break;
          case XK_A: tmp = XK_Aacute; break;
          case XK_E: tmp = XK_Eacute; break;
          case XK_I: tmp = XK_Iacute; break;
          case XK_O: tmp = XK_Oacute; break;
          case XK_U: tmp = XK_Uacute; break;
          case XK_a: tmp = XK_aacute; break;
          case XK_e: tmp = XK_eacute; break;
          case XK_i: tmp = XK_iacute; break;
          case XK_o: tmp = XK_oacute; break;
          case XK_u: tmp = XK_uacute; break;
          case XK_y: tmp = XK_yacute; break;
          default: return 0;
         }
         buffer[current_buffer]->text->InsertChar((int)tmp);
         SmallUpdate(1);
         SetUpdateTimer();
         buffer[current_buffer]->text->sel_cursor = 
            buffer[current_buffer]->text->cursor;
      }
      break;
    case LFUN_TILDE:
      {
         int l = keyseq.length;
         if(l<0) l = -l;
         KeySym tmp = keyseq.getsym();
         switch(tmp)
         {
          case XK_space:     tmp = XK_asciitilde; break;
          case XK_asciitilde: tmp = XK_asciitilde; break;
          case XK_A: tmp = XK_Atilde; break;
          case XK_N: tmp = XK_Ntilde; break;
          case XK_O: tmp = XK_Otilde; break;
          case XK_a: tmp = XK_atilde; break;
          case XK_n: tmp = XK_ntilde; break;
          case XK_o: tmp = XK_otilde; break;
          default: return 0;
         }
         buffer[current_buffer]->text->InsertChar((int)tmp);
         SmallUpdate(1);
         SetUpdateTimer();
         buffer[current_buffer]->text->sel_cursor = 
            buffer[current_buffer]->text->cursor;
      }
      break;
    case LFUN_CEDILLA:
      {
         int l = keyseq.length;
         if(l<0) l = -l;
         KeySym tmp = keyseq.getsym();
         switch(tmp)
         {
          case XK_space:   tmp = XK_cedilla; break;
          case XK_cedilla: tmp = XK_cedilla; break;
          case XK_C: tmp = XK_Ccedilla; break;
          case XK_c: tmp = XK_ccedilla; break;
          default: return 0;
         }
         buffer[current_buffer]->text->InsertChar((int)tmp);
         SmallUpdate(1);
         SetUpdateTimer();
         buffer[current_buffer]->text->sel_cursor = 
            buffer[current_buffer]->text->cursor;
      }
      break;
    case LFUN_MACRON:
      {
         int l = keyseq.length;
         if(l<0) l = -l;
         KeySym tmp = keyseq.getsym();
         switch(tmp)
         {
          case XK_space:   tmp = XK_macron; break;
          case XK_macron:  tmp = XK_macron; break;
          default: return 0;
         }
         buffer[current_buffer]->text->InsertChar((int)tmp);
         SmallUpdate(1);
         SetUpdateTimer();
         buffer[current_buffer]->text->sel_cursor = 
            buffer[current_buffer]->text->cursor;
      }
      break;

   case LFUN_INSERT_INSET_LATEX:
     {
       Inset *new_inset = new InsetLatex(argument);
       BeforeChange();
       buffer[current_buffer]->text->InsertInset(new_inset);
       Update(1);
     }
     break;

   case LFUN_INSERT_INSET_LATEX_DEL:
     {
       Inset *new_inset = new InsetLatexDel(argument);
       BeforeChange();
       buffer[current_buffer]->text->InsertInset(new_inset);
       Update(1);
     }
     break;

   case LFUN_INSERT_INSET_FORMULA_LATEX:
     {
       Inset *new_inset = new InsetFormulaLatex(argument);
       BeforeChange();
       buffer[current_buffer]->text->InsertInset(new_inset);
       Update(1);
     }
     break;

   case LFUN_INSERT_INSET_FORMULA_LATEX_DEL:
     {
       Inset *new_inset = new InsetFormulaLatexDel(argument);
       BeforeChange();
       buffer[current_buffer]->text->InsertInset(new_inset);
       Update(1);
     }
     break;

    case LFUN_SELFINSERT:
      if (num_bytes == 1) {
	buffer[current_buffer]->text->InsertChar((int)s_r[0]);
	SmallUpdate(1);
	SetUpdateTimer();
	buffer[current_buffer]->text->sel_cursor =
	  buffer[current_buffer]->text->cursor;
      } else {
	char buf[100];
	LyXBell();
	keyseq.print(buf,100);
	SetMinibuffer("Command cannot be self-inserted: " ,buf,"");
      }
      break;
    default: {
      if(action==-1 && keyseq.length<-1) // unknown key sequence...
      {
         char buf[100];
         LyXBell();
         keyseq.print(buf,100);
         SetMinibuffer("Unknown command:",buf,"");
	 break;
      }
      static char isochar;
      isochar = keyseq.getiso(); // now we call getiso only once
      if(argument ||
         (!(keyevent->state&ControlMask) &&
         !(keyevent->state&Mod1Mask) &&
         (isochar && keysym_return < 0xF000)))
      {
         BeforeChange();
         if(argument)
         {
            char *p;
            for(p = argument; *p; p++)
            {
               /* the quotes */ 
/*             if(*p == '\"')
                  InsertCorrectQuote();
               else
                  buffer[current_buffer]->text->InsertChar(*p); */
		  TranslateAndInsert(*p, buffer[current_buffer]->text);
            }
         }
         else
         {
/*          if(isochar == '\"')
               InsertCorrectQuote();
            else
               buffer[current_buffer]->text->InsertChar(isochar); */
	       TranslateAndInsert(isochar, buffer[current_buffer]->text);
         }

         SmallUpdate(1);
         SetUpdateTimer();
         /* printf("%c  \n",string_return[0]);*/ 
         buffer[current_buffer]->text->sel_cursor =
            buffer[current_buffer]->text->cursor;
         return 0;
      }}
   }

   return 0;
}



/* ---F+------------------------------------------------------------------ *\
    Function  : LyxDefaultKeyBindings
    Called by : init()
    Purpose   : define the default key bindings for LyX
    Parameters: none
    Returns   : nothing
\* ---F------------------------------------------------------------------- */

void LyxDefaultKeyBindings()
{
   kb_sequence k;

   toplevel_keymap = new kb_keymap;
   kb_keymap *map = toplevel_keymap;

#  define BIND(seq, func) \
   if(!k.parse(seq)) map->defkey(&k, func); k.delseq()

   BIND("C-x u", LFUN_MENULOADAUTO);
   BIND("C-x a", LFUN_AUTOSAVE);
   BIND("C-x f", LFUN_MENUOPEN);
   BIND("C-x d", LFUN_MENUNEW);
   BIND("C-x s", LFUN_MENUWRITE);
   BIND("C-x w", LFUN_MENUWRITEAS);
   BIND("C-x c", LFUN_MENUQUIT);
   BIND("C-x k", LFUN_CLOSEBUFFER);
   BIND("C-x r", LFUN_RUNLATEX);
   BIND("C-x t", LFUN_TOCUPDATE);
   BIND("C-x g", LFUN_PREVIEWPS);
   BIND("C-x p", LFUN_PREVIEW);
   BIND("C-x b", LFUN_PREVBUFFER);
   BIND("C-x C-u", LFUN_MENULOADAUTO);
   BIND("C-x C-a", LFUN_AUTOSAVE);
   BIND("C-x C-f", LFUN_MENUOPEN);
   BIND("C-x C-d", LFUN_MENUNEW);
   BIND("C-x C-s", LFUN_MENUWRITE);
   BIND("C-x C-w", LFUN_MENUWRITEAS);
   BIND("C-x C-c", LFUN_MENUQUIT);
   BIND("C-x C-k", LFUN_CLOSEBUFFER);
   BIND("C-x C-r", LFUN_RUNLATEX);
   BIND("C-x C-t", LFUN_RUNLATEX);
   BIND("C-x C-g", LFUN_PREVIEWPS);
   BIND("C-x C-p", LFUN_PREVIEW);
   BIND("C-x C-b", LFUN_PREVBUFFER);

   BIND("F1", LFUN_LAYOUT1);   BIND("F2", LFUN_LAYOUT2);
   BIND("F3", LFUN_LAYOUT3);   BIND("F4", LFUN_LAYOUT4);
   BIND("F5", LFUN_LAYOUT5);   BIND("F6", LFUN_LAYOUT6);
   BIND("F7", LFUN_LAYOUT7);   BIND("F8", LFUN_LAYOUT8);
   BIND("F9", LFUN_LAYOUT9);   BIND("F10", LFUN_LAYOUT10);

   BIND("F11", LFUN_LAYOUT11); BIND("F12", LFUN_LAYOUT12);
   BIND("F13", LFUN_LAYOUT13); BIND("F14", LFUN_LAYOUT14);
   BIND("F15", LFUN_LAYOUT15); BIND("F16", LFUN_LAYOUT16);
   BIND("F17", LFUN_LAYOUT17); BIND("F18", LFUN_LAYOUT18);
   BIND("F19", LFUN_LAYOUT19); BIND("F20", LFUN_LAYOUT20);

   BIND("F21", LFUN_LAYOUT21); BIND("F22", LFUN_LAYOUT22);
   BIND("F23", LFUN_LAYOUT23); BIND("F24", LFUN_LAYOUT24);
   BIND("F25", LFUN_LAYOUT25); BIND("F26", LFUN_LAYOUT26);
   BIND("F27", LFUN_LAYOUT27); BIND("F28", LFUN_LAYOUT28);
   BIND("F29", LFUN_LAYOUT29); BIND("F30", LFUN_LAYOUT30);

   BIND("F31", LFUN_LAYOUT31); BIND("F32", LFUN_LAYOUT32);
   BIND("F33", LFUN_LAYOUT33); BIND("F34", LFUN_LAYOUT34);
   BIND("F35", LFUN_LAYOUT35);

   BIND("Right",   LFUN_RIGHT);
   BIND("C-f",     LFUN_RIGHT);
   BIND("S-Right", LFUN_RIGHTSEL);
   BIND("S-C-F",   LFUN_RIGHTSEL);

   BIND("Left",    LFUN_LEFT);
   BIND("C-b",     LFUN_LEFT);
   BIND("S-Left",  LFUN_LEFTSEL);
   BIND("S-C-B",   LFUN_LEFTSEL);

   BIND("Up",      LFUN_UP);
   BIND("C-Up",    LFUN_UP);
   BIND("C-p",     LFUN_UP);
   BIND("S-Up",    LFUN_UPSEL);
   BIND("S-C-P",   LFUN_UPSEL);

   BIND("Down",    LFUN_DOWN);
   BIND("C-Down",  LFUN_DOWN);
   BIND("C-n",     LFUN_DOWN);
   BIND("S-Down",  LFUN_DOWNSEL);
   BIND("S-C-N",   LFUN_DOWNSEL);

   BIND("C-Right", LFUN_WORDRIGHT);
   BIND("M-f",     LFUN_WORDRIGHT);
   BIND("S-C-Right", LFUN_WORDRIGHTSEL);
   BIND("S-M-F",   LFUN_WORDRIGHTSEL);

   BIND("C-Left",  LFUN_WORDLEFT);
   BIND("M-b",     LFUN_WORDLEFT);
   BIND("S-C-Left",LFUN_WORDLEFTSEL);
   BIND("S-M-B",   LFUN_WORDLEFTSEL);

   BIND("Home",    LFUN_HOME);
   BIND("C-a",     LFUN_HOME);
   BIND("S-Home",  LFUN_HOMESEL);
   BIND("S-C-A",   LFUN_HOMESEL);

   BIND("End",     LFUN_END);
   BIND("C-e",     LFUN_END);
   BIND("S-End",   LFUN_ENDSEL);
   BIND("S-C-E",   LFUN_ENDSEL);

   BIND("Prior",   LFUN_PRIOR);
   BIND("C-Prior", LFUN_PRIOR);
   BIND("M-v",     LFUN_PRIOR);
   BIND("S-Prior", LFUN_PRIORSEL);
   BIND("S-M-V",   LFUN_PRIORSEL);

   BIND("Next",    LFUN_NEXT);
   BIND("C-Next",  LFUN_NEXT);
   BIND("S-Next",  LFUN_NEXTSEL);
   BIND("C-v",     LFUN_NEXT);
   BIND("S-C-V",   LFUN_NEXTSEL);

   BIND("C-Home",  LFUN_BEGINNINGBUF);
   BIND("S-C-Home", LFUN_BEGINNINGBUFSEL);
   //BIND("M-<",     LFUN_BEGINNINGBUF);

   BIND("C-End",   LFUN_ENDBUF);
   BIND("S-C-End", LFUN_ENDBUFSEL);
   //BIND("S-M->",     LFUN_ENDBUF);

   BIND("C-s",     LFUN_MENUSEARCH);

   BIND("C-space", LFUN_SETMARK);
   BIND("C-y",     LFUN_PASTE);
   BIND("C-7",     LFUN_PASTE);
   BIND("C-w",     LFUN_CUT);
   BIND("C-6",     LFUN_CUT);
   BIND("M-w",     LFUN_COPY);

   BIND("C-1",     LFUN_EMPH);
   BIND("C-2",     LFUN_NOUN);
   BIND("C-g",     LFUN_GOTOERROR);
   BIND("C-o",     LFUN_OPENSTUFF);
   BIND("C-h",     LFUN_HYPHENATION);
   BIND("C-i",     LFUN_HFILL);
   BIND("C-0",     LFUN_DEPTH);
   //BIND("Tab",     LFUN_DEPTH); // !!! what about this one?
   BIND("C-3",     LFUN_FREE);
   BIND("C-8",     LFUN_TEX);
   BIND("C-9",     LFUN_FORMULA);
   BIND("C-4",     LFUN_FOOTMELT);
   BIND("C-5",     LFUN_MARGINMELT);

   BIND("Return",  LFUN_NEWLINE);
   BIND("C-Return", LFUN_PROTECTEDNEWLINE);
   BIND("M-space",  LFUN_PROTECTEDSPACE);

   BIND("Delete",  LFUN_DELETE);
   BIND("S-Delete",LFUN_DELETE);
   BIND("C-d",     LFUN_DELETE);
   BIND("S-C-D",   LFUN_DELETE);
   BIND("BackSpace", LFUN_BACKSPACE);
   BIND("C-l",     LFUN_CENTER);

   BIND("C-q",     LFUN_QUOTE);
   BIND("S-C-quotedbl", LFUN_QUOTE);

   // self-insert keys are handled in the 'default:' section of
   // WorkAreaKeyPress - so we don't have to define them all
   // CHANGE: keys explicit decleared as self-insert are now
   // handled seperatly.

#undef BIND

   keyseq.stdmap = keyseq.curmap = toplevel_keymap;
}


/* === End of File: keybind.C ============================================ */
