/* ======================================================================= *\
    File   : keybind.h, $Id: keybind.h,v 1.1 1995/11/05 18:18:58 chb Exp chb $
    Author : chb, 2.Nov.1995
    Docu   : To add a new function:
             - add a new enum constant immediately before LFUN_LASTACTION
             - add an appropriate line to lyx_func_table
             - add a branch to the switch in WorkAreaKeyPress
    Purpose: definitions for LyX keysequences
\* ======================================================================= */

#ifndef _KEYBIND_H_
#define _KEYBIND_H_

#include "kbmap.h"

/* --- external variables ------------------------------------------------ */

extern kb_keymap *toplevel_keymap;

/* --- LyX Edit Actions -------------------------------------------------- */

enum kb_action {
   LFUN_NOACTION = 0,
   LFUN_PREFIX,
   LFUN_PREFIXARG, /* not used now */
   LFUN_CENTER,
   LFUN_MENUNEW,
   LFUN_MENUOPEN,
   LFUN_MENUWRITE,
   LFUN_MENUWRITEAS,
   LFUN_RUNLATEX,
   LFUN_TOCUPDATE,
   LFUN_PREVIEW,
   LFUN_PREVIEWPS,
   LFUN_CLOSEBUFFER,
   LFUN_MENUQUIT,
   LFUN_AUTOSAVE,
   LFUN_MENULOADAUTO,
   LFUN_MENUSEARCH,
   LFUN_PASTE,
   LFUN_CUT,
   LFUN_COPY,
   LFUN_GOTOERROR,
   LFUN_OPENSTUFF,
   LFUN_HYPHENATION,
   LFUN_HFILL,
   LFUN_DEPTH,
   LFUN_FREE,
   LFUN_TEX,
   LFUN_FORMULA,
   LFUN_FOOTMELT,
   LFUN_MARGINMELT,
   LFUN_PREVBUFFER,
   LFUN_LAYOUT1,   LFUN_LAYOUT2,   LFUN_LAYOUT3,   LFUN_LAYOUT4,
   LFUN_LAYOUT5,   LFUN_LAYOUT6,   LFUN_LAYOUT7,   LFUN_LAYOUT8,
   LFUN_LAYOUT9,   LFUN_LAYOUT10,  LFUN_LAYOUT11,  LFUN_LAYOUT12,
   LFUN_LAYOUT13,  LFUN_LAYOUT14,  LFUN_LAYOUT15,  LFUN_LAYOUT16,
   LFUN_LAYOUT17,  LFUN_LAYOUT18,  LFUN_LAYOUT19,  LFUN_LAYOUT20,
   LFUN_LAYOUT21,  LFUN_LAYOUT22,  LFUN_LAYOUT23,  LFUN_LAYOUT24,
   LFUN_LAYOUT25,  LFUN_LAYOUT26,  LFUN_LAYOUT27,  LFUN_LAYOUT28,
   LFUN_LAYOUT29,  LFUN_LAYOUT30,  LFUN_LAYOUT31,  LFUN_LAYOUT32,
   LFUN_LAYOUT33,  LFUN_LAYOUT34,  LFUN_LAYOUT35,
   LFUN_EMPH,
   LFUN_NOUN,
   LFUN_RIGHT,
   LFUN_LEFT,
   LFUN_UP,
   LFUN_DOWN,
   LFUN_PRIOR,
   LFUN_NEXT,
   LFUN_HOME,
   LFUN_END,
   LFUN_WORDRIGHT,
   LFUN_WORDLEFT,
   LFUN_BEGINNINGBUF,
   LFUN_ENDBUF,
   LFUN_RIGHTSEL,
   LFUN_LEFTSEL,
   LFUN_UPSEL,
   LFUN_DOWNSEL,
   LFUN_PRIORSEL,
   LFUN_NEXTSEL,
   LFUN_HOMESEL,
   LFUN_ENDSEL,
   LFUN_WORDRIGHTSEL,
   LFUN_WORDLEFTSEL,
   LFUN_BEGINNINGBUFSEL,
   LFUN_ENDBUFSEL,
   LFUN_PROTECTEDNEWLINE,
   LFUN_PROTECTEDSPACE,
   LFUN_SETMARK,
   LFUN_DELETE,
   LFUN_BACKSPACE,
   LFUN_NEWLINE,
   LFUN_QUOTE,
   LFUN_DIAERESIS,
   LFUN_CIRCUMFLEX,
   LFUN_GRAVE,
   LFUN_ACUTE,
   LFUN_TILDE,
   LFUN_CEDILLA,
   LFUN_MACRON,
   LFUN_SELFINSERT,
   LFUN_GETBUFNAME,
   LFUN_GETXY,
   LFUN_SETXY,
   LFUN_CHARATCURSOR,
   LFUN_LINEATCURSOR,
   LFUN_GETLAYOUT,
   LFUN_GETFONT,
   LFUN_GETLATEX,
   LFUN_GETNAME,
   LFUN_NOTIFY,
   LFUN_INSERT_INSET_LATEX,
   LFUN_INSERT_INSET_LATEX_DEL,
   LFUN_INSERT_INSET_FORMULA_LATEX,
   LFUN_INSERT_INSET_FORMULA_LATEX_DEL,
   LFUN_LASTACTION  /* this marks the end of the table */
};

typedef struct {
   char *name;
   kb_action action;
} kb_func_table;

#ifdef _KEYBIND_C_

kb_func_table lyx_func_table[] = {
   { "",                           LFUN_NOACTION, },
   { "command-prefix",             LFUN_PREFIX, },
   { "prefix-arg",                 LFUN_PREFIXARG, },
   { "recenter",                   LFUN_CENTER, },
   { "new-buffer",                 LFUN_MENUNEW, },
   { "open-buffer",                LFUN_MENUOPEN, },
   { "write-buffer",               LFUN_MENUWRITE, },
   { "write-buffer-as",            LFUN_MENUWRITEAS, },
   { "run-latex",                  LFUN_RUNLATEX, },
   { "toc-update",                 LFUN_TOCUPDATE, },
   { "preview",                    LFUN_PREVIEW, },
   { "preview-ps",                 LFUN_PREVIEWPS, },
   { "close-buffer",               LFUN_CLOSEBUFFER, },
   { "quit-lyx",                   LFUN_MENUQUIT, },
   { "auto-save",                  LFUN_AUTOSAVE, },
   { "load-auto-save",             LFUN_MENULOADAUTO, },
   { "search",                     LFUN_MENUSEARCH, },
   { "paste",                      LFUN_PASTE, },
   { "cut",                        LFUN_CUT, },
   { "copy",                       LFUN_COPY, },
   { "goto-error",                 LFUN_GOTOERROR, },
   { "open-stuff",                 LFUN_OPENSTUFF, },
   { "hyphenation-point",          LFUN_HYPHENATION, },
   { "insert-hfill",               LFUN_HFILL, },
   { "set-depth",                  LFUN_DEPTH, },
   { "set-free",                   LFUN_FREE, },
   { "set-tex",                    LFUN_TEX, },
   { "set-formula",                LFUN_FORMULA, },
   { "set-foot",                   LFUN_FOOTMELT, },
   { "set-margin",                 LFUN_MARGINMELT, },
   { "previous-buffer",            LFUN_PREVBUFFER, },
   { "layout-standard",            LFUN_LAYOUT1, }, // !!! is this ok?
   { "layout-section",             LFUN_LAYOUT2, },
   { "layout-subsection",          LFUN_LAYOUT3, },
   { "layout-subsubsection",       LFUN_LAYOUT4, },
   { "layout-paragraph",           LFUN_LAYOUT5, },
   { "layout-subparagraph",        LFUN_LAYOUT6, },
   { "layout-quotation",           LFUN_LAYOUT7, },
   { "layout-quote",               LFUN_LAYOUT8, },
   { "layout-list",                LFUN_LAYOUT9, },
   { "layout-enumerate",           LFUN_LAYOUT10, },
   { "layout-itemize",             LFUN_LAYOUT11, },
   { "layout-description",         LFUN_LAYOUT12, },
   { "layout-verse",               LFUN_LAYOUT13, },
   { "layout-right-address",       LFUN_LAYOUT14, },
   { "layout-address",             LFUN_LAYOUT15, },
   { "layout-formula",             LFUN_LAYOUT16, },
   { "layout-title",               LFUN_LAYOUT17, },
   { "layout-bibsection",          LFUN_LAYOUT18, },
   { "layout-caption",             LFUN_LAYOUT19, },
   { "layout-section-star",        LFUN_LAYOUT20, },
   { "layout-subsection-star",     LFUN_LAYOUT21, },
   { "layout-subsubsection-star",  LFUN_LAYOUT22, },
   { "layout-latex",               LFUN_LAYOUT23, },
   { "layout-lyxcode",             LFUN_LAYOUT24, },
   { "layout-chapter",             LFUN_LAYOUT25, },
   { "layout-bibchapter",          LFUN_LAYOUT26, },
   { "layout-layout27",            LFUN_LAYOUT27, },
   { "layout-layout28",            LFUN_LAYOUT28, },
   { "layout-layout29",            LFUN_LAYOUT29, },
   { "layout-layout30",            LFUN_LAYOUT30, },
   { "layout-layout31",            LFUN_LAYOUT31, },
   { "layout-layout32",            LFUN_LAYOUT32, },
   { "layout-layout33",            LFUN_LAYOUT33, },
   { "layout-layout34",            LFUN_LAYOUT34, },
   { "layout-layout35",            LFUN_LAYOUT35, },
   { "set-emph",                   LFUN_EMPH, },
   { "set-noun",                   LFUN_NOUN, },
   { "right",                      LFUN_RIGHT, },
   { "left",                       LFUN_LEFT, },
   { "up",                         LFUN_UP, },
   { "down",                       LFUN_DOWN, },
   { "prior",                      LFUN_PRIOR, },
   { "next",                       LFUN_NEXT, },
   { "home",                       LFUN_HOME, },
   { "end",                        LFUN_END, },
   { "word-right",                 LFUN_WORDRIGHT, },
   { "word-left",                  LFUN_WORDLEFT, },
   { "beginning-of-buffer",        LFUN_BEGINNINGBUF, },
   { "end-of-buffer",              LFUN_ENDBUF, },
   { "right-select",               LFUN_RIGHTSEL, },
   { "left-select",                LFUN_LEFTSEL, },
   { "up-select",                  LFUN_UPSEL, },
   { "down-select",                LFUN_DOWNSEL, },
   { "prior-select",               LFUN_PRIORSEL, },
   { "next-select",                LFUN_NEXTSEL, },
   { "home-select",                LFUN_HOMESEL, },
   { "end-select",                 LFUN_ENDSEL, },
   { "word-right-select",          LFUN_WORDRIGHTSEL, },
   { "word-left-select",           LFUN_WORDLEFTSEL, },
   { "beginning-of-buffer-select", LFUN_BEGINNINGBUFSEL, },
   { "end-of-buffer-select",       LFUN_ENDBUFSEL, },
   { "protected-newline",          LFUN_PROTECTEDNEWLINE, },
   { "protected-space",            LFUN_PROTECTEDSPACE, },
   { "set-mark",                   LFUN_SETMARK, },
   { "delete-forward",             LFUN_DELETE, },
   { "delete-backward",            LFUN_BACKSPACE, },
   { "insert-newline",             LFUN_NEWLINE, },
   { "insert-quote",               LFUN_QUOTE, },
   { "accent-diaeresis",           LFUN_DIAERESIS, },
   { "accent-circumflex",          LFUN_CIRCUMFLEX, },
   { "accent-grave",               LFUN_GRAVE, },
   { "accent-acute",               LFUN_ACUTE, },
   { "accent-tilde",               LFUN_TILDE, },
   { "accent-cedilla",             LFUN_CEDILLA, },
   { "accent-macron",              LFUN_MACRON, },
   { "insert-self",                LFUN_SELFINSERT },
   { "get-buffer-name",            LFUN_GETBUFNAME },
   { "char-after",                 LFUN_CHARATCURSOR },
   { "line-at",                    LFUN_LINEATCURSOR },
   { "get-xy",                     LFUN_GETXY },
   { "set-xy",                     LFUN_SETXY },
   { "get-layout",                 LFUN_GETLAYOUT },
   { "get-font",                   LFUN_GETFONT },
   { "get-latex",                  LFUN_GETLATEX },
   { "get-name",                   LFUN_GETNAME },
   { "notify",                     LFUN_NOTIFY },
   { "insert-inset-latex",               LFUN_INSERT_INSET_LATEX },
   { "insert-inset-latex-del",           LFUN_INSERT_INSET_LATEX_DEL },
   { "insert-inset-formula-latex",       LFUN_INSERT_INSET_FORMULA_LATEX },
   { "insert-inset-formula-latex-del",   LFUN_INSERT_INSET_FORMULA_LATEX_DEL },
   { NULL, (kb_action)0 }
};
#else
extern kb_func_table lyx_func_table[];
#endif

/* --- prototypes -------------------------------------------------------- */

char *DispatchFunction(kb_action action, char *argument);

#endif /* _KEYBIND_H_ */

/* === End of File: keybind.h ============================================ */
