/* This file is part of
 * ======================================================
 * 
 *           LyX, the High Level Word Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *
 *======================================================*/

/* 
 * Modified for external style definition by Pascal Andr <andre@via.ecp.fr>
 * 
 * Modified to use binary search and a small pseudo lexical analizer
 * by Alejandro Aguilar Sierra <asierra@servidor.unam.mx>
 */
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include <config.h>
#include <definitions.h>
#include <layout.h>

#define TRUE 1
#define FALSE 0

/* extern */
extern char system_lyxdir [255];
/* global varaibles : textclass tab and number */
static LyXTextClass *lyxstyle;
static number_of_defined_classes = 0;

/*   
 *  The order of the LayoutTags enum and array is important
 *  to the binary search algorithm. If you add a new tag, 
 *  insert it at the proper alphabetical place.
 *  
 *  lyx7p4-d6 Nov 15, 1995 Alejandro Aguilar Sierra
 */

/* tags indexes */
enum _LayoutTags {
  LT_ALIGN, LT_ALIGNPOSSIBLE, LT_BAR, LT_BASELINE, LT_BLOCK, LT_BOLD, 
  LT_BORDER, LT_BOTTOMSEP, LT_CENTER, LT_COLUMNS, LT_COMMAND, 
  LT_COUNTER_CHAPTER, LT_COUNTER_ENUMI, LT_COUNTER_ENUMII, 
  LT_COUNTER_ENUMIII, LT_COUNTER_ENUMIV, LT_COUNTER_PARAGRAPH, 
  LT_COUNTER_SECTION, LT_COUNTER_SUBPARAGRAPH, LT_COUNTER_SUBSECTION, 
  LT_COUNTER_SUBSUBSECTION, LT_DYNAMIC, LT_EMPTY, LT_END, LT_ENDFONT, 
  LT_ENVIRONEMENT, LT_FAMILY, LT_FILL_BOTTOM, LT_FILL_TOP, 
  LT_FIRST_COUNTER, LT_FIRST_DYNAMIC, LT_FONT, LT_FORMULA, 
  LT_FORMULA_LATEX, LT_GIANT, LT_HEADINGS, LT_HUGE, LT_ITALIC, 
  LT_ITEM_ENVIRONEMENT, LT_ITEMSEP, LT_LABELFONT, LT_LABELINDENT, 
  LT_LABELSEP, LT_LABELSTRING, LT_LABELTYPE, LT_LARGE, LT_LARGER, 
  LT_LARGEST, LT_LATEX, LT_LATEXNAME, LT_LATEXTYPE, LT_LAYOUT, 
  LT_LEFT, LT_LEFTBORDER, LT_LIST_ENVIRONEMENT , LT_MANUAL, 
  LT_MAXCOUNTER, LT_MEDIUM, LT_MODERN, LT_NEWLINE, LT_NEXTNOINDENT, 
  LT_NO_BAR, LT_NO_LABEL, LT_NO_LATEX, LT_NORMAL, LT_OPTIONS, LT_PAGESTYLE, 
  LT_PARAGRAPH, LT_PARINDENT, LT_PARSEP, LT_PARSKIP, LT_PLAIN, 
  LT_PROTECTED, LT_RIGHT, LT_RIGHT_ADDRESS_BOX, LT_RIGHTBORDER, 
  LT_ROMAN, LT_SANS, LT_SENSITIVE, LT_SERIES, LT_SHAPE, LT_SIDES, 
  LT_SIZE, LT_SMALL, LT_SMALLCAPS, LT_STATIC, LT_STYLE, LT_TINY, 
  LT_TOP_ENVIRONEMENT, LT_TOPLABELLAYOUT, LT_TOPSEP, LT_TYPEWRITER, LT_UP, 
  LT_UNDER_BAR,
  LT_LAST
};

/* tags strings */
char *LayoutTags[LT_LAST] = {
  "align", "alignpossible", "bar", "baselinestretch", "block", "bold", 
  "border", "bottomsep", "center", "columns", "command", 
  "counter_chapter", "counter_enumi", "counter_enumii", 
  "counter_enumiii", "counter_enumiv", "counter_paragraph", 
  "counter_section", "counter_subparagraph", "counter_subsection", 
  "counter_subsubsection", "dynamic", "empty", "end", "endfont", 
  "environement", "family", "fill_bottom", "fill_top", 
  "first_counter", "first_dynamic", "font", "formula", 
  "formula_latex", "giant", "headings", "huge", "italic", 
  "item_environement", "itemsep", "labelfont", "labelindent", 
  "labelsep", "labelstring", "labeltype", "large", "larger", 
  "largest", "latex", "latexname", "latextype", "layout", "left", 
  "leftborder", "list_environement", "manual", "maxcounter", 
  "medium", "modern", "newline", "nextnoindent", "no_bar", "no_label", 
  "no_latex", "normal", "options", "pagestyle", "paragraph", 
  "parindent", "parsep", "parskip", "plain", "protected", "right", 
  "right_address_box", "rightborder", "roman", "sans", "sensitive", 
  "series", "shape", "sides", "size", "small", "smallcaps", "static", 
  "style", "tiny", "top_environement", "toplabellayout", "topsep", 
  "typewriter", "up", "underbar", 
};

/* gets a tags index */
/* Function modified to use binary searh. 
 * lyx7p4-d6 Nov 15, 95 Alejandro Aguilar Sierra
 */
static int UncodeTag(char *tag)
{
   int m, k, l= 0, r=LT_LAST-1;

   if (!tag) return -2;

   while (l < r) {
      m = (l+r)/2;
      k = strcasecmp(LayoutTags[m], tag);
      if (k==0)
      	return m;
      else
      	if (k<0) l = m+1; else r = m;
   }
   
   return -1;
}

/*
 * Very simple pseudo-lexical analizer to simplify def files parsing.
 * A real lexical analizer should return basic information about the
 * token (is a number, a word).
 * lyx7p4-d6 Nov 15, 95 Alejandro Aguilar Sierra
 */ 
char *next_token(FILE *file)
{
   char c;
   static char buf[200];
   int i;
   
   while (!feof(file)) { 
      c = getc(file);
      if (c=='#')
      	do c = getc(file); while (c!='\n');
      if (c==',')
      	continue;              /* Skip ','s */
      if (c > ' ' && !feof(file))  {
	 i = 0;
	 do {
	    buf[i++] = c;
	    c = getc(file);
	 } while (c > ' ' && !feof(file));
	 buf[i] = '\0';
	 return &buf[0];
      }
   }
   
   return NULL;
}

/* Same as above but stops on end of line */
/* lyx7p4-d6 Nov 15, 95 Alejandro Aguilar Sierra */
char *next_token_line(FILE *file)
{
   char c;
   static char buf[200];
   int i;
   
   while (!feof(file)) { 
      c = getc(file);
      if (c=='#')
      	return NULL;
      
      if (c=='\n')
      	return NULL;
      
      if (c==',')
      	continue;              /* Skip ','s */

      if (c > ' ' && !feof(file))  {
	 i = 0;
	 do {
	    buf[i++] = c;
	    c = getc(file);
	 } while (c > ' ' && !feof(file));
	 buf[i] = '\0';
	 ungetc(c, file);    /* This character could be a '\n' */
	 return &buf[0];
	 
      }
   }
   
   return NULL;
}

/* gets textclass number from name */
char NumberOfClass(const char* textclass) 
{
  int i = 0;
   
  while (i < number_of_defined_classes &&
	 !StringEqual(textclass, lyxstyle[i].name))
    i++;
   
  if (i >= number_of_defined_classes)
    i = 0;
   
  return i;
}

/* gets layout structure from number and textclass number */
LyXLayout *style(char textclass, char layout) 
{
  if (layout < lyxstyle[textclass].number_of_defined_layouts)
    return &lyxstyle[textclass].style[layout];
  else {
    return &lyxstyle[textclass].style[0];
  }
}

/* gets layout number from name and textclass number */
char NumberOfLayout(char textclass, const char* name) 
{
   int i = 0;

   while (i < lyxstyle[textclass].number_of_defined_layouts &&
	  !StringEqual(name, lyxstyle[textclass].style[i].name))
      i++;
   
   if (i >= lyxstyle[textclass].number_of_defined_layouts) {
      if (StringEqual(name, "dummy"))
      	 i = LYX_DUMMY_LAYOUT;
      else
      	 i = 0;
   }

   return i;
}

/* gets a layout (style) name from number and textclass number */
char* NameOfLayout(char textclass, char layout) 
{
  if (layout < lyxstyle[textclass].number_of_defined_layouts)
    return lyxstyle[textclass].style[layout].name;
  else
    return "dummy";
}

/* gets a textclass name from number */
char* NameOfClass(char number) 
{
  if (number < number_of_defined_classes)
    return lyxstyle[number].name;
  else
    return "dummy";
}

/* gets a textclass structure from number */
LyXTextClass * TextClass(char textclass) 
{
  if (textclass < number_of_defined_classes)
    return &lyxstyle[textclass];
  else
    return &lyxstyle[0];
}

/* gets a font from given file */
static int ReadFont(FILE *file, LyXFont& font)
{
  int family = LYX_ROMAN_FAMILY;
  int series = LYX_MEDIUM_SERIES;
  int shape = LYX_UP_SHAPE;
  int size = LYX_SIZE_NORMAL;
  int latex = LYX_NO_LATEX;
  int bar = LYX_NO_BAR;
  char buffer[256];
  char *temp;
  int error = 0;

  while (!feof(file) && !error) {
     temp = next_token(file);
     switch(UncodeTag(temp)) {

      case -2:                                    /* empty line */
	break;

      case -1:                                    /* parse error */
	fprintf(stderr, "Parse Error : Unknown tag %s.\n", temp);
	error = 1;
	break;

      case LT_ENDFONT:                            /* end of structure */
	error = -1;
	break;

      case LT_FAMILY:                             /* font family definition */
	temp = next_token(file);
	switch (UncodeTag(temp)) {
	case LT_ROMAN:
	  family = LYX_ROMAN_FAMILY;
	  break;
	case LT_SANS:
	  family = LYX_SANS_FAMILY;
	  break;
	case LT_TYPEWRITER:
	  family = LYX_TYPEWRITER_FAMILY;
	  break;
	case LT_MODERN:
	  family = LYX_MODERN_FAMILY;
	  break;
	}
	break;

      case LT_SERIES:                             /* font series definition */
	temp = next_token(file);
	switch (UncodeTag(temp)) {
	case LT_MEDIUM:
	  series = LYX_MEDIUM_SERIES;
	  break;
	case LT_BOLD:
	  series = LYX_BOLD_SERIES;
	  break;
	}
	break;

      case LT_SHAPE:                             /* font shape definition */
	temp = next_token(file);
	switch (UncodeTag(temp)) {
	case LT_UP:
	  shape = LYX_UP_SHAPE;
	  break;
	case LT_ITALIC:
	  shape = LYX_ITALIC_SHAPE;
	  break;
	case LT_SMALLCAPS:
	  shape = LYX_SMALLCAPS_SHAPE;
	  break;
	}
	break;

      case LT_SIZE:                              /* font size definition */
	temp = next_token(file);
	switch (UncodeTag(temp)) {
	case LT_TINY:
	  size = LYX_SIZE_TINY;
	  break;
	case LT_SMALL:
	  size = LYX_SIZE_SMALL;
	  break;
	case LT_NORMAL:
	  size = LYX_SIZE_NORMAL;
	  break;
	case LT_LARGE:
	  size = LYX_SIZE_LARGE;
	  break;
	case LT_LARGER:
	  size = LYX_SIZE_LARGER;
	  break;
	case LT_LARGEST:
	  size = LYX_SIZE_LARGEST;
	  break;
	case LT_HUGE:
	  size = LYX_SIZE_HUGE;
	  break;
	case LT_GIANT:
	  size = LYX_SIZE_GIANT;
	  break;
	}
	break;

      case LT_LATEX:                             /* font latex flag */
	temp = next_token(file);
	switch (UncodeTag(temp)) {
	case LT_NO_LATEX:
	  latex = LYX_NO_LATEX;
	  break;
	case LT_LATEX:
	  latex = LYX_LATEX;
	  break;
	case LT_FORMULA:
	  latex = LYX_FORMULA_LATEX;
	  break;
	case LT_PROTECTED:
	  latex = LYX_PROTECTED_LATEX;
	  break;
	case LT_FORMULA_LATEX:
	  latex = LYX_PROTECTED_FORMULA_LATEX;
	  break;
	}
	break;

     case LT_BAR:                             /* font bar definition */
	temp = next_token(file);
	switch (UncodeTag(temp)) {
	case LT_NO_BAR:
	  bar = LYX_NO_BAR;
	  break;
	case LT_UNDER_BAR:
	  bar = LYX_UNDER_BAR;
	  break;
	}
	break;

      default:                                    /* context error */
	fprintf(stderr, "Context Error : Tag %s not allowed in font.\n", temp);
	error = 1;
	break;
     }
  }

   if (error == -1) error = 0;
   if (!error) font = GetLyXFont(family, series, shape, size, latex, bar);
   
   return error;
}

/* reads a layout definition from file */
static int ReadLayout(FILE *file, LyXLayout& layout)
{
  char buffer[256];
  char *temp;
  int error=0;
  static LyXFont dummy_font = GetLyXFont(LYX_ROMAN_FAMILY, LYX_MEDIUM_SERIES,
					 LYX_UP_SHAPE, LYX_SIZE_NORMAL, 
					 LYX_NO_LATEX, LYX_NO_BAR);

  /* structure initialisation */
  layout.bordertype = BORDER_STATIC;
  layout.latextype = LATEX_PARAGRAPH;
  layout.latexname = "";
  layout.font = dummy_font;
  layout.labelfont = dummy_font;
  layout.nextnoindent = 0;
  layout.leftborder = "";
  layout.rightborder = "";
  layout.labelsep = "";
  layout.labelindent = "";
  layout.parindent = "";
  layout.parskip = 0.0;
  layout.itemsep = 0;
  layout.topsep = 0.0;
  layout.bottomsep = 0.0;
  layout.parsep = 0;
  layout.align = LYX_ALIGN_BLOCK;
  layout.alignpossible = LYX_ALIGN_BLOCK;
  layout.labeltype = LABEL_NO_LABEL;
  layout.baselinestretch = 0.0;
  layout.labelstring = "";
  layout.fill_top = 0;
  layout.fill_bottom = 0;
  layout.newline_allowed = 1;

  /* parse style section */
  while (!feof(file) && !error) {
      temp= next_token(file);
      switch(UncodeTag(temp)) {

      case -2:
	break;

      case -1:                                    /* parse error */
	fprintf(stderr, "Parse Error : Unknown tag %s.\n", temp);
	error = 1;
	break;

      case LT_END:                                /* end of structure */
	error = -1;
	break;

      case LT_BORDER:                             /* border style definition */
	temp= next_token(file);
	switch(UncodeTag(temp)) {
	case LT_STATIC:
	  layout.bordertype = BORDER_STATIC;
	  break;
	case LT_MANUAL:
	  layout.bordertype = BORDER_MANUAL;
	  break;
	case LT_DYNAMIC:
	  layout.bordertype = BORDER_DYNAMIC;
	  break;
	case LT_FIRST_DYNAMIC:
	  layout.bordertype = BORDER_FIRST_DYNAMIC;
	  break;
	case LT_RIGHT_ADDRESS_BOX:
	  layout.bordertype = BORDER_RIGHT_ADDRESS_BOX;
	  break;
	default:
	  layout.bordertype = 0;
	  break;
	}
	break;

      case LT_LATEXTYPE:                          /* latex style definition */
	temp= next_token(file);
	switch (UncodeTag(temp)) {
	case LT_PARAGRAPH:
	  layout.latextype=LATEX_PARAGRAPH;
	  break;
	case LT_COMMAND:
	  layout.latextype=LATEX_COMMAND;
	  break;
	case LT_ENVIRONEMENT:
	  layout.latextype=LATEX_ENVIRONEMENT;
	  break;
	case LT_ITEM_ENVIRONEMENT:
	  layout.latextype=LATEX_ITEM_ENVIRONEMENT;
	  break;
	case LT_LIST_ENVIRONEMENT:
	  layout.latextype=LATEX_LIST_ENVIRONEMENT;
	  break;
	default:
	  layout.latextype=0;
	  break;
	}
	break;

      case LT_FONT:
	ReadFont(file, layout.font);
	break;

      case LT_LABELFONT:
	ReadFont(file, layout.labelfont);
	break;

      case LT_NEXTNOINDENT:                       /* indent next paragraph ? */
	if ((temp= next_token(file)) && (atoi(temp)))
	  layout.nextnoindent = TRUE;
	else
	  layout.nextnoindent = FALSE;
	break;

      case LT_LATEXNAME:                          /* latex name */
	if ((temp= next_token(file))!=NULL)
	  layout.latexname = strdup(temp);
	break;

      case LT_LABELTYPE:                            /* label type */
	temp= next_token(file);
	switch (UncodeTag(temp)) {
	case LT_NO_LABEL:
	  layout.labeltype = LABEL_NO_LABEL;
	  break;
	case LT_MANUAL:
	  layout.labeltype = LABEL_MANUAL;
	  break;
	case LT_TOP_ENVIRONEMENT:
	  layout.labeltype = LABEL_TOP_ENVIRONMENT;
	  break;
	case LT_STATIC:
	  layout.labeltype = LABEL_STATIC;
	  break;
	case LT_SENSITIVE:
	  layout.labeltype = LABEL_SENSITIVE;
	  break;
	case LT_COUNTER_CHAPTER:
	  layout.labeltype = LABEL_COUNTER_CHAPTER;
	  break;
	case LT_COUNTER_SECTION:
	  layout.labeltype = LABEL_COUNTER_SECTION;
	  break;
	case LT_COUNTER_SUBSECTION:
	  layout.labeltype = LABEL_COUNTER_SUBSECTION;
	  break;
	case LT_COUNTER_SUBSUBSECTION:
	  layout.labeltype = LABEL_COUNTER_SUBSUBSECTION;
	  break;
	case LT_COUNTER_PARAGRAPH:
	  layout.labeltype = LABEL_COUNTER_PARAGRAPH;
	  break;
	case LT_COUNTER_SUBPARAGRAPH:
	  layout.labeltype = LABEL_COUNTER_SUBPARAGRAPH;
	  break;
	case LT_COUNTER_ENUMI:
	  layout.labeltype = LABEL_COUNTER_ENUMI;
	  break;
	case LT_COUNTER_ENUMII:
	  layout.labeltype = LABEL_COUNTER_ENUMII;
	  break;
	case LT_COUNTER_ENUMIII:
	  layout.labeltype = LABEL_COUNTER_ENUMIII;
	  break;
	case LT_COUNTER_ENUMIV:
	  layout.labeltype = LABEL_COUNTER_ENUMIV;
	  break;
	}
	break;

      case LT_LEFTBORDER:                         /* left border type */
	if ((temp= next_token(file))!=NULL)
	  layout.leftborder = strdup(temp);
	break;

      case LT_RIGHTBORDER:                        /* right border type */
	if ((temp= next_token(file))!=NULL)
	  layout.rightborder = strdup(temp);
	break;

      case LT_LABELINDENT:                        /* label indenting flag */
	if ((temp= next_token(file))!=NULL)
	  layout.labelindent = strdup(temp);
	break;

      case LT_PARINDENT:                          /* paragraph indent. flag */
	if ((temp= next_token(file))!=NULL)
	layout.parindent = strdup(temp);
	break;

      case LT_PARSKIP:                            /* paragraph skip size */
	if ((temp= next_token(file))!=NULL)
	  layout.parskip = atof(temp);
	break;

      case LT_ITEMSEP:                            /* item separation size */
	if ((temp= next_token(file))!=NULL)
	  layout.itemsep = atof(temp);
	break;

      case LT_TOPSEP:                             /* top separation size */
	if ((temp= next_token(file))!=NULL)
	  layout.topsep = atof(temp);
	break;

      case LT_BOTTOMSEP:                          /* bottom separation size */
	if ((temp= next_token(file))!=NULL)
	  layout.bottomsep = atof(temp);
	break;

      case LT_LABELSEP:                           /* label separator */
	if ((temp= next_token(file))!=NULL) {
	  layout.labelsep = strdup(temp);
	  while ((temp=index(layout.labelsep, 'x'))!=NULL) 
	    *temp=' ';
	}
	break;

      case LT_PARSEP:                             /* par. separation size */
	if ((temp= next_token(file))!=NULL)
	  layout.parsep = atof(temp);
	break;

      case LT_FILL_TOP:                             /* fill top flag */
	if ((temp= next_token(file))!=NULL)
	  layout.fill_top = atoi(temp);
	break;

      case LT_FILL_BOTTOM:                          /* fill bottom flag */
	if ((temp= next_token(file))!=NULL)
	  layout.fill_bottom = atoi(temp);
	break;

      case LT_NEWLINE:                            /* newlines allowed ? */
	if ((temp= next_token(file))!=NULL)
	  layout.newline_allowed = atoi(temp);
	break;

      case LT_ALIGN:                              /* paragraph align */
	temp= next_token(file);
	switch (UncodeTag(temp)) {
	case LT_BLOCK:
	  layout.align = LYX_ALIGN_BLOCK;
	  break;
	case LT_LEFT:
	  layout.align = LYX_ALIGN_LEFT;
	  break;
	case LT_RIGHT:
	  layout.align = LYX_ALIGN_RIGHT;
	  break;
	case LT_CENTER:
	  layout.align = LYX_ALIGN_CENTER;
	  break;
	case LT_FORMULA:
	  layout.align = LYX_ALIGN_FORMULA;
	  break;
	case LT_LAYOUT:
	  layout.align = LYX_ALIGN_LAYOUT;
	  break;
	}
	break;

      case LT_ALIGNPOSSIBLE:                      /* paragraph allowed align */
	layout.alignpossible = 0;
       
	while ((temp= next_token_line(file))!=NULL)
	  switch (UncodeTag(temp)) {
	  case LT_BLOCK:
	    layout.alignpossible |= LYX_ALIGN_BLOCK;
	    break;
	  case LT_LEFT:
	    layout.alignpossible |= LYX_ALIGN_LEFT;
	    break;
	  case LT_RIGHT:
	    layout.alignpossible |= LYX_ALIGN_RIGHT;
	    break;
	  case LT_CENTER:
	    layout.alignpossible |= LYX_ALIGN_CENTER;
	    break;
	  case LT_FORMULA:
	    layout.alignpossible |= LYX_ALIGN_FORMULA;
	    break;
	  case LT_LAYOUT:
	    layout.alignpossible |= LYX_ALIGN_LAYOUT;
	    break;
	  }

	break;

      case LT_BASELINE:                           /* base line stretch */
	if ((temp= next_token(file))!=NULL)
	  layout.baselinestretch = atof(temp);
	break;

      case LT_LABELSTRING:                        /* label string definition */
	if ((temp= next_token(file))!=NULL)
	  layout.labelstring = strdup(temp);
	break;

      default:                                    /* context error */
	fprintf(stderr, "Context Error : Tag %s not allowed in layout.\n", 
		temp);
	error = 1;
	break;
      }
  }
  

  if (error== -1) error = 0;
  return error;
}

/* reads a textclass structure from file */
static int ReadTextClass(LyXTextClass& textclass, char *filename)
{
  FILE *file;
  char buffer[256];
  char *temp;
  int error = 0;
  int layout = 0;

  if ((file = fopen(filename, "r"))==NULL) return -2; 

  /* structure initialisation */
  textclass.style = (LyXLayout *)calloc(100, sizeof(LyXLayout));
  textclass.options = "";
  textclass.columns = 1;
  textclass.sides = 1;
  textclass.pagestyle = PAGESTYLE_PLAIN;
  textclass.maxcounter = LABEL_COUNTER_CHAPTER;
  textclass.top_label_layout = "";

  /* parsing */
  while (!feof(file) && !error) {
     temp= next_token(file);     

     switch(UncodeTag(temp)) {
      case -2:
	break;
	
      case -1:                                 
	fprintf(stderr, "Parse Error : Unknown tag %s.\n", temp);
	error = 1;
	break;

      case LT_STYLE:
	if ((temp= next_token(file))!=NULL) {
	   textclass.style[layout].name = strdup(temp);
	   while ((temp=index(textclass.style[layout].name, '_')) != NULL)
	     *temp = ' ';
	  if (!ReadLayout(file, textclass.style[layout])) 
	    layout++;
	  else {
	    fprintf(stderr, "Parse Error : Error parsing style %s.\n", 
		    textclass.style[layout].name);
	    error = 1;
	  }
	}
	else {
	  fprintf(stderr, "Parse Error : No name given for style.\n");
	   error = 1;
	}
	break;

      case LT_OPTIONS:
	if ((temp= next_token(file))!=NULL)
	  textclass.options = strdup(temp);
	break;

      case LT_COLUMNS:
	if ((temp= next_token(file))!=NULL)
	  textclass.columns = atoi(temp);
	break;

      case LT_SIDES:
	if ((temp= next_token(file))!=NULL)
	  textclass.sides = atoi(temp);
	break;

      case LT_PAGESTYLE:
	temp= next_token(file);
	switch (UncodeTag(temp)) {
	case LT_PLAIN:
	   textclass.pagestyle = PAGESTYLE_PLAIN;
	   break;
	 case LT_HEADINGS:
	   textclass.pagestyle = PAGESTYLE_HEADINGS;
	   break;
	 case LT_EMPTY:
	   textclass.pagestyle = PAGESTYLE_EMPTY;
	   break;
	}
	break;

      case LT_MAXCOUNTER:
	temp= next_token(file);
	switch (UncodeTag(temp)) {
	 case LT_COUNTER_CHAPTER:
	   textclass.maxcounter = LABEL_COUNTER_CHAPTER;
	   break;
	 case LT_COUNTER_SECTION:
	   textclass.maxcounter = LABEL_COUNTER_SECTION;
	   break;
	 case LT_COUNTER_SUBSECTION:
	   textclass.maxcounter = LABEL_COUNTER_SUBSECTION;
	   break;
	 case LT_COUNTER_SUBSUBSECTION:
	   textclass.maxcounter = LABEL_COUNTER_SUBSUBSECTION;
	   break;
	 case LT_COUNTER_PARAGRAPH:
	   textclass.maxcounter = LABEL_COUNTER_PARAGRAPH;
	   break;
	 case LT_COUNTER_SUBPARAGRAPH:
	   textclass.maxcounter = LABEL_COUNTER_SUBPARAGRAPH;
	   break;
	 case LT_COUNTER_ENUMI:
	   textclass.maxcounter = LABEL_COUNTER_ENUMI;
	   break;
	 case LT_COUNTER_ENUMII:
	   textclass.maxcounter = LABEL_COUNTER_ENUMII;
	   break;
	 case LT_COUNTER_ENUMIII:
	   textclass.maxcounter = LABEL_COUNTER_ENUMIII;
	   break;
	 case LT_COUNTER_ENUMIV:
	   textclass.maxcounter = LABEL_COUNTER_ENUMIV;
	   break;
	}
	break;

      case LT_TOPLABELLAYOUT:
	if ((temp= next_token(file))!=NULL)
	  textclass.top_label_layout = strdup(temp);
	break;
	
      default:
	fprintf(stderr, "Parse Error : Out of context tag %s.\n", temp);
	break;
     }
  }
   

  if (error==-2) error = 0;

  if (error) {
    free(textclass.style);
    textclass.number_of_defined_layouts = 0;
  }
  else {
    textclass.style = (LyXLayout *)realloc(textclass.style, 
					   layout*sizeof(LyXLayout));
    textclass.number_of_defined_layouts = layout;
  }

  fclose(file);

  return error;
}

/* gets LyX textclass config from file */
static int GetLyXTextClass(char *path)
{
  char real_file[256];
  char buffer[256];
  char *temp;
  FILE *file;
  int error = 0;
  int textclass = 0;
  char tmppath[256];

  strcpy(tmppath, path);

  // first try cwd, then LYX_DIR/layouts, then cwd/layouts
  strcpy(real_file, tmppath);
  strcat(real_file, "/textclass.def");
  if (!(file = fopen(real_file, "r"))) {
    fprintf(stderr,"GetLyXTextClass: unable to open `%s',\n", real_file);
    strcpy(real_file, system_lyxdir);
    strcat (real_file, "/layouts/textclass.def");
    fprintf(stderr,"GetLyXTextClass: trying `%s' instead.\n", real_file);
    strcpy(tmppath, system_lyxdir);
    strcat (tmppath, "/layouts/");
    if (!(file = fopen(real_file, "r"))) {
      fprintf(stderr,"GetLyXTextClass: unable to open `%s',\n", real_file);
      strcpy(tmppath, path);
      strcat(tmppath, "/layouts/");
      sprintf(real_file, "%s/textclass.def", tmppath);
      fprintf(stderr,"GetLyXTextClass: trying `%s' instead.\n", real_file);
      if (!(file = fopen(real_file, "r"))) {
	fprintf(stderr,"GetLyXTextClass: unable to open `%s'. Exiting.\n",
		real_file);
	return -1;
      }
    }
  }

  strcpy(path, tmppath);

  if ((lyxstyle = (LyXTextClass *)calloc(100, sizeof(LyXTextClass)))==NULL) {
    fprintf(stderr,"GetLyXTextClass: calloc: out of memory!\n");
    return 1;
  }

  while (!feof(file) && !error) {
    bzero(buffer, sizeof(buffer));
    if ((fgets(buffer, sizeof(buffer), file)!=NULL) && (buffer[0]!='#')) {
      while ((temp=index(buffer, '\n'))!=NULL) 
	*temp=' ';

      /* get rid of comments */
      if ((temp = index(buffer, '#'))!=NULL) 
	*temp=0;

      temp = strtok(buffer, " \t");
      /* ignore blank and whitespace-only lines */
      if (temp==NULL) continue;
      lyxstyle[textclass].name = strdup(temp);
      temp = strtok(NULL, " \t");
      if (temp==NULL) error = 1;
      else { 
	strcpy(real_file, path);
	strcat(real_file, temp);
	if (ReadTextClass(lyxstyle[textclass], real_file)) {
	   fprintf(stderr,"GetLyXTextClass: error from ReadTextClass\n");
	   error = 1; 
	}
	else textclass++;
      }
    }
  }

  if (error) {
    number_of_defined_classes = 0;
    free(lyxstyle);
  }
  else {
    number_of_defined_classes = textclass;
    lyxstyle = (LyXTextClass *) realloc(lyxstyle, 
					textclass*sizeof(LyXTextClass));
  }

  fclose(file);
  return error;
};


void LyXSetStyle(char *path) 
{
  char real_path[256];
  int e;
  strcpy(real_path, path);
  strcat(real_path, "/");

  fprintf(stderr, "LyXSetStyle : parsing configuration...\n");
  if ((e=GetLyXTextClass(real_path))!=0) {
    fprintf(stderr, "LyXSetStyle : an error occured during parsing.\n"
	            "              Exiting. [%d]\n", e);
    exit(1);
  }
  fprintf(stderr, "LyXSetStyle : configuration parsed.\n");
}
