/* This file is part of
* ======================================================
* 
*           LyX, the High Level Word Processor
* 	 
*	    Copyright (C) 1995 Matthias Ettrich
*
*======================================================*/
#include <ctype.h>
#include <unistd.h>
#include <signal.h>

// why do these includes use <> instead of "" ????
// <> is giving the wrong message!
#include <config.h>
#include <file.h>
#include <lyx.h>
#include <spellchecker.h>
#include <lyxserver.h>
#include <version.h>
#include "lyx_cb.h"
#include "docu.h"
#include "credits.h"
#include "intl.h"

extern void show_symbols_form(void);
extern FD_form_main *fd_form_main;
extern FD_form_title *fd_form_title;
extern FD_form_paragraph *fd_form_paragraph;
extern FD_form_search *fd_form_search;
extern FD_form_character *fd_form_character;
extern FD_form_document *fd_form_document;
extern FD_form_quotes *fd_form_quotes;
extern FD_form_preamble *fd_form_preamble;
extern FD_form_table *fd_form_table;
extern FD_form_print *fd_form_print;
extern FD_form_figure *fd_form_figure;
extern FD_form_screen *fd_form_screen;
extern FD_form_toc *fd_form_toc;
extern FD_form_ref *fd_form_ref;

extern void ShowFileMenu(FL_OBJECT*,long);
extern void ShowEditMenu(FL_OBJECT*,long);
extern void ShowLayoutMenu(FL_OBJECT*,long);
extern void ShowSpecialMenu(FL_OBJECT*,long);
extern void ShowOptionsMenu(FL_OBJECT*,long);
extern void ShowMathMenu(FL_OBJECT*,long);
extern void ShowBufferMenu(FL_OBJECT*,long);
extern void ShowHelpMenu(FL_OBJECT*,long);
extern void ShowNoMenu(FL_OBJECT*, long);

extern char* roman_font_name;
extern char* sans_font_name;
extern char* typewriter_font_name;
extern char* font_norm;
extern char font_sizes[];

// can't be static since keybind.C wants it.
int work_area_height = 0;
static int work_area_width = 0;

static long current_scrollbar_value = 0;
static int current_layout = 0;
static char current_textclass = 255;

LyXRC lyxrc;

char system_lyxdir [255];
char *document_path = NULL;
static char ascii_type; /* for selection notify callbacks */

Buffer *buffer[NUMBER_OF_BUFFERS];
int current_buffer = 0;
int old_current_buffer = 0;
void InsertCorrectQuote();

char work_area_active = True;	       /* boolean flag  */
char scrolling = False;	       /* boolean flag  */

char updatetimer = 0;
char input_prohibited = False;   /* wether the work area should get callbacks */ 

char selection_possible = False;		  /* just a boolean flag */
/* the selection possible is needed, that only motion events are 
* used, where the bottom press event was on the drawing area too */


/* the cursor */
Cursor xterm_cursor;

/* some function prototypes */ 
int WorkAreaButtonPress(FL_OBJECT *ob,
			Window win,
			int w, int h,
			XEvent *ev, void *d);
int WorkAreaButtonRelease(FL_OBJECT *ob,
			  Window win,
			  int w, int h,
			  XEvent *ev, void *d);
extern int WorkAreaKeyPress(FL_OBJECT *ob,
			    Window win,
			    int w, int h,
			    XEvent *ev, void *d);
int WorkAreaMotionNotify(FL_OBJECT *ob,
			 Window win,
			 int w, int h,
			 XEvent *ev, void *d);
extern void LyxDefaultKeyBindings();
int ResizeCurrentBuffer(FL_OBJECT*,
			Window*,
			int, int, XEvent*, void*);
void ResizeCurrentBufferPseudoExpose();
void InitTextParameters(LyXTextParameters &p);
void GotoError();
void GotoNote();
void OpenStuff();
void HyphenationPoint();
void HFill();
void Newline();
void ProtectedBlank();
void CopyCB();
void LoadLyXFile(const char*);
// LyXBell could perhaps be inlined?
void LyXBell() {
   XBell(fl_display, 20);
}
int RunLaTeX();
int RunLinuxDoc(int, const char *);
void MenuWrite(int buf);
void CloseBuffer(); 
void MenuPreview();
void MenuPreviewPS();
void MenuLayoutCharacter();
void MenuLayoutParagraph();
void MenuLayoutDocument();
void MenuLayoutQuotes();
void MenuLayoutPreamble();
void UpdateMenuBar();

void error_handler(int err_sig);

static char minibuffer_text[200] = "Welcome to LyX!";

const char *GetLanguage()
{
   return (const char*)buffer[current_buffer]->parameters.language;
}

const char *LastFiles:: GetFileName(int nr)
{
   return (const char*)files[nr-1];
}

void LastFiles::ReadFile(const char *filename) 
{
   FILE *fd;
   if ((fd=fopen(filename,"r"))) 
     {
       char *inn=new char(80);
       // some more error checking is needed
       inn[0]=0;
       fscanf(fd,"file1 %s\n",inn);
       if (inn[0])
	 files[0] = StringCopy(inn);
       inn[0]=0;
       fscanf(fd,"file2 %s\n",inn);
       if (inn[0])
	 files[1] = StringCopy(inn);
       inn[0]=0;
       fscanf(fd,"file3 %s\n",inn);
       if (inn[0])
	 files[2] = StringCopy(inn);
       inn[0]=0;
       fscanf(fd,"file4 %s\n",inn);
       if (inn[0])
	 files[3] = StringCopy(inn);
       fclose(fd); // is a free needed when fclose has been used?
       delete inn;
     } else fprintf(stderr,"Unable to read LastFiles: %s\n",filename);
}
void LastFiles::WriteFile(const char *filename) 
{
   FILE *fd;
   if ((fd=fopen(filename,"w"))) 
      {
	 if (files[0])
	   fprintf(fd,"file1 %s\n",files[0]);
	 if (files[1])
	   fprintf(fd,"file2 %s\n",files[1]);
	 if (files[2])
	   fprintf(fd,"file3 %s\n",files[2]);
	 if (files[3])
	   fprintf(fd,"file4 %s\n",files[3]);
	 fclose(fd);
      } else
	 fprintf(stderr,"Warning: unable to print LastFiles `%s'!\n",filename);
}
   
LastFiles::LastFiles(const char *filename) 
{
   int n=0;
   for(n=0;n<4;n++) files[n]= NULL;
   ReadFile(filename);
}

void LastFiles::newfile(const char *file) 
{
   int n=0,exists=0;
   for(n=0;n<=3;n++) if (StringEqual(file,files[n])) exists=1;

   // the files are like a stack with a hole in the bottom
   if (exists==0) {   
      delete files[3];
      files[3]=files[2];
      files[2]=files[1];
      files[1]=files[0];
      files[0]=StringCopy(file);
   }
}

void LastFiles::MakeMenu(int m) 
{
   int n;
   for (n=0;n<=3;n++) {
     if (files[n]) {
	char tmp[strlen(files[n])+4];
	sprintf(tmp,"%s%%x%d",files[n],n+31);
	//fprintf(stderr,"adding to menu: %s\n",tmp);
	fl_addtopup(m,tmp);
     }
     
     // fprintf(stderr,"file%d %s\n",n,files[n]);
   }
}

LastFiles *lastfiles;

int MinibufferExpose(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
   static int width = 0;
   static int height = 0;
   static FL_OBJECT *ob2 = NULL;
   static Window win2 = 0;
   static Font fixed_font =  XLoadQueryFont(fl_display, "fixed")->fid;
   static GC gc = LyXGetCopyGC();

   if (w && h) {
      width = w;
      height = h;
      ob2 = ob;
      win2 = win;
   }
   if (!win2)
      return 0;
   
   XFillRectangle(fl_display, win2, LyXGetLightedGC(), 0, 0, width, height);
   XFlush(fl_display);
   
   int i=0;
   for (i=0;minibuffer_text[i]; i++);
   LyXFont font = GetLyXFont(LYX_SANS_FAMILY, 
			     LYX_MEDIUM_SERIES,
			     LYX_UP_SHAPE,
			     LYX_SIZE_SMALL,
			     LYX_NO_LATEX,
			     LYX_NO_BAR);
   XSetFont(fl_display, gc, fixed_font);
   
   XDrawString(fl_display,
	       win2,
	       gc,
	       5,
	       2* height/3,
	       minibuffer_text, i);
   XFlush(fl_display);
   return 0;
}

void SetMinibuffer(const char* s1, const char* s2, const char* s3) {
   sprintf(minibuffer_text,"%s %s %s", s1, s2, s3);
   MinibufferExpose(NULL, 0, 0, 0, NULL, NULL);
   fl_set_timer(fd_form_main->timer_minibuffer, 4);
}

void InitMinibuffer() {
   char* s="";
   // keyboard_mode = 0;
   if (buffer[current_buffer]->text){
      if (!buffer[current_buffer]->lyx_current)
      	 s = "(unsaved changes)";
      sprintf(minibuffer_text,"current buffer: %s %s, Textclass: %s", 
	      buffer[current_buffer]->filename, 
	      s, 
	      NameOfClass(buffer[current_buffer]->parameters.textclass));
   }
   else
      sprintf(minibuffer_text,"* no buffer * ");
   MinibufferExpose(NULL, 0, 0, 0, NULL, NULL);
}

void MinibufferTimerCB(FL_OBJECT *ob, long data) {
   InitMinibuffer();
}


void MenuPasteSelection(char at) {
  if (!buffer[current_buffer]->screen)
    return;


  ascii_type = at;
  
  Atom data_prop = XInternAtom(fl_display, 
			       "LyX_Primary",
			       False);
  XConvertSelection(fl_display,
		    XA_PRIMARY, XA_STRING, data_prop, 
		    fl_get_canvas_id(fd_form_main->work_area), 0); 
}


void UpdateLayoutChoice() {
   /* update the layout display */ 
   int i;
   
   if (current_textclass != buffer[current_buffer]->parameters.textclass) {
      fl_clear_choice(fd_form_main->choice_layouts);
      for (i = 0; !StringEqual(NameOfLayout(buffer[current_buffer]->parameters.textclass, i),"dummy"); i++) {
	 fl_addto_choice(fd_form_main->choice_layouts, NameOfLayout(buffer[current_buffer]->parameters.textclass,i));
      }
      current_textclass = buffer[current_buffer]->parameters.textclass;
      current_layout = 0;
   }
   
   if (!buffer[current_buffer]->text)
      return;
   char layout = buffer[current_buffer]->text->cursor.par->GetLayout();
   
   if (layout !=  current_layout){
      fl_set_choice(fd_form_main->choice_layouts, layout + 1);
      current_layout = layout;
      
   }
}


// this is wrong after move to pups
// I will write it anew and move it to menus.C
void UpdateMenuBar() 
{
}

void ProhibitInput() {
   input_prohibited = True;
   if (buffer[current_buffer]->screen)
     buffer[current_buffer]->screen->HideCursor();

   static Cursor cursor;
   static char cursor_undefined = 1;
   
   if (cursor_undefined)
     cursor = XCreateFontCursor(fl_display, XC_watch);
   
   
   /* set the cursor to the watch for for all forms and the canvas */ 
   XDefineCursor(fl_display, fd_form_main->form_main->window, cursor);
   if (fd_form_paragraph->form_paragraph->visible)
     XDefineCursor(fl_display, fd_form_paragraph->form_paragraph->window, cursor);
   XDefineCursor(fl_display, fl_get_canvas_id(fd_form_main->work_area), cursor); 
   XDefineCursor(fl_display, fl_get_canvas_id(fd_form_main->minibuffer), cursor);


   XFlush(fl_display);
   
   /* remove the canvxas handler */ 
   fl_remove_canvas_handler(fd_form_main->work_area, KeyPress, WorkAreaKeyPress);
   fl_remove_canvas_handler(fd_form_main->work_area, ButtonPress, WorkAreaButtonPress);
   fl_remove_canvas_handler(fd_form_main->work_area, ButtonRelease, WorkAreaButtonRelease);
   fl_remove_canvas_handler(fd_form_main->work_area, MotionNotify, WorkAreaMotionNotify);
   
   if (buffer[current_buffer]->screen)
     buffer[current_buffer]->screen->HideCursor();
   
}

void AllowInput() {
   input_prohibited = False;
   if (buffer[current_buffer]->screen && work_area_active)
     buffer[current_buffer]->screen->ShowCursor();

   /* reset the cursor from the watch for all forms and the canvas */
   
   XUndefineCursor(fl_display, fd_form_main->form_main->window);
   if (fd_form_paragraph->form_paragraph->visible)
     XUndefineCursor(fl_display, fd_form_paragraph->form_paragraph->window);
   XUndefineCursor(fl_display, fl_get_canvas_id(fd_form_main->work_area));
   XDefineCursor(fl_display, fl_get_canvas_id(fd_form_main->work_area), 
		 xterm_cursor);
   XDefineCursor(fl_display, fl_get_canvas_id(fd_form_main->minibuffer),
		 xterm_cursor);

   
   
   /* re-install the canvas handler */ 
   fl_add_canvas_handler(fd_form_main->work_area, KeyPress, WorkAreaKeyPress, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, ButtonPress, WorkAreaButtonPress, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, ButtonRelease, WorkAreaButtonRelease, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, MotionNotify, WorkAreaMotionNotify, fd_form_main);

   if (buffer[current_buffer]->screen && work_area_active)
     buffer[current_buffer]->screen->ShowCursor();

}



void WriteAlert(const char* s1, const char* s2, const char* s3){
   ProhibitInput();
   SetMinibuffer(s1, s2, s3);
   fl_show_alert(s1, s2,
		s3, 0);
   AllowInput();
}


void FreeUpdateTimer() {
   /* a real free timer would be better but I don't know 
    * how to do this with xforms */
   updatetimer = 0;
}

void SetUpdateTimer() {
   fl_set_timer(fd_form_main->timer_update, 0.5);
   updatetimer = 1;
}


void BeforeChange() {
   buffer[current_buffer]->screen->ToggleSelection();
   buffer[current_buffer]->text->ClearSelection();
   FreeUpdateTimer();
}

int IsLinuxDoc()
{
   /* checks for right textclass */
   int textclass = buffer[current_buffer]->parameters.textclass;

   if (!StringEqual(NameOfClass(textclass), "linuxdoc")) return 0;
   else return 1;
}

void UpdateScrollbar() {
   /* if the text is smaller then the working area, the scrollbar
    * maximum must be the working area height. No scrolling will 
    * be possible */
   static long max2 = 0;
   static long height2 = 0;

   long cbth = 0;
   long cbsf = 0;
   
   if (buffer[current_buffer]->text)
     cbth = buffer[current_buffer]->text->height;
   if (buffer[current_buffer]->screen)
     cbsf = buffer[current_buffer]->screen->first;
   if (max2 == cbth &&
       height2 == work_area_height &&
       current_scrollbar_value == cbsf)
     return;			       /* no update necessary */
   max2 = cbth;
   height2 = work_area_height;
   current_scrollbar_value = cbsf;
   
   long maximum_height = work_area_height * 3/4;
   long value = 0;
   
   if (buffer[current_buffer]->text)
      maximum_height += buffer[current_buffer]->text->height;
   if (buffer[current_buffer]->screen)
      value = buffer[current_buffer]->screen->first;
   
   /* set the scrollbar */
   double hfloat = work_area_height;
   double maxfloat = maximum_height;
   
   fl_set_slider_value(fd_form_main->scrollbar, value);
   fl_set_slider_bounds(fd_form_main->scrollbar, 0, maximum_height - work_area_height);
   /* fl_set_slider_step(fd_form_main->scrollbar, 1); */
   if (maxfloat>0){
     if ((hfloat/maxfloat) * (float) height2 < 3)
       fl_set_slider_size(fd_form_main->scrollbar, 3/(float)height2);
     else
       fl_set_slider_size(fd_form_main->scrollbar, hfloat/maxfloat);
   }
   else
      fl_set_slider_size(fd_form_main->scrollbar, hfloat);
   fl_set_slider_precision(fd_form_main->scrollbar, 0);
}


void Update(signed char f) {
   UpdateLayoutChoice();

   buffer[current_buffer]->text->FullRebreak();
   buffer[current_buffer]->screen->Update();
   buffer[current_buffer]->screen->FitCursor();
   UpdateScrollbar();
       
   if (f==1 || f==-1) {
      if (buffer[current_buffer]->lyx_current) {
	 buffer[current_buffer]->lyx_current = 0;
	 buffer[current_buffer]->bak_current = 0;
	 buffer[current_buffer]->dvi_current = 0;
	 fl_set_timer(fd_form_main->timer_minibuffer, 4);
      }
      else {
	 buffer[current_buffer]->lyx_current = 0;
	 buffer[current_buffer]->bak_current = 0;
	 buffer[current_buffer]->dvi_current = 0;
      }
   }
   if (f >= 0 && work_area_active)
     buffer[current_buffer]->screen->ShowCursor();
}

void SmallUpdate(signed char f) {
   buffer[current_buffer]->screen->SmallUpdate();
   buffer[current_buffer]->screen->FitCursor();
   UpdateScrollbar();
   
   if (f==1 || f==-1) {
      if (buffer[current_buffer]->lyx_current) {
	 buffer[current_buffer]->lyx_current = 0;
	 buffer[current_buffer]->bak_current = 0;
	 buffer[current_buffer]->dvi_current = 0;
	 fl_set_timer(fd_form_main->timer_minibuffer, 4);
      }
      else {
	 buffer[current_buffer]->lyx_current = 0;
	 buffer[current_buffer]->bak_current = 0;
	 buffer[current_buffer]->dvi_current = 0;
      }
   }
   if (f >= 0 && work_area_active)
     buffer[current_buffer]->screen->ShowCursor();
}

int ResizeCurrentBuffer(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
     /* if there is no text loaded, I do not have to resize anything */ 
   if (!buffer[current_buffer]->parameters.paragraph) {
      XFillRectangle(fl_display, win, LyXGetCopyGC(), 0, 0, w, h);
      UpdateScrollbar();
      return 0;
   }

   LyXParagraph *par = NULL;
   LyXParagraph *selstartpar = NULL;
   LyXParagraph *selendpar = NULL;
   int pos = 0;
   int selstartpos = 0;
   int selendpos = 0;
   int selection = 0;
   int mark_set = 0;

   ProhibitInput();

   SetMinibuffer("Resizing buffer...","","");   

   if (buffer[current_buffer]->text) {
      par = buffer[current_buffer]->text->cursor.par;
      pos = buffer[current_buffer]->text->cursor.pos;
      selstartpar = buffer[current_buffer]->text->sel_start_cursor.par;
      selstartpos = buffer[current_buffer]->text->sel_start_cursor.pos;
      selendpar = buffer[current_buffer]->text->sel_end_cursor.par;
      selendpos = buffer[current_buffer]->text->sel_end_cursor.pos;
      selection = buffer[current_buffer]->text->selection;
      mark_set = buffer[current_buffer]->text->mark_set;
      delete(buffer[current_buffer]->text);
   }
   
   buffer[current_buffer]->text = new LyXText(w,
					      &buffer[current_buffer]->parameters);
   if (buffer[current_buffer]->screen)
      delete(buffer[current_buffer]->screen);
   buffer[current_buffer]->screen = new LyXScreen(win, h, w, buffer[current_buffer]->text);
   
   if (par) {
      buffer[current_buffer]->text->selection = True;    /* at this point just
						  * to avoid the Delete-
						  * Empty-Paragraph
						  * Mechanism when
						  * setting the cursor */
      buffer[current_buffer]->text->mark_set = mark_set;
      if (selection) {
	 buffer[current_buffer]->text->SetCursor(selstartpar, selstartpos);
	 buffer[current_buffer]->text->sel_cursor = buffer[current_buffer]->text->cursor;
	 buffer[current_buffer]->text->SetCursor(selendpar, selendpos);
	 buffer[current_buffer]->text->SetSelection();
	 buffer[current_buffer]->text->SetCursor(par, pos);
      }
      else {
	 buffer[current_buffer]->text->SetCursor(par, pos);
	 buffer[current_buffer]->text->sel_cursor = buffer[current_buffer]->text->cursor;
	 buffer[current_buffer]->text->selection = False;
      }
      
   }
   buffer[current_buffer]->screen->first  = 
   buffer[current_buffer]->screen->TopCursorVisible(); /* this will scroll the screen that 
						* the cursor becomes visible */ 
   UpdateScrollbar();
   buffer[current_buffer]->screen->Redraw();
   
   InitMinibuffer();
   AllowInput();

   // update menubar to reflect new buffer state
   UpdateMenuBar();
   
   // now if the title form still exist kill it
   TimerCB(NULL,0);
   
   return 0;
}

int WorkAreaSelectionNotify(FL_OBJECT *ob, Window win, int w, int h, XEvent *event, void *d) {
  if (event->type != SelectionNotify)
    return 0;
  Atom tmpatom;
  unsigned long ul1;
  unsigned long ul2;
  unsigned char* uc = NULL;
  int tmpint;
  if (work_area_active)
    buffer[current_buffer]->screen->HideCursor();
  BeforeChange();
  if (event->xselection.type == XA_STRING
      && event->xselection.property) {
    
    XGetWindowProperty(
		       fl_display            /* display */,
		       win /* w */,
		       event->xselection.property                /* property */,
		       0                /* long_offset */,
		       0                /* long_length */,
		       False                /* delete */,
		       XA_STRING                /* req_type */,
		       &tmpatom               /* actual_type_return */,
		       &tmpint                /* actual_format_return */,
		       &ul1      /* nitems_return */,
		       &ul2      /* bytes_after_return */,
		       &uc     /* prop_return */
		       );
    
      XGetWindowProperty(
		       fl_display           /* display */,
		       win              /* w */,
		       event->xselection.property                /* property */,
		       0                /* long_offset */,
		       ul2/4+1                /* long_length */,
		       True                /* delete */,
		       XA_STRING                /* req_type */,
		       &tmpatom               /* actual_type_return */,
		       &tmpint                /* actual_format_return */,
		       &ul1      /* nitems_return */,
		       &ul2      /* bytes_after_return */,
		       &uc     /* prop_return */
		       );
        
    if (!ascii_type)
      buffer[current_buffer]->text->InsertStringA((char*)uc);
    else
      buffer[current_buffer]->text->InsertStringB((char*)uc);
      
    delete(uc);
    Update(1);
  }
  if (work_area_active)
    buffer[current_buffer]->screen->ShowCursor();
  return 0;
}

int WorkAreaExpose(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
   static FL_OBJECT *ob2 = NULL;
   static Window win2 = 0;
   static XEvent *ev2 = NULL;
   static void *d2 = NULL;
   
   /* FD_form_main *ui = (FD_form_main*) d; */ 
   
   if (ob == NULL && !buffer[current_buffer]->screen) {
      /* resize current buffer */ 
      ResizeCurrentBuffer(ob2,win2,work_area_width,work_area_height,ev2,d2);
   }
   else if (w != work_area_width || h != work_area_height) {
      work_area_width = w;
      work_area_height = h;
      ob2 = ob;
      win2 = win;
      ev2 = ev;
      d2 = d;
      /* resize */ 
      /* all buffers will need resize */ 
      for (int i =0; i < NUMBER_OF_BUFFERS; i++) {
	 if (buffer[i]->screen)
	    delete(buffer[i]->screen);
	 buffer[i]->screen = NULL;
      }
      /* resize current buffer */ 
      ResizeCurrentBuffer(ob,win,w,h,ev,d);
   }
   else {
      /* redraw */
      if (buffer[current_buffer]->screen) 
      	buffer[current_buffer]->screen->Redraw(); 
      else {
	 XFillRectangle(fl_display, win, LyXGetCopyGC(), 0, 0, w, h);
      }
   }
   UpdateLayoutChoice();
   return 0;
}

void ResizeCurrentBufferPseudoExpose() {
   FL_OBJECT *ob2 = NULL;
   Window win2 = 0;
   XEvent *ev2 = NULL;
   void *d2 = NULL;
   if (buffer[current_buffer]->screen)
     UpdateScrollbar();
   /* resize current buffer through Pseudo-Expose*/
   WorkAreaExpose(ob2,win2,work_area_width,work_area_height,ev2,d2);

}


void RedoCurrentBuffer() {
   if (buffer[current_buffer]->text){
      if (buffer[current_buffer]->screen)
      	 delete(buffer[current_buffer]->screen);
      buffer[current_buffer]->screen = NULL;
      ResizeCurrentBufferPseudoExpose();
   }
}

void MenuSearch() {
   fl_show_form(fd_form_search->form_search,
		FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		"Search & Replace");
}

void MenuQuit() {
   ProhibitInput();
   int i;
   for (i=0; i < NUMBER_OF_BUFFERS; i++) {
	 if (buffer[i]->parameters.paragraph && !buffer[i]->lyx_current) {
	    if (fl_show_question("Unsaved changes in buffer:",
				 buffer[i]->filename,
				 "Save buffer?"))
	       MenuWrite(i);
	 }
	 
      }
   if (fl_show_question("", "Quit LyX?","")) {
      lastfiles->WriteFile(lyxrc.lastfiles);
      exit(0);
   }
   
   AllowInput();
}

void MenuPrint() {
   const char* filename = NULL;
   
   if (!buffer[current_buffer]->text) {
      filename = "No_File";
   }
   else
      filename = buffer[current_buffer]->filename;
   
   
   /* calculate ps filename */ 
   int dot=0;
   int path = 0;
   int n=0;
   int n2=0;
   int i=0;
   int ext_len = 0;
   /* find the end of the string */ 
   for (n=0; filename[n]; n++);
   
   
   /* go back to the first dot */ 
   for (n2=n; n2 && filename[n2]!='.'
	&& filename[n2]!='/'; n2--);
   
   /* when you find a dot, use this as extension */ 
   if (filename[n2]=='.')
      dot = n2;
   else
      dot = n;
   
   for (path=dot; path && filename[path]!='/';path--);
   if (filename[path]=='/')
      path++;
   
   for (ext_len = 0; lyxrc.print_file_extension [ext_len]; ext_len++);
   char *ps = new char[dot+ext_len];
   
   
   for (i=path; i<dot; i++) 
      ps[i-path] = filename[i];
   for (n = 0; n < ext_len; n++) 
         ps[i+n-path] = lyxrc.print_file_extension[n];
   if (lyxrc.printer)
      fl_set_input(fd_form_print->input_printer, lyxrc.printer);
   else
      fl_set_input(fd_form_print->input_printer, "lp");
   
   fl_set_input(fd_form_print->input_file, ps);
   
   
   /* set theother default entries */ 
   fl_set_button(fd_form_print->radio_order_normal, 1);
   fl_set_button(fd_form_print->radio_all_pages, 1);
   fl_set_button(fd_form_print->radio_printer, 1);
   fl_set_button(fd_form_print->radio_file, 0);
   fl_set_button(fd_form_print->radio_odd_pages, 0);
   fl_set_button(fd_form_print->radio_even_pages, 0);
   fl_set_button(fd_form_print->radio_order_reverse, 0);

   fl_show_form(fd_form_print->form_print,
		FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		"Print");
   delete(ps);

}

void MenuRunLaTeX() {
   
   int ret;

   if (IsLinuxDoc()) ret = RunLinuxDoc(1, buffer[current_buffer]->filename);
   else ret = RunLaTeX();
   
   if (ret == 0) {
      /* fl_show_message("","Compilation successfull!",""); */ 
   }
   else if (ret > 0) {
      char * s = new char [50];
      if (ret == 1)
      	 sprintf(s, "%d error detected",ret);
      else
      	 sprintf(s, "%d errors detected",ret);
      WriteAlert("LaTeX Error!", s, "");
      delete(s);
   }
}

void AutoSave() {
   char *fname;
   int n = 0;
   int i=0;
   
   if (buffer[current_buffer]->text) {
      
      if (!buffer[current_buffer]->bak_current) {
	 SetMinibuffer("Auto save current buffer...","","");
	 /* create autosave filename */
	 for (n=0; buffer[current_buffer]->filename[n]; n++);
	 
	 fname = new char[n+10];
	 
	 for (i=0; buffer[current_buffer]->filename[i]; i++)
      	    fname[i] = buffer[current_buffer]->filename[i];
	 
	 fname[i] = '.';
	 fname[i+1] = 'b';
	 fname[i+2] = 'a';
	 fname[i+3] = 'k';   
	 fname[i+4] = '\0';
	 
	 /* make a backup */ 
	 char *s = new char[n+5];
	 for (i=0; fname[i]; i++) 
      	    s[i] = fname[i];
	 s[i] = '~';
	 s[i+1] = '\0';
	 rename(fname, s);
	 delete(s);
	 
	 buffer[current_buffer]->parameters.cursor = buffer[current_buffer]->text->cursor;
	 
	 WriteFile(fname, 
		   buffer[current_buffer]->parameters);
	 buffer[current_buffer]->bak_current = 1;
	 delete(fname);
	 SetMinibuffer("Auto save current buffer done!","","");
      }
      else {
	 SetMinibuffer("Auto save not neccessary!","","");
      }
   }
   
   if (lyxrc.autosave)
      fl_set_timer(fd_form_main->timer_autosave, lyxrc.autosave);
}

void MenuLoadAutoSave() {
   char *fname;
   int n = 0;
   int i=0;

   if (lyxrc.autosave)
      fl_set_timer(fd_form_main->timer_autosave, lyxrc.autosave);

   if (!buffer[current_buffer]->text)
      return;
   
   /* create autosave filename */
   for (n=0; buffer[current_buffer]->filename[n]; n++);
   
   fname = new char[n+10];
   
   for (i=0; buffer[current_buffer]->filename[i]; i++)
      fname[i] = buffer[current_buffer]->filename[i];
   
   fname[i] = '.';
   fname[i+1] = 'b';
   fname[i+2] = 'a';
   fname[i+3] = 'k';   
   fname[i+4] = '\0';
   
   SetMinibuffer("Loading autosave file ",fname,"...");   
   
   if (!OpenAllFile(fname) && !OpenAllFile(buffer[current_buffer]->filename)) {
      WriteAlert("Operation not possible!", "No autosave file",
		 fname);
   } else {
      
      /* InitTextParameters(buffer[current_buffer]->parameters); */ 
      
      LyXParagraph *par = buffer[current_buffer]->parameters.paragraph;
      LyXParagraph *tmppar = par;
      while(par){
	 tmppar = par->next;
	 delete(par);
	 par = tmppar;
      }
      
      ReadFile(buffer[current_buffer]->parameters);
      buffer[current_buffer]->lyx_current = 0;
      buffer[current_buffer]->bak_current = 0;
      buffer[current_buffer]->dvi_current = 0;
      CloseFile();
      
      buffer[current_buffer]->text->cursor = buffer[current_buffer]->parameters.cursor;
      buffer[current_buffer]->text->ClearSelection();
      
      
      /* use the backup as new autosave file */
      rename(fname, ".lyxtmp");
      char *s = new char[n+5];
      for (i=0; fname[i]; i++) 
      	 s[i] = fname[i];
      s[i] = '~';
      s[i+1] = '\0';
      rename(s, fname);
      rename(".lyxtmp", s);
      delete(s);
      if (buffer[current_buffer]->screen) {
	 delete(buffer[current_buffer]->screen);
	 buffer[current_buffer]->screen = NULL;
      }
      ResizeCurrentBufferPseudoExpose();
      SetMinibuffer("Autosave file loaded.","","");   
   }
   
   delete(fname);
}

void MenuOpen() {
   const char *got_fname;
   char *s;
   int i=0, file_type=0;

   ProhibitInput();
   got_fname = fl_show_file_selector("File To Load", document_path,"*.lyx","");
   /* flushes Xlib queue */
   XFlush(fl_get_display());
   /* shorten the filename if it begins with the path */
   AllowInput();
   
   if (!got_fname)
      return;
   
   char *fname = new char[strlen(got_fname)+5];
   strcpy(fname, got_fname);
   char *path = new char[500];
   getcwd(path, 500);
   for (i=0; fname[i] == path[i]; i++);
   
   if (i && path[i] == '\0' && fname[i]=='/')
     fname += i + 1;
   delete(path);
   
   if (document_path){
     delete(document_path);
     document_path = NULL;
   }
   if (fname == NULL) return;
 
   SetMinibuffer("Loading file",fname,"...");   

   LoadLyXFile(fname);
   ResizeCurrentBufferPseudoExpose();
   
}

void LoadLyXFile(const char *filename)
{
   char *s;
   int i;
   
   if (IsSGMLFilename(filename)) 
      if (!RunLinuxDoc(-1, filename)) {
         char *temp = rindex(filename, '.');
         if (temp != NULL) {
            *(++temp) = 'l';
            *(++temp) = 'y';
            *(++temp) = 'x';
            *(++temp) = 0;
         }
      }
   

   if (!OpenFile(filename)) {

      // fprintf(stderr, "Cannot open specified file %s \n", filename);
      WriteAlert("Error!", "Cannot open specified file:", filename);
   } else {
      // file already open?
      for (i=0; i < NUMBER_OF_BUFFERS &&
	      !StringEqual(buffer[i]->filename,filename); i++);
      
      if (i<NUMBER_OF_BUFFERS) {
	 CloseFile();
	 WriteAlert("Error!", "Buffer already exists:",
		    filename);
      }
      else {
	 
	 //find a free buffer
	 i=0;
	 while (i < NUMBER_OF_BUFFERS - 1 && buffer[i]->parameters.paragraph) i++;
	 
	 // fprintf(stderr,"assigning to buffer %d\n",i);

	 int a=0;
	 for (a=0;filename[a];a++);
	 s = new char[a+5];
	 for (a=0;filename[a];a++)
	    s[a]= filename[a];    
	 s[a] = '\0';

	 buffer[i]->filename = s;
	 
	 /* InitTextParameters(buffer[i]->parameters); */ 

	 ReadFile(buffer[i]->parameters);

	 buffer[i]->lyx_current = 1;
	 buffer[i]->bak_current = 1;
	 buffer[i]->dvi_current = 0;

	 old_current_buffer = current_buffer;
	 current_buffer = i;
	 CloseFile();
	 
	 //fl_addto_menu(fd_form_main->menu_buffer,
	 //		buffer[current_buffer]->filename);
	 
	 lastfiles->newfile(buffer[current_buffer]->filename);	 

	 //buffer[current_buffer]->menu_position =
	 //  fl_get_menu_maxitems(fd_form_main->menu_buffer);
      }
   }
}

void InsertAsciiFile(char flag) {
   const char *fname;
   LyXParagraph *tmppar;
   
   if (!buffer[current_buffer]->screen)
     return;
   
   ProhibitInput();
   fname = fl_show_file_selector("File To Insert","","*","");
   AllowInput();
   if (fname == NULL) return;
   if (!OpenAllFile(fname)) {
     WriteAlert("Error!", "Cannot open specified file:",
		fname);
      
   } else {
      tmppar = new LyXParagraph();
      ReadSimpleWholeFile(tmppar);
      /* set the end of the string */
      tmppar->InsertChar(tmppar->last-1,'\0');
      CloseFile();
      /* insert the string */
      if (work_area_active)
      	buffer[current_buffer]->screen->HideCursor();
      
      /* clear the selection */
      BeforeChange();
      if (!flag)
      	buffer[current_buffer]->text->InsertStringA(tmppar->text);
      else
      	buffer[current_buffer]->text->InsertStringB(tmppar->text);
      delete(tmppar);
      Update(1);
   }
}

void MenuNew() {
   const char *fname;

   char *s;
   int i;
   ProhibitInput();
   fname = fl_show_input("Insert Filename:", "newfile.lyx");
   AllowInput();
   if (fname == NULL) return;
   

   /* make sure the filename ends with .lyx */
   for (i=0;fname[i];i++);
   s = new char[i+5];
   for (i=0;fname[i];i++)
     s[i]=fname[i];
   s[i]='\0';
   if (!IsLyXFilename(s)) {
      s[i]='.';i++;
      s[i]='l';i++;
      s[i]='y';i++;
      s[i]='x';i++;
      s[i]='\0';
  }
  
  /* check wether the file exists */
  if (OpenFile(s)) {
    CloseFile();
    WriteAlert("Error!", "File exists:",
		  s);
    delete(s);
  }
  else {
     /* file already open? */
     for (i=0; i < NUMBER_OF_BUFFERS && 
	  !StringEqual(buffer[i]->filename,
		       s);
	  i++);
     
     if (i<NUMBER_OF_BUFFERS) {
	CloseFile();
	WriteAlert("Error!", "Buffer already exists:",
		      s);
     }
     else {
	/* the template stuff */
	ProhibitInput();
	int i;
	if (document_path)
	  delete(document_path);
	document_path = StringCopy(fl_get_directory());
	fname = fl_show_file_selector("Choose Template",lyxrc.template_path,"*.lyx","");
	AllowInput();

	/* find a free buffer */ 
	i=0;
	while (i < NUMBER_OF_BUFFERS - 1 && buffer[i]->parameters.paragraph)
	  i++;

	/* InitTextParameters(buffer[i]->parameters); */

	if (fname && OpenFile(fname)){
	   ReadFile(buffer[i]->parameters);
	   CloseFile();
	}
	else {
	  buffer[i]->parameters.paragraph = new LyXParagraph();
	}

	buffer[i]->filename = s;
	old_current_buffer = current_buffer;
	current_buffer = i;
	buffer[current_buffer]->lyx_current = 0;
	buffer[current_buffer]->bak_current = 0;
	buffer[current_buffer]->dvi_current = 0;
	
	ResizeCurrentBufferPseudoExpose();
	// fl_addto_menu(fd_form_main->menu_buffer, buffer[current_buffer]->filename);
	lastfiles->newfile(buffer[current_buffer]->filename);
	//buffer[current_buffer]->menu_position = fl_get_menu_maxitems(fd_form_main->menu_buffer);
   }
  }      
}

void MenuWrite(int buf) {
   int n, i;
   if (buffer[buf]->parameters.paragraph) {
      SetMinibuffer("Write buffer", buffer[buf]->filename,"...");
      /* make a backup */ 
      for (n=0; buffer[buf]->filename[n]; n++);
      char *s = new char[n+5];
      for (i=0; i<n; i++) 
      	s[i] = buffer[buf]->filename[i];
      s[i] = '~';
      s[i+1] = '\0';
      rename(buffer[buf]->filename, s);
      delete(s);
      
      if (buffer[buf]->text)
      	 buffer[buf]->parameters.cursor = buffer[buf]->text->cursor;
      
      WriteFile(buffer[buf]->filename, 
		buffer[buf]->parameters);
      buffer[buf]->lyx_current = 1;
      SetMinibuffer("Wrote buffer", buffer[buf]->filename,"!");
      
   }
}

void MenuMakeLaTeX() {
   int n, n2, i;
   if (buffer[current_buffer]->text) {
      
      /* get LaTeX-Filename */
      
      /* find the end of the string */ 
      for (n=0; buffer[current_buffer]->filename[n]; n++);
      
      /* go back to the first dot */ 
      for (n2=n; n2 && buffer[current_buffer]->filename[n2]!='.'
	   && buffer[current_buffer]->filename[n2]!='/'; n2--);
      
      /* when you find a dot, use this as extension */ 
      if (buffer[current_buffer]->filename[n2]=='.')
      	n = n2;
      char *s = new char[n+10];
      for (i=0; i<n; i++) 
      	s[i] = buffer[current_buffer]->filename[i];
      s[i] = '.';
      s[i+1] = 't';
      s[i+2] = 'e';
      s[i+3] = 'x';
      s[i+4] = '\0';

      if (IsLinuxDoc()) RunLinuxDoc(0, buffer[current_buffer]->filename);
      else MakeLaTeXFile(s, buffer[current_buffer]->parameters, 65);

      delete(s);
      buffer[current_buffer]->dvi_current = 0;
      
   }
   
}

void MenuMakeLinuxDoc() {
   int n, n2, i;
   int textclass;

   if (buffer[current_buffer]->text) {

      if (!IsLinuxDoc()) {
         WriteAlert("Error!", "Textclass must be linuxdoc.", "");
         return;
      }

      /* get LinuxDoc-Filename */
      
      /* find the end of the string */ 
      for (n=0; buffer[current_buffer]->filename[n]; n++);
      
      /* go back to the first dot */ 
      for (n2=n; n2 && buffer[current_buffer]->filename[n2]!='.'
	   && buffer[current_buffer]->filename[n2]!='/'; n2--);
      
      /* when you find a dot, use this as extension */ 
      if (buffer[current_buffer]->filename[n2]=='.')
      	n = n2;
      char *s = new char[n+10];
      for (i=0; i<n; i++) 
      	s[i] = buffer[current_buffer]->filename[i];
      s[i] = '.';
      s[i+1] = 's';
      s[i+2] = 'g';
      s[i+3] = 'm';
      s[i+4] = 'l';
      s[i+5] = '\0';
      SetMinibuffer("Building LinuxDoc SGML file", s,"...");
      MakeLinuxDocFile(s,
		    buffer[current_buffer]->parameters,
		    65);
      SetMinibuffer("LinuxDoc SGML file", s,"build !");
      delete(s);
   }
}

void MenuWriteAs() {
   const char *fname;

   char *s = NULL;
   int i;
   if (!buffer[current_buffer]->text)
      return;
   
   ProhibitInput();
   fname = fl_show_input("Insert New Filename:", buffer[current_buffer]->filename);
   AllowInput();
   if (fname == NULL) return;
   
   /* make sure the filename ends with .lyx */
   for (i=0;fname[i];i++);
   s = new char[i+5];
   for (i=0;fname[i];i++)
     s[i]=fname[i];
   s[i]='\0';
   if (!IsLyXFilename(s)) {
      s[i]='.';i++;
      s[i]='l';i++;
      s[i]='y';i++;
      s[i]='x';i++;
      s[i]='\0';
  }
  
  /* check wether the file exists */
  if (OpenFile(s)) {
    CloseFile();
    WriteAlert("Error!", "File exists:",
		  s);
    delete(s);
  }
  else {
     /* file already open? */
     for (i=0; i < NUMBER_OF_BUFFERS && 
	  !StringEqual(buffer[i]->filename,
		       s);
	  i++);
     
     if (i<NUMBER_OF_BUFFERS) {
	CloseFile();
	WriteAlert("Error!", "Buffer already exists:",
		      s);
     }
     else {
	if (buffer[current_buffer]->filename) {
	   delete (buffer[current_buffer]->filename);
	}
	
	buffer[current_buffer]->filename = s;
	
	buffer[current_buffer]->lyx_current = 0;
	buffer[current_buffer]->dvi_current = 0;
	
	
	/* update the buffer menu menu */
	int pos = buffer[current_buffer]->menu_position;
	
       fl_replace_menu_item(fd_form_main->menu_buffer, 
			   pos,
			   buffer[current_buffer]->filename);
       }
       
     MenuWrite(current_buffer);
  }
}    


void MenuShowTableOfContents() {
   TocUpdateCB(NULL, 0);
   fl_show_form(fd_form_toc->form_toc,
		FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		"Table Of Contents");
}

void MenuInsertLabel(){
  ProhibitInput();
  char *tmp = StringCopy(fl_show_input("Insert new label",""));
  InsetLabel *new_inset = new InsetLabel(tmp);
  buffer[current_buffer]->text->InsertInset(new_inset);
  Update(-1);
  delete (tmp);
  AllowInput();
}

void MenuInsertRef() {
   RefUpdateCB(NULL, 0);
   fl_set_button(fd_form_ref->radio_ref, 1);
   fl_set_button(fd_form_ref->radio_pageref, 0);
   fl_show_form(fd_form_ref->form_ref,
		FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		"Insert Reference");
}

int WorkAreaButtonRelease(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
   selection_possible = False;
   return 0;
}

int WorkAreaButtonPress(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
   work_area_active = True;
   if (!buffer[current_buffer]->screen)
     return 0 ;
   selection_possible = True;
   buffer[current_buffer]->screen->HideCursor(); 
   /* clear the selection */ 
   buffer[current_buffer]->screen->ToggleSelection();
   buffer[current_buffer]->text->ClearSelection();
   buffer[current_buffer]->text->FullRebreak();
   buffer[current_buffer]->screen->Update();
   UpdateScrollbar();
   buffer[current_buffer]->text->SetCursorFromCoordinates(ev->xbutton.x,
							  ev->xbutton.y +
							  buffer[current_buffer]->screen->first);
   buffer[current_buffer]->text->sel_cursor = buffer[current_buffer]->text->cursor;
   buffer[current_buffer]->text->cursor.x_fix = buffer[current_buffer]->text->cursor.x;
   
   /* i will use the middle button to edit insets or toggle footnotes */ 
   if (ev->xbutton.button == 2) {
      buffer[current_buffer]->text->ToggleFootnote();
      Update(0);

   }
	 

   /* the right button will select a word */
   
   if (ev->xbutton.button == 3) {
      buffer[current_buffer]->text->SelectWord();
      buffer[current_buffer]->screen->ToggleSelection();
   }
   
   Update(0); /* this will fit the cursor on the screen
	       * if necessary */ 
   
   return 0;
}


// can't be static since keybind.C wants it
void CursorPrevious() {
  if (buffer[current_buffer]->text->cursor.row->previous){
    long y = buffer[current_buffer]->screen->first;
    long cursory = buffer[current_buffer]->text->cursor.y - y;
    Row* row = buffer[current_buffer]->text->GetRowNearY(y);
    y = y - work_area_height + row->height;
    buffer[current_buffer]->screen->Draw(y);
    buffer[current_buffer]->text->SetCursorFromCoordinates(buffer[current_buffer]->text->cursor.x,
							   cursory + y);
  }
}
// can't be static since keybind.C wants it
void CursorNext() {
  if (buffer[current_buffer]->text->cursor.row->next){
    long y = buffer[current_buffer]->screen->first;
    long cursory = buffer[current_buffer]->text->cursor.y - y;
    y += work_area_height;
    buffer[current_buffer]->text->GetRowNearY(y);
    buffer[current_buffer]->screen->Draw(y);
    buffer[current_buffer]->text->SetCursorFromCoordinates(buffer[current_buffer]->text->cursor.x_fix,
							   cursory + y);
  }
}


int WorkAreaMotionNotify(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
   if (!buffer[current_buffer]->screen)
     return 0;
   
   /* only use motion with button 1 */
   if (ev->xmotion.state & Button3MotionMask
       || ev->xmotion.state & Button2MotionMask)
      return 0; 
    
   
   /* the selection possible is needed, that only motion events are 
    * used, where the bottom press event was on the drawing area too */
   if (selection_possible) {
      
      buffer[current_buffer]->screen->HideCursor();
      /* buffer[current_buffer]->screen->ToggleSelection(); */

      buffer[current_buffer]->text->SetCursorFromCoordinates(ev->xbutton.x,
						     ev->xbutton.y +
						     buffer[current_buffer]->screen->first);
      if (!buffer[current_buffer]->text->selection)
      	Update(-2);       /* maybe an empty line was deleted  */
      buffer[current_buffer]->text->SetSelection();
      buffer[current_buffer]->screen->ToggleToggle();
      if (buffer[current_buffer]->screen->FitCursor())
      	UpdateScrollbar(); 
      buffer[current_buffer]->screen->ShowCursor();
   } 
   return 0;
}
int WorkAreaEnterNotify(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
   work_area_active = True;
   if (buffer[current_buffer]->screen && !input_prohibited)
     buffer[current_buffer]->screen->ShowCursor();
   return 0;
}
int WorkAreaLeaveNotify(FL_OBJECT *ob, Window win, int w, int h, XEvent *ev, void *d) {
   work_area_active = False;
   if (buffer[current_buffer]->screen) {
      buffer[current_buffer]->screen->HideCursor();
      buffer[current_buffer]->text->FullRebreak();
      buffer[current_buffer]->screen->Update();
      UpdateScrollbar();
   }
   return 0;
}




void CursorToggleCB(FL_OBJECT *ob, long data) {
   if (work_area_active && buffer[current_buffer]->screen)
     buffer[current_buffer]->screen->CursorToggle();
   fl_set_timer(ob, 0.4);
}



void InitTextParameters(LyXTextParameters &p) {
   p = lyxrc.parameters;
   p.language = StringCopy(p.language);
   p.fonts = StringCopy(p.fonts);
   p.fontenc = StringCopy(p.fontenc);
   p.inputenc = StringCopy(p.inputenc);
   p.preamble = StringCopy(p.preamble);
   p.options = StringCopy(p.options);
}

void CloseBuffer() {
   int i;

   /* This doesn't make sense if there is no document */
   if (!buffer[current_buffer]->parameters.paragraph)
     return;
						    

   /*doit save buffer etc */
   i = current_buffer;
   if (buffer[i]->parameters.paragraph && !buffer[i]->lyx_current) {
      ProhibitInput();
      if (fl_show_question("Unsaved changes in buffer:",
			   buffer[i]->filename,
			   "Save buffer?"))
      	MenuWrite(i);
      AllowInput();
   }
   
   if (buffer[current_buffer]->filename) {
      delete(buffer[current_buffer]->filename);
      buffer[current_buffer]->filename = NULL;
   }
   
   LyXParagraph *par = buffer[current_buffer]->parameters.paragraph;
   LyXParagraph *tmppar = par;
   while(par){
      tmppar = par->next;
      delete(par);
      par = tmppar;
   }
   buffer[current_buffer]->parameters.paragraph = NULL;
   
   if (buffer[current_buffer]->text) {
      delete(buffer[current_buffer]->text);
      buffer[current_buffer]->text = NULL;
   }
   if (buffer[current_buffer]->screen) {
      delete(buffer[current_buffer]->screen);
      buffer[current_buffer]->screen = NULL;
   }
   
   buffer[current_buffer]->lyx_current = 0;
   buffer[current_buffer]->dvi_current = 0;
   buffer[current_buffer]->bak_current = 0;

   delete(buffer[current_buffer]->parameters.language);
   delete(buffer[current_buffer]->parameters.fonts);
   delete(buffer[current_buffer]->parameters.fontenc);
   delete(buffer[current_buffer]->parameters.inputenc);
   delete(buffer[current_buffer]->parameters.preamble);
   delete(buffer[current_buffer]->parameters.options);
   InitTextParameters(buffer[current_buffer]->parameters);

   // current buffer must be reset here
   i=0;
   while(i<NUMBER_OF_BUFFERS && !buffer[i]->text) i++;
   current_buffer = i;
   if (i==NUMBER_OF_BUFFERS)
     current_buffer = 0;
   
   ResizeCurrentBufferPseudoExpose();
   UpdateMenuBar();
}


void init(int argc, char **argv) {
   int i, a;
   lyxrc.parameters.paragraph_separation = LYX_PARSEP_INDENT;
/*   lyxrc.parameters.quotes_language = GERMAN; */ 
   lyxrc.parameters.quotes_language = ENGLISH;
   lyxrc.parameters.quotes_times = 2;
/*  lyxrc.parameters.papersize = PAPER_A4PAPER; */ 
   lyxrc.parameters.papersize = PAPER_USLETTER;
   lyxrc.parameters.fontsize = 10;
   lyxrc.parameters.baseline = 1.0;
   lyxrc.parameters.secnumdepth = 3;
   lyxrc.parameters.tocdepth = 3;
/*    lyxrc.parameters.language = "german"; */ 
   lyxrc.parameters.language = StringCopy("default");
   lyxrc.parameters.fonts = StringCopy("default");
   lyxrc.parameters.fontenc = StringCopy("default");
   lyxrc.parameters.inputenc = StringCopy("default");
   lyxrc.parameters.preamble = new char; 
   lyxrc.parameters.preamble[0] = 0;
   lyxrc.parameters.options = new char; 
   lyxrc.parameters.options[0] = 0;
   lyxrc.parameters.textclass = 0;
   lyxrc.parameters.epsfig = StringCopy("none");
   
   lyxrc.parameters.sides = 1;
   lyxrc.parameters.columns = 1;
   lyxrc.parameters.orientation = ORIENTATION_PORTRAIT;
   lyxrc.parameters.pagestyle = PAGESTYLE_PLAIN;
   
   lyxrc.printer = StringCopy("lp");
   lyxrc.print_command = "dvips";
   lyxrc.print_evenpage_flag = "-B";
   lyxrc.print_oddpage_flag ="-A";
   lyxrc.print_reverse_flag = "-r";
   lyxrc.print_landscape_flag = "-t landscape";
   lyxrc.print_to_printer = "-P";
   lyxrc.print_to_file = "-o";
   lyxrc.print_file_extension = ".ps";
   lyxrc.print_extra_options = "";
   lyxrc.xdvi_extra_options = "";
   lyxrc.sgml_extra_options = "";
   lyxrc.lyx_layout_lib = ".";
   lyxrc.template_path = "";
   lyxrc.latex_command = "latex";
   
   lyxrc.default_zoom = 14;
   lyxrc.autosave = 120;
   lyxrc.parameters.cursor.par = NULL;
   lyxrc.parameters.cursor.pos = 0;
   
   lyxrc.lastfiles = NULL;
   /* put the text classes in the coice */ 
   for (i = 0; !StringEqual(NameOfClass(i),"dummy"); i++) {
      fl_addto_choice(fd_form_document->choice_class, NameOfClass(i));
   }

   
   /* define the slider bounds */ 
   fl_set_slider_bounds(fd_form_paragraph->slider_spaces_top,0,20);
   fl_set_slider_bounds(fd_form_paragraph->slider_spaces_bottom,0,20);
   fl_set_slider_bounds(fd_form_document->slider_baselinestretch,0.75,2);
   fl_set_slider_bounds(fd_form_document->slider_secnumdepth,-1,5);
   fl_set_slider_bounds(fd_form_document->slider_tocdepth,-1,5);
   fl_set_slider_bounds(fd_form_table->slider_rows,1,50);
   fl_set_slider_bounds(fd_form_table->slider_columns,1,50);
   fl_set_slider_value(fd_form_table->slider_rows, 5);
   fl_set_slider_value(fd_form_table->slider_columns, 5);
   fl_set_slider_precision(fd_form_document->slider_secnumdepth, 0);
   fl_set_slider_precision(fd_form_document->slider_tocdepth, 0);
   fl_set_slider_precision(fd_form_table->slider_rows, 0);
   fl_set_slider_precision(fd_form_table->slider_columns, 0);
   
   fl_set_button(fd_form_figure->radio_postscript, 1);
   fl_set_button(fd_form_figure->radio_latex, 0);
   
   fl_add_canvas_handler(fd_form_main->work_area, Expose, WorkAreaExpose, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, KeyPress, WorkAreaKeyPress, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, ButtonPress, WorkAreaButtonPress, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, ButtonRelease, WorkAreaButtonRelease, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, MotionNotify, WorkAreaMotionNotify, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, EnterNotify, WorkAreaEnterNotify, fd_form_main);
   fl_add_canvas_handler(fd_form_main->work_area, LeaveNotify, WorkAreaLeaveNotify, fd_form_main);

   fl_add_canvas_handler(fd_form_main->work_area, SelectionNotify, WorkAreaSelectionNotify, fd_form_main);
   
   fl_add_canvas_handler(fd_form_main->minibuffer, Expose, MinibufferExpose, fd_form_main);
   

   /* activate the timer for the cursor */ 
   fl_set_timer(fd_form_main->timer_cursor, 0.4);
   
   if (lyxrc.autosave)
      fl_set_timer(fd_form_main->timer_autosave, lyxrc.autosave);
   
   /* set a nice cursor for the drawing area */ 
   xterm_cursor = XCreateFontCursor(fl_display ,XC_xterm);
   XDefineCursor(fl_display, fl_get_canvas_id(fd_form_main->work_area), 
		 xterm_cursor);


   
   /* read the rc file */ 
   char *cpwd = new char[500];
   getcwd(cpwd, 500);
   const char* home = getenv("HOME");
   chdir(home);
   const char * lyxdir = getenv ("LYX_DIR");
   if ((lyxdir != NULL) && (strlen (lyxdir) < 255)) { 
     strcpy (system_lyxdir, lyxdir);
   }
   else {
     strcpy (system_lyxdir, LYX_DIR);
   }

   if (!lyxrc.template_path[0]){
     char* tempchar = new char[300];
     strcpy(tempchar, system_lyxdir);
     strcat (tempchar, "/templates");
     lyxrc.template_path = StringCopy(tempchar);
     delete(tempchar);
   }
   if (!lyxrc.lastfiles){
     char* tempchar = new char[300];
     strcpy(tempchar, home);
     strcat (tempchar, "/.lyx_lastfiles");
     lyxrc.lastfiles = StringCopy(tempchar);
     delete(tempchar);
   }
   LyxDefaultKeyBindings();
   if (ReadRCFile(".lyxrc", lyxrc) < 0) { 
      fprintf (stderr, 
        "User has no .lyxrc file: reading defaults from %s/system.lyxrc\n",
         system_lyxdir);
      chdir (system_lyxdir);
      if (ReadRCFile("system.lyxrc", lyxrc) < 0) {
        fprintf (stderr, "Could not read system.lyxrc. Using defaults.\n");
      }
   }    
   chdir(cpwd);
   delete(cpwd);


   /* load the layouts in the layout choice  */
   LyXSetStyle(lyxrc.lyx_layout_lib);

   SetZoom(lyxrc.default_zoom);
   
   /* init all buffers */
   // if the buffers was a linked list this
   // wouldn't be beeded. now the buffer struct inits itself.
   for (i=0; i < NUMBER_OF_BUFFERS; i++) {
      buffer[i] = new Buffer();
   }
   current_buffer = 0;
   old_current_buffer = current_buffer;

   /* set the textclass choice */
   for (i = 0; !StringEqual(NameOfClass(i),"dummy"); i++) {
      fl_addto_choice(fd_form_document->choice_class, NameOfClass(i));
   }    

   
   /* install the signal handlers */ 
   signal(SIGHUP, error_handler);
   signal(SIGFPE, error_handler);
   signal(SIGSEGV, error_handler);

   /* install the file handlers */
   LyxServerInit();
   UpdateMenuBar();
   
   // load the lastfiles mini-database
   lastfiles = new LastFiles(lyxrc.lastfiles);
   
   /* load the files specified in the command line */ 
   
   for (int argi = 1; argi < argc; argi++) {
      LoadLyXFile(argv[argi]);
   }
   old_current_buffer = current_buffer;
   current_buffer = 0;
}
 

   
/* callbacks for form form_main */
void FootCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Inserting Footnote...","","");
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->InsertFootnoteEnvironement(LYX_FOOTNOTE);
      Update(1);
   }
}

void LayoutsCB(FL_OBJECT *ob, long data) {
   if (!buffer[current_buffer]->screen)
     return;
   buffer[current_buffer]->screen->HideCursor();
   Update(-2);
   int tmplayout = fl_get_choice(ob) - 1;
   
   if (current_layout != tmplayout) {
      current_layout = tmplayout;
      buffer[current_buffer]->text->SetLayout(current_layout);
   }
   Update(1);
}

void WorkAreaCB(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void ScrollCB(FL_OBJECT *ob, long data)
{
   current_scrollbar_value = (long)fl_get_slider_value(ob);
   if (current_scrollbar_value < 0)
      current_scrollbar_value = 0;
   
   if (!buffer[current_buffer]->screen)
     return;
   
   buffer[current_buffer]->screen->Draw(current_scrollbar_value);
}

/* build either TeX (flag = 0), dvi (flag = 1), ... output */
int RunLinuxDoc(int flag, const char *filename) {
   int n, n2, i;
   int a=0;
   int dot = 0;
   int path = 0;
   // int errors = -1;
   TeXErrors errors;
   
   char *s = NULL, *s2 = NULL, *s4 = NULL, *cpwd = NULL;
   char add_flags[3];

   if ((flag == -1) || (buffer[current_buffer]->text)) {
      ProhibitInput();

      /* find the end of the string */ 
      for (n=0; filename[n]; n++);
      
      /* go back to the first dot */
      for (n2=n; n2 && filename[n2]!='.' && filename[n2]!='/'; n2--);
      
      /* when you find a dot, use this as extension */ 
      if (filename[n2]=='.')
         dot = n2;

      s = new char[dot+10];
      s2 = new char[n+100]; /* used to print command strings into */
      s4 = new char[dot+10];
      cpwd = new char[500];

      for (path=dot; path && filename[path]!='/'; 
         path--);
      if (filename[path]=='/') path++;
      
      for (i=0; i<path; i++)
         s4[i] = filename[i];
      s4[path] = '\0';

      for (i=path; i<dot; i++) 
         s[i-path] = filename[i];
      s[i-path] = '.';
      s[i+1-path] = 's';
      s[i+2-path] = 'g';
      s[i+3-path] = 'm';
      s[i+4-path] = 'l';
      s[i+5-path] = '\0';

      getcwd(cpwd, 500);
      if (s4[0]) chdir(s4);

      if (flag != -1) 
         MakeLinuxDocFile(s, buffer[current_buffer]->parameters, 0);

      *add_flags = 0;
      if (lyxrc.parameters.papersize == PAPER_A4) strcpy(add_flags, "-a");

      switch (flag) {
      case -1:			/* Import file */
         SetMinibuffer("Importing LinuxDoc SGML file :", filename, "...");
         sprintf(s2, "sgml2lyx %s %s", lyxrc.sgml_extra_options, s);
         if (system(s2)) return 1;
         break;
      case 0:			/* TeX output asked */
         SetMinibuffer("Converting LinuxDoc SGML to TeX file...","","");   
         sprintf(s2, "sgml2latex %s %s %s", add_flags, 
		 lyxrc.sgml_extra_options, s);
         if (system(s2)) return 1;
         break;
      case 1:			/* dvi output asked */
         SetMinibuffer("Converting LinuxDoc SGML to dvi file...","","");   
         sprintf(s2, "sgml2latex %s -d %s %s", add_flags,
		 lyxrc.sgml_extra_options, s);
         if (system(s2)) return 1;
         buffer[current_buffer]->dvi_current = 1;
         break;
      default:			/* unknown output */
         break;
      }

      AllowInput();

      chdir(cpwd);

      delete(s);
      delete(s2);
      delete(s4);
      delete(cpwd);
   }

   /* reserved for future error handling */
   return 0;
}

int RunLaTeX() {
   int n, n2, i;
   int a=0;
   int dot = 0;
   int path = 0;
   // int errors = -1;
   TeXErrors errors;
   
   char *s = NULL, *s2 = NULL, *s4 = NULL, *cpwd = NULL;

   if (buffer[current_buffer]->text) {
      ProhibitInput();

      /* find the end of the string */ 
      for (n=0; buffer[current_buffer]->filename[n]; n++);

      
      /* go back to the first dot */
      for (n2=n; n2 && buffer[current_buffer]->filename[n2]!='.'
           && buffer[current_buffer]->filename[n2]!='/'; n2--);
      
      /* when you find a dot, use this as extension */ 
      if (buffer[current_buffer]->filename[n2]=='.')
         dot = n2;

      s = new char[dot+10];
      s2 = new char[n+100]; /* used to print command strings into */
      s4 = new char[dot+10];
      cpwd = new char[500];

      for (path=dot; path && buffer[current_buffer]->filename[path]!='/';
           path--);
      if (buffer[current_buffer]->filename[path]=='/')
          path++;
      
      for (i=0; i<path; i++)
         s4[i] = buffer[current_buffer]->filename[i];
      s4[path] = '\0';

      getcwd(cpwd, 500);
      if (s4[0]) chdir(s4);

      SetMinibuffer("Running LaTeX...","","");   

      /* get LaTeX-Filename */
            
      for (i=path; i<dot; i++) 
      	 s[i-path] = buffer[current_buffer]->filename[i];
      s[i-path] = '.';
      s[i+1-path] = 't';
      s[i+2-path] = 'e';
      s[i+3-path] = 'x';
      s[i+4-path] = '\0';
      dot = i;
      
      /*Remove all errors */

      LyXParagraph *par = buffer[current_buffer]->parameters.paragraph;
      a = 0;
      while (par) {
         a += par->AutoDeleteInsets();
         par = par->next;
      }
      
      if (!buffer[current_buffer]->dvi_current) {
 	 MakeLaTeXFile(s, buffer[current_buffer]->parameters, 0);
	 
	 buffer[current_buffer]->dvi_current = 0;
      }
      
      /* new stuff */ 
      sprintf(s2, "%s %s", lyxrc.latex_command, s);
      system(s2);
      buffer[current_buffer]->dvi_current = 1;
   
      /* scanning errors */ 
      s[dot-path] = '.';
      s[dot+1-path] = 'l';
      s[dot+2-path] = 'o';
      s[dot+3-path] = 'g';
      s[dot+4-path] = '\0';

      SetMinibuffer("Scanning log-file ...","","");   

      ScanLogFile(s,&errors);

      SetMinibuffer("Done.","","");
      
      if (a || errors.number_of_errors)
         RedoCurrentBuffer();
      
      if (errors.status == NO_LOGFILE) {
         WriteAlert("LaTeX did not work!", "Missing log file: ", s);
	 buffer[current_buffer]->dvi_current = 0;
      }
      
      AllowInput();

      chdir(cpwd);

      delete(s);
      delete(s2);
      delete(s4);
      delete(cpwd);
   }
      
   return errors.number_of_errors;
}

/* 1 means preview, 0 means print */ 
void RunScript(char flag, char* command) {
   int n, n2, i;
   int ret = 0;
   int dot=0;
   int path=0;
   if (buffer[current_buffer]->text) {
      
      /* get DVI-Filename */
      
      /* find the end of the string */ 
      for (n=0; buffer[current_buffer]->filename[n]; n++);

      
      /* go back to the first dot */ 
      for (n2=n; n2 && buffer[current_buffer]->filename[n2]!='.'
	   && buffer[current_buffer]->filename[n2]!='/'; n2--);
      
      /* when you find a dot, use this as extension */ 
      if (buffer[current_buffer]->filename[n2]=='.')
      	dot = n2;

      char *s = new char[dot+10];
      char *s4 = new char[dot+10];
      
      for (path=dot; path && buffer[current_buffer]->filename[path]!='/';path--);
      if (buffer[current_buffer]->filename[path]=='/')
      	path++;
      
      for (i=0; i<path; i++)
      	s4[i] = buffer[current_buffer]->filename[i];
      s4[path] = '\0';
      
      for (i=path; i<dot; i++) 
      	s[i-path] = buffer[current_buffer]->filename[i];
      s[i-path] = '.';
      s[i+1-path] = 'd';
      s[i+2-path] = 'v';
      s[i+3-path] = 'i';
      s[i+4-path] = '\0';
      dot = i;
      char *s2 = new char[n+100]; /* used to print command strings into */
      char *cpwd = new char[500];
      getcwd(cpwd, 500);
      
      if (!buffer[current_buffer]->dvi_current)
        if (IsLinuxDoc()) ret= RunLinuxDoc(1, 
           buffer[current_buffer]->filename);
      	else ret = RunLaTeX();
      
      if (s4[0])
      	chdir(s4);
      delete (s4);
      
      if (!OpenAllFile(s)) {
	 if (flag)
	    WriteAlert("Error!", "Cannot preview: No LaTeX-Output (dvi)","");
	 else
	    WriteAlert("Error!", "Cannot print: No LaTeX-Output (dvi)","");
      }
      else {
	 CloseFile();
      }
      
      if (ret > 0) {
	 char *s3 = new char [50];
	 if (ret == 1)
	    sprintf(s3, "%d error detected",ret);
	 else
	    sprintf(s3, "%d errors detected",ret);
	 WriteAlert("LaTeX Error!", s3, "");
	 delete(s3);
      }

      
      if (flag==3) {
      SetMinibuffer("Executing command:", command,"");
      	 sprintf(s2, "%s ", command);
      }
      else {
	 SetMinibuffer("Executing command:", command, s);   

	 if (flag == 2)
      	    sprintf(s2, "%s %s ", command, s);
	 else
      	    sprintf(s2, "%s %s &", command, s);
      }
      
      system(s2); 
      
      chdir(cpwd);
      delete(cpwd);
      
      delete(s2);	
      
      delete(s);
   }
   
   
}

void MenuPreviewPS() {
   
   if (!buffer[current_buffer]->text)
      return;

   ProhibitInput();
   /* calculate ps filename */ 
   char* filename = buffer[current_buffer]->filename;
   
   int dot=0;
   int path = 0;
   int n=0;
   int n2=0;
   int i=0;
   
   /* find the end of the string */ 
   for (n=0; filename[n]; n++);
   
   /* go back to the first dot */ 
   for (n2=n; n2 && filename[n2]!='.'
	&& filename[n2]!='/'; n2--);
   
   /* when you find a dot, use this as extension */ 
   if (filename[n2]=='.')
      dot = n2;
   else
      dot = n;
   
   for (path=dot; path && filename[path]!='/';path--);
   if (filename[path]=='/')
      path++;
   

   char *ps = new char[dot+15];
   
   
   for (i=path; i<dot; i++) 
      ps[i-path] = filename[i];
   ps[i-path] = '.';
   ps[i+1-path] = 'p';
   ps[i+2-path] = 's';
   ps[i+3-path] = '_';
   ps[i+4-path] = 't';
   ps[i+5-path] = 'm';
   ps[i+6-path] = 'p';
   ps[i+7-path] = '\0';

   char* paper = "a4";
   
   switch (buffer[current_buffer]->parameters.papersize) {
    case PAPER_USLETTER:
      paper = "letter";
      break;
    case PAPER_A5PAPER:
      paper = "a5";
      break;
    case PAPER_B5PAPER:
      paper = "b5";
      break;
    case PAPER_EXECUTIVEPAPER:
      paper = "foolscap";
      break;
    case PAPER_LEGALPAPER:
      paper = "legal";
      break;
   }

   
   char* orientationflag = "";
   if (buffer[current_buffer]->parameters.orientation == ORIENTATION_LANDSCAPE)
      orientationflag = "-t landscape";
   
   char *command = new char[n + 100];
   
   sprintf(command, "dvips -o %s -t %s %s", ps, paper, orientationflag);
   
   RunScript(2, command);
   sprintf(command, "ghostview -swap %s &", ps);
   RunScript(3, command);
   AllowInput();
   delete(command);
   delete(ps);
}


void MenuPreview() {
   if (!buffer[current_buffer]->text)
      return;
   
   char* command = new char[50];
   
   char* paper = "a4";
   
   switch (buffer[current_buffer]->parameters.papersize) {
    case PAPER_USLETTER:
      paper = "us";
      break;
    case PAPER_A5PAPER:
      paper = "a5";
      break;
    case PAPER_B5PAPER:
      paper = "b5";
      break;
    case PAPER_EXECUTIVEPAPER:
      paper = "foolscap";
      break;
    case PAPER_LEGALPAPER:
      paper = "legal";
      break;
   }
   
   char* rotate ="";
   
   if (buffer[current_buffer]->parameters.orientation == ORIENTATION_LANDSCAPE)
      rotate ="r";
   
   sprintf(command, "xdvi -paper %s%s %s", paper, rotate, 
	   lyxrc.xdvi_extra_options);
   
   RunScript(1, command);
   
   delete(command);
}


//void set_pup_return(int i)
//{
//   char buf[25];
//   fprintf(stderr,"Pop-up Returned %d", i);
   //fl_set_object_label(pret, buf);
//}


void FileCB(FL_OBJECT *ob, long data) {
   ShowFileMenu(ob,data);
}


void AllFloats(char flag) {
   if (!buffer[current_buffer]->text)
      return;
   
   if (!flag && buffer[current_buffer]->text->cursor.par->footnoteflag != LYX_NO_FOOTNOTE) {
      OpenStuff();
   }
   else
      BeforeChange();
   LyXParagraph *par = buffer[current_buffer]->parameters.paragraph;
   while (par) {
      if (flag) {
	 if (par->footnoteflag == LYX_CLOSED_FOOTNOTE)
	    par->footnoteflag = LYX_OPEN_FOOTNOTE;
      }
      else  {
	 if (par->footnoteflag == LYX_OPEN_FOOTNOTE)
	    par->footnoteflag = LYX_CLOSED_FOOTNOTE;
      }
      par = par->next;
   }
   if (flag){
      buffer[current_buffer]->text->cursor.par =   
      buffer[current_buffer]->text->cursor.par->ParFromPos(buffer[current_buffer]->text->cursor.pos);
      buffer[current_buffer]->text->cursor.pos = 
      buffer[current_buffer]->text->cursor.par->last;
   }
   else {
      if (buffer[current_buffer]->text->cursor.par->layout == LYX_DUMMY_LAYOUT){
	 buffer[current_buffer]->text->cursor.par = buffer[current_buffer]->text->cursor.par->FirstPhysicalPar();
	 buffer[current_buffer]->text->cursor.pos = 0;
      }
   }
   
   RedoCurrentBuffer();
}



void EditCB(FL_OBJECT *ob, long data) {
  ShowEditMenu(ob,data);

}


void MenuLayoutCharacter() {
   fl_show_form(fd_form_character->form_character,
		FL_PLACE_CENTER | FL_FREE_SIZE,FL_FULLBORDER,
		"Character Layout");
}
void MenuLayoutParagraph() {
      if (buffer[current_buffer]->text) {
	 fl_set_input(fd_form_paragraph->input_labelwidth,
		      buffer[current_buffer]->text->cursor.par->GetLabelWidthString());
	 fl_set_button(fd_form_paragraph->radio_align_right, 0);
	 fl_set_button(fd_form_paragraph->radio_align_left, 0);
	 fl_set_button(fd_form_paragraph->radio_align_center, 0);
	 fl_set_button(fd_form_paragraph->radio_align_block, 0);
	 int align = buffer[current_buffer]->text->cursor.par->GetAlign();
	 if (align == LYX_ALIGN_LAYOUT)
	   align = style(buffer[current_buffer]->parameters.textclass, buffer[current_buffer]->text->cursor.par->GetLayout())->align;
	 switch (align) {
	  case LYX_ALIGN_RIGHT: fl_set_button(fd_form_paragraph->radio_align_right, 1); break;
	  case LYX_ALIGN_LEFT: fl_set_button(fd_form_paragraph->radio_align_left, 1); break;
	  case LYX_ALIGN_CENTER: fl_set_button(fd_form_paragraph->radio_align_center, 1); break;
	  default: fl_set_button(fd_form_paragraph->radio_align_block, 1); break;
	 }
	 fl_set_button(fd_form_paragraph->check_lines_top,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->line_top);
	 fl_set_button(fd_form_paragraph->check_lines_bottom,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->line_bottom);
	 fl_set_button(fd_form_paragraph->check_pagebreaks_top,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->pagebreak_top);
	 fl_set_button(fd_form_paragraph->check_pagebreaks_bottom,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->pagebreak_bottom);
	 fl_set_button(fd_form_paragraph->check_noindent,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->noindent);
	 fl_set_button(fd_form_paragraph->check_fill_top,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->fill_top);
	 fl_set_button(fd_form_paragraph->check_fill_bottom,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->fill_bottom);
	 fl_set_slider_value(fd_form_paragraph->slider_spaces_top,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->added_space_top);
	 fl_set_slider_value(fd_form_paragraph->slider_spaces_bottom,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->added_space_bottom);
	 fl_set_button(fd_form_paragraph->check_noindent,
		       buffer[current_buffer]->text->cursor.par->FirstPhysicalPar()->noindent);
      }
      fl_show_form(fd_form_paragraph->form_paragraph,
		   FL_PLACE_CENTER | FL_FREE_SIZE,FL_FULLBORDER,
		   "Paragraph Layout");
}

void MenuLayoutDocument() {
   fl_set_choice_text(fd_form_document->choice_class, 
		      NameOfClass(buffer[current_buffer]->parameters.textclass));
   fl_set_choice_text(fd_form_document->choice_language, 
		      buffer[current_buffer]->parameters.language);
   fl_set_choice_text(fd_form_document->choice_fonts, 
		      buffer[current_buffer]->parameters.fonts);
   fl_set_choice_text(fd_form_document->choice_inputenc, 
		      buffer[current_buffer]->parameters.inputenc);
   fl_set_choice_text(fd_form_document->choice_epsfig, 
		      buffer[current_buffer]->parameters.epsfig);
   fl_set_choice(fd_form_document->choice_fontsize, 
		 buffer[current_buffer]->parameters.fontsize - 9);
   fl_set_choice(fd_form_document->choice_papersize,
		 buffer[current_buffer]->parameters.papersize + 1);
   fl_set_choice(fd_form_document->choice_pagestyle,
		 buffer[current_buffer]->parameters.pagestyle + 1);
   fl_set_button(fd_form_document->radio_indent, 0);
   fl_set_button(fd_form_document->radio_skip, 0);
   if (buffer[current_buffer]->parameters.paragraph_separation == LYX_PARSEP_INDENT)
      fl_set_button(fd_form_document->radio_indent, 1);
   else
      fl_set_button(fd_form_document->radio_skip, 1);
   fl_set_button(fd_form_document->radio_portrait, 0);
   fl_set_button(fd_form_document->radio_landscape, 0);
   if (buffer[current_buffer]->parameters.orientation == ORIENTATION_LANDSCAPE)
      fl_set_button(fd_form_document->radio_landscape, 1);
   else
      fl_set_button(fd_form_document->radio_portrait, 1);
   fl_set_button(fd_form_document->radio_sides_one, 0);
   fl_set_button(fd_form_document->radio_sides_two, 0);
   if (buffer[current_buffer]->parameters.sides == 2)
      fl_set_button(fd_form_document->radio_sides_two, 1);
   else
      fl_set_button(fd_form_document->radio_sides_one, 1);
   fl_set_button(fd_form_document->radio_columns_one, 0);
   fl_set_button(fd_form_document->radio_columns_two, 0);
   if (buffer[current_buffer]->parameters.columns == 2)
      fl_set_button(fd_form_document->radio_columns_two, 1);
   else
      fl_set_button(fd_form_document->radio_columns_one, 1);
   fl_set_slider_value(fd_form_document->slider_baselinestretch, 
		       buffer[current_buffer]->parameters.baseline);
   fl_set_slider_value(fd_form_document->slider_secnumdepth, 
		       buffer[current_buffer]->parameters.secnumdepth);
   fl_set_slider_value(fd_form_document->slider_tocdepth, 
		       buffer[current_buffer]->parameters.tocdepth);
   fl_set_input(fd_form_document->input_extra,
		buffer[current_buffer]->parameters.options);
   
   fl_show_form(fd_form_document->form_document,
		FL_PLACE_CENTER | FL_FREE_SIZE,FL_FULLBORDER,
		"Document Layout");
}

void MenuLayoutQuotes() {
     if (buffer[current_buffer]->text){
       fl_set_choice(fd_form_quotes->choice_quotes_language,
		     buffer[current_buffer]->parameters.quotes_language + 1);
       fl_set_button(fd_form_quotes->radio_single, 0);
       fl_set_button(fd_form_quotes->radio_double, 0);
       
       if (buffer[current_buffer]->parameters.quotes_times == 1)
	 fl_set_button(fd_form_quotes->radio_single, 1);
       else
	 fl_set_button(fd_form_quotes->radio_double, 1);
     }
     fl_show_form(fd_form_quotes->form_quotes,
			FL_PLACE_CENTER | FL_FREE_SIZE,FL_FULLBORDER,
			"Quotes");
}

void MenuLayoutPreamble() {
   if (buffer[current_buffer]->text){
      fl_set_input(fd_form_preamble->input_preamble,
		     buffer[current_buffer]->parameters.preamble);
   }
   fl_show_form(fd_form_preamble->form_preamble,
		FL_PLACE_CENTER | FL_FREE_SIZE,FL_FULLBORDER,
		"LaTeX Preamble");
}


void LayoutCB(FL_OBJECT *ob, long data){
  ShowLayoutMenu(ob,data);
}


/* only usable from special cb!!! */ 
void InsertLaTeXTable(char* table) {
   buffer[current_buffer]->text->BreakParagraph();
   Update(-1);
   
   if (buffer[current_buffer]->text->cursor.par->Last()) {
      buffer[current_buffer]->text->CursorLeft();
      
      buffer[current_buffer]->text->BreakParagraph();
      Update(-1);
   }
   
   buffer[current_buffer]->text->SetLayout(NumberOfLayout(buffer[current_buffer]->parameters.textclass, "LaTeX"));
   
   buffer[current_buffer]->text->SetParagraph(0, 0,
						0, 0,
						0, 0,
						0, 0,
						LYX_ALIGN_LAYOUT, 
						NULL,
						0);
   Update(-1);
   
   buffer[current_buffer]->text->current_font.latex = LYX_NO_LATEX;
   
   InsetLatexDel *new_inset = new InsetLatexDel(table);
   buffer[current_buffer]->text->InsertInset(new_inset);
   Update(-1);
   
   buffer[current_buffer]->text->CursorRight();
}

/* callbacks added for LinuxDoc support (Pascal Andr) : 
 *   - marks and cross references TeX macros
 *   - url display macros
 *   - htmlurl macros
 */

/* inserts an inset (used for LinuxDoc SGML related callbacks) */
void InsertLatexDelInset(char *macro)
{
  InsetLatexDel *new_inset = new InsetLatexDel(macro);
  buffer[current_buffer]->text->InsertInset(new_inset);
  Update(-1);
}

/* inserts a cross reference mark skeleton */ 
void CrossMarkCB() {
  //  InitLatexDelInsert();
   
  InsertLatexDelInset("\\label{");
  InsertLatexDelInset("}");
  buffer[current_buffer]->text->InsertChar(' ');
}

/* inserts a cross reference skeleton */ 
void CrossReferenceCB() {
  //  InitLatexDelInsert();
   
  InsertLatexDelInset("\\ref{");
  InsertLatexDelInset("}{");
  InsertLatexDelInset("}");
  buffer[current_buffer]->text->InsertChar(' ');
}

/* inserts an url skeleton */ 
void UrlCB() {
  //  InitLatexDelInsert();
   
  InsertLatexDelInset("\\url{");
  InsertLatexDelInset("}{");
  InsertLatexDelInset("}");
  buffer[current_buffer]->text->InsertChar(' ');
}

/* inserts an html url skeleton */ 
void HtmlUrlCB() {
  //  InitLatexDelInsert();
   
  InsertLatexDelInset("\\htmlurl{");
  InsertLatexDelInset("}{");
  InsertLatexDelInset("}");
  buffer[current_buffer]->text->InsertChar(' ');
}

/* end of LinuxDoc SGML special CB */

void NoteCB() {
   InsetInfo *new_inset = new InsetInfo();
   char string[200];
   sprintf(string, "%s %s\n", getenv("USER"), date());
   new_inset->Set(string);
   buffer[current_buffer]->text->InsertInset(new_inset);
   Update(-1);
}


void SpecialCB(FL_OBJECT *ob, long data) {
  if (!StringEqual(NameOfClass(current_textclass), "linuxdoc"))
    ShowSpecialMenu(ob,0);
  else
    ShowSpecialMenu(ob,1);
  
}


void OptionsCB(FL_OBJECT *ob, long data)
{
  ShowOptionsMenu(ob,data);
}


void BufferCB(FL_OBJECT *ob, long data){
   ShowBufferMenu(ob,data);
}

void MakeBufferMenu(int pup)
{
   int n,ant=0;
   // this would be so much easier if the
   // we had a linked list of buffers
   for(n=0;n<NUMBER_OF_BUFFERS;n++){
      if(buffer[n]->parameters.paragraph){
	 fl_addtopup(pup,buffer[n]->filename);
	 ant++;
      }
   }
   if (ant ==0) fl_addtopup(pup,"No Files Loaded!%t");
}

void HandleBufferMenu(int choice){
   // shoulf perhaps move this func to menus.C   
   //fprintf(stderr,"I got choice = %d\n",choice);
   
   int i=0;
   while (choice--) {
      while (!buffer[i++]->parameters.paragraph);
   }
   i--; // this is ok since choice never == 0
   
   //fprintf(stderr,"Hmm, I think you want to look at buffer...%d!\n",i);
   
   if (i>=NUMBER_OF_BUFFERS)
     return;

   old_current_buffer = current_buffer;
   current_buffer = i;
   ResizeCurrentBufferPseudoExpose();
   UpdateMenuBar();
}

void HelpCB(FL_OBJECT *ob, long data) {
   ShowHelpMenu(ob,data);
}


void OpenStuff() {
   if (buffer[current_buffer]->text)  {
     SetMinibuffer("Open/Close...","","");
      buffer[current_buffer]->screen->HideCursor();
      BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->ToggleFootnote();
      Update(0);
   }
}

void HyphenationPoint() {
   if (buffer[current_buffer]->text)  {
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      InsetLatexDel *new_inset = new InsetLatexDel("\\-");
      buffer[current_buffer]->text->InsertInset(new_inset);
      Update(0);
   }
}
void Newline() {
   if (buffer[current_buffer]->text)  {
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->InsertChar(LYX_META_NEWLINE);
      Update(0);
   }
}
void ProtectedBlank() {
   if (buffer[current_buffer]->text)  {
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->InsertChar(LYX_META_PROTECTED_SEPARATOR);
      Update(0);
   }
}
void HFill() {
   if (buffer[current_buffer]->text)  {
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->InsertChar(LYX_META_HFILL);
      Update(0);
   }
}


void EmphCB(FL_OBJECT *ob, long data)
{   
   if (buffer[current_buffer]->text) { 
      SetMinibuffer("Toggle Emphasize!","","");
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->ToggleEmphasize();
      Update(1);
   }
}

void NounCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Toggle Noun!","","");
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->ToggleNoun();
      Update(1);
   }
}

void MarginCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Inserting marginpar...","","");
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->InsertFootnoteEnvironement(LYX_MARGIN);
      Update(1);
   }
}

void FigureCB(FL_OBJECT *ob, long data) {
   fl_show_form(fd_form_figure->form_figure,
		FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		"Figure");
}



void TableCB(FL_OBJECT *ob, long data){
   fl_show_form(fd_form_table->form_table,
		FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		"Table");
}


void CopyCB() {
  if (buffer[current_buffer]->text) {
     buffer[current_buffer]->text->CopySelection();
      SetMinibuffer("Copy!","","");
   }
}

void CutCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->CutSelection();
      Update(1);
      SetMinibuffer("Cut!","","");
   }
}

void PasteCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Paste!","","");
      buffer[current_buffer]->screen->HideCursor();
      /* clear the selection */ 
      buffer[current_buffer]->screen->ToggleSelection();
      buffer[current_buffer]->text->ClearSelection();
      Update(-2);
      
      /* paste */ 
      buffer[current_buffer]->text->PasteSelection();
      Update(1);
   }
}

void MeltCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Melt!","","");
      buffer[current_buffer]->screen->HideCursor();
      BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->MeltFootnoteEnvironement();
      Update(1);
   }
}

void TexCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Toggle TeX!","","");
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->ToggleLaTeX();
      Update(1);
   }
}

void FormulaCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Toggle Math!","","");
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->ToggleFormLaTeX();
      Update(1);
   }
}

void MathCB(FL_OBJECT *ob, long data){
  ShowMathMenu(ob,data);
}


void DepthCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      buffer[current_buffer]->text->IncDepth();
      Update(1);
      SetMinibuffer("changed environment depth (in possible range, maybe not)", "", "");
   }
}

void FreeCB(FL_OBJECT *ob, long data)
{
  SetMinibuffer("Toggle User style","","");

  LyXFont font = { 
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE
    };
  
   int pos;
  
  pos = fl_get_choice(fd_form_character->choice_family);
  if (pos > 1)
    font.family = pos - 2;
  
  pos = fl_get_choice(fd_form_character->choice_series);  if (pos > 1)
    font.series = pos - 2;
  
  pos = fl_get_choice(fd_form_character->choice_shape);  if (pos > 1)
    font.shape = pos - 2;

  pos = fl_get_choice(fd_form_character->choice_bar); if (pos > 1)
     font.bar = pos - 2;
  
  pos = fl_get_choice(fd_form_character->choice_size);  if (pos > 1)
    font.size = pos - 1; /* heres a -1 because of the
			  * unaccesible tiny font, that I need
			  * for smallcaps in small font */
  
  if (buffer[current_buffer]->text) { 
     buffer[current_buffer]->text->ToggleFree(font);
      Update(1);
    }
}



/* callbacks for form form_title */
void TimerCB(FL_OBJECT *ob, long data)
{
   // only if the form still exists
   if (fd_form_title->form_title != NULL) {
      // fprintf(stderr,"TimerCB: deleting existing title form [%d]\n",(int)ob);      
      fl_hide_form(fd_form_title->form_title);
      fl_free_form(fd_form_title->form_title);
      fd_form_title->form_title = NULL;
   }
   // else fprintf(stderr,"TimerCB: title form already deleted!\n");   
}

void UpdateTimerCB(FL_OBJECT *ob, long data) {
   if (!buffer[current_buffer]->screen || !updatetimer)
     return;
   buffer[current_buffer]->screen->HideCursor();
   Update(-2);
   /* this update can happen, even when the work area has lost
    * the focus. So supress the cursor in that case */ 
   updatetimer = 0;
}

void AutosaveTimerCB(FL_OBJECT *ob, long data) {
   AutoSave();
}



/* callbacks for form form_paragraph */
void ParagraphApplyCB(FL_OBJECT *ob, long data)
{
   if (!buffer[current_buffer]->text)
     return;
   char line_top, line_bottom;
   char pagebreak_top, pagebreak_bottom;
   float space_top, space_bottom;
   char fill_top, fill_bottom;
   char align;
   const char* labelwidthstring;
   char noindent;
   
   line_top = fl_get_button(fd_form_paragraph->check_lines_top);
   line_bottom = fl_get_button(fd_form_paragraph->check_lines_bottom);
   pagebreak_top = fl_get_button(fd_form_paragraph->check_pagebreaks_top);
   pagebreak_bottom = fl_get_button(fd_form_paragraph->check_pagebreaks_bottom);
   space_top = fl_get_slider_value(fd_form_paragraph->slider_spaces_top);
   space_bottom = fl_get_slider_value(fd_form_paragraph->slider_spaces_bottom);
   fill_top = fl_get_button(fd_form_paragraph->check_fill_top);
   fill_bottom = fl_get_button(fd_form_paragraph->check_fill_bottom);
   
   if (fl_get_button(fd_form_paragraph->radio_align_left))
     align = LYX_ALIGN_LEFT;
   else if (fl_get_button(fd_form_paragraph->radio_align_right))
     align = LYX_ALIGN_RIGHT;
   else if (fl_get_button(fd_form_paragraph->radio_align_center))
     align = LYX_ALIGN_CENTER;
   else 
     align = LYX_ALIGN_BLOCK;
   
   labelwidthstring = fl_get_input(fd_form_paragraph->input_labelwidth);
   noindent = fl_get_button(fd_form_paragraph->check_noindent);
   
   buffer[current_buffer]->text->SetParagraph(line_top, line_bottom,
				      pagebreak_top, pagebreak_bottom,
				      space_top, space_bottom,
				      fill_top, fill_bottom,
				      align, 
				      labelwidthstring,
				      noindent);
   Update(1);
   SetMinibuffer("user defined paragraph layout set","","");
}

void ParagraphCancelCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_paragraph->form_paragraph);
}

void ParagraphOKCB(FL_OBJECT *ob, long data)
{
   ParagraphApplyCB(ob, data);
   ParagraphCancelCB(ob, data);
}


/* callbacks for form form_search */
void SearchCancelCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_search->form_search);
}

void SearchReplaceCB(FL_OBJECT *ob, long data){
   if (!buffer[current_buffer]->screen)
     return;
   char* replacestring = (char *) fl_get_input(fd_form_search->input_replace);
   buffer[current_buffer]->screen->HideCursor();
   Update(-2);
   /* clear the selection (if there is any) */ 
   buffer[current_buffer]->screen->ToggleSelection();
   buffer[current_buffer]->text->ReplaceSelectionWithString(replacestring);
   
   buffer[current_buffer]->text->SetSelectionOverString(replacestring);
   Update(1);
}

void SearchForwardCB(FL_OBJECT *ob, long data){
   if (!buffer[current_buffer]->screen)
     return;
   
   buffer[current_buffer]->screen->HideCursor();
   Update(-2);
   char* searchstring = (char*) fl_get_input(fd_form_search->input_search);
   
   if (buffer[current_buffer]->text->selection) 
     buffer[current_buffer]->text->cursor = buffer[current_buffer]->text->sel_end_cursor;
   
   if (!searchstring[0] || buffer[current_buffer]->text->SearchForward(searchstring)) {
      Update(-2);
      /* clear the selection (if there is any) */ 
      buffer[current_buffer]->screen->ToggleSelection();
      buffer[current_buffer]->text->ClearSelection();
      /* set the new selection */
      buffer[current_buffer]->text->SetSelectionOverString(searchstring);
      buffer[current_buffer]->screen->ToggleSelection();
   }
   else {
      LyXBell();
   }
   if (work_area_active)
     buffer[current_buffer]->screen->ShowCursor();
}

void SearchBackwardCB(FL_OBJECT *ob, long data){
   if (!buffer[current_buffer]->screen)
     return;
   
   char* searchstring = (char *) fl_get_input(fd_form_search->input_search);
   
   buffer[current_buffer]->screen->HideCursor();
   Update(-2);

   if (buffer[current_buffer]->text->selection) 
     buffer[current_buffer]->text->cursor = buffer[current_buffer]->text->sel_start_cursor;

   if (!searchstring[0] || buffer[current_buffer]->text->SearchBackward(searchstring)) {
      Update(-2);
      /* clear the selection (if there is any) */ 
      buffer[current_buffer]->screen->ToggleSelection();
      buffer[current_buffer]->text->ClearSelection();
      /* set the new selection */
      buffer[current_buffer]->text->SetSelectionOverString(searchstring);
      buffer[current_buffer]->screen->ToggleSelection();
   }
   else {
      LyXBell();
   }
   if (work_area_active)
     buffer[current_buffer]->screen->ShowCursor();
}



/* callbacks for form form_character */

void CharacterApplyCB(FL_OBJECT *ob, long data) {
  

  LyXFont font = { 
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE,
    LYX_NO_FONT_CHANGE
    };
  
   int pos;
  
  pos = fl_get_choice(fd_form_character->choice_family);
  if (pos > 1)
    font.family = pos - 2;
  
  pos = fl_get_choice(fd_form_character->choice_series);  if (pos > 1)
    font.series = pos - 2;
  
  pos = fl_get_choice(fd_form_character->choice_shape);  if (pos > 1)
    font.shape = pos - 2;
  
  pos = fl_get_choice(fd_form_character->choice_bar);  if (pos > 1)
    font.bar = pos - 2;
  
  pos = fl_get_choice(fd_form_character->choice_size);  if (pos > 1)
    font.size = pos - 1; /* heres a -1 because of the
			  * unaccesible tiny font, that I need
			  * for smallcaps in small font */
  
   if (buffer[current_buffer]->text) { 
      buffer[current_buffer]->text->SetFont(font);
      Update(1);
      SetMinibuffer("user defined character layout set","","");
   }
}

void CharacterCancelCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_character->form_character);
}
void CharacterOKCB(FL_OBJECT *ob, long data){
   CharacterApplyCB(ob,data);
   CharacterCancelCB(ob,data);
}



/* callbacks for form form_document */

void ChoiceClassCB(FL_OBJECT *ob, long data) {
   ProhibitInput();
   if (fl_show_question("Should I set the defaults of textclass",
			fl_get_choice_text(ob),
			"for Columns, Sides, Pagestyle and Extra Options?")) {
      
      fl_set_choice(fd_form_document->choice_pagestyle, 
		    TextClass(fl_get_choice(ob) - 1)->pagestyle + 1);
      fl_set_button(fd_form_document->radio_sides_one, 0);
      fl_set_button(fd_form_document->radio_sides_two, 0);
      if (TextClass(fl_get_choice(ob) - 1)->sides == 2)
      	 fl_set_button(fd_form_document->radio_sides_two, 1);
      else
      	 fl_set_button(fd_form_document->radio_sides_one, 1);
      fl_set_button(fd_form_document->radio_columns_one, 0);
      fl_set_button(fd_form_document->radio_columns_two, 0);
      if (TextClass(fl_get_choice(ob) - 1)->columns == 2)
	   fl_set_button(fd_form_document->radio_columns_two, 1);
	 else
	   fl_set_button(fd_form_document->radio_columns_one, 1);
      fl_set_input(fd_form_document->input_extra,
		   TextClass(fl_get_choice(ob) - 1)->options);
   }
   AllowInput();
}




void DocumentApplyCB(FL_OBJECT *ob, long data){
   char redo = 0;
   delete (buffer[current_buffer]->parameters.language);
   delete(buffer[current_buffer]->parameters.fonts);
   delete(buffer[current_buffer]->parameters.inputenc);
   buffer[current_buffer]->parameters.language = StringCopy(fl_get_choice_text(fd_form_document->choice_language));

   /* this shouldn't be done automatically IMO. For example I write german documents with
      an american keyboard very often. Matthias */
   
   /* ChangeKeymap(buffer[current_buffer]->parameters.language, TRUE, FALSE,
	fl_get_choice(fd_form_document->choice_language)); */
   buffer[current_buffer]->parameters.fonts = StringCopy(fl_get_choice_text(fd_form_document->choice_fonts));
   buffer[current_buffer]->parameters.inputenc = StringCopy(fl_get_choice_text(fd_form_document->choice_inputenc));
   buffer[current_buffer]->parameters.fontsize = 9 + fl_get_choice(fd_form_document->choice_fontsize);
   buffer[current_buffer]->parameters.papersize = -1 + fl_get_choice(fd_form_document->choice_papersize);
   buffer[current_buffer]->parameters.pagestyle = -1 + fl_get_choice(fd_form_document->choice_pagestyle);
   buffer[current_buffer]->parameters.epsfig = StringCopy(fl_get_choice_text(fd_form_document->choice_epsfig));
   
   if (!buffer[current_buffer]->text)
     return;

   char new_class = fl_get_choice(fd_form_document->choice_class) - 1;
   if (buffer[current_buffer]->parameters.textclass != new_class) {
      redo = 1;
      SetMinibuffer("Converting document to new textclass... ","","");
      int ret = buffer[current_buffer]->text->SwitchLayoutsBetweenClasses(buffer[current_buffer]->parameters.textclass,
									  new_class,
									  buffer[current_buffer]->parameters.paragraph);
      
      if (ret){
	 char *s = new char[50];
	 sprintf(s, "%d paragraph(s) couldn't be converted", ret);
	 WriteAlert("Conversion Errors!",s,"into choosen textclass");
	 delete(s);
      }
      
      buffer[current_buffer]->parameters.textclass = new_class;
   }

   char tmpsep = buffer[current_buffer]->parameters.paragraph_separation;
   if (fl_get_button(fd_form_document->radio_indent))
     buffer[current_buffer]->parameters.paragraph_separation = LYX_PARSEP_INDENT;
   else
      buffer[current_buffer]->parameters.paragraph_separation = LYX_PARSEP_SKIP;
   if (tmpsep != buffer[current_buffer]->parameters.paragraph_separation)
      redo = 1;
   
   if (fl_get_button(fd_form_document->radio_landscape))
      buffer[current_buffer]->parameters.orientation = ORIENTATION_LANDSCAPE;
   else
      buffer[current_buffer]->parameters.orientation = ORIENTATION_PORTRAIT;
   if (fl_get_button(fd_form_document->radio_columns_two))
      buffer[current_buffer]->parameters.columns = 2;
   else
      buffer[current_buffer]->parameters.columns = 1;
   if (fl_get_button(fd_form_document->radio_sides_two))
      buffer[current_buffer]->parameters.sides = 2;
   else
      buffer[current_buffer]->parameters.sides = 1;

   if (buffer[current_buffer]->parameters.baseline != fl_get_slider_value(fd_form_document->slider_baselinestretch))
      redo = 1;
   buffer[current_buffer]->parameters.baseline = fl_get_slider_value(fd_form_document->slider_baselinestretch);

   
   double tmpdouble = fl_get_slider_value(fd_form_document->slider_secnumdepth);
   signed char tmpchar = 0;
   
   if (tmpdouble >= 0)
      tmpchar = (signed char) (tmpdouble + 0.5);
   else
      tmpchar = (signed char) (tmpdouble - 0.5);
      
   
   if (buffer[current_buffer]->parameters.secnumdepth != tmpchar)
      redo = 1;
   buffer[current_buffer]->parameters.secnumdepth = tmpchar;
   
   tmpdouble = fl_get_slider_value(fd_form_document->slider_tocdepth);
   
   if (tmpdouble >= 0)
      tmpchar = (signed char) (tmpdouble + 0.5);
   else
      tmpchar = (signed char) (tmpdouble - 0.5);
   buffer[current_buffer]->parameters.tocdepth = tmpchar;
      
   if (redo)
      RedoCurrentBuffer();
   
   SetMinibuffer("user defined document layout set","","");
   buffer[current_buffer]->lyx_current = 0;
   buffer[current_buffer]->dvi_current = 0;
   buffer[current_buffer]->bak_current = 0;

   if (buffer[current_buffer]->parameters.options)
      delete(buffer[current_buffer]->parameters.options);
   buffer[current_buffer]->parameters.options = StringCopy(fl_get_input(fd_form_document->input_extra));
   
}

void DocumentCancelCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_document->form_document);
}


void DocumentOKCB(FL_OBJECT *ob, long data)
{
   DocumentCancelCB(ob,data);
   DocumentApplyCB(ob,data);
}







void GotoError() {
   if (!buffer[current_buffer]->screen)
     return;
   buffer[current_buffer]->screen->HideCursor();
   BeforeChange();
   Update(-2);
   LyXCursor tmp;
   
   if (!buffer[current_buffer]->text->GotoNextError()) {
      if (buffer[current_buffer]->text->cursor.pos 
	  || buffer[current_buffer]->text->cursor.par != buffer[current_buffer]->text->FirstParagraph()) {
	     tmp = buffer[current_buffer]->text->cursor;
	     buffer[current_buffer]->text->cursor.par = buffer[current_buffer]->text->FirstParagraph();
	     buffer[current_buffer]->text->cursor.pos = 0;
	     if (!buffer[current_buffer]->text->GotoNextError()) {
		buffer[current_buffer]->text->cursor = tmp;
		SetMinibuffer("No more errors!","","");
		LyXBell();
	     }
	  }
      else {
	 SetMinibuffer("No more errors!","","");
	 LyXBell();
      }
   }
   Update(0);
   buffer[current_buffer]->text->sel_cursor = buffer[current_buffer]->text->cursor;
}


void GotoNote() {
   if (!buffer[current_buffer]->screen)
     return;
   buffer[current_buffer]->screen->HideCursor();
   BeforeChange();
   Update(-2);
   LyXCursor tmp;
   
   if (!buffer[current_buffer]->text->GotoNextNote()) {
      if (buffer[current_buffer]->text->cursor.pos 
	  || buffer[current_buffer]->text->cursor.par != buffer[current_buffer]->text->FirstParagraph()) {
	     tmp = buffer[current_buffer]->text->cursor;
	     buffer[current_buffer]->text->cursor.par = buffer[current_buffer]->text->FirstParagraph();
	     buffer[current_buffer]->text->cursor.pos = 0;
	     if (!buffer[current_buffer]->text->GotoNextNote()) {
		buffer[current_buffer]->text->cursor = tmp;
		SetMinibuffer("No more notes!","","");
		LyXBell();
	     }
	  }
      else {
	 SetMinibuffer("No more notes!","","");
	 LyXBell();
      }
   }
   Update(0);
   buffer[current_buffer]->text->sel_cursor = buffer[current_buffer]->text->cursor;
}


void InsertCorrectQuote() {
   Inset *new_inset;
   char *s = new char[4];

   switch (buffer[current_buffer]->parameters.quotes_language) {
    case GERMAN:
      s[0]='g';
      break;
    case FRENCH:
      s[0]='f';
      break;
    default:
      s[0]='e';
      break;
   }
   
   
   /* first decide wether left or right */ 
   
   
   char c;
   if  (buffer[current_buffer]->text->cursor.pos )
     c = buffer[current_buffer]->text->cursor.par->GetChar(buffer[current_buffer]->text->cursor.pos -1);
   else c = ' ';
   
   if (c != ' ' &&
       c != '(' &&
       c != '{' &&
       c != '[' &&
       c != '-' &&
       c != ':' &&
       c != LYX_META_HFILL &&
       c != LYX_META_PROTECTED_SEPARATOR &&
       c != LYX_META_NEWLINE) 
     s[1]='r';  /* right quote */
   else
     s[1]='l';     /* left quote */ 
   
   
   if (buffer[current_buffer]->parameters.quotes_times == 1)
     s[2]='s';			       /* single */
   else
     s[2]='d';			       /* double */
   
   s[3]='\0';
   
   new_inset = new InsetQuotes(s);
   
   buffer[current_buffer]->text->InsertInset(new_inset);
   
}

/* callbacks for form form_quotes */
void QuotesApplyCB(FL_OBJECT *ob, long data)
{
  if (buffer[current_buffer]->text){
     SetMinibuffer("user defined quotes set","","");
     buffer[current_buffer]->parameters.quotes_language = fl_get_choice(fd_form_quotes->choice_quotes_language) - 1;
   
     if (fl_get_button(fd_form_quotes->radio_single))   
     	buffer[current_buffer]->parameters.quotes_times = 1;
     else
     	buffer[current_buffer]->parameters.quotes_times = 2;
  }
}
void QuotesCancelCB(FL_OBJECT *ob, long data)
{
  fl_hide_form(fd_form_quotes->form_quotes);
}
void QuotesOKCB(FL_OBJECT *ob, long data)
{
  QuotesApplyCB(ob, data);
  QuotesCancelCB(ob, data);
}




/* callbacks for form form_preamble */
void PreambleCancelCB(FL_OBJECT *ob, long data)
{
  fl_hide_form(fd_form_preamble->form_preamble);
}

void PreambleApplyCB(FL_OBJECT *ob, long data)
{
   if (buffer[current_buffer]->text) {
      if (buffer[current_buffer]->parameters.preamble)
      	 delete(buffer[current_buffer]->parameters.preamble);
      buffer[current_buffer]->parameters.preamble = StringCopy(fl_get_input(fd_form_preamble->input_preamble));
      buffer[current_buffer]->lyx_current = 0;
      buffer[current_buffer]->bak_current = 0;
      buffer[current_buffer]->dvi_current = 0;
      SetMinibuffer("user defined LaTeX preamble set","","");
   }
}

void PreambleOKCB(FL_OBJECT *ob, long data)
{
  PreambleApplyCB(ob, data);
  PreambleCancelCB(ob, data);
}


/* callbacks for form form_table */
void TableApplyCB(FL_OBJECT *ob, long data){
   int xsize,ysize;
   if (!buffer[current_buffer]->screen)
     return;
   SetMinibuffer("Inssert table...","","");

   ysize = (int) (fl_get_slider_value(fd_form_table->slider_rows) + 0.5);
   xsize = (int) (fl_get_slider_value(fd_form_table->slider_columns) + 0.5);
   
   
   buffer[current_buffer]->screen->HideCursor();
   BeforeChange();
   Update(-2);
   
   int i, i2;
   
   buffer[current_buffer]->text->BreakParagraph();
   Update(-1);
   
   if (buffer[current_buffer]->text->cursor.par->Last()) {
      buffer[current_buffer]->text->CursorLeft();
      
      buffer[current_buffer]->text->BreakParagraph();
      Update(-1);
   }
   buffer[current_buffer]->text->current_font.latex = LYX_NO_LATEX;
   
   buffer[current_buffer]->text->SetLayout(NumberOfLayout(buffer[current_buffer]->parameters.textclass, "LaTeX"));

   if (buffer[current_buffer]->text->cursor.par->footnoteflag == LYX_NO_FOOTNOTE) {
      buffer[current_buffer]->text->SetParagraph(0, 0,
						 0, 0,
						 0.3 * buffer[current_buffer]->parameters.baseline,
						 0.3 * buffer[current_buffer]->parameters.baseline,
						 0, 0,
						 LYX_ALIGN_CENTER,
						 NULL,
						 0);
   }
   else
     buffer[current_buffer]->text->SetParagraph(0, 0,
						0, 0,
						0, 0,
						0, 0,
						LYX_ALIGN_CENTER, 
						NULL,
						0);
   Update(-1);
   
   
   InsetLatex *new_inset = new InsetLatex("\\begin{tabular}{");
   buffer[current_buffer]->text->InsertInset(new_inset);
   SmallUpdate(-1);
   
   LyXCursor tmpcursor = buffer[current_buffer]->text->cursor;
   
   buffer[current_buffer]->text->InsertChar('|');
   SmallUpdate(-1);
   
   for (i=0; i<xsize; i++) {
      buffer[current_buffer]->text->InsertChar('c');
      SmallUpdate(-1);
      buffer[current_buffer]->text->InsertChar('|');
      SmallUpdate(-1);
   }
   
   new_inset = new InsetLatex("}");
   buffer[current_buffer]->text->InsertInset(new_inset);
   SmallUpdate(-1);
   
   buffer[current_buffer]->text->InsertChar(LYX_META_NEWLINE);
   SmallUpdate(-1);
   
   new_inset = new InsetLatex("\\hline ");
   buffer[current_buffer]->text->InsertInset(new_inset);
   
   buffer[current_buffer]->text->InsertChar(LYX_META_NEWLINE);
   SmallUpdate(-1);
   
   for (i2=0; i2<ysize;i2++) {
      for (i=1; i<xsize; i++) {
	 new_inset = new InsetLatex("&");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
      }
      
      new_inset = new InsetLatex("\\\\");
      buffer[current_buffer]->text->InsertInset(new_inset);
      SmallUpdate(-1);
      buffer[current_buffer]->text->InsertChar(LYX_META_NEWLINE);
      SmallUpdate(-1);
      
      new_inset = new InsetLatex("\\hline ");
      buffer[current_buffer]->text->InsertInset(new_inset);
      SmallUpdate(-1);
      
      buffer[current_buffer]->text->InsertChar(LYX_META_NEWLINE);
      SmallUpdate(-1);
      
      if (i2==0) {
	 new_inset = new InsetLatex("\\hline ");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 
	 buffer[current_buffer]->text->InsertChar(LYX_META_NEWLINE);
	 SmallUpdate(-1);
      }
      
   }
   
   new_inset = new InsetLatex("\\end{tabular}");
   buffer[current_buffer]->text->InsertInset(new_inset);
   
   buffer[current_buffer]->text->SetCursor(tmpcursor.par, tmpcursor.pos);
   Update(1);
   SetMinibuffer("Table inserted","","");
}

void TableCancelCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_table->form_table);
}

void TableOKCB(FL_OBJECT *ob, long data)
{
   TableApplyCB(ob,data);
   TableCancelCB(ob,data);
}

/* callbacks for form form_print */
void PrintCancelCB(FL_OBJECT *ob, long data){
   fl_hide_form(fd_form_print->form_print);
}

void PrintApplyCB(FL_OBJECT *ob, long data){
   int i;
   if (!buffer[current_buffer]->text)
      return;

   char* pageflag = "";
   if (fl_get_button(fd_form_print->radio_even_pages))
       pageflag = lyxrc.print_evenpage_flag;
   else if (fl_get_button(fd_form_print->radio_odd_pages))
       pageflag = lyxrc.print_oddpage_flag;
   
   char* reverseflag = "";
   if (fl_get_button(fd_form_print->radio_order_reverse))
      reverseflag = lyxrc.print_reverse_flag;
   
   char* orientationflag = "";
   if (buffer[current_buffer]->parameters.orientation == ORIENTATION_LANDSCAPE)
      orientationflag = lyxrc.print_landscape_flag;
   
   const char* ps_file = fl_get_input(fd_form_print->input_file);
   
   for (i=0; ps_file[i]; i++);
   
   char *command = new char[i+100]; 
   
   
   const char* printer = fl_get_input(fd_form_print->input_printer);
   
   
   
   if (printer) {
      if (lyxrc.printer)
      	 delete(lyxrc.printer);
      lyxrc.printer = StringCopy(printer);
   }
   
   if (fl_get_button(fd_form_print->radio_file)) 
      sprintf(command, "%s %s %s %s %s %s \"%s\"", lyxrc.print_command, 
             pageflag, reverseflag, orientationflag, lyxrc.print_extra_options, 
             lyxrc.print_to_file, ps_file);
   else {
      while (printer[0] == ' ')
      	 printer++;
      if (printer[0])
      	 sprintf(command, "%s %s %s %s %s %s%s", lyxrc.print_command, 
                 pageflag, reverseflag, orientationflag, lyxrc.print_extra_options,
                 lyxrc.print_to_printer, printer);
      else
      	 sprintf(command, "%s %s %s %s %s", lyxrc.print_command, pageflag, 
                  reverseflag, orientationflag, 
                  lyxrc.print_extra_options);
   }
   fprintf (stderr, "command: %s\n", command);
   RunScript(0, command);
   delete(command);
}


void PrintOKCB(FL_OBJECT *ob, long data){
   PrintCancelCB(ob, data);  
   PrintApplyCB(ob,data);
}


/* callbacks for form form_figure */
void FigureApplyCB(FL_OBJECT *ob, long data){
   if (buffer[current_buffer]->text) {
      SetMinibuffer("Insert figure...","","");
      buffer[current_buffer]->screen->HideCursor();
      Update(-2);
      BeforeChange();
      
      buffer[current_buffer]->text->BreakParagraph();
      Update(-1);
      
      if (buffer[current_buffer]->text->cursor.par->Last()) {
	 buffer[current_buffer]->text->CursorLeft();
	 
	 buffer[current_buffer]->text->BreakParagraph();
	 Update(-1);
      }
      
      buffer[current_buffer]->text->SetLayout(NumberOfLayout(buffer[current_buffer]->parameters.textclass, "LaTeX"));
      
      if (buffer[current_buffer]->text->cursor.par->footnoteflag == LYX_NO_FOOTNOTE) {
	 buffer[current_buffer]->text->SetParagraph(0, 0,
						    0, 0,
						    0.3 * buffer[current_buffer]->parameters.baseline,
						    0.3 * buffer[current_buffer]->parameters.baseline,
						    0, 0,
						    LYX_ALIGN_CENTER, 
						    NULL,
						    0);
      }
      else
      	buffer[current_buffer]->text->SetParagraph(0, 0,
						   0, 0,
						   0, 0,
						   0, 0,
						   LYX_ALIGN_CENTER, 
						   NULL,
						   0);
      
      Update(-1);
      
      buffer[current_buffer]->text->current_font.latex = LYX_LATEX;
      
      InsetLatex *new_inset = NULL;
      
      if (fl_get_button(fd_form_figure->radio_latex)) {
	 new_inset = new InsetLatex("\\include ");
	 buffer[current_buffer]->text->InsertInset(new_inset);
      }
      else {
	 new_inset = new InsetLatex("\\epsfig{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetLatex("}");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 Update(-1);
	 buffer[current_buffer]->text->CursorLeft();
      }
      
      Update(0);
      SetMinibuffer("Figure inserted","","");
   }

}

void FigureCancelCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_figure->form_figure);
}

void FigureOKCB(FL_OBJECT *ob, long data)
{
   FigureApplyCB(ob,data);
   FigureCancelCB(ob,data);
}

void error_handler(int err_sig) {
   switch (err_sig) {
    case SIGHUP:
      fprintf(stderr, "\nlyx: SIGHUP signal trapped\n");
      break;
    case SIGFPE:
      fprintf(stderr, "\nlyx: SIGFPE signal trapped\n");
	break;
    case SIGSEGV:
	fprintf(stderr, "\nlyx: SIGSEGV signal trapped\n");
        fprintf(stderr, "(Sorry. If you can repeat this violation and have the current version of lyx,\n please send a bug report. Thanks.)\n");
	break;
    }
   
   
   /* deinstall the signal handlers */ 
   signal(SIGHUP, SIG_DFL);
   signal(SIGFPE, SIG_DFL);
   signal(SIGSEGV, SIG_DFL);

   char *s = NULL;
   int n;
   int i,a;
   int buf;
   for (buf=0; buf < NUMBER_OF_BUFFERS; buf++) {
	 if (buffer[buf]->parameters.paragraph && !buffer[buf]->lyx_current) {
	    
	    for (n=0; buffer[buf]->filename[n]; n++);
	    s = new char[n+5];
	    for (i=0; i<n; i++) 
	       s[i] = buffer[buf]->filename[i];
	    s[i] = '#';
	    s[i+1] = '\0';
	    for (a=i+2; a && s[a-1]!='/'; a--) {
	       s[a] = s[a-1];
	    }
	    s[a]='#';
	    
	    fprintf(stderr, "lyx: attempting to save buffer %s in %s...\n", 
		    buffer[buf]->filename, s);

	    if (buffer[buf]->text)
	       buffer[buf]->parameters.cursor = buffer[buf]->text->cursor;
	    
	    WriteFile(s, 
		      buffer[buf]->parameters);
	    buffer[buf]->lyx_current = 1;
	 
	    delete(s);
	 }
   }
   fprintf(stderr, "bye.\n");
   if(err_sig!=SIGHUP && getenv("LYXDEBUG"))
      abort();
   exit(0);
}

void ScreenApplyCB(FL_OBJECT *ob, long data){

   roman_font_name = (char*) fl_get_input(fd_form_screen->input_roman);
   sans_font_name = (char*) fl_get_input(fd_form_screen->input_sans);
   typewriter_font_name = (char*) fl_get_input(fd_form_screen->input_typewriter);
   font_norm = (char*) fl_get_input(fd_form_screen->input_font_norm);
   

    font_sizes[LYX_SIZE_TINY] = atoi(fl_get_input(fd_form_screen->intinput_tiny));
    font_sizes[LYX_SIZE_SMALL] = atoi(fl_get_input(fd_form_screen->intinput_small));
    font_sizes[LYX_SIZE_NORMAL] = atoi(fl_get_input(fd_form_screen->intinput_normal));
    font_sizes[LYX_SIZE_LARGE] = atoi(fl_get_input(fd_form_screen->intinput_large));
    font_sizes[LYX_SIZE_LARGER] = atoi(fl_get_input(fd_form_screen->intinput_larger));
    font_sizes[LYX_SIZE_LARGEST] = atoi(fl_get_input(fd_form_screen->intinput_largest));
    font_sizes[LYX_SIZE_HUGE] = atoi(fl_get_input(fd_form_screen->intinput_huge));
    font_sizes[LYX_SIZE_GIANT] = atoi(fl_get_input(fd_form_screen->intinput_giant));

   lyxrc.default_zoom = atoi(fl_get_input(fd_form_screen->intinput_size));
   SetZoom(lyxrc.default_zoom);
   
   /* all buffers will need resize */ 
   for (int i =0; i < NUMBER_OF_BUFFERS; i++) {
      if (buffer[i]->screen)
      	 delete(buffer[i]->screen);
      buffer[i]->screen = NULL;
   }
   RedoCurrentBuffer();
   SetMinibuffer("New screen options set","","");
}

void ScreenCancelCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_screen->form_screen);
}


void ScreenOKCB(FL_OBJECT *ob, long data)
{
   ScreenCancelCB(ob,data);
   ScreenApplyCB(ob,data);
}

/* these functions are for the spellchecker */ 
char* NextWord(float &value) {
  if (!buffer[current_buffer]->text){
    value = 1;
    return NULL;
  }
   
  char* string =  buffer[current_buffer]->text->SelectNextWord(value);

  return string;
}

void SelectLastWord(){
   if (!buffer[current_buffer]->text)
      return;
   if (work_area_active)
      buffer[current_buffer]->screen->HideCursor();
   BeforeChange(); 
   buffer[current_buffer]->text->SelectSelectedWord();
   buffer[current_buffer]->screen->ToggleSelection();
   Update(0);
}


void EndOfSpellCheck(){
   if (!buffer[current_buffer]->text)
      return;
   if (work_area_active)
      buffer[current_buffer]->screen->HideCursor();
   BeforeChange(); 
   buffer[current_buffer]->text->SelectSelectedWord();
   buffer[current_buffer]->text->ClearSelection();
   Update(0);
}


void ReplaceWord(const char* replacestring) {
   if (!buffer[current_buffer]->screen)
      return;
   buffer[current_buffer]->screen->HideCursor();
   Update(-2);
   /* clear the selection (if there is any) */ 
   buffer[current_buffer]->screen->ToggleSelection();
   buffer[current_buffer]->text->ReplaceSelectionWithString((char*) replacestring);
   
   buffer[current_buffer]->text->SetSelectionOverString((char*) replacestring);
   Update(1);
}


static int ScrollUpCB(long time) {
   if (!buffer[current_buffer]->screen)
      return 0;
   
   double value= fl_get_slider_value(fd_form_main->scrollbar);
   
   if (value == 0)
      return 0;
   
   float add_value =  (buffer[current_buffer]->text->DefaultHeight()
		       + (float)(time) * (float)(time) * 0.125);
   if (add_value > work_area_height)
      add_value = (float) (work_area_height - buffer[current_buffer]->text->DefaultHeight());
   
   value -= add_value;

   if (value < 0)
      value = 0;
   
   fl_set_slider_value(fd_form_main->scrollbar, value);
   
   ScrollCB(fd_form_main->scrollbar,0); 
   return 0;
}

static int ScrollDownCB(long time) {
   if (!buffer[current_buffer]->screen)
      return 0;
   
   double value= fl_get_slider_value(fd_form_main->scrollbar);
   double min, max;
   fl_get_slider_bounds(fd_form_main->scrollbar, &min, &max);

   if (value == max)
      return 0;
   
   float add_value =  (buffer[current_buffer]->text->DefaultHeight()
		       + (float)(time) * (float)(time) * 0.125);
   if (add_value > work_area_height)
      add_value = (float) (work_area_height - buffer[current_buffer]->text->DefaultHeight());
   
   value += add_value;
   
   if (value > max)
      value = max;
   
   fl_set_slider_value(fd_form_main->scrollbar, value);
   
   ScrollCB(fd_form_main->scrollbar,0); 
   return 0;
}


static void ScrollUpOnePageCB(long time) {
   if (!buffer[current_buffer]->screen)
      return;
   long y = buffer[current_buffer]->screen->first;
   if (!y)
      return;
   Row* row = buffer[current_buffer]->text->GetRowNearY(y);
   y = y - work_area_height + row->height;
   buffer[current_buffer]->screen->Draw(y);
   UpdateScrollbar();
}

static void ScrollDownOnePageCB(long time) {
   if (!buffer[current_buffer]->screen)
      return;
   double min, max;
   fl_get_slider_bounds(fd_form_main->scrollbar, &min, &max);
   long y = buffer[current_buffer]->screen->first;
   if (y > buffer[current_buffer]->text->height - work_area_height)
      return;
   y += work_area_height;
   buffer[current_buffer]->text->GetRowNearY(y);
   buffer[current_buffer]->screen->Draw(y);
   UpdateScrollbar();
}

void DownCB(FL_OBJECT *ob, long data) {
   const XEvent*ev2;
   static unsigned int button = Button1;
   static long time = 0;
   ev2 = fl_last_event();
   if (ev2->type == ButtonPress || ev2->type == ButtonRelease) {
     button = ev2->xbutton.button;
     time = 0;
   }
   
   switch (button) {
   case Button2: ScrollUpOnePageCB(time++); break;
   case Button3: ScrollDownOnePageCB(time++); break;
   default: ScrollDownCB(time++); break;
   }
}

void UpCB(FL_OBJECT *ob, long data) {
   const XEvent*ev2;
   static unsigned int button = Button1;
   static long time = 0;
   ev2 = fl_last_event();
   if (ev2->type == ButtonPress || ev2->type == ButtonRelease) {
     button = ev2->xbutton.button;
     time = 0;
   }
   
   switch (button) {
   case Button3: ScrollUpOnePageCB(time++); break;
   case Button2: ScrollDownOnePageCB(time++); break;
   default: ScrollUpCB(time++); break;
   }
}


struct TocList {
  char counter[6];
  TocList *next;
};

static TocList* toclist = NULL;


void TocSelectCB(FL_OBJECT *ob, long data) {
   if (!buffer[current_buffer]->text) 
      return;
   TocList* tmptoclist = toclist;
   int i = fl_get_browser(ob);
   int a = 0;

   for (a=1; a<i && tmptoclist->next; a++){
     tmptoclist = tmptoclist->next;
   }

   if (!tmptoclist)
     return;
     

   LyXParagraph *par = buffer[current_buffer]->parameters.paragraph;
   while (par && 
	  (
	   par->GetCounter(0) != tmptoclist->counter[0] ||
	   par->GetCounter(1) != tmptoclist->counter[1] ||
	   par->GetCounter(2) != tmptoclist->counter[2] ||
	   par->GetCounter(3) != tmptoclist->counter[3] ||
	   par->GetCounter(4) != tmptoclist->counter[4] ||
	   par->GetCounter(5) != tmptoclist->counter[5]
	   )
	  ) {
      par = par->LastPhysicalPar()->Next();
   }
   
   if (par) {
      BeforeChange();
      buffer[current_buffer]->text->SetCursor(par, 0);
      buffer[current_buffer]->text->sel_cursor = buffer[current_buffer]->text->cursor;
      Update(0);
   }
   else {
      WriteAlert("Error", "Couldn't find this label", "in current document.");
   }
	  
}


void TocCancelCB(FL_OBJECT *ob, long data) {
   fl_hide_form(fd_form_toc->form_toc);
}

void TocUpdateCB(FL_OBJECT *ob, long data){
   static LyXParagraph* stapar = NULL;
   TocList *tmptoclist = NULL;
   
   /* deletet the toclist */ 
   if (toclist){
     while (toclist){
       tmptoclist = toclist->next;
       delete(toclist);
       toclist = tmptoclist;
     }
   }
   toclist = NULL;
   tmptoclist = toclist;


   fl_clear_browser(fd_form_toc->browser_toc);
   if (!buffer[current_buffer]->text) {
      fl_add_browser_line(fd_form_toc->browser_toc, "*** No Document ***");
      return;
   }
   fl_hide_object(fd_form_toc->browser_toc);
   /* get the table of contents */ 
   LyXParagraph *par = buffer[current_buffer]->parameters.paragraph;
   char labeltype;
   char* line = new char[200];
   int i = 0;
   int pos = 0;
   unsigned char c;
   int topline = 0;
   
   if (stapar == par)
      topline = fl_get_browser_topline(fd_form_toc->browser_toc);
   stapar = par;
   
   while (par) {
      labeltype = style(buffer[current_buffer]->parameters.textclass, 
			par->GetLayout())->labeltype;   
      if (labeltype >= LABEL_COUNTER_CHAPTER && labeltype <= LABEL_COUNTER_CHAPTER + buffer[current_buffer]->parameters.tocdepth) {
	 /* inset this into the table of contents */ 
	  /* first indent a little bit */ 

	 for (pos=0; 
	      pos < (labeltype - 
		     TextClass(buffer[current_buffer]->parameters.textclass)->maxcounter) * 4 + 2;
	      pos++)
	    line[pos] = ' ';
	 
	 /* then the labelstring */ 
	 
	 i = 0;
	 if (par->labelstring) {
	    while (pos < 199 && par->labelstring[i]) {
	       line[pos] = par->labelstring[i];
	       i++;
	       pos++;
	    }
	 }
	 
	 line[pos] = ' ';
	 pos++;
	 
	 /* now the contents */ 
	 i = 0;
	 while (pos < 199 && i < par->last) {
	    c = par->GetChar(i);
	    if (isprint(c) || c >= 128) {
	       line[pos] = c;
	       pos++;
	    }
	    i++;
	 }
	 line[pos] = '\0';
	 fl_add_browser_line(fd_form_toc->browser_toc, line);
	 
	 /* make a toclist entry */
	 
	 if (!tmptoclist){
	   tmptoclist = new TocList;
	   toclist = tmptoclist;
	 }
	 else {
	   tmptoclist->next = new TocList;
	   tmptoclist = tmptoclist->next;
	 }

	 tmptoclist->next = NULL;
	 int a = 0;
	 for (a=0; a<6; a++){
	   tmptoclist->counter[a] = par->GetCounter(a);
	 }
      }
      par = par->LastPhysicalPar()->Next();
	 
   }
   delete (line);
   fl_set_browser_topline(fd_form_toc->browser_toc, topline);
   fl_show_object(fd_form_toc->browser_toc);
}


int AtCloseMainFormCB(FL_FORM *f, void *p)
{
  MenuQuit();
  return FL_IGNORE;
};


/* callbacks for form form_ref */
void RefSelectCB(FL_OBJECT *ob, long data)
{
   int i;
   if (!buffer[current_buffer]->text)
      return;
   char * s = StringCopy(fl_get_browser_line(fd_form_ref->browser_ref, fl_get_browser(fd_form_ref->browser_ref)));
   
   if (!StringEqual(NameOfClass(current_textclass), "linuxdoc")) {
      char string[200];
      if (fl_get_button(fd_form_ref->radio_ref))
      	 sprintf(string, "\\ref{%s}", s);
      else
      	 sprintf(string, "\\pageref{%s}", s);
      InsertLatexDelInset(string);
   }
   else {
      /* special hack for linuxdoc sgml references */ 
      InsertLatexDelInset("\\ref{");
      for (i=0; s[i]; i++) 
	 buffer[current_buffer]->text->InsertChar(s[i]);
      InsertLatexDelInset("}{");
      InsertLatexDelInset("}");
      buffer[current_buffer]->text->InsertChar(' ');
   }
}

void RefUpdateCB(FL_OBJECT *ob, long data)
{
  fl_clear_browser(fd_form_ref->browser_ref);
  if (!buffer[current_buffer]->text)
    return;

  LyXParagraph *par = buffer[current_buffer]-> parameters.paragraph;
  int pos;
  Inset* inset;
  while (par){
    pos = -1;
    while ((inset = par->ReturnNextInsetPointer(pos))){
      if (inset->LyxCode() == LYX_LABEL_CODE)
	fl_add_browser_line(fd_form_ref->browser_ref, 
			    ((InsetLabel*)inset)->GetLabel());
      pos++;
    }
    par = par->next;
  }
}

void RefHideCB(FL_OBJECT *ob, long data)
{
   fl_hide_form(fd_form_ref->form_ref);

}
