/*
 *  This file is part of
 * ==================================================
 *
 *           LyX, the High Level Word Processor
 *           Copyright (C) 1995 Matthias Ettrich
 *
 *           This file is Copyright (C) 1995
 *           Lars Gullik Bjnnes
 * ==================================================
 */

/* This file contains all the menus and sub menus declarations
   the call backs are in lyx_cb.C */

/*
 * REMEMBER :
 * XFORMS can handle only ten - 10 - ten popups at the same time
 * so when you are finished looking at a pup free it (fl_freepup)
 */

#include <forms.h>
#include "lyx.h"
#include "lyx_cb.h"
#include "file.h"
#include "version.h"

extern int current_buffer;
extern Buffer *buffer[];
extern LyXRC lyxrc;

extern FD_form_screen *fd_form_screen;
extern void ResizeCurrentBufferPseudoExpose();

extern void GotoError();
extern void GotoNote();
extern void OpenStuff();
extern void HyphenationPoint();
extern void HFill();
extern void Newline();
extern void ProtectedBlank();
extern void CopyCB();
extern void LoadLyXFile(const char*);
// LyXBell could perhaps be inlined?
extern void LyXBell();

extern char* roman_font_name;
extern char* sans_font_name;
extern char* typewriter_font_name;
extern char* font_norm;
extern char font_sizes[];

extern int RunLaTeX();
extern int RunLinuxDoc(int, const char *);
extern void MenuWrite(int buf);
extern void MenuWriteAs();
extern void MenuRunLaTeX();
extern void MenuPrint();
extern void MenuMakeLaTeX();
extern void MenuMakeLinuxDoc();
extern void MenuOpen();
extern void MenuNew();
extern void MenuQuit();
extern void InsertAsciiFile(char);

extern void CloseBuffer(); 
extern void MenuPreview();
extern void MenuPreviewPS();
extern void MenuLayoutCharacter();
extern void MenuLayoutParagraph();
extern void MenuLayoutDocument();
extern void MenuLayoutQuotes();
extern void MenuLayoutPreamble();
extern void UpdateMenuBar();
extern void MenuDocu();
extern void ShowCredits();
extern void show_symbols_form(void);
extern void Update(signed char f);
extern void SmallUpdate(signed char f);
extern void MenuInsertLabel();
extern void MenuInsertRef();
extern void NoteCB();
extern void HtmlUrlCB();
extern void UrlCB();
extern void InsertLaTeXTable(char* table);
extern void BeforeChange();

extern void ProhibitInput();
extern void AllowInput();


extern FD_form_main *fd_form_main;
extern LastFiles *lastfiles;


void ShowFileMenu(FL_OBJECT *ob, long data)
{
   //static int FileMenu=0;
   //static int SubFileInsertAscii=0;
   //static int SubFileLast4Files=0;

   int SubFileLast4Files =fl_newpup(FL_ObjWin(ob));
   fl_addtopup(SubFileLast4Files,
	       "Last 4 Files%t");
   lastfiles->MakeMenu(SubFileLast4Files);
      
   int SubFileInsertAscii = fl_defpup(FL_ObjWin(ob),
   " Insert Ascii file %t"
   "| Lines %x21"
   "| Paragraphs %x22"
   "| Verbatim %x23");

   int FileMenu = fl_defpup(FL_ObjWin(ob), 
   " Open... (C-x C-f) "
   "| New... (C-x C-d) %l"
   "| Save (C-x C-s) "
   "| Save As... (C-x C-w) "
   "| Run LaTeX (C-x r) "
   "| Preview (xdvi) (C-x p) "
   "| Preview (ghostsview) (C-x g) %l"
   "| Print... %l"
   "| Make (nice) LaTeX-file "
   "| Make LinuxDoc file %l"
   "| Insert Ascii File %m%l"
   "| Close Buffer (C-x k) %l"
   "| Quit (C-x C-c) %l"
   "| Last 4 Files %m",
   SubFileInsertAscii,SubFileLast4Files);
   
   fl_setpup_position(
		      fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);   
   int choice=fl_dopup(FileMenu);
   //fprintf(stderr,"FileMenu returned with %d\n",choice);
   
   switch (choice) {
   case 1: MenuOpen(); break;
   case 2: MenuNew(); break;
   case 3: MenuWrite(current_buffer); break;
   case 4: MenuWriteAs(); break;
   case 5: MenuRunLaTeX(); break;
   case 6: MenuPreview(); break;
   case 7: MenuPreviewPS(); break;
   case 8: MenuPrint(); break;
   case 9: MenuMakeLaTeX(); break;
   case 10: MenuMakeLinuxDoc(); break;
      // case 11: // insert ascii
      //break;
   case 12: CloseBuffer(); break;
   case 13: MenuQuit(); break;
      // case 14: //lastfiles      
      //break;
   case 21: InsertAsciiFile(0);
      break;
   case 22: InsertAsciiFile(1);
      break;
   case 23:
      break;
   case 31:                 
   case 32:
   case 33:
   case 34:
      //fprintf(stderr,
      //      "HandleSubFileLast4Files: called with choice %d\n",choice-30);
      LoadLyXFile(lastfiles->GetFileName(choice-30));
      ResizeCurrentBufferPseudoExpose();
      break;
   }
   fl_freepup(FileMenu);
   fl_freepup(SubFileInsertAscii);
   fl_freepup(SubFileLast4Files);
   AllowInput();
}


extern void CutCB(FL_OBJECT *ob, long data);
extern void PasteCB(FL_OBJECT *ob, long data);
extern void MenuLoadAutoSave();
extern void AutoSave();
extern void MenuSearch();
extern void MenuKeymap();
extern void ShowSpellChecker();
extern bool SpellCheckerOptions();
extern void MenuShowTableOfContents();
extern void MenuPasteSelection(char at);
extern void AllFloats(char flag);

void ShowEditMenu(FL_OBJECT *ob, long data)
{
   //int EditMenu=0;
   //int SubEditFloats=0;

   
   int SubEditFloats=fl_defpup(FL_ObjWin(ob),
   " Floats %t"
   "| Open/Close (C-o) %x21"
   "| Open All Floats %x22"
   "| Close All Floats %x23");

   int EditMenu=fl_defpup(FL_ObjWin(ob),
   " Load autosave file (C-x u) "
   "| Force autosave (C-x a) %l"
   "| Cut (C-w)"
   "| Copy (M-w) "
   "| Paste (C-y)%l"
   "| Search & Replace... (C-s) "
   "| Goto Error (C-g) "
   "| Goto Note "
   "| Floats %m"
   "| Spell Checker... "
   "| Table Of Contents...(C-x t) %l"
   "| Paste Primary Selection (Lines) "
   "| Paste Primary Selection (Paragraphs)",
   SubEditFloats);
      

   //char* tmpstring = new char[50];

   fl_setpup_position(fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);   
   int choice = fl_dopup(EditMenu);

   switch (choice) {
   case 1: MenuLoadAutoSave(); break;
   case 2: AutoSave(); break;
   case 3: CutCB(ob,data);
      break;
   case 4: CopyCB(); break;
   case 5: PasteCB(ob,data);
      break;
   case 6: MenuSearch(); break;
   case 7: GotoError(); break;
   case 8: GotoNote(); break;
   case 9: // floats
      break;
   case 10:
     ShowSpellChecker();
      break;
   case 11:
      MenuShowTableOfContents();
      break;
   case 12:
      MenuPasteSelection(0);
      break;
   case 13:
      MenuPasteSelection(1);
      break;
      // floats sub-menu
   case 21: OpenStuff();
      break;
   case 22: AllFloats(1);
      break;
   case 23: AllFloats(0);
      break;
   }
   
   //delete (tmpstring);
   fl_freepup(EditMenu);
   fl_freepup(SubEditFloats);
}


extern void EmphCB(FL_OBJECT *ob, long data);
extern void NounCB(FL_OBJECT *ob, long data);
extern void TexCB(FL_OBJECT *ob, long data);
extern void FormulaCB(FL_OBJECT *ob, long data);
extern void DepthCB(FL_OBJECT *ob, long data);


void ShowLayoutMenu(FL_OBJECT *ob, long data)
{
   //static int LayoutMenu=0;
   
   int LayoutMenu = fl_defpup(FL_ObjWin(ob),
   " Character... "
   "| Paragraph... "
   "| Document... "
   "| Quotes... %l"
   "| Emphasize Style %b"
   "| Noun Style %b"
   "| TeX Style %b"
   "| Math Style %b%l"
   "| Change Environment Depth "
   "| LaTeX Preamble...");
   
   fl_setpup_position(
		      fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);
   int choice = fl_dopup(LayoutMenu);
   switch (choice) {
   case 1: MenuLayoutCharacter(); break;
   case 2: MenuLayoutParagraph(); break;
   case 3: MenuLayoutDocument(); break;
   case 4: MenuLayoutQuotes(); break;
   case 5: EmphCB(ob, data);
      break;
   case 6: NounCB(ob, data);
      break;
   case 7: TexCB(ob, data);
      break;
   case 8: FormulaCB(ob, data);
      break;
   case 9: DepthCB(ob, data);
      break;
   case 10: MenuLayoutPreamble(); break;
   }
   fl_freepup(LayoutMenu); 
}


extern void FigureCB(FL_OBJECT *ob, long data);
extern void TableCB(FL_OBJECT *ob, long data);
extern void FootCB(FL_OBJECT *ob, long data);
extern void MarginCB(FL_OBJECT *ob, long data);
extern void MeltCB(FL_OBJECT *ob, long data);


void ShowSpecialMenu(FL_OBJECT *ob, long data)
{
   int SpecialMenu;
   
   if (data) {
      /* for linuxdoc sgml */ 
      SpecialMenu = fl_defpup(FL_ObjWin(ob),
      " Figure "
      "| Table "
      "| Figure Float "
      "| Table Float "
      "| Footnote Float "
      "| Marginpar Float "
      "| Melt Float %l"
      "| Table of Contents "
      "| List of Figures "
      "| List of Tables %l"
      "| HFill (C-i) "
      "| Hyphenation Point (C-h) "
      "| Protected Blank (M-space) "
      "| Newline (C-enter) "
      "| Note "
      "| Label (Cross-Mark) "
      "| Reference (cross-Reference)"
      "| URL "
      "| HTML URL ");
   } else {
      SpecialMenu = fl_defpup(FL_ObjWin(ob),
      " Figure "
      "| Table "
      "| Figure Float "
      "| Table Float "
      "| Footnote Float "
      "| Marginpar Float "
      "| Melt Float %l"
      "| Table of Contents "
      "| List of Figures "
      "| List of Tables %l"
      "| HFill (C-i) "
      "| Hyphenation Point (C-h) "
      "| Protected Blank (M-space) "
      "| Newline (C-enter) "
      "| Note "
      "| Label "
      "| Reference ");
   }

   fl_setpup_position(
		      fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);
   
   int choice = fl_dopup(SpecialMenu);

   if (buffer[current_buffer]->text){
      buffer[current_buffer]->screen->HideCursor();
      BeforeChange();
      Update(-2);
      switch (choice) {
      case 1: FigureCB(ob, data);
	 break;
      case 2: TableCB(ob, data);
	 break;
      case 3:
	 buffer[current_buffer]->text->InsertFootnoteEnvironement(LYX_FIG);
	 break;
      case 4:
	 buffer[current_buffer]->text->InsertFootnoteEnvironement(LYX_TAB);
	 break;
      case 5: FootCB(ob, data);
	 break;
      case 6: MarginCB(ob, data);
	 break;
      case 7: MeltCB(ob, data);
	 break;
      case 8: InsertLaTeXTable("\\tableofcontents"); break;
      case 9: InsertLaTeXTable("\\listoffigures"); break;
      case 10: InsertLaTeXTable("\\listoftables"); break;
      case 11: HFill(); break;
      case 12: HyphenationPoint(); break; 
      case 13: ProtectedBlank(); break; 
      case 14: Newline(); break; 
	 // case 15: CrossMarkCB(); break;
	 // case 16: CrossReferenceCB(); break;
      case 15: NoteCB(); break;
      case 16: MenuInsertLabel(); break;
      case 17: MenuInsertRef(); break;
      case 18: UrlCB(); break;
      case 19: HtmlUrlCB(); break;
      }
      Update(1);
   }
   fl_freepup(SpecialMenu);
}


void ShowMathMenu(FL_OBJECT *ob, long data)
{
   // static int MathMenu=0;

   int MathMenu = fl_defpup(FL_ObjWin(ob), 
   " frac "
   "| sqrt "
   "| ^ (Exponent) "
   "| _ (Index) "
   "| sum "
   "| int %l"
   "| Symbols... ");
   
   fl_setpup_position(
		      fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);   
   int choice = fl_dopup(MathMenu);  

   InsetFormulaLatex *new_inset = NULL;
   if (buffer[current_buffer]->text)  {
      buffer[current_buffer]->screen->HideCursor();
      BeforeChange();
      Update(-2);
      buffer[current_buffer]->text->current_font.latex = LYX_FORMULA_LATEX;
      
      switch (choice) {
      case 1: /* frac */
	 new_inset = new InsetFormulaLatex("\\frac{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 Update(-1);
	 buffer[current_buffer]->text->CursorLeft();
	 buffer[current_buffer]->text->CursorLeft();
	 break;
      case 2: /* sqrt */
	 new_inset = new InsetFormulaLatex("\\sqrt[");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("]{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 Update(-1);
	 buffer[current_buffer]->text->CursorLeft();
	 break;
      case 3: /* Exponent  */ 
	 new_inset = new InsetFormulaLatex("^{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 Update(-1);
	 buffer[current_buffer]->text->CursorLeft();
	 break;
      case 4: /* Index  */
	 new_inset = new InsetFormulaLatex("_{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 Update(-1);
	 buffer[current_buffer]->text->CursorLeft();
	 break;
      case 5: /* sum */ 
	 new_inset = new InsetFormulaLatex("\\sum_{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}^{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 Update(-1);
	 buffer[current_buffer]->text->CursorLeft();
	 buffer[current_buffer]->text->CursorLeft();
	 break;
      case 6: /* int */
	 new_inset = new InsetFormulaLatex("\\int_{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}^{");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 SmallUpdate(-1);
	 new_inset = new InsetFormulaLatex("}");
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 Update(-1);
	 buffer[current_buffer]->text->CursorLeft();
	 buffer[current_buffer]->text->CursorLeft();
	 break;
	 /*************************************************************
	  *  Code to display math symbols form.
	  *  Alejandro Aguilar-Sierra, Nov 1995
	  */
      case 7: /* Symbols */
	 show_symbols_form();
	 break;
	 /**** End of my insertion. ***********************************/       
      }
      Update(0);
   } 
   fl_freepup(MathMenu);
};


void ShowOptionsMenu(FL_OBJECT *ob, long data)
{
   //static int OptionsMenu=0;

   int OptionsMenu = fl_defpup(FL_ObjWin(ob),
   " Screen Options... "
   "| Spellchecker Options... "
   "| Keyboard..." );

   //char* tmpstring = new char[50];
   fl_setpup_position(
		      fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);   
   int choice = fl_dopup(OptionsMenu);
   switch (choice){
   case 1: 
      {
	 fl_set_input(fd_form_screen->input_roman, roman_font_name);
	 fl_set_input(fd_form_screen->input_sans, sans_font_name);
	 fl_set_input(fd_form_screen->input_typewriter, typewriter_font_name);
	 fl_set_input(fd_form_screen->input_font_norm, font_norm);
	 char *tmpstring = new char[50];
	 sprintf(tmpstring,"%d", lyxrc.default_zoom);
	 fl_set_input(fd_form_screen->intinput_size, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_TINY]);
	 fl_set_input(fd_form_screen->intinput_tiny, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_SMALL]);
	 fl_set_input(fd_form_screen->intinput_small, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_NORMAL]);
	 fl_set_input(fd_form_screen->intinput_normal, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_LARGE]);
	 fl_set_input(fd_form_screen->intinput_large, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_LARGER]);
	 fl_set_input(fd_form_screen->intinput_larger, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_LARGEST]);
	 fl_set_input(fd_form_screen->intinput_largest, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_HUGE]);
	 fl_set_input(fd_form_screen->intinput_huge, tmpstring);
	 sprintf(tmpstring,"%d", font_sizes[LYX_SIZE_GIANT]);
	 fl_set_input(fd_form_screen->intinput_giant, tmpstring);
	 fl_show_form(fd_form_screen->form_screen,
		      FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER,
		      "Screen Options");
	 delete(tmpstring);
      }
      break;
   case 2:       
      if(SpellCheckerOptions())
	 ShowSpellChecker();
      break;
   case 3:
     MenuKeymap();
     break;
   default: break;
   }   
   fl_freepup(OptionsMenu);
}


extern void HandleBufferMenu(int);
extern void MakeBufferMenu(int);

void ShowBufferMenu(FL_OBJECT *ob, long data)
{
   //static int BufferMenu=0;
   
   int BufferMenu = fl_newpup(FL_ObjWin(ob));
   MakeBufferMenu(BufferMenu);
   
   fl_setpup_position(
		      fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);   
   int choice = fl_dopup(BufferMenu);
   if (choice > 0) HandleBufferMenu(choice);
   
   fl_freepup(BufferMenu);
}


void ShowHelpMenu(FL_OBJECT *ob, long data)
{
   //static int HelpMenu=0;
   
   int HelpMenu = fl_defpup(FL_ObjWin(ob),
   " Documentation... "
   "| Copyright... "
   "| Warranty... "
   "| Credits... "
   "| Version... ");

   fl_setpup_position(
		      fd_form_main->form_main->x + ob->x,
		      fd_form_main->form_main->y + ob->y + ob->h + 10);   
   int choice = fl_dopup(HelpMenu);
   switch (choice) {
   case 1:
      ProhibitInput();
      MenuDocu();
      AllowInput();
      break;
   case 2: 
      ProhibitInput();
      fl_show_message("","LyX is Copyright (C) 1995 by Matthias Ettrich\n(ettrich@informatik.uni-tuebingen.de)\n\nThis program is free software; you can redistribute it and/or modify it under the\nterms of the GNU General Public License as published by the Free Software\nFoundation; either version 2 of the License, or (at your option) any later version.","");
      AllowInput();
      break;
   case 3: 
      ProhibitInput();
      fl_show_message("","LyX is distributed in the hope that it will be useful, but WITHOUT ANY\nWARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n FOR A PARTICULAR PURPOSE. See the GNU General Public License for more\ndetails. You should have received a copy of the GNU General Public License\nalong with this program; if not, write to the Free Software Foundation,\n Inc., 675 Mass Ave, Cambridge, MA 02139, USA.","");
      AllowInput();
      break;
   case 4:
      ShowCredits();
      break;
   case 5:
      ProhibitInput();
      fl_show_message("","This is LyX " VERSION " of " RELEASE,"");
      AllowInput();
      break;
   }
   fl_freepup(HelpMenu);
}
