/* This file is part of
* ======================================================
* 
*           LyX, the High Level Word Processor
* 	 
*	    Copyright (C) 1995 Matthias Ettrich
*
*======================================================*/

#include "xdefinitions.h"

/* Konstruktor */ 
LyXScreen::LyXScreen(Window drawingarea, int h, int w, LyXText *text_ptr) {
   canvas = drawingarea;
   
   text = text_ptr;
   
   /* get the dimensions of the canvas */ 
   height = h;
   width = w;

   first = 0;
   
   /* the cursor isnt yet visible */ 
   cursor_visible = 0;
   screen_refresh_y = -1;
   
}
   
void LyXScreen::Redraw() {
   DrawFromTo(0, height);
   screen_refresh_y = -1;
   if (cursor_visible) {
      /* XFreePixmap(fl_display, cursor_pixmap); */ 
      cursor_visible = 0;
      ShowCursor();
   }
}



	  






void LyXScreen::DrawFromTo(int y1, int y2) {
   Row* row;
   long  y_text;
   Pixmap pm;
   long y;
   
   y_text = first + y1;
   
   /* get the first needed row */ 
   row = text->GetRowNearY(y_text);
   /* y_text is now the real beginning of the row */
   
   y = y_text - first;
   /* y1 is now the real beginning of row on the screen */
   
   
   
   while (row != NULL && y < y2) {
	 
      pm = text->GetVisibleRow(row, y + first);
      if (pm) {
	 XCopyArea (fl_display,
		    pm,
		    canvas,
		    LyXGetCopyGC(),
		    0, 0,
		    width, row->height,
		    0, y);
	 
	 
	 XFreePixmap (fl_display,pm);
      }
      
      XFlush(fl_display);
      y += row->height;
      row = row -> next;
      

      
   }
   
   /* maybe we have to clear the screen at the bottom */ 
   if (y < y2) {
      XFillRectangle(fl_display, canvas, LyXGetLightedGC(),
		     0, y, width, y2 - y);
   }



}
void LyXScreen::DrawOneRow(Row *row, long &y_text) {
   Pixmap pm;
   int y;
   
   
   y = y_text - first;
      
   if (y + row->height > 0 && y - row->height <= height) {
      /* ok there is anything visible */
      
      pm = text->GetVisibleRow(row, y + first);
      if (pm) {
	 XCopyArea (fl_display,
		    pm,
		    canvas,
		    LyXGetCopyGC(),
		    0, 0,
		    width, row->height,
		    0, y);
	 
	 XFreePixmap (fl_display,pm);
      }
      XFlush(fl_display);

   }
 
   y_text+=row->height;
   
   
}

/* draws the screen, starting with textposition y. uses as much already
* printed pixels as possible */
void LyXScreen::Draw(long  y) {
   long old_first;
   if (y<0)
      y = 0;
   old_first = first;
   first = y;
   /* is any optimiziation possible? */ 
   if ((y - old_first) < height 
       && (old_first - y) < height) {
      if (first < old_first) {
	 XCopyArea (fl_display,
		    canvas,
		    canvas,
		    LyXGetCopyGC(),
		    0, 0, 
		    width , height - old_first + first,
		    0, old_first - first);
	DrawFromTo(0, old_first - first);
      }
      else  {
	 XCopyArea (fl_display,
		    canvas,
		    canvas,
		    LyXGetCopyGC(),
		    0, first - old_first, 
		    width , old_first + height - first, 
		    0, 0);
	 DrawFromTo(height + old_first - first, height);
      }
	
   }
   else {
      /* make a dump new-draw */ 
      DrawFromTo(0, height);
   }
     
}

void LyXScreen::ShowCursor() {
   long  x1,x2,y1,y2;

   
   if (cursor_visible)
     return;
   
   x1 = text->cursor.x;
   x2 = text->cursor.x;
   y1 = text->cursor.y - LyXMaxAscent(text->real_current_font) - first + 1;
   y2 = text->cursor.y + LyXMaxDescent(text->real_current_font) - first - 1;
   
   if (y2 > 0 && y1 < height) {
      XDrawLine(fl_display,
		canvas,
		LyXGetSelectGC(),
		x1,y1,x2,y2);
      cursor_visible = 1;
   }
}

void LyXScreen::HideCursor() {
   if (cursor_visible) {
      cursor_visible = 0;
      ShowCursor();
      cursor_visible = 0;
   }
}



void LyXScreen::CursorHi() {
   if (!cursor_visible)
     ShowCursor();
}
void LyXScreen::CursorLo() {
   if (cursor_visible)
     HideCursor();
}
void LyXScreen::CursorToggle() {
   if (cursor_visible)
     HideCursor();
   else
     ShowCursor();
}

/* returns a new top so that the cursor is visible */ 
long LyXScreen::TopCursorVisible() {
   long  newtop = first;
   
   /* The old style, the new one is better but needs a greater scroll region than the text 
    if (text->cursor.y - text->cursor.row->baseline + text->cursor.row->height
	- first >= height
	&& first + height < text->height) {
       newtop = text->cursor.y - 3*height / 4;
       if (newtop > text->height - height)
       	  newtop = text->height - height;
    }
    else if (text->cursor.y - text->cursor.row->baseline <= first 
	     && first > 0) {
       newtop = text->cursor.y - height / 4;
       if (newtop < 0)
       	  newtop = 0;
    }
    * */ 
   
   if (text->cursor.y - text->cursor.row->baseline + text->cursor.row->height
       - first >= height)
     newtop = text->cursor.y - 3*height / 4;   /* the scroll region must be so big!! */
   else if (text->cursor.y - text->cursor.row->baseline < first 
	    && first > 0) {
     newtop = text->cursor.y - height / 4;
   }
   if (newtop < 0)
     newtop = 0;
   
   
   return newtop;
}


/* scrolls the screen so that the cursor is visible, if necessary.
* returns 1 if a change was made, otherwise 0 */ 
int  LyXScreen::FitCursor() {
   /* is a change necessary */ 
   
   long  newtop = TopCursorVisible();
   if (newtop != first)
     Draw(newtop);
   return ( newtop != first);
}

   
void  LyXScreen::Update() {
   long y;
   
   
   if (text->status == LYX_NEED_MORE_REFRESH || 
       screen_refresh_y > -1 ) {
	  if (screen_refresh_y > -1 && screen_refresh_y < text->refresh_y)
	    y = screen_refresh_y;
	  else
	    y = text->refresh_y;
	  
	  if (y < first)
	    y = first;
	  DrawFromTo(y - first, height);
	  text->refresh_y = 0;
	  text->status = LYX_UNCHANGED;
	  screen_refresh_y = -1;
   }
   else if (text->status == LYX_NEED_VERY_LITTLE_REFRESH) {
      /* ok I will update the current cursor row */
      
      DrawOneRow(text->refresh_row, text->refresh_y);
      text->status = LYX_UNCHANGED;
      
   }
   
}

void LyXScreen::SmallUpdate() {
   Row *row;
   long y;
   if (text->status == LYX_NEED_MORE_REFRESH){
      /* ok I will update till the current cursor row */
      row = text->refresh_row;
      y = text->refresh_y;
      
      if (y > text->cursor.y) {
	 Update();
	 return;
      }
	 
      while (row && row != text->cursor.row && y < first + height) {
	 DrawOneRow(row, y);
	 row = row->next;
      }
      
      DrawOneRow(row, y);
      screen_refresh_y = y;
      screen_refresh_row = row->next;
      text->status = LYX_UNCHANGED;
   }
   else if (text->status == LYX_NEED_VERY_LITTLE_REFRESH) {
      /* ok I will update the current cursor row */
      row = text->refresh_row;
      y = text->refresh_y;
      
      DrawOneRow(row, y);
      text->status = LYX_UNCHANGED;
   }
}



void LyXScreen::ToggleSelection() {
   
   long top;
   long bottom;
   
   
   /* only if there is a selection */ 
   if (!text->selection)
     return;
   
  
   
   /* selection only in one row ?*/ 
   if (text->sel_start_cursor.y == text->sel_end_cursor.y) {

      /* only if anything is visible */ 
      if (text->sel_start_cursor.y - text->cursor.row->baseline - first < height
	  && text->sel_start_cursor.y - text->sel_start_cursor.row->baseline + 
	  text->sel_start_cursor.row->height - first > 0)
      	XFillRectangle(fl_display,canvas,
		       LyXGetSelectGC(),
		       text->sel_start_cursor.x, 
		       text->sel_start_cursor.y - text->sel_start_cursor.row->baseline - first,
		       text->sel_end_cursor.x  - text->sel_start_cursor.x,
		       text->sel_start_cursor.row->height);
   
     
   }
   else  {
      /* the sel_start_cursor row first */ 
        /* only if anything is visible */ 
      if (text->sel_start_cursor.y - text->sel_start_cursor.row->baseline - first < height
	  && text->sel_start_cursor.y - text->sel_start_cursor.row->baseline + 
	  text->sel_start_cursor.row->height - first > 0)
      	XFillRectangle(fl_display,canvas,
		       LyXGetSelectGC(),
		       text->sel_start_cursor.x, 
		       text->sel_start_cursor.y - text->sel_start_cursor.row->baseline - first,
		       width - text->sel_start_cursor.x,
		       text->sel_start_cursor.row->height);
      
      /* the main body */ 
      
      if (text->sel_start_cursor.row->next != text->sel_end_cursor.row) {
	 top = text->sel_start_cursor.y - text->sel_start_cursor.row->baseline + text->sel_start_cursor.row->height;
	 bottom = text->sel_end_cursor.y - text->sel_end_cursor.row->baseline;
	 
	 if (top - first < 0)
	   top = first;
	 
	 if (bottom - first > height)
	   bottom = first + height;
	 
	 XFillRectangle(fl_display,canvas,
			LyXGetSelectGC(),
			0, 
			top - first,
			width,
			bottom - top);
	 
	 
      }
      
      
      
      
      
      /* the sel_end_cursor row last */ 
      if (text->sel_end_cursor.y - text->sel_end_cursor.row->baseline - first < height
	  && text->sel_end_cursor.y - text->sel_end_cursor.row->baseline + 
	  text->sel_end_cursor.row->height - first > 0)
      	XFillRectangle(fl_display,canvas,
		       LyXGetSelectGC(),
		       0, 
		       text->sel_end_cursor.y - text->sel_end_cursor.row->baseline - first,
		       text->sel_end_cursor.x,
		       text->sel_end_cursor.row->height);
   }
   
   XFlush(fl_display);
   
}
   
   
void LyXScreen::ToggleToggle() {
   
   long top;
   long bottom;
   
   
   /* only if there is a selection */ 
   /* 
   if (!text->selection)
     return;
   */ 
  
   
   
   /* selection only in one row ?*/ 
   if (text->toggle_cursor.y == text->toggle_end_cursor.y) {

      /* only if anything is visible */ 
      if (text->toggle_cursor.y - text->toggle_cursor.row->baseline - first < height
	  && text->toggle_cursor.y - text->toggle_cursor.row->baseline + 
	  text->toggle_cursor.row->height - first > 0)
      	XFillRectangle(fl_display,canvas,
		       LyXGetSelectGC(),
		       text->toggle_cursor.x, 
		       text->toggle_cursor.y - text->toggle_cursor.row->baseline - first,
		       text->toggle_end_cursor.x  - text->toggle_cursor.x,
		       text->toggle_cursor.row->height);
   
     
   }
   else  {
      /* the toggle_cursor row first */ 
        /* only if anything is visible */ 
      if (text->toggle_cursor.y - text->toggle_cursor.row->baseline - first < height
	  && text->toggle_cursor.y - text->toggle_cursor.row->baseline + 
	  text->toggle_cursor.row->height - first > 0)
      	XFillRectangle(fl_display,canvas,
		       LyXGetSelectGC(),
		       text->toggle_cursor.x, 
		       text->toggle_cursor.y - text->toggle_cursor.row->baseline - first,
		       width - text->toggle_cursor.x,
		       text->toggle_cursor.row->height);
      
      /* the main body */ 
      
      if (text->toggle_cursor.row->next != text->toggle_end_cursor.row) {
	 top = text->toggle_cursor.y - text->toggle_cursor.row->baseline + text->toggle_cursor.row->height;
	 bottom = text->toggle_end_cursor.y - text->toggle_end_cursor.row->baseline;
	 
	 if (top - first < 0)
	   top = first;
	 if (bottom - first < 0)
	   bottom = first;
	 
	 if (bottom - first > height)
	   bottom = first + height;
	 if (top - first > height)
	   top = first + height;
	 
	 if (top != bottom)
	   XFillRectangle(fl_display,canvas,
			  LyXGetSelectGC(),
			  0, 
			  top - first,
			  width,
			  bottom - top);
	 
	 
      }
      
      
      
      
      
      /* the toggle_end_cursor row last */ 
      if (text->toggle_end_cursor.y - text->toggle_end_cursor.row->baseline - first < height
	  && text->toggle_end_cursor.y - text->toggle_end_cursor.row->baseline +
	  text->toggle_end_cursor.row->height - first > 0)
      	XFillRectangle(fl_display,canvas,
		       LyXGetSelectGC(),
		       0, 
		       text->toggle_end_cursor.y - text->toggle_end_cursor.row->baseline - first,
		       text->toggle_end_cursor.x,
		       text->toggle_end_cursor.row->height);
   }
   
   XFlush(fl_display);

}
   
   
