#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <ctype.h>
#include "forms.h"
#include "sp_form.h"
#include "spellchecker.h"

#define ISP_OK 1
#define ISP_ROOT 2
#define ISP_COMPOUNDWORD 3
#define ISP_UNKNOWN 4
#define ISP_MISSED 5
#define ISP_IGNORE 6

#define WRITE_DICT 1
#define NO_WRITE 0

extern char *StringCopy(const char*);
extern char *GetLanguage();
static bool RunSpellChecker(const char*);

static FILE *in, *out;  /* streams to communicate with ispell */
static pid_t isp_pid;

static FD_form_spell_options *fd_form_spell_options = NULL;
static FD_form_spell_check *fd_form_spell_check = NULL;

static unsigned int isp_status = 2;
static int isp_fd;

#define ISP_LANG 1
#define ISP_COMPOUND 2
#define ISP_PDICT 4
#define ISP_ROOTAFFIX 8
#define ISP_ESC 16

static const char Compound[2][3] = { "-B", "-C" };
static char *pdict, *Language, *Esc_Chars;
static char *isp_argv[5];

static void sigchldhandler(int sig);

typedef struct isp_result {
  int flag;
  int count;
  char *string;
  char **misses;
} isp_result;


/* callbacks for form form_spell_options */
static void language_cb(FL_OBJECT *ob, long data)
{
  isp_status = data ? isp_status&~ISP_LANG:isp_status|ISP_LANG;
}

static void pdict_cb(FL_OBJECT *ob, long data)
{
  isp_status = fl_get_button(ob) ? isp_status|ISP_PDICT:isp_status&~ISP_PDICT;
}

static void compound_cb(FL_OBJECT *ob, long data)
{
  isp_status = fl_get_button(ob) ? isp_status|ISP_COMPOUND:isp_status&~ISP_COMPOUND;
}

static void rootaffix_cb(FL_OBJECT *ob, long data)
{
  isp_status = fl_get_button(ob) ? isp_status|ISP_ROOTAFFIX:isp_status&~ISP_ROOTAFFIX;
}

static void esc_chars_cb(FL_OBJECT *ob, long data)
{
  isp_status = fl_get_button(ob) ? isp_status|ISP_ESC:isp_status&~ISP_ESC;
}


static FD_form_spell_check *create_form_form_spell_check(void)
{
  FL_OBJECT *obj;
  FD_form_spell_check *fdui = (FD_form_spell_check *) fl_calloc(1, sizeof(FD_form_spell_check));

  fdui->form_spell_check = fl_bgn_form(FL_NO_BOX, 550, 350);
  obj = fl_add_box(FL_UP_BOX,0,0,550,350,"");
  obj = fl_add_text(FL_NORMAL_TEXT,55,300,380,40,"Ispell-frontend for LyX");
    fl_set_object_lsize(obj,FL_HUGE_SIZE);
    fl_set_object_lstyle(obj,FL_FIXEDBOLDITALIC_STYLE+FL_ENGRAVED_STYLE);
  fdui->text = obj = fl_add_text(FL_NORMAL_TEXT,60,260,330,30,"");
    fl_set_object_boxtype(obj,FL_DOWN_BOX);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fdui->input = obj = fl_add_input(FL_NORMAL_INPUT,60,215,330,30,"Replace");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fdui->browser = obj = fl_add_browser(FL_SELECT_BROWSER,60,70,330,135,"Near    \nMisses ");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lalign(obj,FL_ALIGN_LEFT);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fdui->options = obj = fl_add_button(FL_NORMAL_BUTTON,420,260,100,30,"Options...");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,FL_TIMESBOLD_STYLE);
  fdui->start = obj = fl_add_button(FL_NORMAL_BUTTON,420,215,100,30,"Start");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_lcol(obj,FL_BLUE);
  fdui->insert = obj = fl_add_button(FL_NORMAL_BUTTON,420,180,100,25,"I\010nsert");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMESBOLD_STYLE);
  fdui->accept = obj = fl_add_button(FL_NORMAL_BUTTON,420,145,100,25,"A\010ccept");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMESBOLD_STYLE);
  fdui->replace = obj = fl_add_button(FL_NORMAL_BUTTON,420,110,100,25,"R\010eplace");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMESBOLD_STYLE);
  fdui->stop = obj = fl_add_button(FL_NORMAL_BUTTON,420,70,100,30,"Stop");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_lcol(obj,FL_YELLOW);
  fdui->quit = obj = fl_add_button(FL_NORMAL_BUTTON,420,20,100,30,"Quit");
    fl_set_object_lsize(obj,FL_LARGE_SIZE);
    fl_set_object_lstyle(obj,FL_TIMESBOLD_STYLE);
    fl_set_object_lcol(obj,FL_RED);
  fdui->slider = obj = fl_add_slider(FL_HOR_FILL_SLIDER,10,30,390,20,"");
    fl_set_object_color(obj,FL_COL1,FL_BLUE);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lalign(obj,FL_ALIGN_LEFT);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  obj = fl_add_box(FL_NO_BOX,0,10,50,20,"0 %");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  obj = fl_add_box(FL_NO_BOX,355,10,50,20,"100 %");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fl_end_form();

  return fdui;
}


static FD_form_spell_options *create_form_form_spell_options(void)
{
  FL_OBJECT *obj;
  FD_form_spell_options *fdui = (FD_form_spell_options *) fl_calloc(1, sizeof(FD_form_spell_options));

  fdui->form_spell_options = fl_bgn_form(FL_NO_BOX, 340, 370);
  obj = fl_add_box(FL_UP_BOX,0,0,340,370,"");

  fdui->lang_buts = fl_bgn_group();
  obj = fl_add_checkbutton(FL_RADIO_BUTTON,30,290,30,30,"Use buffer language for dictionary");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_callback(obj,language_cb,1);
    fl_set_button(obj, 1);
  obj = fl_add_checkbutton(FL_RADIO_BUTTON,30,260,30,30,"Use alternate dictionary:");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_callback(obj,language_cb,0);
  fl_end_group();

  fdui->dict = obj = fl_add_input(FL_NORMAL_INPUT,60,230,230,30,"");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fdui->compounds = obj = fl_add_checkbutton(FL_PUSH_BUTTON,30,200,30,30,"Treat run-together words as legal compounds");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_callback(obj,compound_cb,0);
  fdui->rootaffix = obj = fl_add_checkbutton(FL_PUSH_BUTTON,30,170,30,30,"Report words made from root/affix combinations.");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_callback(obj,rootaffix_cb,0);
  fdui->ok = obj = fl_add_button(FL_NORMAL_BUTTON,20,10,140,30,"Ok.");
    fl_set_button_shortcut(obj,"^O^O",1);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fdui->start = obj = fl_add_button(FL_NORMAL_BUTTON,180,10,140,30,"Start Spellchecker.");
    fl_set_button_shortcut(obj,"^S^S",1);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,50,320,250,40,"ispell options");
    fl_set_object_lsize(obj,FL_HUGE_SIZE);
    fl_set_object_lstyle(obj,FL_FIXEDBOLDITALIC_STYLE+FL_ENGRAVED_STYLE);
  fdui->pdict = obj = fl_add_checkbutton(FL_PUSH_BUTTON,30,140,30,30,"Use alternate personal dictionary:");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_callback(obj,pdict_cb,0);
  fdui->esc_chars = obj = fl_add_checkbutton(FL_PUSH_BUTTON,30,80,30,30,"Escape characters:");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
    fl_set_object_callback(obj,esc_chars_cb,0);
  fdui->pdict_input = obj = fl_add_input(FL_NORMAL_INPUT,60,110,230,30,"");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fdui->esc_chars_input = obj = fl_add_input(FL_NORMAL_INPUT,60,50,230,30,"");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMES_STYLE);
  fl_end_form();

  return fdui;
}
/*---------------------------------------*/



bool SpellCheckerOptions(void) {
  FL_OBJECT *obj;
  bool start=false;
  
  if (fd_form_spell_options ==  NULL) {
    fd_form_spell_options = create_form_form_spell_options();
    fl_set_button(fd_form_spell_options->compounds, 
		  (isp_status&ISP_COMPOUND)? 1:0);
    fl_set_button(fd_form_spell_options->pdict, 
		  (isp_status&ISP_PDICT) ? 1:0);
    fl_set_button(fd_form_spell_options->esc_chars,
		  (isp_status&ISP_ESC) ? 1:0);
  }
  fl_show_form(fd_form_spell_options->form_spell_options,
	       FL_PLACE_CENTER,FL_FULLBORDER,
	       "Spellchecker Options");
  while(1) {
     obj = fl_do_forms();
     if (obj == fd_form_spell_options->ok) break;
     if	(obj == fd_form_spell_options->start){
	start=true;
	break;
     }
  }
  
  if(isp_status&ISP_LANG) {
    if (Language) {
      free(Language);
    }
    Language = strdup(fl_get_input(fd_form_spell_options->dict));
  }
  if(isp_status&ISP_PDICT) {
    if (pdict) {
      free(pdict);
    }
    pdict = strdup(fl_get_input(fd_form_spell_options->pdict));
  }
  if(isp_status&ISP_ESC) {
    if (Esc_Chars) {
      free(Esc_Chars);
    }
    Esc_Chars = (char *)malloc(strlen(fl_get_input(fd_form_spell_options->esc_chars_input))+3);
    sprintf(Esc_Chars, "\"%s\"", fl_get_input(fd_form_spell_options->esc_chars_input));
  }    
  fl_hide_form(fd_form_spell_options->form_spell_options);
  return start;
}

static void create_ispell_pipe(char *lang) {
  int oldfd;
  char buf[255];
  int pipein[2], pipeout[2];
  char * argv[12];
  int argc;

  isp_pid = -1;

  if(pipe(pipein)==-1 ||pipe(pipeout)==-1) {
/*    fl_show_message("Can't create pipe!", "",""); */
    return;
  }

  if ((out = fdopen(pipein[1], "w"))==NULL) {
/*    fl_show_message("Can't create stream for pipe!", "", "");*/
    return;
  }

  if ((in = fdopen(pipeout[0], "r"))==NULL) {
/*    fl_show_message("Can't create stream for pipe!", "", "");*/
    return;
  }

  setvbuf(out, NULL, _IOLBF, 0); 
  isp_fd = pipeout[0];

  isp_pid = fork();

  if(isp_pid==-1) {
/*    fl_show_message("Can't create child process!", "", "");*/
    return;
  }


  if(isp_pid==0) {        /* child process */
    dup2(pipein[0], STDIN_FILENO);
    dup2(pipeout[1], STDOUT_FILENO);
    close(pipein[0]);
    close(pipein[1]);
    close(pipeout[0]);
    close(pipeout[1]);

    argc = 0;
    argv[argc++] = "ispell";
    argv[argc++] = "-a";
    argv[argc++] = "-t";
    argv[argc++] = "-d";
    argv[argc++] = lang;
    argv[argc++] = StringCopy(Compound[(isp_status&ISP_COMPOUND)? 1:0]);
    if (isp_status&ISP_ESC) {
      argv[argc++] = "-w";
      argv[argc++] = Esc_Chars;
    }
    if (isp_status&ISP_PDICT) {
      argv[argc++] = "-p";
      argv[argc++] = pdict;
    }
    argv[argc++] = NULL;


    execvp("ispell", argv);

    fprintf(stderr, "Failed to start ispell!\n");
    exit(1);
  }

  /* parent process 
     read ispells identification message */
  fgets(buf, 256, in);

}


static isp_result *ispell_check_word(char *word) {
  isp_result *result;
  char buf[1024], *p, *nb;
  int count, i;

  result = (isp_result *)malloc(sizeof(isp_result));
  result->flag = ISP_UNKNOWN;
  result->count = 0;
  result->string = (char *)NULL;

  fputs(word, out); 
  fputc('\n', out);
  
  fgets(buf, 1024, in); 
  
  /* I think we have to check if ispell is still alive here because
     the signal-handler could have disabled blocking on the fd */
  if (isp_pid == -1) return (isp_result *) NULL;
  
  switch (*buf) {
  case '*':
    result->flag = ISP_OK;
    break;
  case '+':
    result->flag = ISP_ROOT;
    break;
  case '-':
    result->flag = ISP_COMPOUNDWORD;
    break;
  case '\n':
    result->flag = ISP_IGNORE;  /* This happens in case of numbers */
    break;
  case '#':
    result->flag = ISP_UNKNOWN;
    break;
  case '?':
  case '&':
    result->flag = ISP_MISSED;
    nb = strdup(buf);
    result->string = nb;
    p = strpbrk(nb+2, " ");
    sscanf(p, "%d", &count);
    result->count = count;
    if (count) result->misses = (char **)malloc(count*sizeof(char*));
    p = strpbrk(nb, ":");
    p +=2;
    for (i = 0; i<count; i++) {
      result->misses[i] = p;
      p = strpbrk(p, ",\n");
      *p = 0;
      p+=2;
    }
    break;
  default:
    result->flag = ISP_UNKNOWN;
  }

  *buf = 0;
  if (result->flag!=ISP_IGNORE) {
    while (*buf!='\n') fgets(buf, 255, in); /* wait for ispell to finish */
  }
  return result;
}

static inline void ispell_terminate() {
  /* tell ispell to write the dictionary here */
  fputs("#\n", out);

  fflush(out);
  fclose(out);
}

static inline void ispell_insert_word(const char *word) {
  fputc('*', out);
  fputs(word, out);
  fputc('\n', out);
}

static inline void ispell_accept_word(const char *word) 
{
   fputc('@', out);
   fputs(word, out);
   fputc('\n', out);
}

void ShowSpellChecker(){
  FL_OBJECT *obj;
  int ret;
  if (fd_form_spell_check == NULL) {
    fd_form_spell_check = create_form_form_spell_check();
  }

  fl_set_slider_value(fd_form_spell_check->slider, 0);
  fl_set_slider_bounds(fd_form_spell_check->slider, 0, 100);

  fl_show_form(fd_form_spell_check->form_spell_check,
	       FL_PLACE_CENTER,FL_FULLBORDER,
	       "Spellchecker");
  fl_deactivate_object(fd_form_spell_check->slider); 
  


  // deactivate insert, accept, replace, and stop
  fl_deactivate_object(fd_form_spell_check->insert);
  fl_deactivate_object(fd_form_spell_check->accept);
  fl_deactivate_object(fd_form_spell_check->replace);
  fl_deactivate_object(fd_form_spell_check->stop);
  fl_deactivate_object(fd_form_spell_check->input);
  fl_deactivate_object(fd_form_spell_check->browser);
  
  while(1){   
     obj = fl_do_forms();
     if (obj == fd_form_spell_check->options){
	fl_deactivate_form(fd_form_spell_check->form_spell_check);	
	SpellCheckerOptions();
	fl_activate_form(fd_form_spell_check->form_spell_check);
     }
     if (obj == fd_form_spell_check->start){
	// activate insert, accept, replace, and stop
	fl_activate_object(fd_form_spell_check->insert);
	fl_activate_object(fd_form_spell_check->accept);
	fl_activate_object(fd_form_spell_check->replace);
	fl_activate_object(fd_form_spell_check->stop);
	fl_activate_object(fd_form_spell_check->input);
	fl_activate_object(fd_form_spell_check->browser);
	// deactivate options and start
	fl_deactivate_object(fd_form_spell_check->options);
	fl_deactivate_object(fd_form_spell_check->start);

	ret = RunSpellChecker(GetLanguage());
	
	// deactivate insert, accept, replace, and stop
	fl_deactivate_object(fd_form_spell_check->insert);
	fl_deactivate_object(fd_form_spell_check->accept);
	fl_deactivate_object(fd_form_spell_check->replace);
	fl_deactivate_object(fd_form_spell_check->stop);
	fl_deactivate_object(fd_form_spell_check->input);
	fl_deactivate_object(fd_form_spell_check->browser);
	// activate options and start
	fl_activate_object(fd_form_spell_check->options);
	fl_activate_object(fd_form_spell_check->start);

	// if RunSpellChecker returns false quit spellchecker
	if (!ret) break;
     }
     if (obj == fd_form_spell_check->quit) break;
  }
   fl_hide_form(fd_form_spell_check->form_spell_check);
   EndOfSpellCheck();
   return;
}

static bool RunSpellChecker(const char *lang)
{
   isp_result *result;
   char *word;
   int i, oldval;
   float newval;
   char *tmp;
   FL_OBJECT *obj;
   char *language = StringCopy(lang);
   
   tmp = (isp_status&ISP_LANG) ? Language:language;

   oldval = 0;  /* used for updating slider only when needed */
	
   /* create ispell process */
   signal(SIGCHLD, sigchldhandler);

   create_ispell_pipe(tmp);

   if (isp_pid == -1) {
      fl_show_message("\n\nThe ispell-process has died for some reason. *One* possible reason\ncould be you do not have installed a dictionary file\nfor the language of this buffer. Check /usr/lib/ispell or set another\ndictionary in the Spell Checker Options menu.", "", "");
      fclose(out);
      return true;
   }

   while(1) {
      word = NextWord(newval);
      if (word==NULL) break;
      if (!isalpha(*word)) continue;
      if(20*5.0*newval>oldval) {
	 oldval = (int)(20*5.0*newval);
	 fl_set_slider_value(fd_form_spell_check->slider, oldval);
      }

      result = ispell_check_word(word);
      if (isp_pid==-1) break;

      obj =  fl_check_forms();
      if (obj == fd_form_spell_check->stop) {
	 free(result);
	 free(word);
	 ispell_terminate();
	 return true;
      }
      if (obj == fd_form_spell_check->quit) {
	 if (fl_show_question("Do you really want to quit the spellchecker?",
			      "","")) {
	    free(result);
	    free(word);
	    ispell_terminate(); 
	    // this will be done in ShowSpellChecker (Matthias)
	    // fl_hide_form(fd_form_spell_check->form_spell_check);
	    // EndOfSpellCheck();
	    return false;
	 }
      }
    
      switch (result->flag) {
      case ISP_UNKNOWN:
      case ISP_MISSED:
	 SelectLastWord();
	 fl_set_object_label(fd_form_spell_check->text, word);
	 fl_set_input(fd_form_spell_check->input, word);
	 fl_clear_browser(fd_form_spell_check->browser);
	 for (i=0; i<result->count; i++) {
	    fl_add_browser_line(fd_form_spell_check->browser, result->misses[i]);
	 }
	 if (result->count) {
	    free(result->misses);
	 }
	 free(result->string);
      
	 while(1) {
	    obj = fl_do_forms();
	    if (obj==fd_form_spell_check->insert) {
	       ispell_insert_word(word);
	       break;
	    }
	    if (obj==fd_form_spell_check->accept) {
	       ispell_accept_word(word);
	       break;
	    }
	  

	    if (obj==fd_form_spell_check->replace || 
		obj==fd_form_spell_check->input) {
	       ReplaceWord(fl_get_input(fd_form_spell_check->input));
	       ispell_insert_word(fl_get_input(fd_form_spell_check->input));
	       break;
	    }
	    if (obj==fd_form_spell_check->browser) {
	       fl_set_input(fd_form_spell_check->input, 
			    fl_get_browser_line(fd_form_spell_check->browser,
						fl_get_browser(fd_form_spell_check->browser)));
	    }
	    if (obj==fd_form_spell_check->stop) {
	       free(result);
	       free(word);
	       ispell_terminate();
	       return true;
	    }
	    
	    if (obj==fd_form_spell_check->quit) {
	       if(fl_show_question("Do you really want to quit the spellchecker?", "", "")) { 
		  free(result);
		  free(word);
		  ispell_terminate();
		  // this will be done in ShowSpellChecker (Matthias)
		  // fl_hide_form(fd_form_spell_check->form_spell_check);
		  // EndOfSpellCheck();
		  return false;
	       }
	    }
	 }
      default:
	 free(result);
	 free(word);
      }
   }
   if(isp_pid!=-1) {
      ispell_terminate();
      fl_show_message("Spell-checking completed!", "", "Goodbye.");
   } else {
      fl_show_message("The ispell-process has died for some reason. Maybe\nit has been killed.", "", "");
      fclose(out);
   }
   return true;
}

static void sigchldhandler(int sig){ 
  int status ;
  if (isp_pid>0)
    if (waitpid(isp_pid,&status,WNOHANG)==isp_pid) {
	isp_pid=-1;
	fcntl(isp_fd, F_SETFL, O_NONBLOCK); /* set the file descriptor
					       to nonblocking so we can 
					       continue */
      }
}
