/* This file is part of
* ======================================================
* 
*           LyX, the High Level Word Processor
* 	 
*	    Copyright (C) 1995 Matthias Ettrich
*
*======================================================*/

#include "inset.h"
#include "layout.h"
#include "lyx.h"
#include <ctype.h>
#include <forms.h>

extern FD_form_main *fd_form_main;

extern  WriteAlert(const char* s1, const char* s2, const char* s3);
extern  SetMinibuffer(const char* s1, const char* s2, const char* s3);


#define LYX_PAPER_MARGIN 15

static int IsSeparatorChar(char c) {
   return (c == ' ' || c == LYX_META_PROTECTED_SEPARATOR);
}

static int IsLineSeparatorChar(char c) {
   return (c == ' ');
}

static int IsSeparator(LyXParagraph  *par, int pos) {
   return IsSeparatorChar(par->GetChar(pos));
}

static int IsLineSeparator(LyXParagraph  *par, int pos) {
   return IsLineSeparatorChar(par->GetChar(pos));
}

static int IsNewlineChar(char c) {
   return (c == LYX_META_NEWLINE);
}

static int IsNewline(LyXParagraph  *par, int pos) {
   return IsNewlineChar(par->GetChar(pos));
}

static int ClearParagraph(LyXParagraph *par){
  int i=0;
  if (par->layout != LYX_DUMMY_LAYOUT){
    while (par->Last()
	   && (IsNewline(par,0) 
	       || IsLineSeparator(par,0))){
      par->Erase(0);
      i++;
    }
  }
  return i;
}



void LyXText::DeleteEmptyParagraphMechanism(LyXCursor old_cursor) {
  char deleted = False;

  /* this is the delete-empty-paragraph-mechanism. */ 
  if (selection)
    return;

  if (old_cursor.par != cursor.par) {
    if ( (old_cursor.par->Last() == 0
	  || (old_cursor.par->Last() == 1
	      && (IsLineSeparator(old_cursor.par,0))))
	 && old_cursor.par->FirstPhysicalPar() == old_cursor.par->LastPhysicalPar()
	 && (style(parameters->textclass, old_cursor.par->GetLayout())->labeltype!=LABEL_SENSITIVE
	     || (old_cursor.par->footnoteflag == LYX_NO_FOOTNOTE
		 || (old_cursor.par->footnotekind != LYX_FIG
		     && old_cursor.par->footnotekind != LYX_TAB)))) {
	     
      /* ok, we will delete anything */ 
	     
      /* make shure that you do not delete any environements */ 
      if ((old_cursor.par->footnoteflag != LYX_OPEN_FOOTNOTE &&
	   !(old_cursor.row->previous 
	     && old_cursor.row->previous->par->footnoteflag == LYX_OPEN_FOOTNOTE)
	   && !(old_cursor.row->next 
		&& old_cursor.row->next->par->footnoteflag == LYX_OPEN_FOOTNOTE))
	  || 
	  (old_cursor.par->footnoteflag == LYX_OPEN_FOOTNOTE &&
	   ((old_cursor.row->previous 
	     && old_cursor.row->previous->par->footnoteflag == LYX_OPEN_FOOTNOTE)
	    || 
	    (old_cursor.row->next
	     && old_cursor.row->next->par->footnoteflag == LYX_OPEN_FOOTNOTE))
	   )){
	status = LYX_NEED_MORE_REFRESH;
	deleted = True;
	 if (parameters->paragraph == old_cursor.par) {
	    parameters->paragraph = parameters->paragraph->next;
	 }
	 
	 if (old_cursor.row->previous) {
	    refresh_row = old_cursor.row->previous;
	    refresh_y = old_cursor.y - old_cursor.row->baseline - refresh_row->height;
	    /* delete old row */ 
	    RemoveRow(old_cursor.row);
	    /* delete old par */ 
	    delete(old_cursor.par);
	    
	    /* Breakagain the next par. Needed because of
	     * the parindent that can occur or dissappear.
	     * The next row can change its height, if there
	     * is another layout before */
	    if (refresh_row->next) {
	       BreakAgain(refresh_row->next);
	       UpdateCounters(refresh_row);
	    }
	    SetHeightOfRow(refresh_row);
	 }
	 else {
	    refresh_row = old_cursor.row->next;
	    refresh_y = old_cursor.y - old_cursor.row->baseline;
	    
	    /* delete old row */ 
	    RemoveRow(old_cursor.row);
	    /* delete old par */ 
	    delete(old_cursor.par);
	    
	    /* Breakagain the next par. Needed because of
	     * the parindent that can occur or dissappear.
	     * The next row can change its height, if there
	     * is another layout before */ 
	    if (refresh_row) {
	    BreakAgain(refresh_row);
	       UpdateCounters(refresh_row);
	    }
	 }
	 
	 /* correct cursor y */
	 SetCursor(cursor.par, cursor.pos);
		     
	 /* if (cursor.y > old_cursor.y)
	     cursor.y -= old_cursor.row->height; */ 
	 
	 if (sel_cursor.par  == old_cursor.par
	     && sel_cursor.pos == sel_cursor.pos) {
	    /* correct selection*/ 
	    sel_cursor = cursor;
	 }
      }
       
    }
     if (!deleted){
	if (ClearParagraph(old_cursor.par)){
	   RedoParagraphs(old_cursor, old_cursor.par->Next());
	   /* correct cursor y */
	   SetCursor(cursor.par, cursor.pos);
	}
     }
  }
}



	 
      
LyXFont LyXText::GetFont(LyXParagraph* par, int pos){
   LyXFont layoutfont, tmpfont;

   LyXLayout *layout = style(parameters->textclass, par->GetLayout());
   
   if (pos < BeginningOfMainBody(par))
     layoutfont = layout->labelfont;
   else
     layoutfont = layout->font;
   
   
   tmpfont = par->GetFontSettings(pos);
   
   if (tmpfont.family == LYX_LAYOUT_DEFAULT)
     tmpfont.family = layoutfont.family;
   if (tmpfont.series == LYX_LAYOUT_DEFAULT)
     tmpfont.series = layoutfont.series;
   if (tmpfont.shape == LYX_LAYOUT_DEFAULT)
     tmpfont.shape = layoutfont.shape;
   if (tmpfont.size == LYX_LAYOUT_DEFAULT)
     tmpfont.size = layoutfont.size;
   if (tmpfont.bar == LYX_LAYOUT_DEFAULT)
     tmpfont.bar = layoutfont.bar;
   if (tmpfont.latex == LYX_LAYOUT_DEFAULT)
     tmpfont.latex = layoutfont.latex;
   
   /* if this paragraph is an open footnote, just make the font smaller */ 
   if (par->footnoteflag == LYX_OPEN_FOOTNOTE
       && par->footnotekind == LYX_FOOTNOTE) {
	  if (tmpfont.size > LYX_SIZE_SMALL) tmpfont.size--;
       }
   
   
   return tmpfont;
};
void LyXText::SetCharFont(LyXParagraph *par, int pos, LyXFont font){
   LyXFont layoutfont;

   /* let the insets convert their font */ 
   if (par->GetChar(pos) == LYX_META_INSET) {
      if (par->GetInset(pos))
      	 font = par->GetInset(pos)->ConvertFont(font);
   }
   
   LyXLayout *layout = style(parameters->textclass, par->GetLayout());
   
   if (pos < BeginningOfMainBody(par))
     layoutfont = layout->labelfont;
   else
     layoutfont = layout->font;

   if (font.family == layoutfont.family)
     font.family = LYX_LAYOUT_DEFAULT;
   if (font.series == layoutfont.series)
     font.series = LYX_LAYOUT_DEFAULT;
   if (font.shape == layoutfont.shape)
     font.shape = LYX_LAYOUT_DEFAULT;
   if (font.size == layoutfont.size)
     font.size = LYX_LAYOUT_DEFAULT;
   if (font.bar == layoutfont.bar)
     font.bar = LYX_LAYOUT_DEFAULT;
   if (font.latex == layoutfont.latex)
     font.latex = LYX_LAYOUT_DEFAULT;
   par->SetFont(pos, font);
};



	  

/* inserts a new row behind the specified row, increments
* the touched counters */
void LyXText::InsertRow(Row *row, LyXParagraph *par, int pos) {
   Row *tmprow = new Row;
   if (!row) {
      tmprow->previous = NULL;
      tmprow->next = firstrow;
      firstrow = tmprow;
   }
   else {
      tmprow->previous = row;
      tmprow->next = row->next;
      row->next = tmprow;
   }
   
   if (tmprow->next)
     tmprow->next->previous = tmprow;
   
   if (tmprow->previous)
     tmprow->previous->next = tmprow;
   
   
   tmprow->par = par;
   tmprow->pos = pos;
   
   if (row == lastrow)
     lastrow = tmprow;
   number_of_rows++;			       /* one more row  */
}

/* removes the row and reset the touched counters */
void LyXText::RemoveRow(Row *row) {
   /* this must not happen before the currentrow for clear reasons.
      so the trick is just to set the current row onto the previous
      row of this row */
   long unused_y;
   GetRow(row->par, row->pos, unused_y);
   currentrow = currentrow->previous;
   if (currentrow)
     currentrow_y -= currentrow->height;
   else
     currentrow_y = 0;
   
   if (row->next)
     row->next->previous = row->previous;
   if (!row->previous) {
      firstrow = row->next;
   }
   else  {
      row->previous->next = row->next;
   }
   if (row == lastrow)
     lastrow = row->previous;
   
   height -= row->height;	       /* the text becomes smaller  */
   
   delete(row);
   number_of_rows--;			       /* one row less  */
}
     

/* remove all following  rows of the paragraph of the specified row. */
void LyXText::RemoveParagraph(Row *row) {
   LyXParagraph *tmppar;
   Row *tmprow;

   tmppar = row->par;
   row = row->next;
    
   while (row && row->par == tmppar) {
      tmprow = row->next;
      RemoveRow(row);
      row = tmprow;
   }
}
   
   


  
/* insert the specified paragraph behind the specified row */
void LyXText::InsertParagraph(LyXParagraph *par, Row *row) {
   InsertRow(row, par, 0);	       /* insert a new row, starting 
					* at postition 0 */

   SetCounter(par);		       /* set the counters  */
   
   /* and now append the whole paragraph behind the new row */
   if (!row) {
      firstrow->height = 0;
      AppendParagraph(firstrow);
   }
   else {
      row->next->height = 0;
      AppendParagraph(row->next);
   }
}
    
   

/* konstruktor */ 
LyXText::LyXText(int pw, LyXTextParameters *p) {
   firstrow = NULL;
   lastrow = NULL;
   currentrow = NULL;
   currentrow_y = 0;
   paperwidth = pw;
   parameters = p;
   number_of_rows = 0;
   refresh_y= 0;
   status = LYX_UNCHANGED;
   LyXParagraph *par = p->paragraph;
   current_font = GetFont(par, 0);
   
   height = 0;
   
   while (par) {
      InsertParagraph(par, lastrow);
      par = par->Next();
   }
   /* set cursor at the very top position */
   selection = True;		       /* these setting is necessary 
    					* because of the delete-empty-
					* paragraph mechanism in
					* SetCursor */
   SetCursor(firstrow->par, 0);
   sel_cursor = cursor;
   selection = False;
   mark_set = False;
   
   /* no rebreak necessary */ 
   need_break_row = NULL;
   


   
   
}


/* Destruktor */ 
LyXText::~LyXText() {
   Row *tmprow;
   
   /* SetCursor(firstrow->par, 0); */ 
   
   tmprow = firstrow;
   
   while (firstrow) {		       /* delete all rows, this does 
					* not touch the paragraphs! */
      tmprow = firstrow->next;
      delete(firstrow);
      firstrow = tmprow;
   }
}


void LyXText::ToggleFootnote() {
   LyXParagraph *par;
   
   /* this is the same function than editing insets */ 
   if (cursor.pos < cursor.par->Last() 
       && cursor.par->GetChar(cursor.pos) == LYX_META_INSET
       && cursor.par->GetInset(cursor.pos)->Editable()) {
	  cursor.par->GetInset(cursor.pos)->Edit();
	  SetMinibuffer("Opened inset.","","");
       }
   else {
      par = cursor.par->ParFromPos(cursor.pos);
      if (par->next && par->next->footnoteflag == LYX_CLOSED_FOOTNOTE){
      	OpenFootnote();
	SetMinibuffer("Opened float.","","");
      }
      else {
	SetMinibuffer("Close footnote.","","");
      	CloseFootnote();
      }
   }
}





void LyXText::CloseFootnote() {
   LyXParagraph *par, *endpar,*tmppar;
   Row *row;
   
   par = cursor.par->ParFromPos(cursor.pos);
   
   /* if the cursor is not in an open footnote, or 
    * there is no open footnote in this paragraph, just return. */ 
   if (cursor.par->footnoteflag != LYX_OPEN_FOOTNOTE) {
      
      if (!par->next
	  || par->next->footnoteflag != LYX_OPEN_FOOTNOTE) {
	 SetMinibuffer("Nothing to do.","","");
	 return;
      }

   
   
      /* ok, move the cursor right before the footnote */ 
 
      /* just a little faster than using CursorRight() */
      for (cursor.pos=0; cursor.par->ParFromPos(cursor.pos)!=par; cursor.pos++);
      /* now the cursor is at the beginning of the physical par */
      SetCursor(cursor.par, cursor.pos + cursor.par->ParFromPos(cursor.pos)->last);
   }
   else  {
       /* we are in a footnote, so let us move at the beginning */ 
       /*      while (cursor.par->footnoteflag == LYX_OPEN_FOOTNOTE)
	       cursor.par = cursor.par->Previous();
	       
	       SetCursor(cursor.par, cursor.par->Last()); */
       /* this is just faster than using just CursorLeft() */ 
       
       
     tmppar = cursor.par;
     while (tmppar->footnoteflag == LYX_OPEN_FOOTNOTE) {
       /* just a little bit faster than movin the cursor */
	  tmppar = tmppar->Previous();
     }
      SetCursor(tmppar, tmppar->Last());
   }
   
   
   
   /* the cursor must be exactly before the footnote */ 
   par = cursor.par->ParFromPos(cursor.pos);
   
   status = LYX_NEED_MORE_REFRESH;
   refresh_row = cursor.row;
   refresh_y = cursor.y - cursor.row->baseline;
   
   tmppar = cursor.par;
   endpar = par->NextAfterFootnote()->Next();
   row = cursor.row;
   
   tmppar->CloseFootnote(cursor.pos);


   /* set the dimensions of the cursor row */
   /* row->fill = Fill(row, paperwidth);
   SetHeightOfRow(row);  */ 
   
   while (tmppar != endpar) {
      RemoveRow(row->next);
      if (row->next)
      	tmppar = row->next->par;
      else
      	tmppar = NULL;
   }
   
   AppendParagraph(cursor.row);
   
   
   
   SetCursor(cursor.par, cursor.pos);
   sel_cursor = cursor;
   
   /* just necessary */
   if (cursor.row->next)
       SetHeightOfRow(cursor.row->next);
}





/* used in setlayout */
void LyXText::MakeFontEntriesLayoutSpecific(LyXParagraph *par) {
   LyXFont layoutfont, tmpfont;
   
   int pos;
   
   LyXLayout* layout = style(parameters->textclass, par->GetLayout());
   
   
   for (pos = 0; pos < par->Last(); pos++) {
      if (pos < BeginningOfMainBody(par))
      	layoutfont = layout->labelfont;
      else
      	layoutfont = layout->font;
      
      tmpfont = par->GetFontSettings(pos);
   
      if (tmpfont.family == layoutfont.family)
      	tmpfont.family = LYX_LAYOUT_DEFAULT;
      if (tmpfont.series == layoutfont.series)
      	tmpfont.series = LYX_LAYOUT_DEFAULT;
      if (tmpfont.shape == layoutfont.shape)
      	tmpfont.shape = LYX_LAYOUT_DEFAULT;
      if (tmpfont.size == layoutfont.size)
      	tmpfont.size = LYX_LAYOUT_DEFAULT;
      if (tmpfont.bar == layoutfont.bar)  
      	tmpfont.bar = LYX_LAYOUT_DEFAULT;
      if (tmpfont.latex == layoutfont.latex)  
      	tmpfont.latex = LYX_LAYOUT_DEFAULT;
      par->SetFont(pos, tmpfont);
   }
   
}





/* set layout over selection and make a total rebreak of those  paragraphs */
void  LyXText::SetLayout(char layout){
   LyXCursor tmpcursor;
  

   /* if there is no selection just set the layout of the current paragraph  */
   if (!selection) {
      sel_start_cursor = cursor;       /* dummy selection  */
      sel_end_cursor = cursor;
   }

   tmpcursor = cursor;		       /* store the current cursor  */

   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   cursor = sel_start_cursor;
   
   LyXLayout * lyxlayout = style(parameters->textclass, layout);
   
   while (cursor.par != sel_end_cursor.par) {
      if (cursor.par->footnoteflag ==
	  sel_start_cursor.par->footnoteflag) {
	     cursor.par->SetLayout(layout);
	     MakeFontEntriesLayoutSpecific(cursor.par);
	     cursor.par->FirstPhysicalPar()->fill_top = lyxlayout->fill_top;
	     cursor.par->FirstPhysicalPar()->fill_bottom = lyxlayout->fill_bottom;
	     if (lyxlayout->bordertype == BORDER_MANUAL)
	       cursor.par->SetLabelWidthString(lyxlayout->labelstring);
	  }
      cursor.par = cursor.par->Next();
   }
   if (cursor.par->footnoteflag ==
       sel_start_cursor.par->footnoteflag) {
	  cursor.par->SetLayout(layout);
	  MakeFontEntriesLayoutSpecific(cursor.par);
	  cursor.par->FirstPhysicalPar()->fill_top = lyxlayout->fill_top;
	  cursor.par->FirstPhysicalPar()->fill_bottom = lyxlayout->fill_bottom;
	  if (lyxlayout->bordertype == BORDER_MANUAL)
	    cursor.par->SetLabelWidthString(lyxlayout->labelstring);
       }
   
   
   
   LyXParagraph *endpar = sel_end_cursor.par->LastPhysicalPar()->Next();

   if (endpar && endpar->GetDepth()) {
      while (endpar && endpar->GetDepth()) {
	 endpar->depth = 0;
	 endpar = endpar->Next();
      }
   }
   else if (endpar) {
      endpar = endpar->Next();	       /* because of parindents etc.  */
   }
   
   RedoParagraphs(sel_start_cursor, endpar);
   
   
   /* we have to reset the selection, because the
    * geometry could have changed */ 
   SetCursor(sel_start_cursor.par, sel_start_cursor.pos);
   sel_cursor = cursor;
   SetCursor(sel_end_cursor.par, sel_end_cursor.pos);
   UpdateCounters(cursor.row);
   ClearSelection();
   SetSelection();
   SetCursor(tmpcursor.par, tmpcursor.pos);
}


/* increment deph over selection and make a total rebreak of those  paragraphs */
void  LyXText::IncDepth(){
   LyXCursor tmpcursor;
   
   
   /* if there is no selection just set the layout of the current paragraph  */
   if (!selection) {
      sel_start_cursor = cursor;       /* dummy selection  */
      sel_end_cursor = cursor;
   }
   
   tmpcursor = cursor;		       /* store the current cursor  */

   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   cursor = sel_start_cursor;
   
   char anything_changed = 0;
   
   while (cursor.par != sel_end_cursor.par) {
      if (cursor.par->footnoteflag ==
	  sel_start_cursor.par->footnoteflag) {
	     if (cursor.par->FirstPhysicalPar()->Previous()
		 && (cursor.par->FirstPhysicalPar()->Previous()->GetDepth() -
		     cursor.par->GetDepth() > 0
		     ||
		     (cursor.par->FirstPhysicalPar()->Previous()->GetDepth()
		      ==  cursor.par->GetDepth()
		      && style(parameters->textclass, cursor.par->FirstPhysicalPar()->Previous()->GetLayout())->latextype != LATEX_PARAGRAPH
		      && style(parameters->textclass, cursor.par->FirstPhysicalPar()
			       ->Previous()->GetLayout())->latextype != LATEX_COMMAND))) {
		cursor.par->FirstPhysicalPar()->depth++;
		anything_changed = 1;
	     }
	  }
      cursor.par = cursor.par->Next();
   }
   if (cursor.par->footnoteflag ==
       sel_start_cursor.par->footnoteflag) {
	  if (cursor.par->FirstPhysicalPar()->Previous()
	      && (cursor.par->FirstPhysicalPar()->Previous()->GetDepth() -
		  cursor.par->GetDepth() > 0
		  ||
		  (cursor.par->FirstPhysicalPar()->Previous()->GetDepth()
		   ==  cursor.par->GetDepth()
		   && style(parameters->textclass, cursor.par->FirstPhysicalPar()->Previous()->GetLayout())->latextype != LATEX_PARAGRAPH
		   && style(parameters->textclass, cursor.par->FirstPhysicalPar()
			    ->Previous()->GetLayout())->latextype != LATEX_COMMAND))) {
	     cursor.par->FirstPhysicalPar()->depth++;
	     anything_changed = 1;
	  }
   }
   
   /* if nothing changed set all depth to 0 */ 
   if (!anything_changed) {
       cursor = sel_start_cursor;
      while (cursor.par != sel_end_cursor.par) {
	 cursor.par->FirstPhysicalPar()->depth = 0;
	 cursor.par = cursor.par->Next();
      }
      if (cursor.par->footnoteflag == sel_start_cursor.par->footnoteflag)
      	 cursor.par->FirstPhysicalPar()->depth = 0;
   }
   
   
   LyXParagraph *endpar = sel_end_cursor.par->LastPhysicalPar()->Next();
   if (endpar)
     endpar = endpar->Next();	       /* this is because of parindents  */
   
   while (endpar && endpar->GetDepth())
     endpar = endpar->Next();
   
   RedoParagraphs(sel_start_cursor, endpar);
   
   
   /* we have to reset the selection, because the
    * geometry could have changed */ 
   SetCursor(sel_start_cursor.par, sel_start_cursor.pos);
   sel_cursor = cursor;
   SetCursor(sel_end_cursor.par, sel_end_cursor.pos);
   UpdateCounters(cursor.row);
   ClearSelection();
   SetSelection();
   SetCursor(tmpcursor.par, tmpcursor.pos);
}

/* decrement deph over selection and make a total rebreak of those  paragraphs */
void  LyXText::DecDepth(){
   LyXCursor tmpcursor;
   
   
   /* if there is no selection just set the layout of the current paragraph  */
   if (!selection) {
      sel_start_cursor = cursor;       /* dummy selection  */
      sel_end_cursor = cursor;
   }
   
   tmpcursor = cursor;		       /* store the current cursor  */

   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   cursor = sel_start_cursor;
   
   while (cursor.par != sel_end_cursor.par) {
      if (cursor.par->footnoteflag ==
	  sel_start_cursor.par->footnoteflag) {
	     if (cursor.par->FirstPhysicalPar()->depth)
	       cursor.par->FirstPhysicalPar()->depth--;
	  }
      cursor.par = cursor.par->Next();
   }
   if (cursor.par->footnoteflag ==
       sel_start_cursor.par->footnoteflag) {
	  if (cursor.par->FirstPhysicalPar()->depth)
	    cursor.par->FirstPhysicalPar()->depth--;
       }
   
   
   LyXParagraph *endpar = sel_end_cursor.par->LastPhysicalPar()->Next();
   if (endpar)
     endpar = endpar->Next();	       /* this is because of parindents  */
   
   while (endpar && endpar->GetDepth())
     endpar = endpar->Next();
   
   RedoParagraphs(sel_start_cursor, endpar);
   
   
   /* we have to reset the selection, because the
    * geometry could have changed */ 
   SetCursor(sel_start_cursor.par, sel_start_cursor.pos);
   sel_cursor = cursor;
   SetCursor(sel_end_cursor.par, sel_end_cursor.pos);
   UpdateCounters(cursor.row);
   ClearSelection();
   SetSelection();
   SetCursor(tmpcursor.par, tmpcursor.pos);
}



/* set font over selection and make a total rebreak of those paragraphs */
void  LyXText::SetFont(LyXFont font){
   LyXCursor tmpcursor;
  

   /* if there is no selection just set the current_font */
   if (!selection) {
      if (font.family != LYX_NO_FONT_CHANGE)
      	current_font.family = font.family;
      if (font.series != LYX_NO_FONT_CHANGE)
      	current_font.series = font.series;
      if (font.shape != LYX_NO_FONT_CHANGE)
      	current_font.shape = font.shape;
      if (font.size != LYX_NO_FONT_CHANGE)
      	current_font.size = font.size;
      if (font.bar != LYX_NO_FONT_CHANGE)
      	current_font.bar = font.bar;
      
      
      /* set the real current font */ 
      LyXLayout *layout = style(parameters->textclass, cursor.par->GetLayout());
      LyXFont layoutfont;
      if (cursor.pos < BeginningOfMainBody(cursor.par))
      	 layoutfont = layout->labelfont;
      else
      	 layoutfont = layout->font;
      
      real_current_font = current_font;
      
      if (real_current_font.family == LYX_LAYOUT_DEFAULT)
      	 real_current_font.family = layoutfont.family;
      if (real_current_font.series == LYX_LAYOUT_DEFAULT)
      	 real_current_font.series = layoutfont.series;
      if (real_current_font.shape == LYX_LAYOUT_DEFAULT)
      	 real_current_font.shape = layoutfont.shape;
      if (real_current_font.size == LYX_LAYOUT_DEFAULT)
      	 real_current_font.size = layoutfont.size;
      if (real_current_font.latex == LYX_LAYOUT_DEFAULT)
      	 real_current_font.latex = layoutfont.latex;
      if (real_current_font.bar == LYX_LAYOUT_DEFAULT)
      	 real_current_font.bar = layoutfont.bar;
      return;
   }
   
   tmpcursor = cursor;		       /* store the current cursor  */
   
   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   cursor = sel_start_cursor;
   while (cursor.par != sel_end_cursor.par ||
	  cursor.pos < sel_end_cursor.pos) {
	     if (cursor.pos < cursor.par->Last()
		 && cursor.par->footnoteflag
		 == sel_start_cursor.par->footnoteflag) {   /* an open footnote
							    * should behave
							    * like a closed */
		    SetCharFont(cursor.par, cursor.pos, font);
		    cursor.pos++;
		 }
	     else {
		cursor.pos = 0;
		cursor.par = cursor.par->Next();
	     }
	  }
   
   RedoParagraphs(sel_start_cursor, sel_end_cursor.par->Next());
   
   
   /* we have to reset the selection, because the
    * geometry could have changed */ 
   SetCursor(sel_start_cursor.par, sel_start_cursor.pos);
   sel_cursor = cursor;
   SetCursor(sel_end_cursor.par, sel_end_cursor.pos);
   ClearSelection();
   SetSelection();
   SetCursor(tmpcursor.par, tmpcursor.pos);
}

void LyXText::RedoHeightOfParagraph(LyXCursor cursor) {
   Row *tmprow;
   LyXParagraph *first_phys_par;
   long y;
   
   tmprow = cursor.row;
   
   y = cursor.y - tmprow->baseline;
   SetHeightOfRow(tmprow);
   first_phys_par = tmprow->par->FirstPhysicalPar();
   /* find the first row of the paragraph */
   if (first_phys_par != tmprow->par)
     while (tmprow->previous && tmprow->previous->par != first_phys_par)  {
	tmprow = tmprow->previous;
	y -= tmprow->height;
	SetHeightOfRow(tmprow);
     }
   while (tmprow->previous && tmprow->previous->par == first_phys_par)  {
      tmprow = tmprow->previous;
      y -= tmprow->height;
      SetHeightOfRow(tmprow);
      
   }
   
   /* we can set the refreshing parameters now */
   status = LYX_NEED_MORE_REFRESH;
   refresh_y = y;
   refresh_row = tmprow;
   SetCursor(cursor.par, cursor.pos);
						
};

void LyXText::RedoDrawingOfParagraph(LyXCursor cursor) {
   Row *tmprow;
   LyXParagraph *first_phys_par;
   long y;
   
   tmprow = cursor.row;
   
   y = cursor.y - tmprow->baseline;
   SetHeightOfRow(tmprow);
   first_phys_par = tmprow->par->FirstPhysicalPar();
   /* find the first row of the paragraph */
   if (first_phys_par != tmprow->par)
     while (tmprow->previous && tmprow->previous->par != first_phys_par)  {
	tmprow = tmprow->previous;
	y -= tmprow->height;
     }
   while (tmprow->previous && tmprow->previous->par == first_phys_par)  {
      tmprow = tmprow->previous;
      y -= tmprow->height;
      
   }
   
   /* we can set the refreshing parameters now */
   status = LYX_NEED_MORE_REFRESH;
   refresh_y = y;
   refresh_row = tmprow;
   SetCursor(cursor.par, cursor.pos);
						
};





/* deletes and inserts again all paragaphs between the cursor
* and the specified par 
* This function is needed after SetLayout and SetFont etc. */
void LyXText::RedoParagraphs(LyXCursor cursor, LyXParagraph *endpar) {
   Row *tmprow, *tmprow2;
   LyXParagraph *tmppar, *first_phys_par;
   long y;
   
   tmprow = cursor.row;
   
   y = cursor.y - tmprow->baseline;
   first_phys_par = tmprow->par->FirstPhysicalPar();
   /* find the first row of the paragraph */
   if (first_phys_par != tmprow->par)
     while (tmprow->previous && tmprow->previous->par != first_phys_par)  {
	tmprow = tmprow->previous;
	y -= tmprow->height;
     }
   while (tmprow->previous && tmprow->previous->par == first_phys_par)  {
      tmprow = tmprow->previous;
      y -= tmprow->height;
   }
   
   /* we can set the refreshing parameters now */
   status = LYX_NEED_MORE_REFRESH;
   refresh_y = y;
   refresh_row = tmprow->previous;	       /* the real refresh row will
						* be deleted, so I store
						* the previous here */ 
   
   /* remove it */
   if (tmprow->next)
     tmppar = tmprow->next->par;
   else
     tmppar = NULL;
   while (tmppar != endpar) {
      RemoveRow(tmprow->next);
      if (tmprow->next)
      	tmppar = tmprow->next->par;
      else
      	tmppar = NULL;
   }  
   
   /* remove the first one */
   tmprow2 = tmprow;		       /* this is because tmprow->previous
					* can be NULL */
   tmprow = tmprow->previous;
   RemoveRow(tmprow2);
   
   tmppar = first_phys_par;
   
   do {
      if (tmppar) {
	 InsertParagraph(tmppar, tmprow);
	 if (!tmprow)
	   tmprow = firstrow;
	 while (tmprow->next && tmprow->next->par == tmppar)
	   tmprow = tmprow->next;
	 tmppar = tmppar->Next();
	 
      }
   }
   while (tmppar != endpar);
   
   /* this is because of layout changes */ 
   if (refresh_row) {
     refresh_y -= refresh_row->height;
     SetHeightOfRow(refresh_row);   
   }
   else 
     refresh_row = firstrow;
   
   if (tmprow && tmprow->next)
     SetHeightOfRow(tmprow->next);
   
   		       /* restore the correct refresh row  */
/*    if (refresh_row)
     refresh_row = refresh_row->next;
   else
     refresh_row = firstrow;*/ 
   
}



int LyXText::FullRebreak() {
   if (need_break_row) {
      BreakAgain(need_break_row);
      need_break_row = NULL;
      return 1;
   }
   return 0;
}


/* important for the screen */


/* the cursor set functions have a special mechanism. When they
* realize, that you left an empty paragraph, they will delete it.
* They also delet the corresponding row */


   
/* need the selection cursor: */ 
void LyXText::SetSelection() {
   
   if (!selection) {
      last_sel_cursor = sel_cursor;
      sel_start_cursor = sel_cursor;
      sel_end_cursor = sel_cursor;
   }
   
   
   
   selection = True;
   
   
   /* first the toggling area */ 
   if (cursor.y < last_sel_cursor.y ||
       (cursor.y == last_sel_cursor.y && cursor.x < last_sel_cursor.x)) {
	  toggle_end_cursor = last_sel_cursor;
	  toggle_cursor = cursor;
       }
   else {
      toggle_end_cursor = cursor;
      toggle_cursor = last_sel_cursor;
   }
   
   last_sel_cursor = cursor;
   
   
   /* and now the whole selection */ 
   
   
   if (sel_cursor.y < cursor.y ||
       (sel_cursor.y == cursor.y && sel_cursor.x < cursor.x)) {
	  sel_end_cursor = cursor;
	  sel_start_cursor = sel_cursor;
       }
   else {
      sel_end_cursor = sel_cursor; 
      sel_start_cursor = cursor;
   }
   
   
   
   
   
   /* a selection with no contents is not a selection */ 
   if (sel_start_cursor.x == sel_end_cursor.x && 
       sel_start_cursor.y == sel_end_cursor.y)
     selection = False;
   
}


void LyXText::ClearSelection() {
   selection = False;
   mark_set = False;
}
   
   


void  LyXText::CursorUp(){
   SetCursorFromCoordinates(cursor.x_fix, 
			    cursor.y - cursor.row->baseline - 1);
};
void  LyXText::CursorDown(){
   SetCursorFromCoordinates(cursor.x_fix, 
			    cursor.y - cursor.row->baseline
			    + cursor.row->height + 1);
};

void  LyXText::CursorLeft(){
   if (cursor.pos > 0) {
      SetCursor(cursor.par, cursor.pos - 1);
   }
   else if (cursor.par->Previous()) {
      SetCursor(cursor.par->Previous(), cursor.par->Previous()->Last());
   }
};

void  LyXText::CursorRight(){
   if (cursor.pos < cursor.par->Last()) {
      SetCursor(cursor.par, cursor.pos + 1);
   }
   else if (cursor.par->Next()) {
      SetCursor(cursor.par->Next(), 0);
   }
};
   
   

   
void  LyXText::CursorHome(){
  SetCursor(cursor.par, cursor.row->pos);
}

void  LyXText::CursorEnd(){
   if (!cursor.row->next || cursor.row->next->par != cursor.row->par)
     SetCursor(cursor.par, RowLast(cursor.row) + 1);
   else {
      if (cursor.par->Last() && 
	  (cursor.par->GetChar(RowLast(cursor.row)) == ' '
	   || IsNewline(cursor.par, RowLast(cursor.row))))
      	SetCursor(cursor.par, RowLast(cursor.row));
      else
      	SetCursor(cursor.par, RowLast(cursor.row) + 1);
   }
   
}

void  LyXText::CursorTop(){
   while (cursor.par->Previous())
     cursor.par = cursor.par->Previous();
   SetCursor(cursor.par, 0);
}

void  LyXText::CursorBottom(){
   while (cursor.par->Next())
     cursor.par = cursor.par->Next();
   SetCursor(cursor.par, cursor.par->Last());
}
   
   






/* returns a pointer to the row near the specified y-coordinate
* (relative to the whole text). y is set to the real beginning
* of this row */ 
Row* LyXText::GetRowNearY(long& y) {
   
   Row* tmprow;
   long tmpy;
   
   if (currentrow){
     tmprow = currentrow;
     tmpy = currentrow_y;
   }
   else {
     tmprow = firstrow;
     tmpy = 0;
   }

   if (tmpy<=y)
     while (tmprow->next && tmpy + tmprow->height <= y) {
       tmpy += tmprow->height;
       tmprow = tmprow->next;
     }
   else
     while (tmprow->previous && tmpy > y) {
       tmprow = tmprow->previous;
       tmpy -= tmprow->height;
     }


   
   currentrow = tmprow;
   currentrow_y = tmpy;

   
   y = tmpy;			       /* return the real y  */
   return tmprow;

}
   
   
   


void LyXText::ToggleEmphasize() {
   LyXFont italic_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_ITALIC_SHAPE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE
   };
   LyXFont non_italic_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_UP_SHAPE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE
   };
   
   int all_italic; 		       /* boolean flag  */
       
      /* if the whole selection is italic, we will do non-italic,
       * otherwise we do italic. The paragraph himself will 
       * change it to layout-default if possible , however it 
       * looks like */
   

   /* if there is no selection just set the current font  */ 
   if (!selection) {
      current_font.shape =real_current_font.shape;
      if (current_font.shape == LYX_ITALIC_SHAPE) {
	 current_font.shape = LYX_UP_SHAPE;
	 SetMinibuffer("Emphasize style removed.","","");
	 fl_set_menu_item_mode(fd_form_main->menu_layout,5, FL_PUP_BOX);
      }
      else {
	 current_font.shape = LYX_ITALIC_SHAPE;
	 SetMinibuffer("Emphasize style set.","","");
	 fl_set_menu_item_mode(fd_form_main->menu_layout,5, FL_PUP_CHECK);
      }
      real_current_font.shape = current_font.shape;
      return;
   }
   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   LyXCursor tmpcursor = sel_start_cursor;
   all_italic = True;
   while (all_italic && 
	  (tmpcursor.par != sel_end_cursor.par ||
	   tmpcursor.pos < sel_end_cursor.pos)) {
	      if (tmpcursor.pos < tmpcursor.par->Last()) {
		 if (tmpcursor.par->footnoteflag 
		     == sel_start_cursor.par->footnoteflag
		     && GetFont(tmpcursor.par, tmpcursor.pos).shape 
		     != LYX_ITALIC_SHAPE)
		   all_italic = False;
		 tmpcursor.pos++;
	      }
	      else {
		 tmpcursor.pos = 0;
		 tmpcursor.par = tmpcursor.par->Next();
	      }
	   }
   if (all_italic){
     SetFont(non_italic_font);
     SetMinibuffer("Emphasize style removed.","","");
   }
   else{
     SetFont(italic_font);
     SetMinibuffer("Emphasize style set.","","");
   }
         
}


void LyXText::ToggleNoun() {
   LyXFont noun_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_SMALLCAPS_SHAPE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE
   };
   LyXFont non_noun_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_UP_SHAPE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE
   };
   
   int all_noun; 		       /* boolean flag  */
       
      /* if the whole selection is italic, we will do non-italic,
       * otherwise we do italic. The paragraph himself will 
       * change it to layout-default if possible , however it 
       * looks like */
   

   /* if there is no selection just set the current font  */ 
   if (!selection) {
      current_font.shape =real_current_font.shape;
      if (current_font.shape == LYX_SMALLCAPS_SHAPE){
      	current_font.shape = LYX_UP_SHAPE;
	SetMinibuffer("Noun style removed.","","");
      }
      else {
      	current_font.shape = LYX_SMALLCAPS_SHAPE;
	SetMinibuffer("Noun style set.","","");
      }
      real_current_font.shape = current_font.shape;
      return;
   }

   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   LyXCursor tmpcursor = sel_start_cursor;
   all_noun = True;
   while (all_noun && 
	  (tmpcursor.par != sel_end_cursor.par ||
	   tmpcursor.pos < sel_end_cursor.pos)) {
	      if (tmpcursor.pos < tmpcursor.par->Last()) {
		 if (tmpcursor.par->footnoteflag 
		     == sel_start_cursor.par->footnoteflag
		     && GetFont(tmpcursor.par, tmpcursor.pos).shape 
		     != LYX_SMALLCAPS_SHAPE)
		   all_noun = False;
		 tmpcursor.pos++;
	      }
	      else {
		 tmpcursor.pos = 0;
		 tmpcursor.par = tmpcursor.par->Next();
	      }
	   }

   if (all_noun){
     SetFont(non_noun_font);
     SetMinibuffer("Noun style removed.","","");
   }
   else{
     SetFont(noun_font);
     SetMinibuffer("Noun style set.","","");
   }
         
         
}

void LyXText::ToggleLaTeX() {
   LyXFont latex_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_LATEX,
      LYX_NO_FONT_CHANGE
   };
   LyXFont non_latex_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_LATEX,
      LYX_NO_FONT_CHANGE
   };
   
   int all_latex; 		       /* boolean flag  */
       
      /* if the whole selection is italic, we will do non-italic,
       * otherwise we do italic. The paragraph himself will 
       * change it to layout-default if possible , however it 
       * looks like */
   

   /* if there is no selection just set the current font  */ 
   if (!selection) {
      current_font.latex =real_current_font.latex;
      if (current_font.latex == LYX_LATEX) {
      	current_font.latex = LYX_NO_LATEX;
	SetMinibuffer("TeX style removed.","","");}
      else{
      	current_font.latex = LYX_LATEX;
	SetMinibuffer(" TeX style set.","","");}
      real_current_font.latex = current_font.latex;
      return;
   }

   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   LyXCursor tmpcursor = sel_start_cursor;
   all_latex = True;
   while (all_latex && 
	  (tmpcursor.par != sel_end_cursor.par ||
	   tmpcursor.pos < sel_end_cursor.pos)) {
	      if (tmpcursor.pos < tmpcursor.par->Last()) {
		 if (tmpcursor.par->footnoteflag 
		     == sel_start_cursor.par->footnoteflag
		     && GetFont(tmpcursor.par, tmpcursor.pos).latex
		     == LYX_NO_LATEX)
		   all_latex = False;
		 tmpcursor.pos++;
	      }
	      else {
		 tmpcursor.pos = 0;
		 tmpcursor.par = tmpcursor.par->Next();
	      }
	   }
   if (all_latex){
     SetFont(non_latex_font);
     SetMinibuffer("LaTeX style removed.","","");
   }
   else{
     SetFont(latex_font);
     SetMinibuffer("LaTeX style set.","","");
   }
         
}

void LyXText::ToggleFormLaTeX() {
   LyXFont latex_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_FORMULA_LATEX,
      LYX_NO_FONT_CHANGE
   };
   LyXFont non_latex_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_LATEX,
      LYX_NO_FONT_CHANGE
   };
   
   int all_latex; 		       /* boolean flag  */
       
      /* if the whole selection is italic, we will do non-italic,
       * otherwise we do italic. The paragraph himself will 
       * change it to layout-default if possible , however it 
       * looks like */
   

   /* if there is no selection just set the current font  */ 
   if (!selection) {
      current_font.latex =real_current_font.latex;
      if (current_font.latex == LYX_FORMULA_LATEX){
      	current_font.latex = LYX_NO_LATEX;
	SetMinibuffer("Math style removed.","","");}
      else{
      	current_font.latex = LYX_FORMULA_LATEX;
	SetMinibuffer("Math style set.","","");}
      real_current_font.latex = current_font.latex;
      return;
   }

   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   LyXCursor tmpcursor = sel_start_cursor;
   all_latex = True;
   while (all_latex && 
	  (tmpcursor.par != sel_end_cursor.par ||
	   tmpcursor.pos < sel_end_cursor.pos)) {
	      if (tmpcursor.pos < tmpcursor.par->Last()) {
		 if (tmpcursor.par->footnoteflag 
		     == sel_start_cursor.par->footnoteflag
		     && GetFont(tmpcursor.par, tmpcursor.pos).latex
		     == LYX_NO_LATEX)
		   all_latex = False;
		 tmpcursor.pos++;
	      }
	      else {
		 tmpcursor.pos = 0;
		 tmpcursor.par = tmpcursor.par->Next();
	      }
	   }
   if (all_latex){
     SetFont(non_latex_font);
     SetMinibuffer("Math style removed.","","");
   }
   else{
     SetFont(latex_font);
     SetMinibuffer("Math style set.","","");
   }

}

void LyXText::ToggleFree(LyXFont font) {
   LyXFont non_free_font = { 
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE,
      LYX_NO_FONT_CHANGE
   };


   if (font.family == LYX_NO_FONT_CHANGE &&
       font.series == LYX_NO_FONT_CHANGE &&
       font.shape == LYX_NO_FONT_CHANGE &&
       font.size == LYX_NO_FONT_CHANGE &&
       font.latex == LYX_NO_FONT_CHANGE &&
       font.bar == LYX_NO_FONT_CHANGE){
     SetMinibuffer("No user style defined. Use menu entry layout->character to define.","","");
     return;
   }




   if (font.family != LYX_NO_FONT_CHANGE)
     non_free_font.family = LYX_LAYOUT_DEFAULT;
   if (font.series != LYX_NO_FONT_CHANGE)
     non_free_font.series = LYX_LAYOUT_DEFAULT;
   if (font.shape != LYX_NO_FONT_CHANGE)
     non_free_font.shape = LYX_LAYOUT_DEFAULT;
   if (font.size != LYX_NO_FONT_CHANGE)
     non_free_font.size = LYX_LAYOUT_DEFAULT;
   if (font.latex != LYX_NO_FONT_CHANGE)
     non_free_font.latex = LYX_LAYOUT_DEFAULT;
   if (font.bar != LYX_NO_FONT_CHANGE)
     non_free_font.bar = LYX_LAYOUT_DEFAULT;

   
   int all_free; 		       /* boolean flag  */
       
      /* if the whole selection is italic, we will do non-italic,
       * otherwise we do italic. The paragraph himself will 
       * change it to layout-default if possible , however it 
       * looks like */
   

   /* if there is no selection just set the current font  */
   
   if (!selection) {
      if ((font.family != LYX_NO_FONT_CHANGE && real_current_font.family != font.family)
	  ||
	  (font.series != LYX_NO_FONT_CHANGE && real_current_font.series != font.series)
	  ||
	  (font.shape != LYX_NO_FONT_CHANGE && real_current_font.shape != font.shape)
	  ||
	  (font.size != LYX_NO_FONT_CHANGE && real_current_font.size != font.size)
	  ||
	  (font.bar != LYX_NO_FONT_CHANGE && real_current_font.bar != font.bar)
	  )
	{
	   font.latex = current_font.latex;
	   SetFont(font);
	   SetMinibuffer("User style set.","","");
	}
      else{
	 non_free_font.latex = current_font.latex;
	 SetFont(non_free_font);
	 SetMinibuffer("User style removed.","","");
      }
      return;
   }
   
   
      
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */ 
   LyXCursor tmpcursor = sel_start_cursor;
   all_free = True;
   while (all_free && 
	  (tmpcursor.par != sel_end_cursor.par ||
	   tmpcursor.pos < sel_end_cursor.pos)) {
	      if (tmpcursor.pos < tmpcursor.par->Last()) {
		 if (tmpcursor.par->footnoteflag 
		     == sel_start_cursor.par->footnoteflag
		     &&
		     (
		      (font.family != LYX_NO_FONT_CHANGE && GetFont(tmpcursor.par, tmpcursor.pos).family != font.family) ||
		      (font.series != LYX_NO_FONT_CHANGE && GetFont(tmpcursor.par, tmpcursor.pos).series != font.series) ||
		      (font.shape != LYX_NO_FONT_CHANGE && GetFont(tmpcursor.par, tmpcursor.pos).shape != font.shape) ||
		      (font.size != LYX_NO_FONT_CHANGE && GetFont(tmpcursor.par, tmpcursor.pos).size != font.size) ||
		      (font.latex != LYX_NO_FONT_CHANGE && GetFont(tmpcursor.par, tmpcursor.pos).latex != font.latex) ||
		      (font.bar != LYX_NO_FONT_CHANGE && GetFont(tmpcursor.par, tmpcursor.pos).bar != font.bar)
		      )
		     )
		   all_free = False;
		 tmpcursor.pos++;
	      }
	      else {
		 tmpcursor.pos = 0;
		 tmpcursor.par = tmpcursor.par->Next();
	      }
	   }
   if (all_free){
     SetFont(non_free_font);
     SetMinibuffer("User style removed.","","");
   }
   else{
     SetFont(font);
     SetMinibuffer("User style set.","","");
   }

}


int LyXText::BeginningOfMainBody(LyXParagraph *par) {
   if (style(parameters->textclass, par->GetLayout())->labeltype != LABEL_MANUAL)
     return 0;
   else
     return par->BeginningOfMainBody();
}



/* if there is a selection, reset every environement you can find
* in the selection, otherwise just the environement you are in */ 
void LyXText::MeltFootnoteEnvironement() {
   LyXParagraph *tmppar, *firsttmppar;
   
   ClearSelection();
   
   /* is is only allowed, if the cursor is IN an open footnote.
    * Otherwise it is too dangerous */ 
   if (cursor.par->footnoteflag != LYX_OPEN_FOOTNOTE)
     return;
   
   /* ok, move to the beginning of the footnote. */ 
   while (cursor.par->footnoteflag == LYX_OPEN_FOOTNOTE)
     cursor.par = cursor.par->Previous();
   
   SetCursor(cursor.par, cursor.par->Last());
   /* this is just faster than using CursorLeft(); */ 
   
   
   firsttmppar = cursor.par->ParFromPos(cursor.pos);
   tmppar = firsttmppar;
   /* tmppar is now the paragraph right before the footnote */
   

   char first_footnote_par_is_not_empty = tmppar->next->last;
   
   while (tmppar->next && tmppar->next->footnoteflag == LYX_OPEN_FOOTNOTE) {
      tmppar = tmppar->next;	       /* I use next instead of Next(),
					* because there cannot be any
					* footnotes in a footnote
					* environement */
      tmppar->footnoteflag = LYX_NO_FOOTNOTE;
      
      /* remember the captions and empty paragraphs */
      if (style(parameters->textclass, tmppar->GetLayout())->labeltype == LABEL_SENSITIVE
	  || !tmppar->Last())
	tmppar->SetLayout(0);
      
   }
   
   
   /* now we will paste the ex-footnote, if the layouts allow it */
   /* first restore the layout of the paragraph right behind the footnote*/ 
   tmppar->next->SetOnlyLayout(cursor.par->GetLayout());
   
   
   /* first the end */ 
   if (!tmppar->GetLayout() 
       || (tmppar->Next() && (tmppar->Next()->Last()
			      || tmppar->Next()->HasSameLayout(tmppar)))) {
      if (tmppar->Next()->Last() && IsLineSeparator(tmppar->Next(), 0))
      	 tmppar->Next()->Erase(0);
      tmppar->PasteParagraph();
   }

   

   tmppar = tmppar->Next();	       /* make shure tmppar cannot be touched
					* by the pasting of the beginning */

   
   /* then the beginning */ 
   /* if there is no space between the text and the footnote, so we insert
    * a blank 
    * (only if the previous par and the footnotepar are not empty!) */
   if (!firsttmppar->next->GetLayout()
       || firsttmppar->HasSameLayout(firsttmppar->next)) {
      if (firsttmppar->last
	  && !IsSeparator(firsttmppar, firsttmppar->last - 1)
	  && first_footnote_par_is_not_empty) {
	 firsttmppar->next->InsertChar(0, ' ');
      }
      firsttmppar->PasteParagraph();
   }
   
   /* now redo the paragaphs */
   RedoParagraphs(cursor, tmppar);
   
   SetCursor(cursor.par, cursor.pos);
   
   /* sometimes it can happen, that there is a counter change */ 
   Row *row = cursor.row;
   while (row->next && row->par != tmppar && row->next->par != tmppar)
     row = row->next;
   UpdateCounters(row);
   
   
   ClearSelection();
   
}


/* the DTP switches for paragraphs. LyX will store them in the 
* first physicla paragraph. When a paragraph is broken, the top settings 
* rest, the bottom settings are given to the new one. So I can make shure, 
* they do not duplicate themself and you cannnot make dirty things with 
* them!  */ 

void LyXText::SetParagraph(char line_top, char line_bottom,
			   char pagebreak_top, char pagebreak_bottom,
			   float space_top, float space_bottom,
			   char fill_top, char fill_bottom,
			   char align, 
			   const char* labelwidthstring,
			   char noindent) 
     {
	LyXCursor tmpcursor;
	tmpcursor = cursor;
	LyXParagraph *tmppar;
	if (!selection) {
	   sel_start_cursor = cursor;
	   sel_end_cursor = cursor;
	}
	
	tmppar = sel_end_cursor.par;
	while (tmppar != sel_start_cursor.par->FirstPhysicalPar()->Previous()) {
	   SetCursor(tmppar->FirstPhysicalPar(), 0);
	   status = LYX_NEED_MORE_REFRESH;
	   refresh_row = cursor.row;
	   refresh_y = cursor.y - cursor.row->baseline;
	   if (cursor.par->footnoteflag ==
	       sel_start_cursor.par->footnoteflag) {
	      cursor.par->line_top = line_top;
	      cursor.par->line_bottom = line_bottom;
	      cursor.par->pagebreak_top = pagebreak_top;
	      cursor.par->pagebreak_bottom = pagebreak_bottom;
	      cursor.par->added_space_top = space_top;
	      cursor.par->added_space_bottom = space_bottom;
	      cursor.par->fill_top = fill_top;
	      cursor.par->fill_bottom = fill_bottom;
	      /* does the layout allow the new alignment? */ 
	      if (align == LYX_ALIGN_LAYOUT)
	      	 align = style(parameters->textclass, cursor.par->GetLayout())->align;
	      if (align & style(parameters->textclass, cursor.par->GetLayout())->alignpossible) {
		 if (align == style(parameters->textclass, cursor.par->GetLayout())->align)
		    cursor.par->align = LYX_ALIGN_LAYOUT;
		 else
		    cursor.par->align = align;
	      }
	      cursor.par->SetLabelWidthString((char *)labelwidthstring);
	      cursor.par->noindent = noindent;
	   }
	   
	   /* 
	    tmprow = cursor.row;
	    while (tmprow->next && tmprow->next->par->previous != cursor.par->LastPhysicalPar())
	       tmprow = tmprow->next;
	    SetHeightOfRow(tmprow);
	    SetHeightOfRow(cursor.row); */ 
	   tmppar = cursor.par->FirstPhysicalPar()->Previous();
	}
	
	LyXParagraph *endpar = sel_end_cursor.par->LastPhysicalPar()->Next();
	if (endpar)
	  endpar=endpar->Next();	       /* this is because of parindents  */
	RedoParagraphs(sel_start_cursor, endpar);
	
	
	
	ClearSelection();
	SetCursor(sel_start_cursor.par, sel_start_cursor.pos);
	sel_cursor = cursor;
	SetCursor(sel_end_cursor.par, sel_end_cursor.pos);
	SetSelection();
	SetCursor(tmpcursor.par, tmpcursor.pos);
	
     };


/* set the counter of a paragraph. This includes the labels */ 
void LyXText::SetCounter(LyXParagraph *par) {
   int i;
   char *s;
   
   /* this is only relevant for the beginning of paragraph */ 
   par = par->FirstPhysicalPar();

   
   /* copy the prev-counters to this one */
   if (par->Previous()) {
      for (i=0; i<10; i++) {
	 par->counter[i]=par->Previous()->GetCounter(i);
      }
      par->enumdepth = par->Previous()->FirstPhysicalPar()->enumdepth;
      par->itemdepth = par->Previous()->FirstPhysicalPar()->itemdepth;
   }
   else {
      for (i=0; i<10; i++) {
	 par->counter[i]=0;
      }  
      par->enumdepth = 0;
      par->itemdepth = 0;
   }
   
   LyXLayout* layout = style(parameters->textclass, par->GetLayout());
   
   LyXTextClass *textclass = TextClass(parameters->textclass);
   
   /* maybe we have to increment the enumeration depth */ 
   if (par->Previous() &&
       par->Previous()->GetDepth() < par->GetDepth() &&
       style(parameters->textclass, par->Previous()->GetLayout())->labeltype == LABEL_COUNTER_ENUMI
       && par->enumdepth < 3)
     par->enumdepth++;

   /* maybe we have to decrement the enumeration depth */ 
   if (par->Previous() &&
       par->Previous()->GetDepth() > par->GetDepth()) {
	  par->enumdepth = par->DepthHook(par->GetDepth())->enumdepth;
	  par->counter[6 + par->enumdepth] = par->DepthHook(par->GetDepth())->counter[6 + par->enumdepth];
	  /* reset the counters. A depth change is like a breaking layout */ 
	  for (i=6 + par->enumdepth + 1; i<10;i++)
	    par->counter[i]=0;
	  
       }
   
   
   
   if (par->labelstring) {
      delete(par->labelstring);       /* free memory  */
      par->labelstring = NULL;
   }
   
   
   if (layout->bordertype == BORDER_MANUAL) {
      if (!par->labelwidthstring) {
	 par->SetLabelWidthString(layout->labelstring);
      }
   }
   else {
      par->SetLabelWidthString(NULL);
   }
   
   
   
   
   
   /* is it a layout that has an automatic label ? */ 
   if (layout->labeltype >=  LABEL_FIRST_COUNTER) {
      
      i = layout->labeltype - LABEL_FIRST_COUNTER;
      if (i>=0 && i<=parameters->secnumdepth) {
	 par->counter[i]++;	       /* increment the counter  */
	 
	 
	 
	 s = new char[50];
	 
	 
	 if (LABEL_FIRST_COUNTER + i == LABEL_COUNTER_CHAPTER) {
	    sprintf(s, "Chapter %d", par->counter[i]);
	 }
	 else {
	    switch (2 * LABEL_FIRST_COUNTER - textclass->maxcounter + i) {
	     case LABEL_COUNTER_CHAPTER:
	       sprintf(s, "%d", par->counter[i]);
	       break;
	     case LABEL_COUNTER_SECTION:
	       sprintf(s, "%d.%d", par->counter[i - 1], par->counter[i]);
	       break;
	     case LABEL_COUNTER_SUBSECTION:
	       sprintf(s, "%d.%d.%d", par->counter[i-2], par->counter[i-1],par->counter[i]);
	       break;
	     case LABEL_COUNTER_SUBSUBSECTION:
	       sprintf(s, "%d.%d.%d.%d", par->counter[i-3], par->counter[i-2],par->counter[i-1],par->counter[i]);
	       break;
	     case LABEL_COUNTER_PARAGRAPH:
	       sprintf(s, "%d.%d.%d.%d.%d", par->counter[i-4], par->counter[i-3], par->counter[i-2],par->counter[i-1],par->counter[i]);
	       break;
	     case LABEL_COUNTER_SUBPARAGRAPH:
	       sprintf(s, "%d.%d.%d.%d.%d.%d", par->counter[i-5], par->counter[i-4], par->counter[i-3], par->counter[i-2],par->counter[i-1],par->counter[i]);
	       break;
	     default:
	       sprintf(s, "%d.", par->counter[i]);
	       break;
	    }
	 }
	 
	 par->labelstring = StringCopy(s);
	 delete(s);
	 
	 for (i++;i<10;i++)
	   par->counter[i]=0;      /* reset the following counters  */
      }
      else if (layout->labeltype == LABEL_COUNTER_ENUMI) {
	 par->counter[i + par->enumdepth]++;
	 s = new char[25];
	 int number = par->counter[i + par->enumdepth];
	 switch (par->enumdepth) {
	  case 1:
	    sprintf(s, "(%c)", number + 'a' - 1);
	    break;
	  case 2:
	    switch (number) {
	     case 1: sprintf(s, "i."); break;
	     case 2: sprintf(s, "ii."); break;
	     case 3: sprintf(s, "iii."); break;
	     case 4: sprintf(s, "iv."); break;
	     case 5: sprintf(s, "v."); break;
	     case 6: sprintf(s, "vi."); break;
	     case 7: sprintf(s, "vii."); break;
	     case 8: sprintf(s, "viii."); break;
	     case 9: sprintf(s, "ix."); break;
	     case 10: sprintf(s, "x."); break;
	     case 11: sprintf(s, "xi."); break;
	     case 12: sprintf(s, "xii."); break;
	     case 13: sprintf(s, "xiii."); break;
	     default:
	       sprintf(s, "\\roman{%d}.", number);
	       break;
	    }
	    break;
	  case 3:
	    sprintf(s, "%c.", number + 'A' - 1);
	    break;
	  default:
	    sprintf(s, "%d.", number);
	    break;
	 }
	 par->labelstring = StringCopy(s);
	 delete(s);
	 
	 for (i += par->enumdepth + 1;i<10;i++)
	    par->counter[i]=0;      /* reset the following counters  */
	 
      }
   }
   else  {
      s = layout->labelstring;
      
      /* the caption hack: */
      
      if (layout->labeltype == LABEL_SENSITIVE) {
	 if (par->footnoteflag != LYX_NO_FOOTNOTE
	     && par->footnotekind == LYX_FIG)
	   s = "Figure:";
	 else if (par->footnoteflag != LYX_NO_FOOTNOTE
		  && par->footnotekind == LYX_TAB)
	   s = "Table:";
	 else {
	   /* par->SetLayout(0); 
	   s = layout->labelstring;  */
	   s = "Senseless: "; 
	   
	 }
      }
      par->labelstring = StringCopy(s);
      
      /* reset the enumeration counter. They are always resetted
       * when there is any other layout between */ 
      for (i=6 + par->enumdepth; i<10;i++)
      	par->counter[i]=0;
   }
}



/* Updates all counters BEHIND the row. Changed paragraphs
* with a dynamic left border will be rebroken. */ 
void LyXText::UpdateCounters(Row *row) {
   LyXParagraph *par;
   if (!row) {
      row = firstrow;
      par = row->par;
   }
   else
     par = row->par->LastPhysicalPar()->Next();
   while (par) {
     while (row->par != par)
      	row = row->next;
      
      SetCounter(par);
      
      
      /* now  check for the headline layouts. remember that they
       * have a dynamic left border */ 
      if (par->layout != LYX_DUMMY_LAYOUT
	  && ( style(parameters->textclass, par->layout)->bordertype == BORDER_DYNAMIC
	      || style(parameters->textclass, par->layout)->labeltype == LABEL_SENSITIVE)
	  ){
	 
       /* Rebreak the paragraph */ 
       RemoveParagraph(row);
       AppendParagraph(row);
       
       /* think about the damned open footnotes! */ 
       while (par->Next() &&
	      (par->Next()->footnoteflag == LYX_OPEN_FOOTNOTE
	       || par->Next()->layout == LYX_DUMMY_LAYOUT)){
	 par = par->Next();
	 if (par->layout == LYX_DUMMY_LAYOUT) {
	     while (row->par != par)
	       row = row->next;
	     RemoveParagraph(row);
	     AppendParagraph(row);
	   }
       }
     }
     
     par = par->LastPhysicalPar()->Next();
     
   }
   
 }


/* insets an inset. */ 
void LyXText::InsertInset(Inset *inset){
   cursor.par->InsertChar(cursor.pos, LYX_META_INSET);
   cursor.par->InsertInset(cursor.pos, inset);
   InsertChar(LYX_META_INSET);	       /* just to rebreak and refresh correctly.
					* The character will not be inserted a
					* second time */
};



/* this is for the simple cut and paste mechanism */ 
static LyXParagraph *simple_cut_buffer = NULL;
static char simple_cut_buffer_textclass = 0;

static void DeleteSimpleCutBuffer() {
  if (!simple_cut_buffer)
    return;
  LyXParagraph *tmppar;

  while (simple_cut_buffer) {
    tmppar =  simple_cut_buffer;
    simple_cut_buffer = simple_cut_buffer->next;
    delete(tmppar);
  }
  simple_cut_buffer = NULL;

}


void LyXText::CutSelection(){
  /* this doesnt make sense, if there is no selection */ 
   if (!selection) {
      return;
   }
   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */
   LyXParagraph *tmppar;
   LyXParagraph *endpar;
   int i;
   
   /* check wether there are half footnotes in the selection */
   if (sel_start_cursor.par->footnoteflag != LYX_NO_FOOTNOTE
       || sel_end_cursor.par->footnoteflag != LYX_NO_FOOTNOTE){
     tmppar = sel_start_cursor.par;
     while (tmppar != sel_end_cursor.par){
       if (tmppar->footnoteflag != sel_start_cursor.par->footnoteflag){
	 WriteAlert("Impossible operation", "Don't know what to do with these half floats.", "sorry.");
	 return;
       }
       tmppar = tmppar->Next();
     }
   }
   
   /* set the textclass */
   simple_cut_buffer_textclass = parameters->textclass;

   char space_wrapped = IsLineSeparator(sel_end_cursor.par,sel_end_cursor.pos);
   if (sel_end_cursor.pos > 0 
       && IsLineSeparator(sel_end_cursor.par,sel_end_cursor.pos - 1)) {
	  sel_end_cursor.pos--;	       /* please break before a space at
					* the end */
	  space_wrapped = True;
       }
   
   
   /* there are two cases: cut only within one paragraph or more than one paragraph */
   
   if (sel_start_cursor.par->ParFromPos(sel_start_cursor.pos) 
       == sel_end_cursor.par->ParFromPos(sel_end_cursor.pos)) {
     /* only within one paragraph */
     simple_cut_buffer = new LyXParagraph();
     for (i=sel_start_cursor.pos; i< sel_end_cursor.pos; i++){
       sel_start_cursor.par->CutIntoMinibuffer(sel_start_cursor.pos);
       simple_cut_buffer->InsertFromMinibuffer(i - sel_start_cursor.pos);
     }
     if (space_wrapped)
       simple_cut_buffer->InsertChar(i - sel_start_cursor.pos, ' ');
     endpar = sel_end_cursor.par->Next();
   }
   else {
     /* cut more than one paragraph */ 
   
     sel_end_cursor.par->BreakParagraphConservative(sel_end_cursor.pos);
     /* insert a space at the end if there was one */
     if (space_wrapped)
       sel_end_cursor.par->InsertChar(sel_end_cursor.par->Last(), ' ');
   
     sel_end_cursor.par = sel_end_cursor.par->Next();
     sel_end_cursor.pos = 0;
   
   
     cursor = sel_end_cursor;
   
   
     /* please break behind a space, if there is one. The space should
    * be copied too */ 
     if (IsLineSeparator(sel_start_cursor.par,sel_start_cursor.pos))
       sel_start_cursor.pos++;
   
     sel_start_cursor.par->BreakParagraphConservative(sel_start_cursor.pos);
     ClearParagraph(sel_start_cursor.par->Next());
     if (!sel_start_cursor.pos
	 || IsLineSeparator(sel_start_cursor.par,sel_start_cursor.pos - 1)
	 || IsNewline(sel_start_cursor.par,sel_start_cursor.pos - 1)) {
       sel_start_cursor.par->Next()->InsertChar(0, ' ');
     }
   


     /* store the endparagraph for redoing later */
     endpar = sel_end_cursor.par->Next();   /* needed because
							   * the sel_end_
							   * cursor.par
							   * will be pasted!*/
   
   
   
   
   /* delete the simple_cut_buffer */ 
     DeleteSimpleCutBuffer();
   
   
     /*store the selection */ 
     simple_cut_buffer = sel_start_cursor.par->ParFromPos(sel_start_cursor.pos)->next;
     simple_cut_buffer->previous = NULL;
     sel_end_cursor.par->previous->next = NULL;
     /* the cut selection should begin with standard layout */
     simple_cut_buffer->Clear(); 
   

   
   
     /* cut the selection */ 
     sel_start_cursor.par->ParFromPos(sel_start_cursor.pos)->next 
       = sel_end_cursor.par;
   
     sel_end_cursor.par->previous 
       = sel_start_cursor.par->ParFromPos(sel_start_cursor.pos);

   /* paste the paragraphs again, if possible  */
     if (sel_start_cursor.par->FirstPhysicalPar()->HasSameLayout(sel_start_cursor.par->Next())
	 || 
	 (!sel_start_cursor.par->FirstPhysicalPar()->Last()
	  && ! sel_start_cursor.par->Next()->Last()))
       sel_start_cursor.par->ParFromPos(sel_start_cursor.pos)->PasteParagraph();
     else
       ClearParagraph(sel_start_cursor.par->Next());
   
   /* maybe a forgotten blank */
     if (sel_start_cursor.pos 
	 && IsLineSeparator(sel_start_cursor.par, sel_start_cursor.pos)
	 && IsLineSeparator(sel_start_cursor.par, sel_start_cursor.pos - 1)) {
       sel_start_cursor.par->Erase(sel_start_cursor.pos);
     }

   }   
   

   RedoParagraphs(sel_start_cursor, endpar);
   
   ClearSelection();
   /* just to avoid the delete empty paragraph mechanism */
   cursor = sel_start_cursor;
   selection = True;
   SetCursor(cursor.par, cursor.pos);
   sel_cursor = cursor;
   selection = False;
   
}


    
void LyXText::CopySelection(){
  int i=0;
   
   /* this doesnt make sense, if there is no selection */ 
   if (!selection) {
      return;
   }
   

   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */
   LyXParagraph *tmppar;
   
/* check wether there are half footnotes in the selection */
   if (sel_start_cursor.par->footnoteflag != LYX_NO_FOOTNOTE
       || sel_end_cursor.par->footnoteflag != LYX_NO_FOOTNOTE){
     tmppar = sel_start_cursor.par;
     while (tmppar != sel_end_cursor.par){
       if (tmppar->footnoteflag != sel_start_cursor.par->footnoteflag){
	 WriteAlert("Impossible operation", "Don't know what to do with these half floats.", "sorry.");
	 return;
       }
       tmppar = tmppar->Next();
     }
   }
   
   /* delete the simple_cut_buffer */ 
   DeleteSimpleCutBuffer();

   /* set the textclass */
   simple_cut_buffer_textclass = parameters->textclass;

   
   /* there are two cases: copy only within one paragraph or more than one paragraph */
   if (sel_start_cursor.par->ParFromPos(sel_start_cursor.pos) 
       == sel_end_cursor.par->ParFromPos(sel_end_cursor.pos)) {
     /* only within one paragraph */
     simple_cut_buffer = new LyXParagraph();
     for (i=sel_start_cursor.pos; i< sel_end_cursor.pos; i++){
       sel_start_cursor.par->CopyIntoMinibuffer(i);
       simple_cut_buffer->InsertFromMinibuffer(i - sel_start_cursor.pos);
}
   }
   else {
     /* copy more than one paragraph */ 
     /* clone the paragraphs within the selection*/
     tmppar = sel_start_cursor.par->ParFromPos(sel_start_cursor.pos);
     simple_cut_buffer = tmppar->Clone();
     LyXParagraph *tmppar2 = simple_cut_buffer;
     
     while (tmppar != sel_end_cursor.par->ParFromPos(sel_end_cursor.pos)
	    && tmppar->next) {
       tmppar = tmppar->next;
       tmppar2->next = tmppar->Clone();
       tmppar2->next->previous = tmppar2;
       tmppar2=tmppar2->next;
     }
     tmppar2->next = NULL;
     
     /* the simple_cut_buffer paragraph is too big */
     int tmpi2;

     tmpi2 = sel_start_cursor.par->PositionInParFromPos(sel_start_cursor.pos);
     for (;tmpi2;tmpi2--)
       simple_cut_buffer->Erase(0);

     /* now tmppar 2 is too big, delete all after sel_end_cursor.pos */
     
     tmpi2 = sel_end_cursor.par->PositionInParFromPos(sel_end_cursor.pos);
     while (tmppar2->last > tmpi2) {
       tmppar2->Erase(tmppar2->last-1);
     }

   }
}
          
   
   
   

void LyXText::PasteSelection() {
   /* this does not make sense, if there is nothing to paste */ 
   if (!simple_cut_buffer)
     return;

   int i;

   LyXParagraph *tmppar;
   LyXParagraph *endpar;

   LyXCursor tmpcursor;

   /* be carefull with footnotes in footnotes */ 
   if (cursor.par->footnoteflag != LYX_NO_FOOTNOTE) {
      
      /* check wether the cut_buffer includes a footnote */
      tmppar = simple_cut_buffer;
      while (tmppar && tmppar->footnoteflag == LYX_NO_FOOTNOTE)
      	tmppar = tmppar->next;
      
      if (tmppar) {
	WriteAlert("Impossible operation", "Float would include float!", "sorry.");
	return;
      }
   }

   tmpcursor = cursor;

   /* there are two cases: cutbuffer only one paragraph or many */
   if (!simple_cut_buffer->next) {
     /* only within a paragraph */
     
       /* please break behind a space, if there is one */
     while (cursor.par->Last() > cursor.pos && IsLineSeparator(cursor.par,cursor.pos))
       cursor.pos++; 
     
     for (i=simple_cut_buffer->last-1; i>=0; i--){
       simple_cut_buffer->CopyIntoMinibuffer(i);
       cursor.par->InsertFromMinibuffer(cursor.pos);
     }
     tmpcursor.pos += simple_cut_buffer->last;
     endpar = tmpcursor.par->Next();
   }
   else {
     /* many paragraphs */

     /* make a copy of the simple cut_buffer */
     tmppar = simple_cut_buffer;
     LyXParagraph *simple_cut_clone = tmppar->Clone();
     LyXParagraph *tmppar2 = simple_cut_clone;
     
     while (tmppar->next) {
       tmppar = tmppar->next;
       tmppar2->next = tmppar->Clone();
       tmppar2->next->previous = tmppar2;
       tmppar2=tmppar2->next;
     }
     
     /* make sure there is no class difference */ 
     SwitchLayoutsBetweenClasses(simple_cut_buffer_textclass, parameters->textclass, simple_cut_clone);
     
     
     /* make the simple_cut_clone exactly the same layout than
	the cursor paragraph */
     simple_cut_clone->MakeSameLayout(cursor.par);
     
     /* find the end of the buffer */ 
     LyXParagraph *lastbuffer = simple_cut_clone;
     while (lastbuffer->Next())
     lastbuffer=lastbuffer->Next();
     
     
     /* find the physical end of the buffer */ 
     lastbuffer = simple_cut_clone;
     while (lastbuffer->Next())
       lastbuffer=lastbuffer->Next();
     
     /* please break behind a space, if there is one. The space should
      * be copied too */ 
     if (cursor.par->Last() > cursor.pos && IsLineSeparator(cursor.par,cursor.pos))
       cursor.pos++; 
     
     char paste_the_end = 0;

     /* open the paragraph for inserting the simple_cut_clone if necessary */
     if (cursor.par->Last() > cursor.pos || !cursor.par->Next()){
       cursor.par->BreakParagraphConservative(cursor.pos);
       paste_the_end = 1;
     }
     
     /* be careful with double spaces */ 
     if ((!cursor.par->Last()
	|| IsLineSeparator(cursor.par, cursor.pos - 1)
	|| IsNewline(cursor.par, cursor.pos - 1))
	 && simple_cut_clone->last
	 && IsLineSeparator(simple_cut_clone, 0))
       simple_cut_clone->Erase(0);
     

     /* set the end for redoing later */ 
     endpar = cursor.par->ParFromPos(cursor.pos)->next->Next();
     
   /* paste it! */ 
     lastbuffer->ParFromPos(lastbuffer->Last())->next = cursor.par->ParFromPos(cursor.pos)->next;
     cursor.par->ParFromPos(cursor.pos)->next->previous = lastbuffer->ParFromPos(lastbuffer->Last());
     
     cursor.par->ParFromPos(cursor.pos)->next = simple_cut_clone;
     simple_cut_clone->previous = cursor.par->ParFromPos(cursor.pos);
   

     
     if (cursor.par->ParFromPos(cursor.pos)->Next() == lastbuffer)
       lastbuffer = cursor.par;
     
     cursor.par->ParFromPos(cursor.pos)->PasteParagraph();
     
     /* store the new cursor position  */
     tmpcursor.par = lastbuffer;
     tmpcursor.pos = lastbuffer->Last();
     
     
     /* maybe some pasting */ 
     if (lastbuffer->Next() && paste_the_end) {
       if (lastbuffer->Next()->HasSameLayout(lastbuffer)) {
	 
	 /* be careful witth double spaces */ 
	 if ((!lastbuffer->Last()
	      || IsLineSeparator(lastbuffer, lastbuffer->Last() - 1)
	      || IsNewline(lastbuffer, lastbuffer->Last() - 1))
	     && lastbuffer->Next()->Last()
	     && IsLineSeparator(lastbuffer->Next(), 0))
	   lastbuffer->Next()->Erase(0);
	 

	 lastbuffer->ParFromPos(lastbuffer->Last())->PasteParagraph();
	 
       }
       else if (!lastbuffer->Next()->Last()) {
	 lastbuffer->Next()->MakeSameLayout(lastbuffer);
	 
	 /* be careful witth double spaces */ 
	 if ((!lastbuffer->Last()
	      || IsLineSeparator(lastbuffer, lastbuffer->Last() - 1)
	      || IsNewline(lastbuffer, lastbuffer->Last() - 1))
	     && lastbuffer->Next()->Last()
	     && IsLineSeparator(lastbuffer->Next(), 0))
	   lastbuffer->Next()->Erase(0);
	 
	 lastbuffer->ParFromPos(lastbuffer->Last())->PasteParagraph();
	 
       }
       else if (!lastbuffer->Last()) {
	 lastbuffer->MakeSameLayout(lastbuffer->next);
	 
	 /* be careful witth double spaces */ 
	 if ((!lastbuffer->Last()
	      || IsLineSeparator(lastbuffer, lastbuffer->Last() - 1)
	      || IsNewline(lastbuffer, lastbuffer->Last() - 1))
	     && lastbuffer->Next()->Last()
	     && IsLineSeparator(lastbuffer->Next(), 0))
	   lastbuffer->Next()->Erase(0);
	 
	 lastbuffer->ParFromPos(lastbuffer->Last())->PasteParagraph();
	 
       }
       else ClearParagraph(lastbuffer->Next());
     }

   }
   
   RedoParagraphs(cursor, endpar);
   
   SetCursor(cursor.par, cursor.pos);
   ClearSelection();
   
   sel_cursor = cursor;
   SetCursor(tmpcursor.par, tmpcursor.pos);
   SetSelection();
   UpdateCounters(cursor.row);
}
   








/* returns a pointer to the very first LyXParagraph */ 
LyXParagraph* LyXText::FirstParagraph() {
   return parameters->paragraph;
}





/* returns true if the specified string is at the specified position */
char LyXText::IsStringInText(LyXParagraph *par, int pos, char* string) {
  int i;

  if (par) {
    i = 0;
    while (pos+i < par->Last() && string[i] && string[i]==par->GetChar(pos+i)) {
      i++;
    }
    
    if (!string[i])
      return True;
  }
  return False;

}



/* sets the selection over the number of characters of string, no check!! */
void LyXText::SetSelectionOverString(char* string) {
  sel_cursor = cursor;
  int i;
  for (i=0; string[i]; i++)
    CursorRight();
  SetSelection();
}

/* simple replacing. The font of the first selected character is used */
void LyXText::ReplaceSelectionWithString(char* string){
  int i=0;
  if (!selection) { /* create a dummy selection */
    sel_end_cursor = cursor;
    sel_start_cursor = cursor;
  }
 
  /* insert the new string behind the selection */
  LyXFont font = sel_start_cursor.par->GetFontSettings(sel_start_cursor.pos);
  for (i=0; string[i];i++); /* the end of the string */
  while (i) {
    i--;
    sel_end_cursor.par->InsertChar(sel_end_cursor.pos, string[i]);
    sel_end_cursor.par->SetFont(sel_end_cursor.pos, font);
  }

  /* cut the selection*/
  CutSelection();
}

/* if the string can be found: return true and set the cursor to the new position */
char LyXText::SearchForward(char* string) {
  LyXParagraph *par = cursor.par;
  int pos = cursor.pos;

  while (par && !IsStringInText(par,pos,string)) {
    if (pos<par->Last()-1)
      pos++;
    else {
      pos = 0;
      par = par->Next();
    }
  }
  if (par) {
    SetCursor(par,pos);
    return True;
  }
  else
    return False;
}

char LyXText::SearchBackward(char* string) {
  LyXParagraph *par = cursor.par;
  int pos = cursor.pos;

  do {
    if (pos>0)
      pos--;
    else {
      par = par->Previous();
      if (par)
	pos = par->Last()-1;
    }
  } while (par && !IsStringInText(par,pos,string));
  
  if (par) {
    SetCursor(par,pos);
    return True;
  }
  else
    return False;
}




/* needed to insert the selection */
void LyXText::InsertStringA(char* string) {
  LyXParagraph *par = cursor.par;
  int pos = cursor.pos;
   int a = 0;
  LyXParagraph *endpar = cursor.par->Next();
   
   char flag = (style(parameters->textclass, cursor.par->GetLayout())->latextype >= LATEX_ENVIRONEMENT);
   /* only to be shure, should not be neccessary */ 
   ClearSelection();
   
   /* insert the string, don't insert doublespace */ 
  int i=0;

   /* the old idea that was more strict 
  while (string[i]) {
    if (string[i]!='\n') {
       if (string[i]!=' ' || pos > 0) {
	  if ((i<1 && string[i]!=' ')  
	      || (i>0 && (string[i]!=' ' || string[i-1]!=' '))) {
		 if (string[i] != 13) {
		    par->InsertChar(pos,string[i]);
		    pos++;
		 }
	      }
       }
    }
    else {
      if (i>0 && pos>0 && string[i-1]!='\n') {
	 par->BreakParagraph(pos, flag);
	 par = par->Next();
	 pos = 0;
      }
    }  
     i++;
  }
   */ 
   
   
   while (string[i]) {
    if (string[i]!='\n') {
       if (string[i]==' ' && (string[i+1]!=' ')
	   && pos && par->GetChar(pos-1)!=' ') {
	  par->InsertChar(pos,' ');
	  pos++;
       }
       else if (string[i]==' ') {
	  par->InsertChar(pos,LYX_META_PROTECTED_SEPARATOR);
	  pos++;
       }
       else if (string[i]=='\t') {
	  for (a=pos; a<(pos/8+1)*8 ; a++) {
	     par->InsertChar(a,LYX_META_PROTECTED_SEPARATOR);
	  }
	  pos = a;
       }
       else if (string[i]!=13) {
	  par->InsertChar(pos,string[i]);
	  pos++;
       }
    }
      else {
	 if (!par->last) {
	    par->InsertChar(pos,LYX_META_PROTECTED_SEPARATOR);
	    pos++;
	 }
	 par->BreakParagraph(pos, flag);
	 par = par->Next();
	 pos = 0;
      }
      
      i++;
   }
   

   RedoParagraphs(cursor,endpar);
   SetCursor(cursor.par, cursor.pos);
   sel_cursor = cursor;
   SetCursor(par, pos);
   SetSelection();
}

/* turns double-CR to single CR, others where converted into one blank and 13s 
* that are ignored .Double spaces are also converted into one. Spaces at the beginning
* of a paragraph are forbidden. tabs are converted into one space.
* then InsertStringA is called */ 
void LyXText::InsertStringB(char* string) {
   int i=1;
   while (string[i]) {
      if (string[i]=='\t')
      	 string[i] = ' ';
      if (string[i]==' ' && string[i+1]==' ')
      	 string[i] = 13;
      if (string[i]=='\n' && string[i+1]){
	 if (string[i+1]!='\n') {
	    if (string[i-1]!=' ')
	       string[i]=' ';
	    else
	       string[i]= 13;
	 }
	 while (string[i+1] && (string[i+1]==' '
				|| string[i+1]=='\t'
				|| string[i+1]=='\n'
				|| string[i+1]==13)) {
	    string[i+1]=13;
	    i++;
	 }
      }
      i++;
   }
   InsertStringA(string);
}


char LyXText::GotoNextError() {
   LyXCursor res=cursor;
   do {
      if (res.pos < res.par->Last()-1) {
	 res.pos++;
      }
      else  {
	 res.par=res.par->Next();
	 res.pos = 0;
      }
      
   } while (res.par && 
	    !(res.par->GetChar(res.pos)==LYX_META_INSET
	      && res.par->GetInset(res.pos)->AutoDelete()));
   
   if (res.par) {
      SetCursor(res.par, res.pos);
      return 1;
   }
   
   return 0;
   
}
char LyXText::GotoNextNote() {
   LyXCursor res=cursor;
   do {
      if (res.pos < res.par->Last()-1) {
	 res.pos++;
      }
      else  {
	 res.par=res.par->Next();
	 res.pos = 0;
      }
      
   } while (res.par && 
	    !(res.par->GetChar(res.pos)==LYX_META_INSET
	      && res.par->GetInset(res.pos)->LyxCode()==LYX_IGNORE_CODE));
   
   if (res.par) {
      SetCursor(res.par, res.pos);
      return 1;
   }
   
   return 0;
   
}

int LyXText::SwitchLayoutsBetweenClasses(char class1, char class2, LyXParagraph *par) {
   char* name = NULL;
   char* s;
   InsetError * new_inset = NULL;
   int ret = 0;
   if (!par || class1 == class2)
      return ret;
   par = par->FirstPhysicalPar();
   while (par) {
      name = NameOfLayout(class1, par->layout);
      par->layout = NumberOfLayout(class2, name);
      
      if (!StringEqual(name, NameOfLayout(class2, par->layout))) {
	 ret++;
	 s = new char[300];
	 sprintf(s, "Layout had to be changed from\n%s to %s\nbecause of class conversion from\n%s to %s", 
		 name, NameOfLayout(class2, par->layout), NameOfClass(class1), NameOfClass(class2));
	 new_inset = new InsetError(s);
	 par->InsertChar(0, LYX_META_INSET);
	 par->InsertInset(0, new_inset);
	 delete(s);
      }
      
      
      par = par->next;
   }
   return ret;
}

